/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2010 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Felix Metzner, Max Welsch, Patrick Ecker                 *
 *                                                                        *
 * File for untracked private Variables for B -> D(*(*)) l nu             *
 * Created on October 17th 2018                                           *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#include <analysis/VariableManager/OwnVariables.h>
#include <analysis/utility/ReferenceFrame.h>
#include <analysis/VariableManager/Manager.h>
#include <analysis/VariableManager/Variables.h>

#include <analysis/dataobjects/Particle.h>

#include <mdst/dataobjects/MCParticle.h>

#include <framework/datastore/StoreArray.h>
#include <framework/utilities/Conversion.h>

#include <TLorentzVector.h>
#include <TVector3.h>

#include <boost/lexical_cast.hpp>
#include <cmath>

#include <assert.h>

using namespace std;

namespace Belle2 {
  namespace Variable {

    bool isValidBMeson(const Particle* part, unsigned int minDaughters)
    {
      if (part == nullptr) {
        return false;
      }
      int absPdg = abs(part->getPDGCode());

      if (absPdg != 511 and absPdg != 521) {
        return false;
      }

      unsigned int nDaughters = part->getNDaughters();
      if (nDaughters < minDaughters) {
        return false;
      }

      int d1Pdg = abs(part->getDaughter(1)->getPDGCode());
      if ((d1Pdg < 11) or (d1Pdg > 15)) {
        return false;
      }

      if (minDaughters < 3) {
        return true;
      }

      int d2Pdg = abs(part->getDaughter(2)->getPDGCode());
      if ((d2Pdg < 12) or (d2Pdg > 16)) {
        return false;
      }

      return true;
    }

    bool isValidBMesonMC(const MCParticle* mcPart, unsigned int minDaughters)
    {
      if (mcPart == nullptr) {
        return false;
      }
      int absPdg = abs(mcPart->getPDG());

      if (absPdg != 511 and absPdg != 521) {
        return false;
      }

      unsigned int nDaughters = mcPart->getNDaughters();
      if (nDaughters < minDaughters) {
        return false;
      }

      auto daughters = mcPart->getDaughters();

      int d1Pdg = abs(daughters.at(1)->getPDG());
      if ((d1Pdg < 11) or (d1Pdg > 15)) {
        return false;
      }

      if (minDaughters < 3) {
        return true;
      }

      int d2Pdg = abs(daughters.at(2)->getPDG());
      if ((d2Pdg < 12) or (d2Pdg > 16)) {
        return false;
      }

      return true;
    }

    bool isValidUpsilon4S(const Particle* part)
    {
      if (part == nullptr) {
        return false;
      }
      int absPdg = abs(part->getPDGCode());

      if (absPdg != 300553) {
        return false;
      }

      unsigned int nDaughters = part->getNDaughters();
      if (nDaughters != 2) {
        return false;
      }

      int d0Pdg = abs(part->getDaughter(0)->getPDGCode());
      int d1Pdg = abs(part->getDaughter(1)->getPDGCode());
      if ((d0Pdg != 511) and (d0Pdg != 521)) {
        return false;
      }
      if ((d1Pdg != 511) and (d1Pdg != 521)) {
        return false;
      }

      return true;
    }

    double pStarL(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
        tagVec.SetE(-beamEnergy);

        TLorentzRotation lab2BSigRestFrame = TLorentzRotation(-(T.rotateCmsToLab() * -tagVec).BoostVector());

        TLorentzVector lVecInLab = part->getDaughter(1)->getDaughter(1)->get4Vector();

        return (lab2BSigRestFrame * lVecInLab).Vect().Mag();
      }
    }

    double qSquared(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get D(*) meson
        const Particle* daughter0 = part->getDaughter(0);

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pD = frame.getMomentum(daughter0);

        auto q = pB - pD;

        return q.Mag2();
      }
    }

    double qSquaredMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();

      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();

        // get D(*) meson
        const Particle daughter0(mcDaughters.at(0));

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pD = frame.getMomentum(&daughter0);

        auto q = pB - pD;

        return q.Mag2();
      }
    }

    double qSquaredReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        tagVec.SetE(-beamEnergy);

        return (-tagVec - DVec).M2();
      }
    }

    double qSquaredReco2(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        tagVec.SetE(-tagVec.E());

        return (-tagVec - DVec).M2();
      }
    }

    double qSquaredReco3(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        TLorentzVector upsilon4SBoost = T.getBoostVector();

        TLorentzVector tagVec = part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        return (upsilon4SBoost - tagVec - DVec).M2();
      }
    }

    double qSquaredReco4(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        TLorentzVector upsilon4SBoost = T.getBoostVector();
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        tagVec.SetE(beamEnergy);

        return (upsilon4SBoost - tagVec - DVec).M2();
      }
    }

    double qSquaredRecoDStarStar(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(2) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(0)->get4Vector();
        TLorentzVector Pi0Vec = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(2)->get4Vector();

        tagVec.SetE(-beamEnergy);

        return (-tagVec - DVec - Pi0Vec).M2();
      }
    }

    double w(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get D** meson
        const Particle* daughter0 = part->getDaughter(0);

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pDstst = frame.getMomentum(daughter0);

        auto q = pB - pDstst;

        return (pB.M2() + pDstst.M2() - q.M2()) / (2 * pB.M() * pDstst.M());
      }
    }

    double wMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        // get D** meson
        Particle dstst{mcDaughters.at(0)};
        const Particle* daughter0 = &dstst;

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pDstst = frame.getMomentum(daughter0);

        auto q = pB - pDstst;

        return (pB.M2() + pDstst.M2() - q.M2()) / (2 * pB.M() * pDstst.M());
      }
    }

    double wReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pD = T.rotateLabToCms() * sigB->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * sigB->getDaughter(2)->get4Vector();
          pD += pPi;
        }

        pTag.SetE(-beamEnergy);

        TLorentzVector pBSig = -pTag;
        TLorentzVector q = -pTag - pD;

        return (pBSig.M2() + pD.M2() - q.M2()) / (2 * pBSig.M() * pD.M());
      }
    }

    double cosThetaL(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get lepton and neutrino
        const Particle* daughter1 = part->getDaughter(1);
        const Particle* daughter2 = part->getDaughter(2);

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;

        pl.Boost(motherBoost);
        pW.Boost(motherBoost);
        pl.Boost(-pW.BoostVector());

        return cos(pW.Angle(pl.Vect()));
      }
    }

    double cosThetaLMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        // get lepton and neutrino
        Particle d1{mcDaughters.at(1)};
        Particle d2{mcDaughters.at(2)};
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);

        // Adding 4Momentum of PHOTOS Photons to Lepton 4Momentum.
        for (const MCParticle* mcDaughter : mcDaughters) {
          if (mcDaughter->getPDG() == 22) {
            const Particle photon{mcDaughter};
            pl += frame.getMomentum(&photon);
          };
        }

        TLorentzVector pW = pl + pnu;

        pl.Boost(motherBoost);
        pW.Boost(motherBoost);
        pl.Boost(-pW.BoostVector());

        return cos(pW.Angle(pl.Vect()));
      }
    }

    double cosThetaLWOPhotonsMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        // get lepton and neutrino
        Particle d1{mcDaughters.at(1)};
        Particle d2{mcDaughters.at(2)};
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;

        pl.Boost(motherBoost);
        pW.Boost(motherBoost);
        pl.Boost(-pW.BoostVector());

        return cos(pW.Angle(pl.Vect()));
      }
    }

    double cosThetaLReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pD = T.rotateLabToCms() * sigB->getDaughter(0)->get4Vector();
        TLorentzVector pl = T.rotateLabToCms() * sigB->getDaughter(1)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * sigB->getDaughter(2)->get4Vector();
          pD += pPi;
        }

        pTag.SetE(-beamEnergy);

        TLorentzVector pW = -pTag - pD;

        pl.Boost(-pW.BoostVector());

        return cos(pW.Angle(pl.Vect()));
      }
    }

    double cosThetaV(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get D** and daugther(D**, 0)
        const Particle* daughter0 = part->getDaughter(0);
        const Particle* grandDaughter00 = daughter0->getDaughter(0);

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        pDstst.Boost(motherBoost);
        pD.Boost(motherBoost);
        pD.Boost(-pDstst.BoostVector());

        return cos(pDstst.Angle(pD.Vect()));
      }
    }

    double cosThetaVMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        auto mcGrandDaughters = mcDaughters.at(0)->getDaughters();

        // get D** and daugther(D**, 0)
        Particle d0{mcDaughters.at(0)};
        Particle gd00{mcGrandDaughters.at(0)};

        const Particle* daughter0 = &d0;
        const Particle* grandDaughter00 = &gd00;

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        pDstst.Boost(motherBoost);
        pD.Boost(motherBoost);
        pD.Boost(-pDstst.BoostVector());

        return cos(pDstst.Angle(pD.Vect()));
      }
    }

    double cosThetaVReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pDst = sigB->getDaughter(0)->get4Vector();
        TLorentzVector pD = sigB->getDaughter(0)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = sigB->getDaughter(2)->get4Vector();
          pDst += pPi;
        }

        pTag.SetE(-beamEnergy);
        TLorentzVector pBSigLab = T.rotateCmsToLab() * (-pTag);

        pDst.Boost(-pBSigLab.BoostVector());
        pD.Boost(-pBSigLab.BoostVector());
        pD.Boost(-pDst.BoostVector());

        return cos(pDst.Angle(pD.Vect()));
      }
    }

    double chiPrime(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        const Particle* daughter0 = part->getDaughter(0);
        const Particle* daughter1 = part->getDaughter(1);
        const Particle* daughter2 = part->getDaughter(2);
        const Particle* grandDaughter00 = daughter0->getDaughter(0);

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        TVector3 WBoost = -pW.BoostVector();
        TVector3 DststBoost = -pDstst.BoostVector();

        pD.Boost(DststBoost);
        double cosThetaV =
          (-pD.Px() * pDstst.Px() - pD.Py() * pDstst.Py() - pD.Pz() * pDstst.Pz()) / (pD.P() * pDstst.P());

        pl.Boost(WBoost);
        double cosThetaWL = (-pl.Px() * pW.Px() - pl.Py() * pW.Py() - pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Perp() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Perp() * pDstst.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiPrimeMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        // get D** and daugther(D**, 0)
        Particle d0{mcDaughters.at(0)};
        Particle d1{mcDaughters.at(1)};
        Particle d2{mcDaughters.at(2)};

        const Particle* daughter0 = &d0;
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        auto mcGrandDaughters = mcDaughters.at(0)->getDaughters();
        Particle gd00{mcGrandDaughters.at(0)};
        const Particle* grandDaughter00 = &gd00;

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);

        // Adding 4Momentum of PHOTOS Photons to Lepton 4Momentum.
        for (const MCParticle* mcDaughter : mcDaughters) {
          if (mcDaughter->getPDG() == 22) {
            const Particle photon{mcDaughter};
            pl += frame.getMomentum(&photon);
          };
        }

        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        TVector3 WBoost = -pW.BoostVector();
        TVector3 DststBoost = -pDstst.BoostVector();

        pD.Boost(DststBoost);
        double cosThetaV =
          (-pD.Px() * pDstst.Px() - pD.Py() * pDstst.Py() - pD.Pz() * pDstst.Pz()) / (pD.P() * pDstst.P());

        pl.Boost(WBoost);
        double cosThetaWL = (-pl.Px() * pW.Px() - pl.Py() * pW.Py() - pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Perp() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Perp() * pDstst.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiPrimeWOPhotonsMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        // get D** and daugther(D**, 0)
        Particle d0{mcDaughters.at(0)};
        Particle d1{mcDaughters.at(1)};
        Particle d2{mcDaughters.at(2)};

        const Particle* daughter0 = &d0;
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        auto mcGrandDaughters = mcDaughters.at(0)->getDaughters();
        Particle gd00{mcGrandDaughters.at(0)};
        const Particle* grandDaughter00 = &gd00;

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        TVector3 WBoost = -pW.BoostVector();
        TVector3 DststBoost = -pDstst.BoostVector();

        pD.Boost(DststBoost);
        double cosThetaV =
          (-pD.Px() * pDstst.Px() - pD.Py() * pDstst.Py() - pD.Pz() * pDstst.Pz()) / (pD.P() * pDstst.P());

        pl.Boost(WBoost);
        double cosThetaWL = (-pl.Px() * pW.Px() - pl.Py() * pW.Py() - pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Perp() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Perp() * pDstst.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiPrimeReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pDst = sigB->getDaughter(0)->get4Vector();
        TLorentzVector pD = sigB->getDaughter(0)->getDaughter(0)->get4Vector();
        TLorentzVector pl = sigB->getDaughter(1)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = sigB->getDaughter(2)->get4Vector();
          pDst += pPi;
        }

        pTag.SetE(-beamEnergy);
        TLorentzVector pBSigLab = T.rotateCmsToLab() * (-pTag);

        TLorentzVector pW = pBSigLab - pDst;

        pl.Boost(-pW.BoostVector());
        pD.Boost(-pDst.BoostVector());

        double cosThetaV = (-pD.Px() * pDst.Px() - pD.Py() * pDst.Py() - pD.Pz() * pDst.Pz()) / (pD.P() * pDst.P());
        double cosThetaWL = (-pl.Px() * pW.Px() - pl.Py() * pW.Py() - pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Perp() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Perp() * pDst.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double altChi(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        const Particle* daughter0 = part->getDaughter(0);
        const Particle* daughter1 = part->getDaughter(1);
        const Particle* daughter2 = part->getDaughter(2);
        const Particle* grandDaughter00 = daughter0->getDaughter(0);

        TLorentzVector pDs = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        pD.Boost(motherBoost);
        pDs.Boost(motherBoost);
        pl.Boost(motherBoost);
        pW.Boost(motherBoost);

        double cosThetaV = (pD.Px() * pDs.Px() + pD.Py() * pDs.Py() + pD.Pz() * pDs.Pz()) / (pD.P() * pDs.P());
        double cosThetaWL = (pl.Px() * pW.Px() + pl.Py() * pW.Py() + pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Mag() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Mag() * pDs.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        // get D** and daugther(D**, 0)
        Particle d0{mcDaughters.at(0)};
        Particle d1{mcDaughters.at(1)};
        Particle d2{mcDaughters.at(2)};

        const Particle* daughter0 = &d0;
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        auto mcGrandDaughters = mcDaughters.at(0)->getDaughters();
        Particle gd00{mcGrandDaughters.at(0)};
        const Particle* grandDaughter00 = &gd00;

        TLorentzVector pDs = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);

        // Adding 4Momentum of PHOTOS Photons to Lepton 4Momentum.
        for (const MCParticle* mcDaughter : mcDaughters) {
          if (mcDaughter->getPDG() == 22) {
            const Particle photon{mcDaughter};
            pl += frame.getMomentum(&photon);
          };
        }

        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        pD.Boost(motherBoost);
        pDs.Boost(motherBoost);
        pl.Boost(motherBoost);
        pW.Boost(motherBoost);

        double cosThetaV = (pD.Px() * pDs.Px() + pD.Py() * pDs.Py() + pD.Pz() * pDs.Pz()) / (pD.P() * pDs.P());
        double cosThetaWL = (pl.Px() * pW.Px() + pl.Py() * pW.Py() + pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Mag() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Mag() * pDs.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiWOPhotonsMC(const Particle* part)
    {
      auto mcPart = part->getMCParticle();
      if (not isValidBMesonMC(mcPart)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mcDaughters = mcPart->getDaughters();
        // get D** and daugther(D**, 0)
        Particle d0{mcDaughters.at(0)};
        Particle d1{mcDaughters.at(1)};
        Particle d2{mcDaughters.at(2)};

        const Particle* daughter0 = &d0;
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        auto mcGrandDaughters = mcDaughters.at(0)->getDaughters();
        Particle gd00{mcGrandDaughters.at(0)};
        const Particle* grandDaughter00 = &gd00;

        TLorentzVector pDs = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(grandDaughter00);

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        pD.Boost(motherBoost);
        pDs.Boost(motherBoost);
        pl.Boost(motherBoost);
        pW.Boost(motherBoost);

        double cosThetaV = (pD.Px() * pDs.Px() + pD.Py() * pDs.Py() + pD.Pz() * pDs.Pz()) / (pD.P() * pDs.P());
        double cosThetaWL = (pl.Px() * pW.Px() + pl.Py() * pW.Py() + pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Mag() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Mag() * pDs.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pDst = sigB->getDaughter(0)->get4Vector();
        TLorentzVector pD = sigB->getDaughter(0)->getDaughter(0)->get4Vector();
        TLorentzVector pl = sigB->getDaughter(1)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = sigB->getDaughter(2)->get4Vector();
          pDst += pPi;
        }

        pTag.SetE(-beamEnergy);
        TLorentzVector pBSigLab = T.rotateCmsToLab() * (-pTag);

        TLorentzVector pW = pBSigLab - pDst;

        TLorentzVector pTagForBoost = tagB->get4Vector();
        pTagForBoost.SetE(-beamEnergy);
        TVector3 bSigBoost = -(-pTagForBoost).BoostVector();

        pl.Boost(bSigBoost);
        pD.Boost(bSigBoost);
        pDst.Boost(bSigBoost);
        pW.Boost(bSigBoost);

        double cosThetaV = (pD.Px() * pDst.Px() + pD.Py() * pDst.Py() + pD.Pz() * pDst.Pz()) / (pD.P() * pDst.P());
        double cosThetaWL = (pl.Px() * pW.Px() + pl.Py() * pW.Py() + pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 lTrans = pl.Vect() - cosThetaWL * pl.Vect().Mag() * pW.Vect().Unit();
        TVector3 vTrans = pD.Vect() - cosThetaV * pD.Vect().Mag() * pDst.Vect().Unit();

        double chi = atan2(lTrans.Cross(vTrans) * pW.Vect().Unit(), lTrans * vTrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    template<typename Iterator>
    TLorentzVector sumMomentumOfAllDescendantsWithPDG(
      const MCParticle* mcPart,
      Iterator begin,
      Iterator end
    )
    {
      TLorentzVector sumOf4Momenta(0., 0., 0., 0.);
      for (const MCParticle* mcDaughter : mcPart->getDaughters()) {
        if (std::find(begin, end, abs(mcDaughter->getPDG())) != end) {
          sumOf4Momenta += Particle(mcDaughter).get4Vector();
        } else if (mcDaughter->getNDaughters() > 0) {
          sumOf4Momenta += sumMomentumOfAllDescendantsWithPDG(mcDaughter, begin, end);
        }
      }

      return sumOf4Momenta;
    };

    double mcMissingMassSquared(const Particle*)
    {
      StoreArray <MCParticle> mcParticles("MCParticles");
      if (mcParticles.getEntries() == 0) {
        return -999;
      }

      MCParticle* mcUpsilon4S = mcParticles[0];
      if (mcUpsilon4S->getPDG() != 300553) {
        return -999;
      }

      std::vector<MCParticle*> mcU4SDaughters = mcUpsilon4S->getDaughters();
      if (mcU4SDaughters.size() != 2) {
        return -999;
      }

      const MCParticle* firstB = mcU4SDaughters.at(0);
      const MCParticle* secondB = mcU4SDaughters.at(1);

      if ((abs(firstB->getPDG()) != 511 && abs(firstB->getPDG()) != 521)
          || (abs(secondB->getPDG()) != 511 && abs(secondB->getPDG()) != 521)) {
        return -999;
      }

      std::vector<int> nuPDGs {12, 14, 16};

      TLorentzVector firstNuMomentumSum = sumMomentumOfAllDescendantsWithPDG(firstB, nuPDGs.begin(), nuPDGs.end());
      TLorentzVector secondNuMomentumSum = sumMomentumOfAllDescendantsWithPDG(secondB, nuPDGs.begin(), nuPDGs.end());

      if (firstNuMomentumSum.P() == 0. && secondNuMomentumSum.P() == 0) {
        return -2.;
      } else if (firstNuMomentumSum.P() == 0.) {
        return secondNuMomentumSum.M2();
      } else if (secondNuMomentumSum.P() == 0.) {
        return firstNuMomentumSum.M2();
      } else {
        return -1;
      }
    }

    double missingMomentum(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();
      tagVec.SetE(beamEnergy);
      TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);

      return missingMomentum.Vect().Mag();
    }

    double missingMomentumRescaledTag(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      tagVec.SetE(beamEnergy);
      double bMesonMass = part->getDaughter(0)-> getPDGMass();
      double pTagMag = sqrt(pow(beamEnergy, 2) - pow(bMesonMass, 2));
      TVector3 vecTag = tagVec.Vect();
      vecTag.SetMag(pTagMag);
      tagVec.SetVect(vecTag);

      TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);
      return missingMomentum.Vect().Mag();
    }

    double mcMissingMomentum(const Particle*)
    {
      StoreArray <MCParticle> mcParticles("MCParticles");
      if (mcParticles.getEntries() == 0) {
        return -999;
      }

      MCParticle* mcUpsilon4S = mcParticles[0];
      if (mcUpsilon4S->getPDG() != 300553) {
        return -999;
      }

      std::vector<MCParticle*> mcU4SDaughters = mcUpsilon4S->getDaughters();
      if (mcU4SDaughters.size() != 2) {
        return -999;
      }

      const MCParticle* firstB = mcU4SDaughters.at(0);
      const MCParticle* secondB = mcU4SDaughters.at(1);

      if ((abs(firstB->getPDG()) != 511 && abs(firstB->getPDG()) != 521)
          || (abs(secondB->getPDG()) != 511 && abs(secondB->getPDG()) != 521)) {
        return -999;
      }

      std::vector<int> nuPDGs {12, 14, 16};

      TLorentzVector firstNuMomentumSum = sumMomentumOfAllDescendantsWithPDG(firstB, nuPDGs.begin(), nuPDGs.end());
      TLorentzVector secondNuMomentumSum = sumMomentumOfAllDescendantsWithPDG(secondB, nuPDGs.begin(), nuPDGs.end());

      if (firstNuMomentumSum.P() == 0. && secondNuMomentumSum.P() == 0) {
        return -2.;
      } else if (firstNuMomentumSum.P() == 0.) {
        return secondNuMomentumSum.Vect().Mag();
      } else if (secondNuMomentumSum.P() == 0.) {
        return firstNuMomentumSum.Vect().Mag();
      } else {
        return -1;
      }
    }

    double missingMomentumInCMSFrame(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();
      tagVec.SetE(beamEnergy);
      TLorentzVector missingMomentum = tagVec + sigVec;

      return missingMomentum.Vect().Mag();
    }

    double missingMomentumRescaledTagInCMSFrame(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      double bMesonMass = part->getDaughter(0)-> getPDGMass();
      double pTagMag = sqrt(pow(beamEnergy, 2) - pow(bMesonMass, 2));
      TVector3 vecTag = tagVec.Vect();
      vecTag.SetMag(pTagMag);
      tagVec.SetVect(vecTag);

      TLorentzVector missingMomentum = tagVec + sigVec;
      return missingMomentum.Vect().Mag();
    }

    double missingMomentumInBSigRestFrame(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();
      tagVec.SetE(beamEnergy);
      TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);

      TVector3 boost = part->getDaughter(1)->get4Vector().BoostVector();
      TLorentzRotation lab2BSigRestFrame = TLorentzRotation(-boost);

      return (lab2BSigRestFrame * missingMomentum).Vect().Mag();
    }

    double missingMomentumRescaledTagInBSigRestFrame(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      tagVec.SetE(beamEnergy);
      double bMesonMass = part->getDaughter(0)-> getPDGMass();
      double pTagMag = sqrt(pow(beamEnergy, 2) - pow(bMesonMass, 2));
      TVector3 vecTag = tagVec.Vect();
      vecTag.SetMag(pTagMag);
      tagVec.SetVect(vecTag);

      TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);

      TVector3 boost = part->getDaughter(1)->get4Vector().BoostVector();
      TLorentzRotation lab2BSigRestFrame = TLorentzRotation(-boost);

      return (lab2BSigRestFrame * missingMomentum).Vect().Mag();
    }

    double mcMissingMomentumInBSigRestFrame(const Particle*)
    {
      StoreArray <MCParticle> mcParticles("MCParticles");
      if (mcParticles.getEntries() == 0) {
        return -999;
      }

      MCParticle* mcUpsilon4S = mcParticles[0];
      if (mcUpsilon4S->getPDG() != 300553) {
        return -999;
      }

      std::vector<MCParticle*> mcU4SDaughters = mcUpsilon4S->getDaughters();
      if (mcU4SDaughters.size() != 2) {
        return -999;
      }

      const MCParticle* firstB = mcU4SDaughters.at(0);
      const MCParticle* secondB = mcU4SDaughters.at(1);

      if ((abs(firstB->getPDG()) != 511 && abs(firstB->getPDG()) != 521)
          || (abs(secondB->getPDG()) != 511 && abs(secondB->getPDG()) != 521)) {
        return -999;
      }

      std::vector<int> nuPDGs {12, 14, 16};

      TLorentzVector firstNuMomentumSum = sumMomentumOfAllDescendantsWithPDG(firstB, nuPDGs.begin(), nuPDGs.end());
      TLorentzVector secondNuMomentumSum = sumMomentumOfAllDescendantsWithPDG(secondB, nuPDGs.begin(), nuPDGs.end());

      if (firstNuMomentumSum.P() == 0. && secondNuMomentumSum.P() == 0) {
        return -2.;
      } else if (firstNuMomentumSum.P() == 0.) {
        TVector3 boost = Particle(secondB).get4Vector().BoostVector();
        TLorentzRotation lab2SecondBRestFrame = TLorentzRotation(-boost);
        return (lab2SecondBRestFrame * secondNuMomentumSum).Vect().Mag();
      } else if (secondNuMomentumSum.P() == 0.) {
        TVector3 boost = Particle(firstB).get4Vector().BoostVector();
        TLorentzRotation lab2FirstBRestFrame = TLorentzRotation(-boost);
        return (lab2FirstBRestFrame * firstNuMomentumSum).Vect().Mag();
      } else {
        return -1;
      }
    }

    Manager::FunctionPtr angleWithMissingMomentum(const std::vector<std::string>& arguments)
    {
      if (arguments.size() == 2) {
        std::string particle = arguments[0];
        int daughterNumber = 0;
        int daughterOfDaughterNumber = 0;
        try {
          daughterNumber = Belle2::convertString<int>(arguments[0]);
        } catch (boost::bad_lexical_cast&) {
          B2WARNING("First argument of angleWithMissingMass meta function must be integer!");
          return nullptr;
        }
        try {
          daughterOfDaughterNumber = Belle2::convertString<int>(arguments[1]);
        } catch (boost::bad_lexical_cast&) {
          B2WARNING("Second argument of angleWithMissingMass meta function must be integer!");
          return nullptr;
        }

        auto func = [daughterNumber, daughterOfDaughterNumber](const Particle * part) -> double {
          if (not isValidUpsilon4S(part))
          {
            return -999;
          }
          int nDaughters = static_cast<int>(part->getNDaughters());
          if (daughterNumber >= nDaughters)
          {
            return -999;
          }

          int nDaughtersOfDaughter = static_cast<int>(part->getDaughter(daughterNumber)->getNDaughters());
          if (daughterOfDaughterNumber >= nDaughtersOfDaughter)
          {
            return -999;
          }

          PCmsLabTransform T;
          double beamEnergy = T.getCMSEnergy() / 2.;
          TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
          TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();
          tagVec.SetE(beamEnergy);
          TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);

          const Particle* BSig = part->getDaughter(daughterNumber);
          TVector3 boost = BSig->get4Vector().BoostVector();
          TLorentzRotation lab2BSigRestFrame = TLorentzRotation(-boost);

          TVector3 missingMomInRestFrame = (lab2BSigRestFrame * missingMomentum).Vect();

          TVector3 particleV = (lab2BSigRestFrame * part->getDaughter(daughterNumber)->getDaughter(
            daughterOfDaughterNumber)->get4Vector()).Vect();

          return std::cos(particleV.Angle(missingMomInRestFrame));
        };
        return func;
      } else {
        B2FATAL("Wrong number of arguments for meta function angleWithMissingMomentum");
      }
    }

    double m2RecoilSignalSideRescaledTag(const Particle* part)
    {
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();
      tagVec.SetE(-beamEnergy);
      double bMesonMass = part->getDaughter(0)-> getPDGMass();
      double pTagMag = sqrt(pow(beamEnergy, 2) - pow(bMesonMass, 2));
      TVector3 vecTag = tagVec.Vect();
      vecTag.SetMag(pTagMag);
      tagVec.SetVect(vecTag);

      return (-tagVec - sigVec).M2();
    }

    double m2RecoilSignalSideNoScaling(const Particle* part)
    {
      PCmsLabTransform T;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      TLorentzVector y4sVec = TLorentzVector(0, 0, 0, T.getCMSEnergy());
      return (y4sVec - tagVec - sigVec).M2();
    }

    double missingEnergyCMS(const Particle* part)
    {
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      tagVec.SetE(beamEnergy);

      return (tagVec - sigVec).E();
    }

    double missingEnergyNoScalingCMS(const Particle* part)
    {
      PCmsLabTransform T;

      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      return (T.getCMSEnergy() - tagVec.E() - sigVec.E());
    }

    double missingEnergyLab(const Particle* part)
    {
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2;
      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      tagVec.SetE(-beamEnergy);

      return (T.rotateCmsToLab() * (-tagVec - sigVec)).E();
    }

    double missingEnergyNoScalingLab(const Particle* part)
    {
      PCmsLabTransform T;

      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      TLorentzVector y4sVec = TLorentzVector(0, 0, 0, T.getCMSEnergy());
      return (T.rotateCmsToLab() * (y4sVec - tagVec - sigVec)).E();
    }

    double missingEnergyRescaledTagLab(const Particle* part)
    {
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2;

      TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();

      tagVec.SetE(beamEnergy);
      double bMesonMass = part->getDaughter(0)-> getPDGMass();
      double pTagMag = sqrt(pow(beamEnergy, 2) - pow(bMesonMass, 2));
      TVector3 vecTag = tagVec.Vect();
      vecTag.SetMag(pTagMag);
      tagVec.SetVect(vecTag);

      TLorentzVector y4sVec = TLorentzVector(0, 0, 0, T.getCMSEnergy());
      return (T.rotateCmsToLab() * (y4sVec - tagVec - sigVec)).E();
    }

    double wasFSRCorrected(const Particle* part)
    {
      if (part->hasExtraInfo("fsrCorrected")) {
        return 1.;
      } else {
        std::vector<Particle*> daughters = part->getDaughters();
        for (const Particle* daughter : daughters) {
          if (wasFSRCorrected(daughter)) {
            return 1.;
          }
        }
        return 0.;
      }
    }

    TLorentzVector getFSRMomentum(const Particle* part)
    {
      if (abs(part->getPDGCode()) != 11) {
        return TLorentzVector(0, 0, 0, 0);
      }
      const auto& frame = ReferenceFrame::GetCurrent();

      TLorentzVector fsr4Vector(0, 0, 0, 0);
      std::vector<Particle*> daughters = part->getDaughters();

      while (daughters.size() > 0) {
        Particle* currentDaughter = daughters.back();
        daughters.pop_back();
        std::vector<Particle*> currentDaughters = currentDaughter->getDaughters();

        if (currentDaughter->hasExtraInfo("fsrCorrected")) {
          assert((currentDaughter->getExtraInfo("fsrCorrected") + 1) == currentDaughters.size());
          for (const Particle* daughter : currentDaughters) {
            if (daughter->getPDGCode() == currentDaughter->getPDGCode()) {
              fsr4Vector += (frame.getMomentum(currentDaughter) - frame.getMomentum(daughter));
              break;
            }
          }
        } else {
          for (Particle* daughter : currentDaughters) {
            daughters.push_back(daughter);
          }
        }
      }
      return fsr4Vector;
    }

    double m2RecoilSignalSideNoFSRCorrection(const Particle* part)
    {
      if (!wasFSRCorrected(part)) {
        return m2RecoilSignalSide(part);
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;
        TLorentzVector tagVec = T.rotateLabToCms()
                                * part->getDaughter(0)->get4Vector();
        TLorentzVector sigVec = T.rotateLabToCms()
                                * part->getDaughter(1)->get4Vector();

        UseReferenceFrame<CMSFrame> frame;
        sigVec -= getFSRMomentum(part->getDaughter(1)->getDaughter(1));

        tagVec.SetE(-beamEnergy);
        return (-tagVec - sigVec).M2();
      }
    }

    double pStarLNoFSR(const Particle* part)
    {
      if (!wasFSRCorrected(part)) {
        return pStarL(part);
      } else {
        if (not isValidUpsilon4S(part)) {
          return -999;
        } else {
          PCmsLabTransform T;
          double beamEnergy = T.getCMSEnergy() / 2.;

          TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
          tagVec.SetE(-beamEnergy);

          TLorentzRotation lab2BSigRestFrame = TLorentzRotation(-(T.rotateCmsToLab() * -tagVec).BoostVector());

          TLorentzVector lVecInLab = part->getDaughter(1)->getDaughter(1)->get4Vector();
          UseReferenceFrame<LabFrame> frame;
          lVecInLab -= getFSRMomentum(part->getDaughter(1)->getDaughter(1));

          return (lab2BSigRestFrame * lVecInLab).Vect().Mag();
        }
      }
    }

    double fsrMomentumX(const Particle* part)
    {
      if (!wasFSRCorrected(part)) {
        return 0;
      } else {
        return getFSRMomentum(part->getDaughter(1)->getDaughter(1)).X();
      }
    }

    double fsrMomentumY(const Particle* part)
    {
      if (!wasFSRCorrected(part)) {
        return 0;
      } else {
        return getFSRMomentum(part->getDaughter(1)->getDaughter(1)).Y();
      }
    }

    double fsrMomentumZ(const Particle* part)
    {
      if (!wasFSRCorrected(part)) {
        return 0;
      } else {
        return getFSRMomentum(part->getDaughter(1)->getDaughter(1)).Z();
      }
    }

    double fsrE(const Particle* part)
    {
      if (!wasFSRCorrected(part)) {
        return 0;
      } else {
        return getFSRMomentum(part->getDaughter(1)->getDaughter(1)).E();
      }
    }


    VARIABLE_GROUP("MYVARIABLES");
    REGISTER_VARIABLE("pStarL", pStarL,
                      "Magnitude of the lepton momentum in the rest frame of the "
                      "signal B meson extracted via the tag side B's 4-momentum "
                      "for which the energy component is replaced by 1/2 of the "
                      "beam energy.\n"
                      "Must be a calculated from the Upsilon(4S) for which the"
                      "second daughter must be the signal B meson.");
    REGISTER_VARIABLE("qSquared", qSquared,
                      "Squared difference of the four-momenta of the B meson "
                      "and the first daughter of the B meson in the B rest frame. "
                      "The B meson is required to have 3 daughter particles!");
    REGISTER_VARIABLE("qSquaredMC", qSquaredMC,
                      "Squared difference of the four-momenta of the B meson "
                      "and the first daughter of the B meson in the B rest frame. "
                      "The B meson is required to have 3 daughter particles!");
    REGISTER_VARIABLE("qSquaredReco", qSquaredReco,
                      "To be applied on Upsilon(4S) particles to retrieve the "
                      "squared difference of the four-momenta of the signal B meson "
                      "and its first daughter. The signal B meson must be the second "
                      "daughter of the Upsilon(4S) meson! The Upsilon(4S) meson must "
                      "have 2 B mesons as daughters to be valid.");
    REGISTER_VARIABLE("qSquaredReco2", qSquaredReco2, "");
    REGISTER_VARIABLE("qSquaredReco3", qSquaredReco3, "");
    REGISTER_VARIABLE("qSquaredReco4", qSquaredReco4, "");
    REGISTER_VARIABLE("qSquaredRecoDStarStar", qSquaredRecoDStarStar,
                      "To be applied on Upsilon(4S) particles to retrieve the "
                      "squared difference of the four-momenta of the signal B meson "
                      "and its first and third daughter which together form a D** candidate. "
                      "The signal B meson must be the second daughter of the Upsilon(4S) meson! "
                      "The Upsilon(4S) meson must have 2 B mesons as daughters to be valid.");
    REGISTER_VARIABLE("w", w,
                      "product of four velocities of B meson and D** daugther");
    REGISTER_VARIABLE("wMC", wMC,
                      "product of four velocities of B meson and D** daugther");
    REGISTER_VARIABLE("wReco", wReco,
                      "product of four velocities of B meson and D** daugther");
    REGISTER_VARIABLE("costhetal", cosThetaL,
                      "cosine of angle between virtual W boson and lepton in W rest frame");
    REGISTER_VARIABLE("costhetalMC", cosThetaLMC,
                      "cosine of angle between virtual W boson and lepton in W rest frame");
    REGISTER_VARIABLE("costhetalWOPhotonsMC", cosThetaLWOPhotonsMC,
                      "cosine of angle between virtual W boson and lepton in W rest frame  "
                      "Photons photons are NOT added to lepton 4 Momentum!");
    REGISTER_VARIABLE("costhetalReco", cosThetaLReco,
                      "cosine of angle between virtual W boson and lepton in W rest frame calculated "
                      "from reconstructed Upsilon(4S) meson.");
    REGISTER_VARIABLE("costhetav", cosThetaV,
                      "cosine of angle between D** meson and D(*) meson in D(*) rest frame");
    REGISTER_VARIABLE("costhetavMC", cosThetaVMC,
                      "cosine of angle between D** meson and D(*) meson in D(*) rest frame");
    REGISTER_VARIABLE("costhetavReco", cosThetaVReco,
                      "cosine of angle between D** meson and D(*) meson in D(*) rest frame");
    REGISTER_VARIABLE("chi_prime", chiPrime,
                      "angle between decay planes of virtual W boson and D** boson");
    REGISTER_VARIABLE("chi_primeMC", chiPrimeMC,
                      "angle between decay planes of virtual W boson and D** boson");
    REGISTER_VARIABLE("chi_primeWOPhotonsMC", chiPrimeWOPhotonsMC,
                      "angle between decay planes of virtual W boson and D** boson. "
                      "Photons photons are NOT added to lepton 4 Momentum!");
    REGISTER_VARIABLE("chi_primeReco", chiPrimeReco,
                      "angle between decay planes of virtual W boson and D** boson");
    REGISTER_VARIABLE("alt_chi", altChi,
                      "angle between decay planes of virtual W boson and D* boson in the B meson rest frame");
    REGISTER_VARIABLE("chi_MC", chiMC,
                      "angle between decay planes of virtual W boson and D* boson in the B meson rest frame");
    REGISTER_VARIABLE("chiWOPhotons_MC", chiWOPhotonsMC,
                      "angle between decay planes of virtual W boson and D* boson in the B meson rest frame. "
                      "Photons photons are NOT added to lepton 4 Momentum!");
    REGISTER_VARIABLE("chi_Reco", chiReco,
                      "angle between decay planes of virtual W boson and D* boson in the B meson rest frame");
    REGISTER_VARIABLE("mcMissingMassSquared", mcMissingMassSquared,
                      "Returns the true squared Missing Mass of the event.");
    REGISTER_VARIABLE("missingMomentum", missingMomentum,
                      "Returns the Missing Momentum of the event. "
                      "The Signal B meson must be the second daughter of the Upsilon(4s)!\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B.");
    REGISTER_VARIABLE("missingMomentumRescaledTag", missingMomentumRescaledTag,
                      "Returns the Missing Momentum of the event. "
                      "The Signal B meson must be the second daughter of the Upsilon(4s)!\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B "
                      "and rescaling the momentum of the tag-side B accordingly.");
    REGISTER_VARIABLE("mcMissingMomentum", mcMissingMomentum,
                      "Returns the true Missing Momentum of the event in the lab frame.");
    REGISTER_VARIABLE("missingMomentumInCMSFrame", missingMomentumInCMSFrame,
                      "Returns the Missing Momentum of the event in the center of mass frame, "
                      "The signal B meson must be the second daughter of the Upsilon(4s).\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B.");
    REGISTER_VARIABLE("missingMomentumRescaledTagInCMSFrame", missingMomentumRescaledTagInCMSFrame,
                      "Returns the Missing Momentum of the event in the center of mass frame, "
                      "The signal B meson must be the second daughter of the Upsilon(4s).\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B and "
                      "rescaling the momentum of the tag - side B accordingly.");
    REGISTER_VARIABLE("missingMomentumInBSigRestFrame", missingMomentumInBSigRestFrame,
                      "Returns the Missing Momentum of the event in the rest frame of the signal side, "
                      "if signal B is the second daughter of the Upsilon(4s).\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B.");
    REGISTER_VARIABLE("missingMomentumRescaledTagInBSigRestFrame", missingMomentumRescaledTagInBSigRestFrame,
                      "Returns the Missing Momentum of the event in the rest frame of the signal side, "
                      "if signal B is the second daughter of the Upsilon(4s).\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag - side B and "
                      "rescaling the momentum of the tag - side B accordingly.");
    REGISTER_VARIABLE("mcMissingMomentumInBSigRestFrame", mcMissingMomentumInBSigRestFrame,
                      "Returns the true Missing Momentum of the event in the signal B meson's rest frame.");
    REGISTER_VARIABLE("angleWithMissingMomentum(daughterNumber, daughterOfDaughterNumber)", angleWithMissingMomentum,
                      "Returns the cosin of the angle between the missing momentum (hence the function must "
                      "be applied on the Upsilon(4S)!) and the momentum vector of the daughterOfDaughterNumber-th "
                      "daughter of the daughterNumber-th daughter of the particle it is applied to in the rest frame "
                      "of the signal B meson's rest frame.");
    REGISTER_VARIABLE("m2RecoilSignalSideRescaledTag", m2RecoilSignalSideRescaledTag,
                      "returns the squared missing mass of the signal side which is calculated in the CMS frame under the assumption that the signal and "
                      "tag side are produced back to back and the tag side energy equals the beam energy and the tag-side moemntum is resacaled accordingly.\n"
                      "The variable must be applied to the Upsilon and"
                      "the tag side must be the first, the signal side the second daughter!");
    REGISTER_VARIABLE("m2RecoilSignalSideNoScaling", m2RecoilSignalSideNoScaling,
                      "returns the squared missing mass of the signal side which is calculated in the CMS frame under the assumption that the signal and "
                      "tag side are produced back to back.\n"
                      "The variable must be applied to the Upsilon and"
                      "the tag side must be the first, the signal side the second daughter!");
    REGISTER_VARIABLE("missingEnergyCMS", missingEnergyCMS,
                      "Returns the missing energy in the event with the assumption that the energy of the tag-side B meson equals the beamEnergy.\n"
                      "The variable must be applied to the Upsilon and"
                      "the tag side must be the first, the signal side the second daughter!");
    REGISTER_VARIABLE("missingEnergyNoScalingCMS", missingEnergyNoScalingCMS,
                      "Returns the missing energy in the event.\n"
                      "The variable must be applied to the Upsilon and"
                      "the tag side must be the first, the signal side the second daughter!");
    REGISTER_VARIABLE("missingEnergyLab", missingEnergyLab,
                      "Returns the missing energy in the event with the assumption that the energy of the tag-side B meson equals the beamEnergy.\n"
                      "The variable must be applied to the Upsilon and"
                      "the tag side must be the first, the signal side the second daughter!");
    REGISTER_VARIABLE("missingEnergyNoScalingLab", missingEnergyNoScalingLab,
                      "Returns the missing energy in the event.\n"
                      "The variable must be applied to the Upsilon and"
                      "the tag side must be the first, the signal side the second daughter!");
    REGISTER_VARIABLE("missingEnergyRescaledTagLab", missingEnergyRescaledTagLab,
                      "Returns the missing energy in the event with the assumption that the energy of the tag-side B meson equals the beamEnergy "
                      "and the momentum is rescaled accordingly.\n"
                      "The variable must be applied to the Upsilon and"
                      "the tag side must be the first, the signal side the second daughter!");
    REGISTER_VARIABLE("wasFSRCorrected", wasFSRCorrected,
                      "Returns if a given Particle has a FSR corrected Particle somewhere in his decay chain.");
    REGISTER_VARIABLE("m2RecoilSignalSideNoFSRCorrection", m2RecoilSignalSideNoFSRCorrection,
                      "Returns the squared missing mass with no fsr correction applied.")
    REGISTER_VARIABLE("pStarLNoFSR", pStarLNoFSR,
                      "Returns the pStarL variable with no fsr correction applied.")
    REGISTER_VARIABLE("fsrMomentumX", fsrMomentumX, "Returns the X component of the fsr 4-momentum.")
    REGISTER_VARIABLE("fsrMomentumY", fsrMomentumY, "Returns the Y component of the fsr 4-momentum.")
    REGISTER_VARIABLE("fsrMomentumZ", fsrMomentumZ, "Returns the Z component of the fsr 4-momentum.")
    REGISTER_VARIABLE("fsrE", fsrE, "Returns the E component of the fsr 4-momentum.")

  } // OwnVariables namespace
} // Belle2 namespace
