"""
Definition of Continuum Suppression MVA Training

Felix Metzner 2019
"""


import os
import numpy as np

from PyFastBDT import FastBDT
from typing import List, Dict, Union, Optional

from templatefitter.plotter.histogram_variable import HistVariable

from analysistools.dataframe_tools.hdf5_utils import load_hdf5

from rdstar.mva_methods.mva_methods import MVABaseClass, MVAVarInfo

__all__ = [
    "ContinuumSuppression",
    "get_cs_variable_dict",
    "get_cs_variable_infos",
    "cs_additional_variables",
    "run_cs_training",
    "evaluate_cs_training"
]


base_cs_variable_infos = [  # type: List[MVAVarInfo]
    MVAVarInfo(var="spR2__bc", bins=25, min=0., max=1., latex=r"$\mathrm{R}_2$"),
    MVAVarInfo(var="spcosTBTO__bc", bins=25, min=0., max=1., latex=r"$\cos\theta_\mathrm{TBTO}$"),
    MVAVarInfo(var="spcosTBz__bc", bins=25, min=0., max=1., latex=r"$\cos\theta_\mathrm{TBz}$"),
    MVAVarInfo(var="spthrustBm__bc", bins=25, min=0.5, max=1., latex=r"$\mathrm{T}_\mathrm{B}$"),
    MVAVarInfo(var="spthrustOm__bc", bins=25, min=0.5, max=1., latex=r"$\mathrm{T}_\mathrm{O}$"),
    MVAVarInfo(var="spKSFWVariables__boet__bc__bc", bins=25, min=2., max=12., latex=r"$\mathrm{KSFW}\;E_t$"),
    MVAVarInfo(var="spKSFWVariables__bomm2__bc__bc", bins=25, min=-10., max=10., latex=r"$\mathrm{KSFW}\;M_m^2$"),
    MVAVarInfo(var="spKSFWVariables__bohso00__bc__bc", bins=25, min=0, max=1.5, latex=r"$\mathrm{KSFW}\;H^{so}_{0,0}$"),
    # MVAVarInfo(var="spKSFWVariables__bohso01__bc__bc", bins=25, min=0, max=1.5, latex=r"$\mathrm{KSFW}\;H^{so}_{0,1}$"),  # TODO: Not considered atm, but should we include it?
    MVAVarInfo(var="spKSFWVariables__bohso02__bc__bc", bins=25, min=-0.6, max=1.1, latex=r"$\mathrm{KSFW}\;H^{so}_{0,2}$"),
    # MVAVarInfo(var="spKSFWVariables__bohso03__bc__bc", bins=25, min=0, max=1.5, latex=r"$\mathrm{KSFW}\;H^{so}_{0,3}$"),  # TODO: Not considered atm, but should we include it?
    MVAVarInfo(var="spKSFWVariables__bohso04__bc__bc", bins=25, min=-0.4, max=0.6, latex=r"$\mathrm{KSFW}\;H^{so}_{0,4}$"),
    MVAVarInfo(var="spKSFWVariables__bohso10__bc__bc", bins=25, min=0, max=1.25, latex=r"$\mathrm{KSFW}\;H^{so}_{1,0}$"),
    MVAVarInfo(var="spKSFWVariables__bohso12__bc__bc", bins=25, min=-0.2, max=0.55, latex=r"$\mathrm{KSFW}\;H^{so}_{1,2}$"),
    MVAVarInfo(var="spKSFWVariables__bohso14__bc__bc", bins=25, min=-0.2, max=0.3, latex=r"$\mathrm{KSFW}\;H^{so}_{1,4}$"),
    MVAVarInfo(var="spKSFWVariables__bohso20__bc__bc", bins=25, min=0, max=0.7, latex=r"$\mathrm{KSFW}\;H^{so}_{2,0}$"),
    MVAVarInfo(var="spKSFWVariables__bohso22__bc__bc", bins=25, min=-0.2, max=0.5, latex=r"$\mathrm{KSFW}\;H^{so}_{2,2}$"),
    MVAVarInfo(var="spKSFWVariables__bohso24__bc__bc", bins=25, min=-0.2, max=0.3, latex=r"$\mathrm{KSFW}\;H^{so}_{2,4}$"),
    MVAVarInfo(var="spKSFWVariables__bohoo0__bc__bc", bins=25, min=0.0, max=0.3, latex=r"$\mathrm{KSFW}\;H^{oo}_0$"),
    MVAVarInfo(var="spKSFWVariables__bohoo1__bc__bc", bins=25, min=-0.04, max=0.04, latex=r"$\mathrm{KSFW}\;H^{oo}_1$"),
    MVAVarInfo(var="spKSFWVariables__bohoo2__bc__bc", bins=25, min=-0.02, max=0.12, latex=r"$\mathrm{KSFW}\;H^{oo}_2$"),
    MVAVarInfo(var="spKSFWVariables__bohoo3__bc__bc", bins=25, min=-0.04, max=0.04, latex=r"$\mathrm{KSFW}\;H^{oo}_3$"),
    MVAVarInfo(var="spKSFWVariables__bohoo4__bc__bc", bins=25, min=-0.025, max=0.08, latex=r"$\mathrm{KSFW}\;H^{oo}_4$"),
    MVAVarInfo(var="spCleoCone__bo1__bc__bc", bins=30, min=0., max=5., latex=r"$\mathrm{CleoCone}(1)$"),
    MVAVarInfo(var="spCleoCone__bo2__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(2)$"),
    MVAVarInfo(var="spCleoCone__bo3__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(3)$"),
    MVAVarInfo(var="spCleoCone__bo4__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(4)$"),
    MVAVarInfo(var="spCleoCone__bo5__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(5)$"),
    MVAVarInfo(var="spCleoCone__bo6__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(6)$"),
    MVAVarInfo(var="spCleoCone__bo7__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(7)$"),
    MVAVarInfo(var="spCleoCone__bo8__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(8)$"),
    MVAVarInfo(var="spCleoCone__bo9__bc__bc", bins=30, min=0., max=6., latex=r"$\mathrm{CleoCone}(9)$"),
]


def get_cs_variable_dict(signal_side: bool = True) -> Dict[str, HistVariable]:
    if signal_side:
        b_id = "daughter__bo1__cm__"
        b_name = r"$\mathrm{B}_\mathrm{sig}$"
    else:
        b_id = "daughter__bo0__cm__"
        b_name = r"$\mathrm{B}_\mathrm{tag}$"
    separator = r"$\;$"

    return {
        b_id + var_info.var: HistVariable(
            df_label=b_id + var_info.var,
            n_bins=var_info.bins,
            scope=(var_info.min, var_info.max),
            var_name=b_name + separator + var_info.latex,
            unit=var_info.unit,
            use_log_scale=var_info.log
        )
        for var_info in base_cs_variable_infos
    }


def get_cs_variable_infos(signal_side: bool = True) -> List[MVAVarInfo]:
    if signal_side:
        b_id = "daughter__bo1__cm__"
        b_name = r"$\mathrm{B}_\mathrm{sig}$"
    else:
        b_id = "daughter__bo0__cm__"
        b_name = r"$\mathrm{B}_\mathrm{tag}$"
    separator = r"$\;$"

    return [
        MVAVarInfo(
            var=b_id + info.var,
            latex=b_name + separator + info.latex,
            bins=info.bins,
            min=info.min,
            max=info.max
        ) for info in base_cs_variable_infos
    ]


# List of additional variables needed for training process but not as input for the cs mva.
cs_additional_variables = {
    "isNotContinuumEvent",  # Target variable for the mva training.
    "m2RecoilSignalSide",  # Signal variable used to check correlation with input variables for the mva.
    "isSignalAcceptMissingNeutrino",  # Can be used to extract only correctly reconstructed signal events.
    "daughter__bo1__cm__spisSignalAcceptMissingNeutrino__bc",  # Can be used to extract corr. reconstructed sig. events.
    "daughter__bo0__cm__spisSignalAcceptMissingNeutrino__bc",  # Can be used to extract corr. reconstructed sig. events.
}


class ContinuumSuppression(MVABaseClass):
    def __init__(
            self,
            cs_variables: List[str],
            target_variable: str,
            weight: Optional[str] = None
    ) -> None:
        super().__init__(
            variables=cs_variables,
            target_variable=target_variable,
            weight=weight
        )


def run_cs_training(
        train_data_file_path: Union[str, os.fspath],
        variable_file_path: Union[str, os.fspath],
        weight_file_path: Union[str, os.fspath],
        target_variable: str = "isNotContinuumEvent",
        fast_bdt_settings: Optional[Dict[str, str]] = None,
) -> None:
    """
    Trains a FastBDT for Continuum Suppression using a provided, prepared data set,
    and storing the resulting weight files as well as the list of used variables.

    The truth target used for the supervised training process is specified via the
    parameter target_variable.
    The FastBDT settings can be set via the optional parameter fast_bdt_settings,
    which expects a dictionary mapping the setting names to their respective values.

    The variables used as input for the MVA are provided as pickled list loaded from
    a pickle file specified by variable_file_path.

    The resulting classifier weights and the variable list required to apply the
    CS MVA are stored to a path defined by the argument weight_file_path.

    :param train_data_file_path: Path to the training data sample hdf5 file.
    :param variable_file_path: Path to the pickle file containing the variable to be used.
    :param weight_file_path: Path under which the trained weights and variables will be stored.
    :param target_variable: The variable used as truth target in the training.
    :param fast_bdt_settings: Settings used for the FastBDT MVA.
    :return: None
    """
    final_vars = ContinuumSuppression.load_pickled_variables(variable_file_path)
    cs_train_data = load_hdf5(hdf5_file_path=train_data_file_path, hdf5_key="train_sample")

    cs_train_data.loc[:, "cs_weights"] = np.ones(len(cs_train_data.index))

    cs = ContinuumSuppression(
        cs_variables=final_vars,
        target_variable=target_variable,
        weight='cs_weights'
    )

    if fast_bdt_settings is None:
        fast_bdt_settings = {
            "nTrees": 200,
        }

    cs.add_classifier(classifier=FastBDT.Classifier(**fast_bdt_settings))
    cs.train(cs_train_data)

    cs.save_classifier_weights(weight_file_path=weight_file_path, overwrite=True)


def evaluate_cs_training(
        test_data_file_path: Union[str, os.fspath],
        weight_file_path: Union[str, os.fspath],
        output_dir_path: Union[str, os.fspath]
) -> float:
    """
    Evaluates a given trained Continuum Suppression classifier.

    :param test_data_file_path: Path to the hdf5 file containing the test data set.
    :param weight_file_path: Path to the weights of the CS MVA to be tested.
    :param output_dir_path: Path to the directory under which the output of the
                            evaluation is stored.
    :return: Area under the ROC curve.
    """
    cs = ContinuumSuppression.load_classifier(weight_file_path=weight_file_path)
    cs_test_data = load_hdf5(hdf5_file_path=test_data_file_path, hdf5_key="train_sample")

    roc_auc_value = cs.test_evaluation(cs_test_data, output_dir=output_dir_path)

    return roc_auc_value
