#!/usr/bin/python
from __future__ import absolute_import

import argparse
import subprocess
import os
import socket
import sys

script_dir = sys.path[0]

dec_file_dirs = {
    "Dlnu": "{0}/dec_files/b_to_dlnu".format(script_dir),
    "Dstlnu": "{0}/dec_files/b_to_dstlnu".format(script_dir),
    "DlnuMVA": "{0}/dec_files/b_to_dlnu".format(script_dir),
    "DstlnuMVA": "{0}/dec_files/b_to_dstlnu".format(script_dir),
    "Dststlnu": "{0}/dec_files/b_to_dststlnu".format(script_dir),
    "SignalDststlnupiz": "{0}/dec_files/b_to_dststlnu_piz_signal".format(script_dir),
    "SignalDststlnupim": "{0}/dec_files/b_to_dststlnu_pim_signal".format(script_dir),
    "SignalDststlnupizMVA": "{0}/dec_files/b_to_dststlnu_piz_signal".format(script_dir),
    "SignalDststlnupimMVA": "{0}/dec_files/b_to_dststlnu_pim_signal".format(script_dir),
    "Dststtaunu": "{0}/dec_files/b_to_dststtaunu".format(script_dir),
    "SignalDststtaunupiz": "{0}/dec_files/b_to_dststtaunu_piz_signal".format(script_dir),
    "SignalDststtaunupim": "{0}/dec_files/b_to_dststtaunu_pim_signal".format(script_dir),
    "GapD1": "{0}/dec_files/b_to_gap_d1".format(script_dir),
    "GapDEta": "{0}/dec_files/b_to_gap_d_eta".format(script_dir),
    "GapDPiPi": "{0}/dec_files/b_to_gap_d_pipi".format(script_dir),
    "GapDstPiPi": "{0}/dec_files/b_to_gap_dst_pipi".format(script_dir),
    "GapDEtaViaDstst": "{0}/dec_files/b_to_gap_d_eta_via_dstst".format(script_dir),
    "GapDPiPiViaDstst": "{0}/dec_files/b_to_gap_d_pipi_via_dstst".format(script_dir),
    "GapDstPiPiViaDstst": "{0}/dec_files/b_to_gap_dst_pipi_via_dstst".format(script_dir),
}

def check_host():
    """
    Checks if scripts are executed on the KEKCC. If not, the
    execution of the script is terminated.
    """
    if "cc.kek.jp" not in socket.gethostname():
        sys.exit("This script can only be executed on kekcc. Terminating!")
    if "BELLE_TOP_DIR" not in os.environ:
        sys.exit("You must source basf to run this script! Check the README if you don't know what to do...")
    if "BELLE2_LOCAL_DIR" in os.environ:
        sys.exit("If you want to run this script, basf2 must NOT be set up!")

def get_dec_types(dec_file_dir):
    """
    Creates a list of decay types from the .dec files in the
    given directory by removing the file extension.

    :param dec_file_dir: Path to the .dec files.
    :type dec_file_dir: str

    :return: List of decay types.
    """
    return [os.path.splitext(file)[0] for file in os.listdir(dec_file_dir)]


def parse_command_line_args():
    """
    Parses command line arguments.

    :return: Command line arguments.
    """
    parser = argparse.ArgumentParser(description="This script starts the event simulation with GSim for the given"
                                                 " Decay type.")

    parser.add_argument("-t", action="store", dest="dec_type", type=str, required=True,
                        help="Decay type: Currently available is 'Dlnu', 'Dstlnu', 'Dststlnu'.")
    parser.add_argument("-i", action="store", dest="in_path", type=str, required=True,
                        help="Absolute path to the generated events. If the events are stored in "
                             "/some/path/evtgen/gen/DECAYS, you have to provide /some/path.")
    parser.add_argument("-s", action="store", dest="sub_dec_type", default=None, required=False,
                        help="Allows to specify a subset of the decay type, meaning one of the decay files "
                             "associated to the decay type.")

    parser.add_argument("--clean", action="store_true", required=False,
                        help="Tidy up the gsim directory after all jobs ran successfully.")

    args = parser.parse_args()

    if not str(os.path.abspath(args.in_path)).endswith(args.dec_type):
        raise ValueError("Argument in_path\n\t" + args.in_path + "\nshould end in with dec_type " + args.dec_type)

    return args


class EvtSimulationManager:

    def __init__(self, dec_file_dir, in_path, sub_decay_type):
        self._dec_types = get_dec_types(dec_file_dir)

        if not (sub_decay_type in self._dec_types or sub_decay_type is None):
            raise ValueError(
                "The optional argument sub_decay_type must be one of\n\t-"
                + "\n\t-".join(self._dec_types)
                + "\nbut sub_decay_type = " + sub_decay_type + " was provided!"
            )
        self._sub_decay_type = sub_decay_type

        self._path_to_gen = os.path.join(in_path, "evtgen", "gen")
        self._output_path = str(os.path.abspath(in_path)) + "/"

    def simulate(self):

        os.chdir("{0}/gsim".format(sys.path[0]))

        for dec_type in self._dec_types:
            if self._sub_decay_type is not None and self._sub_decay_type != dec_type:
                continue

            abs_path = str(os.path.join(self._path_to_gen, dec_type)) + "/"
            print "\n\n"
            print "\tExecuting detector simulation for decay {0}".format(dec_type)
            print "\tInput from {0}".format(abs_path)
            print "\tOutput will be put in {0}/gsim/mdst/{1}/".format(self._output_path, dec_type)
            print "\n\n"


            subprocess.check_call([
                "./runGsimReco.csh",
                self._output_path,
                abs_path,
                dec_type
            ])


if __name__ == "__main__":
    raise RuntimeError("You want to run the mc_mix_simulation.py script, or!?")
    check_host()

    args = parse_command_line_args()

    if args.clean:
        os.chdir("{0}/gsim".format(sys.path[0]))
        subprocess.check_call(["./clean.csh"], shell=True)
    else:
        sub_decay_info = "(all)" if args.sub_dec_type is None else "({0})".format(args.sub_dec_type)
        print "Starting detector simulation for decay type: {0} {1}".format(args.dec_type, sub_decay_info)
        dec_file_dir_path = dec_file_dirs[args.dec_type]
        gsim_manager = EvtSimulationManager(
            dec_file_dir=dec_file_dir_path,
            in_path=args.in_path,
            sub_decay_type=args.sub_dec_type
        )
        gsim_manager.simulate()
