import logging
import numpy as np
import pandas as pd
from uncertainties import ufloat

from eFFORT.utility import PDG
from eFFORT.SLBToC.BToDLNu import BToDLNuCLN, BToDLNuBGL, BToDLNu

from rdstar.mc_reweighting.utility import FFReweightingBase, get_varied_central_values
from rdstar.offline_analysis.mc_ids import gen_b_ids

logging.getLogger(__name__).addHandler(logging.NullHandler())


class BtoDFFReweighting(FFReweightingBase):
    """
    Applies form factor reweighting for B->Dstlv decays for a given
    DataFrame.
    """
    weight_name = "weight_ff_btod_"
    nom_weight_name = f"{weight_name}nom"
    base_var_weight_name = f"{weight_name}var_"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.weight_name = BtoDFFReweighting.weight_name

        self.B1_is_btodl = None
        self.B2_is_btodl = None

        self.vcb_cln = 41.1e-3
        # self.vcb_bgl = 40.83e-3

        self.B1_reweighting_var = self.b1_w
        self.B2_reweighting_var = self.b2_w

        self.params = np.array([
            ufloat(4.11837e-02, 1.15178e-03),
            ufloat(1.26147e-02, 9.78879e-05),  # bgl fplus coeff 0
            ufloat(-9.62084e-02, 3.34148e-03),  # bgl fplus coeff 1
            ufloat(4.13884e-01, 9.44471e-02),  # bgl fplus coeff 2
            ufloat(-1.73699e-01, 8.91809e-01),  # bgl fplus coeff 3
        ])
        self.param_central_values = np.array([cv.nominal_value for cv in self.params])
        self.param_uncertainties = np.array([cv.std_dev for cv in self.params])
        self.param_correlation_matrix = np.array([
            [1.000, -0.402, -0.238, -0.110, 0.047],
            [-0.402, 1.000, 0.245, -0.161, 0.020],
            [-0.238, 0.245, 1.000, -0.654, 0.272],
            [-0.110, -0.161, -0.654, 1.000, -0.770],
            [0.047, 0.020, 0.272, -0.770, 1.000],
        ])
        self.covariance_matrix = np.outer(
            self.param_uncertainties, self.param_uncertainties) * self.param_correlation_matrix

        self.varied_central_values = get_varied_central_values(
            self.param_central_values,
            self.covariance_matrix
        )

        self._btod_cln = BToDLNuCLN(PDG.m_Bplus, PDG.m_Dzero, self.vcb_cln)
        self._btod_bgl_nom = BToDLNuBGL(
            PDG.m_Bplus, PDG.m_Dzero, self.param_central_values[0],
            bgl_fplus_coefficients=self.param_central_values[1:]
        )
        self._btod_bgl_var = list()

        for variation in self.varied_central_values:
            btod_bgl_var_up = BToDLNuBGL(
                PDG.m_Bplus, PDG.m_Dzero, variation[0][0], bgl_fplus_coefficients=variation[0][1:]
            )
            btod_bgl_var_down = BToDLNuBGL(
                PDG.m_Bplus, PDG.m_Dzero, variation[1][0], bgl_fplus_coefficients=variation[1][1:]
            )

            self._btod_bgl_var.append((btod_bgl_var_up, btod_bgl_var_down))

    def _add_nominal_weights(self, df: pd.DataFrame) -> None:
        self.nominal_weight_key = f"{self.weight_name}nom"
        df.loc[:, self.nominal_weight_key] = np.nan

        b_conditions = [self.B1_is_btodl, self.B2_is_btodl]
        b_variables = [self.B1_reweighting_var, self.B2_reweighting_var]

        for b_condition, b_variable in zip(b_conditions, b_variables):
            df.loc[b_condition, self.nominal_weight_key] = self._calculate_weights(
                old=self._btod_cln,
                new=self._btod_bgl_nom,
                w=df.loc[b_condition, b_variable].values,
            )

        df.loc[df[self.nominal_weight_key].isnull(), self.nominal_weight_key] = 1.0
        logging.info("Nominal B->Dlnu FF weights added.")

    def _add_varied_weights(self, df: pd.DataFrame) -> None:
        self.variation_weight_keys = list()

        b_conditions = [self.B1_is_btodl, self.B2_is_btodl]
        b_variables = [self.B1_reweighting_var, self.B2_reweighting_var]

        for i, variation in enumerate(self._btod_bgl_var):
            up = variation[0]
            down = variation[1]

            up_key = f"{self.weight_name}var_{i}_up"
            down_key = f"{self.weight_name}var_{i}_down"

            self.variation_weight_keys.append(up_key)
            self.variation_weight_keys.append(down_key)

            for var_key, new_var in zip([up_key, down_key], [up, down]):
                df.loc[:, var_key] = np.nan
                for b_condition, b_variable in zip(b_conditions, b_variables):
                    df.loc[b_condition, var_key] = self._calculate_weights(
                        old=self._btod_cln,
                        new=new_var,
                        w=df.loc[b_condition, b_variable].values,
                    )
                df.loc[df[var_key].isnull(), var_key] = 1.0

            logging.info(f"Variation {i} B->Dlnu FF weights added.")

    @staticmethod
    def _calculate_weights(old: BToDLNu, new: BToDLNu, w: np.ndarray) -> np.ndarray:
        old_d_gamma = old.dGamma_dw(w)
        new_d_gamma = new.dGamma_dw(w)

        return (old.Gamma() / new.Gamma()) * (new_d_gamma / old_d_gamma)

    def add_weights(self, df: pd.DataFrame) -> None:
        btodlnu_gen_ids = [gen_b_ids["Dtau"]["id"], gen_b_ids["Dl"]["id"]]

        self.B1_is_btodl = (
                (df.GEN_B1_ID.isin(btodlnu_gen_ids))
                & ((df[self.b_index_var] == 1) | (df[self.b_via_tau_index_var] == 1))
        )
        self.B2_is_btodl = (
                (df.GEN_B2_ID.isin(btodlnu_gen_ids))
                & ((df[self.b_index_var] == 2) | (df[self.b_via_tau_index_var] == 2))
        )

        logging.info(f"Number of candidates to reweight: {len(self.B1_is_btodl.index) + len(self.B2_is_btodl.index)}")

        self._add_nominal_weights(df)
        self._add_varied_weights(df)
        self._write_weight_info(df, conditions=(self.B1_is_btodl, self.B2_is_btodl))

    @staticmethod
    def finalize_weights(
            df: pd.DataFrame,
            weight_base_name: str = "weight_ff_btod_",
            threshold: float = 3.,
            reset_special_to: float = 0.
    ) -> None:
        nom_weight_col = f"{weight_base_name}nom"
        assert nom_weight_col in df.columns

        wcs = [c for c in df.columns if c.startswith(weight_base_name)]

        btodlnu_gen_ids = [gen_b_ids["Dtau"]["id"], gen_b_ids["Dl"]["id"]]

        BtoDFFReweighting.reset_problematic_special_mc(
            df=df,
            weight_base_name=weight_base_name,
            gen_ids=btodlnu_gen_ids,
            reset_special_to=reset_special_to
        )

        assert not pd.isnull(df[wcs]).any().any()
        assert (df.loc[:, nom_weight_col] >= 0.).all()
        assert (df.loc[:, nom_weight_col] < threshold).all()
