import logging
import numpy as np
import pandas as pd

from uncertainties import ufloat
from multiprocessing.pool import Pool
from typing import Optional, Dict, Tuple

from eFFORT.utility import PDG
from eFFORT.SLBToC.BToDstarLNu import BToDstarLNuBGL, BToDstarLNuCLN, BToDstarLNu

from rdstar.mc_reweighting.utility import FFReweightingBase, get_varied_central_values
from rdstar.offline_analysis.mc_ids import gen_b_ids

logging.getLogger(__name__).addHandler(logging.NullHandler())


class BtoDstarFFReweighting(FFReweightingBase):
    """
    Applies form factor reweighting for B->Dstlv decays for a given
    DataFrame.
    """
    weight_name = "weight_ff_btodst_"
    nom_weight_name = f"{weight_name}nom"
    base_var_weight_name = f"{weight_name}var_"

    daughter2_pdg_info_col = "daughter2_pdg"

    def __init__(self, num_cpu: int = 1, cached_gammas: Optional[Dict[str, Dict[int, float]]] = None, **kwargs) -> None:
        super().__init__(**kwargs)

        self.weight_name = BtoDstarFFReweighting.weight_name

        self.num_cpu = num_cpu
        self.cached_gammas = cached_gammas

        self.b1_info_vars.update({BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(0): self.daughter2_pdg_info_col})
        self.b2_info_vars.update({BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(1): self.daughter2_pdg_info_col})

        self.B1_is_btodstl = None
        self.B2_is_btodstl = None

        self.vcb_cln = 37.4e-3
        self.vcb_bgl = 41.6558e-3

        self.params = np.array([
            ufloat(3.79139e-04, 2.49014e-04),  # a0
            ufloat(2.69537e-02, 9.31972e-03),  # a1
            ufloat(5.49846e-04, 2.34728e-05),  # b0
            ufloat(-2.04028e-03, 1.06439e-03),  # b1
            ufloat(-4.32818e-04, 2.63747e-04),  # c0
            ufloat(5.35029e-03, 4.60575e-03),  # c1
        ])
        self.param_central_values = np.array([cv.nominal_value for cv in self.params])
        self.param_uncertainties = np.array([cv.std_dev for cv in self.params])
        self.param_correlation_matrix = np.array([
            [1.000, -0.952, -0.249, 0.417, 0.137, -0.054],
            [-0.952, 1.000, 0.383, -0.543, -0.268, 0.165],
            [-0.249, 0.383, 1.000, -0.793, -0.648, 0.461],
            [0.417, -0.543, -0.793, 1.000, 0.542, -0.333],
            [0.137, -0.268, -0.648, 0.542, 1.000, -0.953],
            [-0.054, 0.165, 0.461, -0.333, -0.953, 1.000],
        ])

        self.B1_reweighting_vars = {
            "w": self.b1_w,
            "cos_l": self.b1_costhetal,
            "cos_v": self.b1_costhetav,
            "chi": self.b1_chi
        }

        self.B2_reweighting_vars = {
            "w": self.b2_w,
            "cos_l": self.b2_costhetal,
            "cos_v": self.b2_costhetav,
            "chi": self.b2_chi
        }

        self.covariance_matrix = np.outer(self.param_uncertainties,
                                          self.param_uncertainties) * self.param_correlation_matrix

        self.varied_central_values = get_varied_central_values(
            central=self.param_central_values,
            covariance_matrix=self.covariance_matrix
        )

        self._btodst_cln = BToDstarLNuCLN(
            PDG.m_Bplus, PDG.m_Dstarzero, self.vcb_cln,
            cached_gammas=self.cached_gammas["cln"] if self.cached_gammas is not None else None
        )
        logging.debug("Initialized CLN.")

        self._btodst_bgl_nom = BToDstarLNuBGL(
            PDG.m_Bplus, PDG.m_Dstarzero, self.vcb_bgl,
            cached_gammas=self.cached_gammas["bgl_nom"] if self.cached_gammas is not None else None
        )
        logging.debug("Initialized nominal BGL.")

        with Pool(processes=min(self.num_cpu, len(self.varied_central_values))) as pool:
            self._btodst_bgl_var = pool.map(self._init_variation_ff_classes, range(len(self.varied_central_values)))

    @staticmethod
    def get_d_daughter_mc_pdg_var(b_index: int) -> str:
        assert b_index in [0, 1]
        v = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmmcDaughter__bo1__cmPDG__bc__bc__bc__bc"
        return v

    def _init_variation_ff_classes(self, variation: int) -> Tuple[BToDstarLNuBGL, BToDstarLNuBGL]:
        btodst_bgl_var_up = BToDstarLNuBGL(
            PDG.m_Bplus, PDG.m_Dstarzero, self.vcb_bgl,
            cached_gammas=self.cached_gammas[f"bgl_var_up{variation}"] if self.cached_gammas is not None else None,
            exp_coeff=self.varied_central_values[variation][0]
        )
        btodst_bgl_var_down = BToDstarLNuBGL(
            PDG.m_Bplus, PDG.m_Dstarzero, self.vcb_bgl,
            cached_gammas=self.cached_gammas[f"bgl_var_down{variation}"] if self.cached_gammas is not None else None,
            exp_coeff=self.varied_central_values[variation][1]
        )
        logging.debug("Initialized BGL variation.")

        return btodst_bgl_var_up, btodst_bgl_var_down

    def get_gamma_integrals(self) -> Dict[str, Dict[int, float]]:
        gamma_integrals = dict()
        gamma_integrals["cln"] = self._btodst_cln.get_gammas()
        gamma_integrals["bgl_nom"] = self._btodst_bgl_nom.get_gammas()

        for i, (up_var, down_var) in enumerate(self._btodst_bgl_var):
            gamma_integrals[f"bgl_var_up{i}"] = up_var.get_gammas()
            gamma_integrals[f"bgl_var_down{i}"] = down_var.get_gammas()

        for val in gamma_integrals.values():
            BToDstarLNuBGL.check_precomputed_gammas_dict(val)

        return gamma_integrals

    def _add_nominal_weights(self, df: pd.DataFrame) -> None:
        self.nominal_weight_key = f"{self.weight_name}nom"
        df.loc[:, self.nominal_weight_key] = np.nan

        b_conditions = [self.B1_is_btodstl, self.B2_is_btodstl]
        variables_for_bs = [self.B1_reweighting_vars, self.B2_reweighting_vars]

        for b_index, condition, variables_for_b in zip([0, 1], b_conditions, variables_for_bs):
            df.loc[condition, self.nominal_weight_key] = self._calculate_weights(
                old=self._btodst_cln,
                new=self._btodst_bgl_nom,
                w=df.loc[condition, variables_for_b["w"]].values,
                costhetal=df.loc[condition, variables_for_b["cos_l"]].values,
                costhetav=df.loc[condition, variables_for_b["cos_v"]].values,
                chiprime=df.loc[condition, variables_for_b["chi"]].values,
                pdg=df.loc[condition, self.get_d_daughter_mc_pdg_var(b_index=b_index)].values
            )

        df.loc[df[self.nominal_weight_key].isnull(), self.nominal_weight_key] = 1.0
        logging.info("Nominal B->Dstlnu FF weights added.")

    def _add_varied_weights(self, df: pd.DataFrame) -> None:
        self.variation_weight_keys = list()

        b_conditions = [self.B1_is_btodstl, self.B2_is_btodstl]
        variables_for_bs = [self.B1_reweighting_vars, self.B2_reweighting_vars]

        for i, variation in enumerate(self._btodst_bgl_var):
            up = variation[0]
            down = variation[1]

            up_key = f"{self.weight_name}var_{i}_up"
            down_key = f"{self.weight_name}var_{i}_down"

            self.variation_weight_keys.append(up_key)
            self.variation_weight_keys.append(down_key)

            for var_key, var_info in zip([up_key, down_key], [up, down]):
                df.loc[:, var_key] = np.nan
                for b_index, condition, variables_for_b in zip([0, 1], b_conditions, variables_for_bs):
                    df.loc[condition, var_key] = self._calculate_weights(
                        old=self._btodst_cln,
                        new=var_info,
                        w=df.loc[condition, variables_for_b["w"]].values,
                        costhetal=df.loc[condition, variables_for_b["cos_l"]].values,
                        costhetav=df.loc[condition, variables_for_b["cos_v"]].values,
                        chiprime=df.loc[condition, variables_for_b["chi"]].values,
                        pdg=df.loc[condition, self.get_d_daughter_mc_pdg_var(b_index=b_index)].values
                    )
                df.loc[df[var_key].isnull(), var_key] = 1.0

            logging.info(f"Variation {i} B->Dstlnu FF weights added.")

    @staticmethod
    def _calculate_weights(
            old: BToDstarLNu,
            new: BToDstarLNu,
            w: np.ndarray,
            costhetal: np.ndarray,
            costhetav: np.ndarray,
            chiprime: np.ndarray,
            pdg: np.ndarray
    ) -> np.ndarray:
        old_d_gamma = old.dGamma_dw_dcosL_dcosV_dChi(w, costhetal, costhetav, chiprime, pdg)
        new_d_gamma = new.dGamma_dw_dcosL_dcosV_dChi(w, costhetal, costhetav, chiprime, pdg)
        return (old.Gamma(pdg) / new.Gamma(pdg)) * (new_d_gamma / old_d_gamma)

    def add_weights(self, df: pd.DataFrame) -> None:
        btodstlnu_gen_ids = [gen_b_ids["Dstl"]["id"]]

        self.B1_is_btodstl = (
                (df.GEN_B1_ID.isin(btodstlnu_gen_ids))
                & ((df[self.b_index_var] == 1) | (df[self.b_via_tau_index_var] == 1))
        )
        self.B2_is_btodstl = (
                (df.GEN_B2_ID.isin(btodstlnu_gen_ids))
                & ((df[self.b_index_var] == 2) | (df[self.b_via_tau_index_var] == 2))
        )

        logging.info(
            f"Number of candidates to reweight: "
            f"{len(self.B1_is_btodstl.index) + len(self.B2_is_btodstl.index)}"
        )

        self._add_nominal_weights(df)
        self._add_varied_weights(df)
        self._write_weight_info(df, conditions=(self.B1_is_btodstl, self.B2_is_btodstl))

    @staticmethod
    def finalize_weights(
            df: pd.DataFrame,
            weight_base_name: str = "weight_ff_btodst_",
            threshold: float = 5.,
            set_to: float = 1.0,
            reset_special_to: float = 0.,
            combo_name: Optional[str] = None
    ) -> None:
        combo_info = ""
        if combo_name is not None:
            combo_info = f"events in {combo_name} combo "

        nom_weight_col = f"{weight_base_name}nom"
        assert nom_weight_col in df.columns

        wcs = [c for c in df.columns if c.startswith(weight_base_name)]

        btodstlnu_gen_ids = [gen_b_ids["Dstl"]["id"]]

        BtoDstarFFReweighting.reset_problematic_special_mc(
            df=df,
            weight_base_name=weight_base_name,
            gen_ids=btodstlnu_gen_ids,
            reset_special_to=reset_special_to
        )

        update_mask = df[nom_weight_col] > threshold
        if update_mask.any():
            for col in [c for c in wcs if c.endswith("up") or c.endswith("down")]:
                if set_to == 0.:
                    df.loc[update_mask, col] = set_to
                else:
                    df.loc[update_mask, col] = df.loc[update_mask, col] / df.loc[update_mask, nom_weight_col]
            df.loc[update_mask, nom_weight_col] = set_to

            print(f"Resetting the {weight_base_name} FF weights for {len(update_mask.index)} {combo_info}"
                  f"to {set_to}, because their weight value was above {threshold}")

        assert not pd.isnull(df[wcs]).any().any()
        assert (df.loc[:, nom_weight_col] >= 0.).all()
        assert (df.loc[:, nom_weight_col] < threshold).all()


if __name__ == "__main__":
    import logging

    bToDstar_BGL = BToDstarLNuBGL(PDG.m_Bplus, PDG.m_Dstarzero, V_cb=41.6558e-3)
    btodstar_cln = BToDstarLNuCLN(PDG.m_Bplus, PDG.m_Dstarzero, V_cb=41.6558e-3)

    w_min = 1
    w_max = (bToDstar_BGL.m_B ** 2 + bToDstar_BGL.m_Dstar ** 2) / (2 * bToDstar_BGL.m_B * bToDstar_BGL.m_Dstar)
    n = 10
    w_range = np.linspace(w_min + 1e-7, w_max - 1e-7, n, endpoint=True)
    cosl_range = np.linspace(-1, 1, n, endpoint=True)
    cosnu_range = np.linspace(-1, 1, n, endpoint=True)
    chi_range = np.linspace(0, 2 * np.pi, n, endpoint=True)
    pdg_codes = np.random.choice([22, 111, 211], len(w_range))

    test_df = pd.DataFrame({
        "GEN_B1_ID": np.full(n, 4),
        "GEN_B2_ID": np.full(n, 5),
        "SIG_ID": np.full(n, 4),
        "genUpsilon4S__bomcDaughter__bo0__cm__spwMC__bc__bc": w_range,
        "genUpsilon4S__bomcDaughter__bo0__cm__spcosthetalMC__bc__bc": cosl_range,
        "genUpsilon4S__bomcDaughter__bo0__cm__spcosthetavMC__bc__bc": cosnu_range,
        "genUpsilon4S__bomcDaughter__bo0__cm__spchi_primeMC__bc__bc": chi_range,
        BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(0): pdg_codes,
        "genUpsilon4S__bomcDaughter__bo1__cm__spwMC__bc__bc": w_range,
        "genUpsilon4S__bomcDaughter__bo1__cm__spcosthetalMC__bc__bc": cosl_range,
        "genUpsilon4S__bomcDaughter__bo1__cm__spcosthetavMC__bc__bc": cosnu_range,
        "genUpsilon4S__bomcDaughter__bo1__cm__spchi_primeMC__bc__bc": chi_range,
        BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(1): pdg_codes,
    })

    dst_rw = BtoDstarFFReweighting()
    dst_rw.add_weights(test_df)
    print(test_df)
