"""
Analysis of the weights generated for the reweighting of the B -> D** l nu samples from
the ISGW2 model to the LLSW model.

Felix Metzner, Maximilian Welsch (2018 - 2020)
"""

import os
import re
import abc
import b2luigi
import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from multiprocessing import pool
from typing import Union, AnyStr, Tuple, List, Dict, Generator

from analysistools.reweighting.rbf_reweighter import RbfReweighter

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.plot_utilities import export, AxesType, FigureType
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot, DataMCHistogramPlot
from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors, xlabel_pos, ylabel_pos

from rdstar.offline_analysis.rdstar_plots import dss_mc_titles
from rdstar.mc_reweighting.BtoDststlnu.event_generation import DststReweightingEventGenerator
from rdstar.mc_reweighting.BtoDststlnu.b_to_dstst_weights_generation_tasks import RbfCreationTask, \
    ReweightingDataMergerTask, MergeIntoCommonDataFrameTask

__all__ = [
    "DststWeightEvaluationWrapperTask"
]

logging.getLogger(__name__).addHandler(logging.NullHandler())

set_matplotlibrc_params()


class DststWeightCalculationTask(b2luigi.Task):
    result_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    log_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    decay_channel = b2luigi.Parameter()  # type: str
    n_events = b2luigi.IntParameter(positional=False)  # type: int

    data_input_key = "data_keys"  # type: str
    rbf_input_key = "rbfs_info_files"  # type: str

    hdf_key = "with_weights"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

        self.base_output_dir = os.path.join(self.result_dir, "isgw2_data")

    def get_isgw2_variations(self) -> List[str]:
        return list(RbfCreationTask.get_model_variations(model_key="ISGW2", decay_channel=self.decay_channel))

    def load_rbf_dict(self) -> Dict[str, RbfReweighter]:
        return {
            key: RbfReweighter.import_from_json(file_path=target.path)
            for key, target in self.input()[self.rbf_input_key].items()
        }

    def get_output_file_path(self, temp: bool = False) -> Union[os.PathLike, AnyStr]:
        prefix = "temp_" if temp else ""
        return os.path.join(self.base_output_dir, f"{prefix}{self.decay_channel}_with_weights.h5")

    def load_isgw2_sample(self) -> pd.DataFrame:
        input_key = ReweightingDataMergerTask.get_df_output_key(decay_channel=self.decay_channel, model="ISGW2")
        file_path = self.input()[self.data_input_key][input_key].path
        df = pd.read_hdf(path_or_buf=file_path, key=ReweightingDataMergerTask.df_key)
        assert isinstance(df, pd.DataFrame), type(df)
        return df

    @staticmethod
    def apply_rbf(dstst_type: str, variation: str, rbf: RbfReweighter, df: pd.DataFrame) -> Tuple[str, np.ndarray]:
        logging.info(f"Calculating weights for {dstst_type}_{variation}")
        return variation, rbf(*[df[var].values for var in rbf.variables])

    def run(self) -> None:
        os.makedirs(name=self.base_output_dir, exist_ok=True)

        isgw2_sample = self.load_isgw2_sample()
        rbf_dict = self.load_rbf_dict()

        args_list = [(self.decay_channel, key, rbf, isgw2_sample[list(rbf.variables)]) for key, rbf in rbf_dict.items()]
        weights_result_tuples = pool.Pool(processes=4).starmap(
            func=DststWeightCalculationTask.apply_rbf,
            iterable=args_list
        )

        for var, weight in weights_result_tuples:
            isgw2_sample.loc[:, var] = weight

        isgw2_sample.to_hdf(path_or_buf=self.get_output_file_path(temp=True), key=self.hdf_key)
        os.rename(src=self.get_output_file_path(temp=True), dst=self.get_output_file_path(temp=False))

    def output(self) -> b2luigi.LocalTarget:
        return b2luigi.LocalTarget(self.get_output_file_path(temp=False))

    def requires(self) -> Dict[str, b2luigi.Task]:
        return {
            self.data_input_key: self.clone(ReweightingDataMergerTask, model_variations=self.get_isgw2_variations()),
            self.rbf_input_key: self.clone(RbfCreationTask)
        }


@b2luigi.inherits(DststWeightCalculationTask)
class DststWeightAnalysisPlotBaseTask(b2luigi.Task):
    base_dir_name = "weight_analysis"  # type: str
    plot_dir_name = "plots"  # type: str

    w = HistVariable(df_label=RbfCreationTask.w_col_name, n_bins=20, scope=(1., 1.45), var_name="w", unit="GeV")
    costhetal = HistVariable(df_label=RbfCreationTask.costhetal_col_name, n_bins=20, scope=(-1., 1.),
                             var_name=r"$\cos (\theta_\ell)$", unit="")
    costhetav = HistVariable(df_label=RbfCreationTask.costhetanu_col_name, n_bins=20, scope=(-1., 1.),
                             var_name=r"$\cos (\theta_\nu)$", unit="")
    chi = HistVariable(df_label=RbfCreationTask.chi_col_name, n_bins=20, scope=(0, 2 * np.pi), var_name=r"$\chi$",
                       unit="")

    # Additional variables for evaluation:
    cols_dict = DststReweightingEventGenerator.get_predefined_root_compatible_variables_dict()  # type: Dict[str, str]
    chi_prime = HistVariable(df_label=cols_dict["chi_primeMC"], n_bins=20, scope=(0, 2 * np.pi), var_name=r"$\chi '$",
                             unit="")
    p_star_l = HistVariable(df_label=cols_dict["p_lepton_star"], n_bins=26, scope=(0, 2.6),
                            var_name=r"$\mathrm{p}_\ell^\ast$", unit="GeV")

    reweighting_vars = [w, costhetal, costhetav, chi, chi_prime, p_star_l]
    vars_with_legend = [w, chi_prime, p_star_l]

    llsw_input_key = "llsw_input"  # type: str
    isgw2_input_key = "isgw2_input"  # type: str
    isgw2_ww_input_key = "isgw2_with_weights_input"  # type: str

    dss_mc_plot_titles = dss_mc_titles  # type: Dict[str, str]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

        assert self.decay_channel in self.dss_mc_plot_titles, (self.decay_channel, self.dss_mc_plot_titles.keys())

        self.base_output_dir = os.path.join(self.result_dir, self.base_dir_name, self.plot_dir_name)

        self.colors = [
            KITColors.kit_blue,
            KITColors.kit_cyan,
            KITColors.kit_maygreen,
            KITColors.kit_green,
            KITColors.kit_red,
            KITColors.kit_purple,
            KITColors.kit_orange,
            KITColors.kit_yellow,
            KITColors.grey
        ]

    @staticmethod
    def create_hist_ratio_figure() -> Tuple[FigureType, Tuple[AxesType, AxesType]]:
        return DataMCHistogramPlot.create_hist_ratio_figure(fig_size=(5, 5), height_ratio=(3.5, 1))

    @staticmethod
    def create_single_figure() -> Tuple[FigureType, AxesType]:
        return plt.subplots(1, 1, figsize=(5, 5), dpi=200)

    @staticmethod
    def map_variable(var: str) -> str:
        var_mapping = DststReweightingEventGenerator.get_inverted_predefined_root_compatible_variables_dict()
        assert var in var_mapping, var
        return var_mapping[var]

    @staticmethod
    def get_diff_vec(h_nom: np.ndarray, h_up: np.ndarray, h_down: np.ndarray) -> np.ndarray:
        """
        Calculates the difference between a nominal histogram and its up and down variations.
        If the down variation has a higher value than the up variation, a relative minus sign
        is added to the difference. This is used for the construction of a covariance matrix
        as outer product of the difference vector, as is sets the correct sign for negatively
        correlated bins.
        """
        sign = np.ones_like(h_up)
        mask = h_up < h_down
        sign[mask] = -1

        diff_up = np.abs(h_up - h_nom)
        diff_down = np.abs(h_down - h_nom)
        diff_sym = (diff_up + diff_down) / 2
        return sign * diff_sym

    def get_llsw_variations(self, remove_model_name: bool = False) -> List[str]:
        variation_keys = list(RbfCreationTask.get_model_variations(model_key="LLSW", decay_channel=self.decay_channel))
        if remove_model_name:
            variation_keys = [k.replace("LLSW_", "").replace("variation_", "") for k in variation_keys]
        return variation_keys

    def load_llsw_data(self) -> Dict[str, pd.DataFrame]:
        input_key = ReweightingDataMergerTask.get_df_output_key(decay_channel=self.decay_channel, model="LLSW")
        path_to_hdf = self.input()[self.llsw_input_key][input_key].path
        data = pd.read_hdf(path_or_buf=path_to_hdf, key=ReweightingDataMergerTask.df_key)
        assert isinstance(data, pd.DataFrame), type(pd.DataFrame)
        variations = list(data[MergeIntoCommonDataFrameTask.variation_col_name].unique())

        return {v: data.query(f"{MergeIntoCommonDataFrameTask.variation_col_name}=='{v}'") for v in variations}

    def get_isgw2_variations(self) -> List[str]:
        return list(RbfCreationTask.get_model_variations(model_key="ISGW2", decay_channel=self.decay_channel))

    def load_isgw2_sample(self) -> pd.DataFrame:
        i_key = ReweightingDataMergerTask.get_df_output_key(decay_channel=self.decay_channel, model="ISGW2")
        path_to_file = self.input()[self.isgw2_input_key][i_key].path
        df = pd.read_hdf(path_or_buf=path_to_file, key=ReweightingDataMergerTask.df_key)
        assert isinstance(df, pd.DataFrame), type(df)
        return df

    def load_isgw2_with_weights_sample(self) -> pd.DataFrame:
        file_path = self.input()[self.isgw2_ww_input_key].path
        df = pd.read_hdf(path_or_buf=file_path, key=DststWeightCalculationTask.hdf_key)
        assert isinstance(df, pd.DataFrame), type(df)
        return df

    def get_isgw2_weights(self, isgw2_sample: pd.DataFrame) -> Dict[str, np.ndarray]:
        variation_col_names = self.get_llsw_variations(remove_model_name=True)
        assert all(var_col_name in isgw2_sample.columns for var_col_name in variation_col_names), \
            [var_col_name for var_col_name in variation_col_names if var_col_name not in isgw2_sample.columns]
        return {variation: isgw2_sample[variation].values for variation in variation_col_names}

    def append_to_output_dict(self, output_dict: Dict[str, Union[AnyStr, os.PathLike]], file_name: str) -> None:
        assert file_name not in output_dict.keys(), file_name
        png_output_key = f"{file_name}_png"
        assert png_output_key not in output_dict.keys(), png_output_key
        output_path = os.path.join(self.base_output_dir, f"{file_name}.pdf")
        output_dict.update({file_name: output_path})
        png_output_path = os.path.join(self.base_output_dir, f"{file_name}.png")
        output_dict.update({png_output_key: png_output_path})

    def run(self) -> None:
        os.makedirs(name=self.base_output_dir, exist_ok=True)
        self.create_plots(dry_run=False)

    def output(self) -> Dict[str, b2luigi.LocalTarget]:
        return {k: b2luigi.LocalTarget(v) for k, v in self.create_plots(dry_run=True).items()}

    @abc.abstractmethod
    def create_plots(self, dry_run: bool = False) -> Dict[str, Union[AnyStr, os.PathLike]]:
        raise NotImplementedError("This base task class should not be used directly!")

    @abc.abstractmethod
    def requires(self) -> Dict[str, b2luigi.Task]:
        raise NotImplementedError("This base task class should not be used directly!")


@b2luigi.inherits(DststWeightCalculationTask)
class LLSWComparisonPlots(DststWeightAnalysisPlotBaseTask):
    plot_dir_name = "llsw_variation_comparison"  # type: str
    plot_name = "llsw_variation_comparison"  # type: str

    def plot_llsw_variations(self, hist_var: HistVariable, sample_dict: Dict[str, pd.DataFrame], ax: AxesType) -> None:
        shp = SimpleHistogramPlot(variable=hist_var)
        for variation, color in zip(sample_dict.keys(), self.colors):
            component_label = self._get_label(variation=variation)
            shp.add_component(label=component_label, data=sample_dict[variation], color=color, hist_type="step")
        shp.plot_on(ax=ax, normed=True)

        ax.legend(frameon=True, loc="best", ncol=3)
        ax.set_title(label="LLSW", loc="left")
        ax.set_title(label=self.dss_mc_plot_titles[self.decay_channel], loc="right")

    @staticmethod
    def _get_label(variation: str) -> str:
        return " ".join(variation.capitalize().split("_"))

    def create_plots(self, dry_run: bool = False) -> Dict[str, Union[AnyStr, os.PathLike]]:
        llsw_sample_dict = None
        if not dry_run:
            llsw_sample_dict = self.load_llsw_data()

        output_dict = {}
        for hist_var in self.reweighting_vars:
            file_name = f"{self.plot_name}_{self.decay_channel}_{self.map_variable(var=hist_var.df_label)}"
            self.append_to_output_dict(output_dict=output_dict, file_name=file_name)

            if not dry_run:
                fig, ax = self.create_single_figure()
                self.plot_llsw_variations(hist_var=hist_var, sample_dict=llsw_sample_dict, ax=ax)

                export(fig=fig, filename=file_name, target_dir=self.base_output_dir, close_figure=True)

        return output_dict

    def requires(self) -> Dict[str, b2luigi.Task]:
        return {self.llsw_input_key: self.clone(ReweightingDataMergerTask, model_variations=self.get_llsw_variations())}


@b2luigi.inherits(DststWeightCalculationTask)
class PreReweightingComparison(DststWeightAnalysisPlotBaseTask):
    plot_dir_name = "pre_reweighting_comparison"  # type: str
    plot_name = "pre_reweighting_comparison"  # type: str

    def compare_isgw2_and_llsw(
            self,
            hist_var: HistVariable,
            isgw2_df: pd.DataFrame,
            llsw_df_dict: Dict[str, pd.DataFrame],
            axs: Tuple[AxesType, AxesType]
    ) -> None:
        ax1, ax2 = axs
        hw = SimpleHistogramPlot(hist_var)
        hw.add_component("LLSW nominal", llsw_df_dict["nominal"], color=KITColors.kit_red, hist_type="step")
        hw.add_component("ISGW2", isgw2_df, color=KITColors.kit_blue, hist_type="step")
        hw.plot_on(ax1, normed=True, draw_legend=False)

        ax1.set_xlabel("")
        ax2.set_xlabel(xlabel=hist_var.x_label, **xlabel_pos)

        llsw_nom, _ = np.histogram(llsw_df_dict[f"nominal"][hist_var.df_label], bins=hw.bin_edges, density=True)
        errors_sq = np.zeros_like(llsw_nom)

        for i in range(int((len(llsw_df_dict.keys()) - 1) / 2)):
            llsw_up, _ = np.histogram(llsw_df_dict[f"up_{i}"][hist_var.df_label], bins=hw.bin_edges, density=True)
            llsw_down, _ = np.histogram(llsw_df_dict[f"down_{i}"][hist_var.df_label], bins=hw.bin_edges, density=True)
            errors_sq += self.get_diff_vec(h_nom=llsw_nom, h_up=llsw_up, h_down=llsw_down) ** 2

        errors = np.sqrt(errors_sq)

        ax1.bar(x=hw.bin_mids, height=2 * errors, width=hw.minimal_bin_width, bottom=llsw_nom - errors,
                label="sys. uncertainty", color=KITColors.kit_red, alpha=0.7, lw=0, fill=True)

        if hist_var in self.vars_with_legend:
            ax1.legend(frameon=False, loc="upper left", ncol=2)
            y_limits = ax1.get_ylim()
            ax1.set_ylim(y_limits[0], 1.25 * y_limits[1])

        isgw2_bc, _ = np.histogram(isgw2_df[hist_var.df_label], bins=hw.bin_edges, density=True)

        ratio = (llsw_nom - isgw2_bc) / llsw_nom
        ratio_error = isgw2_bc / llsw_nom ** 2 * errors
        ax2.axhline(y=0, color=KITColors.dark_grey, alpha=0.8)
        ax2.errorbar(hw.bin_mids, ratio, yerr=ratio_error, ls="", marker=".", color=KITColors.kit_black)
        ax2.set_ylabel(ylabel=r"$\frac{\mathrm{LLSW - ISGW2}}{\mathrm{LLSW}}$", **ylabel_pos)
        ax2.set_ylim(bottom=-1, top=1)

        ax1.set_title(label=self.dss_mc_plot_titles[self.decay_channel], loc="right")
        ax1.set_title(label="Original ISGW2 vs. LLSW", loc="left")

    def create_plots(self, dry_run: bool = False) -> Dict[str, Union[AnyStr, os.PathLike]]:
        llsw_sample, isgw2_sample = None, None
        if not dry_run:
            llsw_sample = self.load_llsw_data()
            isgw2_sample = self.load_isgw2_sample()

        output_dict = {}
        for hist_var in self.reweighting_vars:
            file_name = f"{self.plot_name}_{self.decay_channel}_{self.map_variable(var=hist_var.df_label)}"
            self.append_to_output_dict(output_dict=output_dict, file_name=file_name)

            if not dry_run:
                fig, axs = self.create_hist_ratio_figure()
                plt.subplots_adjust(hspace=0.1)
                self.compare_isgw2_and_llsw(hist_var=hist_var, isgw2_df=isgw2_sample, llsw_df_dict=llsw_sample, axs=axs)

                export(fig=fig, filename=file_name, target_dir=self.base_output_dir, close_figure=True)

        return output_dict

    def requires(self) -> Dict[str, b2luigi.Task]:
        return {
            self.isgw2_input_key: self.clone(ReweightingDataMergerTask, model_variations=self.get_isgw2_variations()),
            self.llsw_input_key: self.clone(ReweightingDataMergerTask, model_variations=self.get_llsw_variations()),
        }


@b2luigi.inherits(DststWeightCalculationTask)
class PlotWeights(DststWeightAnalysisPlotBaseTask):
    plot_dir_name = "weights_plots"  # type: str
    plot_name = "isgw2_weights"  # type: str

    def plot_weights(self, weights: np.ndarray, variation: str, ax: AxesType) -> None:
        color, ecolor = KITColors.kit_blue, KITColors.kit_black
        bc, be, _ = ax.hist(x=weights, bins=30, color=color, edgecolor=ecolor, histtype="stepfilled")
        ax.set_yscale("log")
        bw = be[1] - be[0]
        mean = float(np.mean(weights))
        median = float(np.median(weights))

        if variation == "nominal":
            ax.set_xlabel(xlabel="Nominal Weight", **xlabel_pos)
        elif variation.startswith("down"):
            num = int(re.findall(r"\d+", variation)[0])
            ax.set_xlabel(xlabel=f"Weight Variation Down {num}", **xlabel_pos)
        elif variation.startswith("up"):
            num = int(re.findall(r"\d+", variation)[0])
            ax.set_xlabel(xlabel=f"Weight Variation Up {num}", **xlabel_pos)

        ax.axvline(mean, label=f"Mean: {mean:.2f}", color=KITColors.kit_orange)
        ax.axvline(median, label=f"Median: {median:.2f}", color=KITColors.kit_maygreen, ls="--")
        ax.set_ylabel(ylabel=f"Events / ({bw:.3f})", **ylabel_pos)

        ax.set_title(label=self.dss_mc_plot_titles[self.decay_channel], loc="right")
        ax.set_title(label=r"ISGW2 $\rightarrow$ LLSW Weights", loc="left")
        ax.legend(frameon=False)

    def create_plots(self, dry_run: bool = False) -> Dict[str, Union[AnyStr, os.PathLike]]:
        isgw2_sample = None
        if not dry_run:
            isgw2_sample = self.load_isgw2_with_weights_sample()

        output_dict = {}
        for variation, var_col in zip(self.get_llsw_variations(), self.get_llsw_variations(remove_model_name=True)):
            file_name = f"{self.plot_name}_{self.decay_channel}_{variation}"
            self.append_to_output_dict(output_dict=output_dict, file_name=file_name)

            if not dry_run:
                fig, ax = self.create_single_figure()
                self.plot_weights(weights=isgw2_sample[var_col].values, variation=var_col, ax=ax)

                export(fig=fig, filename=file_name, target_dir=self.base_output_dir, close_figure=True)

        return output_dict

    def requires(self) -> Dict[str, b2luigi.Task]:
        return {self.isgw2_ww_input_key: self.clone(DststWeightCalculationTask)}


@b2luigi.inherits(DststWeightCalculationTask)
class PlotReweightedVsLLSW(DststWeightAnalysisPlotBaseTask):
    plot_dir_name = "reweighted_vs_llsw"  # type: str
    plot_name = "reweighted_vs_llsw"  # type: str

    def compare_reweighed_and_llsw(
            self,
            hist_var: HistVariable,
            isgw2_df: pd.DataFrame,
            isgw2_weights: Dict[str, np.ndarray],
            llsw_df: Dict[str, pd.DataFrame],
            axs: Tuple[AxesType, AxesType]
    ) -> None:
        ax1, ax2 = axs
        column = hist_var.df_label
        hw = SimpleHistogramPlot(variable=hist_var)

        c1, c2 = KITColors.kit_red, KITColors.kit_green
        hw.add_component(label="LLSW nominal", data=llsw_df["nominal"], color=c1, hist_type="step")
        hw.add_component(label="ISGW2 reweighted\nto LLSW", data=isgw2_df, weights=isgw2_weights["nominal"], color=c2,
                         hist_type="step")
        hw.plot_on(ax=ax1, normed=False, draw_legend=False)

        ax1.set_xlabel("")
        ax2.set_xlabel(xlabel=hist_var.x_label, **xlabel_pos)

        llsw_nom, _ = np.histogram(llsw_df[f"nominal"][column], bins=hw.bin_edges)
        llsw_errors_sq = np.zeros_like(llsw_nom, dtype=float)

        for i in range(int((len(llsw_df.keys()) - 1) / 2)):
            llsw_up, _ = np.histogram(llsw_df[f"up_{i}"][column], bins=hw.bin_edges)
            llsw_down, _ = np.histogram(llsw_df[f"down_{i}"][column], bins=hw.bin_edges)
            variation_error_sq = self.get_diff_vec(h_nom=llsw_nom, h_up=llsw_up, h_down=llsw_down) ** 2
            llsw_errors_sq += variation_error_sq

        llsw_errors = np.sqrt(llsw_errors_sq)

        isgw2_nom, _ = np.histogram(isgw2_df[column], weights=isgw2_weights["nominal"], bins=hw.bin_edges)
        isgw2_errors_sq = np.zeros_like(isgw2_nom)

        for i in range(int((len(llsw_df.keys()) - 1) / 2)):
            isgw2_up, _ = np.histogram(isgw2_df[column], weights=isgw2_weights[f"up_{i}"], bins=hw.bin_edges)
            isgw2_dw, _ = np.histogram(isgw2_df[column], weights=isgw2_weights[f"down_{i}"], bins=hw.bin_edges)
            isgw2_errors_sq += self.get_diff_vec(h_nom=isgw2_nom, h_up=isgw2_up, h_down=isgw2_dw) ** 2
        isgw2_errors = np.sqrt(isgw2_errors_sq)

        ax1.bar(x=hw.bin_mids, height=2 * llsw_errors, width=hw.minimal_bin_width, label="sys. uncertainty",
                bottom=llsw_nom - llsw_errors, color=c1, alpha=0.6, lw=0, fill=True)
        ax1.bar(x=hw.bin_mids, height=2 * isgw2_errors, width=hw.minimal_bin_width, label="sys. uncertainty",
                bottom=isgw2_nom - isgw2_errors, color=c2, alpha=0.6, lw=0, fill=True)

        if hist_var in self.vars_with_legend:
            ax1.legend(frameon=False, loc="upper left", ncol=2)
            y_lims = ax1.get_ylim()
            ax1.set_ylim(y_lims[0], 1.5 * y_lims[1])

        ratio = (llsw_nom - isgw2_nom) / llsw_nom
        # TODO: Fix this warning:
        #       /local/scratch/ssd1/fmetzner/RDStar/RDStar/rdstar/mc_reweighting/BtoDststlnu/
        #       b_to_dstst_weights_analysis_tasks.py:467: RuntimeWarning: invalid value encountered in multiply
        #       ratio_error = np.sqrt((isgw2_nom / llsw_nom ** 2) ** 2 * llsw_errors_sq + (1 / llsw_nom) ** 2
        #       * isgw2_errors_sq)

        ratio_error = np.sqrt((isgw2_nom / llsw_nom ** 2) ** 2 * llsw_errors_sq + (1 / llsw_nom) ** 2 * isgw2_errors_sq)
        ax2.axhline(y=0, color=KITColors.dark_grey, alpha=0.8)
        ax2.errorbar(hw.bin_mids, ratio, yerr=ratio_error, ls="", marker=".", color=KITColors.kit_black)
        ax2.set_ylabel(ylabel=r"$\frac{\mathrm{LLSW - ISGW2}}{\mathrm{LLSW}}$", **ylabel_pos)
        ax2.set_ylim(bottom=-1., top=1.)

        ax1.set_title(label=self.dss_mc_plot_titles[self.decay_channel], loc="right")
        # ax1.set_title(label="Reweighted ISGW2 vs. LLSW", loc="left")

    def create_plots(self, dry_run: bool = False) -> Dict[str, Union[AnyStr, os.PathLike]]:
        llsw_sample, isgw2_sample, isgw2_weights = None, None, None
        if not dry_run:
            llsw_sample = self.load_llsw_data()
            isgw2_sample = self.load_isgw2_with_weights_sample()
            isgw2_weights = self.get_isgw2_weights(isgw2_sample=isgw2_sample)

        output_dict = {}
        for hist_var in self.reweighting_vars:
            file_name = f"{self.plot_name}_{self.decay_channel}_{self.map_variable(var=hist_var.df_label)}"
            self.append_to_output_dict(output_dict=output_dict, file_name=file_name)

            if not dry_run:
                fig, axs = self.create_hist_ratio_figure()
                plt.subplots_adjust(hspace=0.1)
                self.compare_reweighed_and_llsw(
                    hist_var=hist_var,
                    isgw2_df=isgw2_sample,
                    isgw2_weights=isgw2_weights,
                    llsw_df=llsw_sample,
                    axs=axs
                )

                export(fig=fig, filename=file_name, target_dir=self.base_output_dir, close_figure=True)

        return output_dict

    def requires(self) -> Dict[str, b2luigi.Task]:
        return {
            self.isgw2_ww_input_key: self.clone(DststWeightCalculationTask),
            self.llsw_input_key: self.clone(ReweightingDataMergerTask, model_variations=self.get_llsw_variations())
        }


@b2luigi.inherits(DststWeightCalculationTask)
class PlotReweightedVsISGW2(DststWeightAnalysisPlotBaseTask):
    plot_dir_name = "reweighted_vs_isgw2"  # type: str
    plot_name = "reweighted_vs_isgw2"  # type: str

    def create_comparison_plot(
            self,
            hist_var: HistVariable,
            isgw2_df: pd.DataFrame,
            isgw2_weights: Dict[str, np.ndarray],
            ax1: AxesType
    ) -> None:
        column = hist_var.df_label
        c1, c2 = KITColors.kit_green, KITColors.kit_blue

        hw = SimpleHistogramPlot(variable=hist_var)
        hw.add_component(label="ISGW2 reweighted\nto LLSW", data=isgw2_df, weights=isgw2_weights["nominal"], color=c1,
                         hist_type="step")
        hw.add_component(label="ISGW2", data=isgw2_df, color=c2, hist_type="step")
        hw.plot_on(ax=ax1, normed=False, draw_legend=False)

        ax1.set_xlabel(xlabel=hist_var.x_label, **xlabel_pos)

        isgw2_nom, _ = np.histogram(isgw2_df[column], weights=isgw2_weights["nominal"], bins=hw.bin_edges)
        isgw2_errors_sq = np.zeros_like(isgw2_nom)

        for i in range(int((len(isgw2_weights.keys()) - 1) / 2)):
            isgw2_up, _ = np.histogram(isgw2_df[column], weights=isgw2_weights[f"up_{i}"], bins=hw.bin_edges)
            isgw2_dw, _ = np.histogram(isgw2_df[column], weights=isgw2_weights[f"down_{i}"], bins=hw.bin_edges)
            isgw2_errors_sq += self.get_diff_vec(h_nom=isgw2_nom, h_up=isgw2_up, h_down=isgw2_dw) ** 2
        isgw2_errors = np.sqrt(isgw2_errors_sq)

        ax1.bar(x=hw.bin_mids, height=2 * isgw2_errors, width=hw.minimal_bin_width, bottom=isgw2_nom - isgw2_errors,
                label="sys. uncertainty", color=c1, alpha=0.6, lw=0, fill=True)

        if hist_var in self.vars_with_legend:
            ax1.legend(frameon=False, loc="upper left", ncol=2)
            ylims = ax1.get_ylim()
            ax1.set_ylim(ylims[0], 1.25 * ylims[1])

        # isgw2_orig, _ = np.histogram(isgw2_df[column], bins=hw.bin_edges)
        # isgw2_orig_errors_sq = isgw2_orig

        # ratio = (isgw2_orig - isgw2_nom) / isgw2_orig
        # ratio_error = np.sqrt(
        #     (isgw2_nom / isgw2_orig ** 2) ** 2 * isgw2_orig_errors_sq + (1 / isgw2_orig) ** 2 * isgw2_errors_sq)
        # ax2.axhline(y=0, color=KITColors.dark_grey, alpha=0.8)
        # ax2.errorbar(hw.bin_mids, ratio, yerr=ratio_error, ls="", marker=".", color=KITColors.kit_black)
        # ax2.set_ylabel(r"$\frac{\mathrm{ISGW2 - LLSW}}{\mathrm{ISGW2}}$")
        # ax2.set_ylim((-1, 1))

        ax1.set_title(label=self.dss_mc_plot_titles[self.decay_channel], loc="right")
        # ax1.set_title(label="Reweighted vs. Original ISGW2", loc="left")

    def create_plots(self, dry_run: bool = False) -> Dict[str, Union[AnyStr, os.PathLike]]:
        isgw2_df, isgw2_weights = None, None
        if not dry_run:
            isgw2_df = self.load_isgw2_with_weights_sample()
            isgw2_weights = self.get_isgw2_weights(isgw2_sample=isgw2_df)

        output_dict = {}
        for hist_var in self.reweighting_vars:
            file_name = f"{self.plot_name}_{self.decay_channel}_{self.map_variable(var=hist_var.df_label)}"
            self.append_to_output_dict(output_dict=output_dict, file_name=file_name)

            if not dry_run:
                fig, ax = self.create_single_figure()
                self.create_comparison_plot(hist_var=hist_var, isgw2_df=isgw2_df, isgw2_weights=isgw2_weights, ax1=ax)

                export(fig=fig, filename=file_name, target_dir=self.base_output_dir, close_figure=True)

        return output_dict

    def requires(self) -> Dict[str, b2luigi.Task]:
        return {self.isgw2_ww_input_key: self.clone(DststWeightCalculationTask)}


@b2luigi.inherits(DststWeightCalculationTask)
class DststWeightEvaluationWrapperTask(b2luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

    def requires(self) -> Generator[b2luigi.Task, None, None]:
        yield self.clone(PlotReweightedVsISGW2)
        yield self.clone(PlotReweightedVsLLSW)
        yield self.clone(PlotWeights)
        yield self.clone(PreReweightingComparison)
        yield self.clone(LLSWComparisonPlots)
