"""
This file calculates 1 sigma variations of the form factor parameters of the
LLSW_nominal model for D^*{**} decays and generates new decay files accordingly.

Maximilian Welsch 2018
"""
import os
import copy
import logging
import itertools
import numpy as np

from typing import Union, Optional, Tuple, List, Dict, AnyStr

logging.basicConfig(format="%(asctime)s %(message)s", level=logging.INFO)

__all__ = [
    "DecayFileWithLLSWVariationsCreator"
]


class ParameterVariation:
    """
    This class performs the parameter variation for the form factor
    parameters of the LLSW_nominal model.
    """

    def __init__(
            self,
            nominal_values: np.array,
            cov_mat: Optional[np.array] = None
    ) -> None:
        """
        Constructor.
        :param nominal_values: Nominal parameter values. Shape (n,).
        :param cov_mat: Parameter covariance matrix. Shape (n, n). Optional. Default is None.
        """
        self._nom_vals = nominal_values  # type: np.ndarray
        self._n_params = nominal_values.shape[0]  # type: int
        self._cov_mat = cov_mat  # type: np.ndarray
        self._stds = None  # type: Optional[np.ndarray]
        self._corr_mat = None  # type: Optional[np.ndarray]
        self._eig_vals = None  # type: Optional[np.ndarray]
        self._eig_vecs = None  # type: Optional[np.ndarray]
        self._variation = None  # type: Optional[np.ndarray]

    def calculate_covariance_matrix(self, stds: np.array, corr_mat: np.array) -> None:
        """
        If no covariance matrix was given in the constructor, the covariance
        matrix can be calculated with this function using the correlation
        matrix and the standard deviations of the parameters.
        :param stds: Parameter standard deviations. Shape (n,).
        :param corr_mat: Parameter correlation matrix. Shape (n, n).
        """
        self._stds = stds

        self._corr_mat = corr_mat
        self._cov_mat = np.matmul(np.diag(stds), np.matmul(corr_mat, np.diag(stds)))

    def _calc_eigvals_and_eigvecs(self) -> None:
        """
        Calculates eigenvalues and vectors of the covariance matrix
        """
        if self._cov_mat is None:
            raise ValueError("Covariance matrix is not initialized yet.")

        self._eig_vals, self._eig_vecs = np.linalg.eig(self._cov_mat)

    def _calc_variation(self) -> None:
        """
        Calculates the 1 sigma variation for each parameter.
        """
        if self._eig_vecs is None or self._eig_vals is None:
            self._calc_eigvals_and_eigvecs()

        # Note that the columns of this array correspond to the variation of the parameters.
        self._variation = np.matmul(self._eig_vecs, np.sqrt(np.diag(self._eig_vals)))

    def calculate_variation(self) -> None:
        """
        Helper function if variations should be calculated manually.
        """
        self._calc_variation()

    def calculate_up_variation(self) -> np.ndarray:
        """
        Calculates 1 sigma upwards variation of the parameters.
        :return: Upwards variation.
        """
        if self._variation is None:
            self._calc_variation()

        nom_val = np.repeat(
            np.reshape(self._nom_vals, (self._n_params, 1)),
            repeats=self._n_params,
            axis=1
        )
        return np.around(nom_val + self._variation, decimals=2)

    def get_up_variation_as_list(self) -> List[List[float]]:
        """
        :return: Up variations as list.
        """
        up_variation = self.calculate_up_variation()
        up_variation_list = up_variation.T.tolist()
        return up_variation_list

    def calculate_down_variation(self) -> np.ndarray:
        """
        Calculates 1 sigma downwards variation of the parameters.
        :return: Downwards variation.
        """
        if self._variation is None:
            self._calc_variation()

        nom_val = np.repeat(
            np.reshape(self._nom_vals, (self._n_params, 1)),
            repeats=self._n_params,
            axis=1
        )
        return np.around(nom_val - self._variation, decimals=2)

    def get_down_variation_as_list(self) -> List[List[float]]:
        """
        :return: Down variations as list.
        """
        down_variation = self.calculate_down_variation()
        down_variation_list = down_variation.T.tolist()
        return down_variation_list

    def get_variation_info_dict(self) -> Dict[str, List[List[float]]]:
        return {
            "up": self.get_up_variation_as_list(),
            "down": self.get_down_variation_as_list()
        }

    def print_summary(self) -> None:
        print(50 * "-")
        print(f"Nominal Values:\n{self._nom_vals}\n")
        print(f"Standard Deviations:\n{self._stds}\n")
        print(f"Correlation Matrix:\n{self._corr_mat}\n")
        print(f"Covariance Matrix:\n{self._cov_mat}\n")
        print(f"Eigen Values:\n{self._eig_vals}\n")
        print(f"Eigen Vectors:\n{self._eig_vecs}\n")
        print(f"1 Sigma Variation up:\n{self.calculate_up_variation()}\n")
        print(f"1 Sigma Variation down:\n{self.calculate_down_variation()}\n")


class DecFileManipulator:
    """
    This class creates new decay file with LLSW_nominal parameters exchanged for the D^{**} decay.
    """

    def __init__(
            self,
            original_file: Union[AnyStr, os.PathLike],
            keyword: str
    ) -> None:
        self._keyword = keyword
        self._original_file = original_file

        with open(original_file, "r") as of:
            self._original_lines = tuple([line for line in of])  # type: Tuple[str]

        self._new_lines = None  # type: Optional[List[str]]

    def replace_parameters(self, params: List[float]) -> None:
        n_params = len(params)
        self._new_lines = list()
        for original_line in self._original_lines:
            line = copy.copy(original_line)
            if self._keyword in line:
                line_wo_params = self._cut_off_old_params(line=line, n_params=n_params)
                line_w_new_params = self._add_new_params(line=line_wo_params, params=params)
                self._new_lines.append(line_w_new_params)
            else:
                self._new_lines.append(line)

    def _cut_off_old_params(self, line: str, n_params: int) -> str:
        """
        Cuts off the parameters in the D^{**} decay command.
        :param str line: Decay line with old parameters.
        :param int n_params: Number of parameters.
        :return str: Decay line w/o parameters
        """
        split_line = line.split(self._keyword)
        assert len(split_line) == 2, (len(split_line), split_line)
        line_wo_parameters = split_line[0] + self._keyword  # type: str
        old_parameters = split_line[1]  # type: str
        assert old_parameters.endswith(";\n"), (old_parameters, old_parameters[-1])
        old_parameters = old_parameters.replace(";\n", "").strip()  # type: str
        assert len(old_parameters.split(" ")) == n_params, (old_parameters.split(" "), n_params)

        return line_wo_parameters

    @staticmethod
    def _add_new_params(line: str, params: List[float]) -> str:
        """
        Adds the new parameters at the end of the line.
        :param str line: Old line w/o parameters.
        :param List[float] params: New parameters.
        :return str: New decay line.
        """
        param_string = " ".join([str(param) for param in params])
        # add end of command character for EvtGen decay files
        param_string += ";\n"
        return " ".join([line, param_string])

    def print_old_file(self) -> None:
        """
        Prints the old decay file.
        """
        for line in self._original_lines:
            print(line)

    def print_new_file(self) -> None:
        """
        Prints the new decay file
        """
        for line in self._new_lines:
            print(line)

    def _compare_to_old_file_contents(self, file_path: Union[AnyStr, os.PathLike]) -> None:
        with open(file=file_path, mode="r") as existing_file:
            existing_lines = [line for line in existing_file]
            assert len(self._new_lines) == len(existing_lines), (len(self._new_lines), len(existing_lines))
            if not all(l1 == l2 for l1, l2 in zip(self._new_lines, existing_lines)):
                logging.error("New file will not contain same info as existing file!")
                for line_number, (l1, l2) in enumerate(zip(self._new_lines, existing_lines)):
                    if l1 != l2:
                        logging.info(f"line {line_number} of new file: {l1}")
                        logging.info(f"line {line_number} of old file: {l2}")
                raise RuntimeError("File contents do not match!")

    @staticmethod
    def get_result_file_name(original_file: Union[AnyStr, os.PathLike], name_extension: str) -> str:
        file_name, file_extension = os.path.splitext(os.path.basename(original_file))
        assert file_name.endswith("_nominal"), file_name
        file_name, _ = file_name.rsplit("_nominal", 1)
        assert not file_name.endswith("_nominal"), file_name
        return f"{file_name}_{name_extension}{file_extension}"

    def save_as(
            self,
            name_extension: str,
            new_dir: Union[AnyStr, os.PathLike],
            force_overwrite: bool = False
    ) -> None:
        new_file_name = self.get_result_file_name(original_file=self._original_file, name_extension=name_extension)
        temp_file_path = os.path.join(new_dir, f"temp_{new_file_name}")
        new_file_path = os.path.join(new_dir, new_file_name)
        if os.path.exists(new_file_path):
            if force_overwrite:
                os.remove(path=new_file_path)
            else:
                self._compare_to_old_file_contents(file_path=new_file_path)
                return

        with open(file=temp_file_path, mode="w") as new_file:
            for line in self._new_lines:
                new_file.write(line)

        os.rename(src=temp_file_path, dst=new_file_path)


class DecayFileWithLLSWVariationsCreator:
    narrow_key = "narrow"  # type: str
    broad_key = "broad"  # type: str
    model_name = "LLSW"  # type: str

    decay_files_dir_name = "decay_files"  # type: str
    decay_files_dir_path = os.path.join(os.path.split(os.path.realpath(__file__))[0], decay_files_dir_name)

    d_meson_types = {
        narrow_key: ("Done", "Dtwostar"),
        broad_key: ("Dprimeone", "Dzerostar")
    }  # type: Dict[str, Tuple[str, ...]]

    dimensions = {
        narrow_key: 4,
        broad_key: 3
    }  # type: Dict[str, int]

    def __init__(self, decay_file_directory: Union[AnyStr, os.PathLike]) -> None:
        assert os.path.exists(decay_file_directory), decay_file_directory
        self.decay_file_dir = decay_file_directory  # type: Union[AnyStr, os.PathLike]

        self.nominal = {
            self.narrow_key: np.array([0.71, -1.6, -0.5, 2.9]),
            self.broad_key: np.array([0.68, -0.2, 0.3])
        }  # type: Dict[str, np.ndarray]

        self.standard_deviation = {
            self.narrow_key: np.array([0.07, 0.2, 0.3, 1.6]),
            self.broad_key: np.array([0.2, 1.2, 0.3])
        }  # type: Dict[str, np.ndarray]

        self.correlation_matrix = {
            self.narrow_key: np.array([[1, -0.83, 0.66, -0.63],
                                       [-0.83, 1, -0.27, 0.20],
                                       [0.66, -0.27, 1, -0.93],
                                       [-0.63, 0.20, -0.93, 1]]),
            self.broad_key: np.array([[1, -0.95, -0.35],
                                      [-0.95, 1, 0.51],
                                      [-0.35, 0.51, 1]])
        }  # type: Dict[str, np.ndarray]

        for key in self.dimensions.keys():
            assert len(self.nominal[key].shape) == 1 and self.nominal[key].shape[0] == self.dimensions[key]
            assert len(self.standard_deviation[key].shape) == 1
            assert self.standard_deviation[key].shape[0] == self.dimensions[key]
            assert len(self.correlation_matrix[key].shape) == 2
            assert self.correlation_matrix[key].shape[0] == self.dimensions[key]
            assert self.correlation_matrix[key].shape[1] == self.dimensions[key]

    def get_base_llsw_decay_files_for_meson(self, d_meson_name: str) -> List[Union[AnyStr, os.PathLike]]:
        assert d_meson_name in [m_name for m_names in self.d_meson_types.values() for m_name in m_names]
        decay_file_dir_path = os.path.join(self.decay_file_dir, f"{self.model_name}_nominal", d_meson_name)
        paths = [os.path.join(decay_file_dir_path, decay_file) for decay_file in os.listdir(decay_file_dir_path)]
        assert all(os.path.isfile(path) for path in paths), [p for p in paths if not os.path.exists(p)]
        return paths

    def get_base_llsw_decay_file_dict(self) -> Dict[str, List[Union[AnyStr, os.PathLike]]]:
        path_dict = {}
        for d_meson_name in [m_name for m_names in self.d_meson_types.values() for m_name in m_names]:
            assert d_meson_name not in path_dict.keys(), (d_meson_name, path_dict.keys())
            path_dict.update({d_meson_name: self.get_base_llsw_decay_files_for_meson(d_meson_name=d_meson_name)})
        return path_dict

    @staticmethod
    def get_number_of_variations_for_meson(d_meson_name: str) -> int:
        d_meson_types_dict = DecayFileWithLLSWVariationsCreator.d_meson_types
        assert d_meson_name in [m_name for m_names in d_meson_types_dict.values() for m_name in m_names]
        assert sum([1 if d_meson_name in d_mesons else 0 for d_mesons in d_meson_types_dict.values()]) == 1, \
            d_meson_types_dict

        for key in d_meson_types_dict.keys():
            if d_meson_name in d_meson_types_dict[key]:
                return DecayFileWithLLSWVariationsCreator.dimensions[key]

    @staticmethod
    def get_variation_str(var_index: int, direction: str) -> str:
        assert direction in ("up", "down"), direction
        return f"variation_{direction}_{var_index}"

    @staticmethod
    def get_llsw_variation_name(var_index: int, direction: str) -> str:
        var_str = DecayFileWithLLSWVariationsCreator.get_variation_str(var_index=var_index, direction=direction)
        return f"LLSW_{var_str}"

    def create_llsw_variations(self, force_overwrite: bool = False) -> None:
        logging.info("Start calculation.")

        for d_meson_type in self.d_meson_types.keys():
            logging.info(f"Processing type {d_meson_type}")
            variation = ParameterVariation(nominal_values=self.nominal[d_meson_type])
            variation.calculate_covariance_matrix(
                stds=self.standard_deviation[d_meson_type],
                corr_mat=self.correlation_matrix[d_meson_type]
            )
            variation.calculate_variation()
            variation.print_summary()
            variations_info_dict = variation.get_variation_info_dict()
            assert all(key in ("up", "down") and len(value) == self.dimensions[d_meson_type]
                       for key, value in variations_info_dict.items())

            for variation_type, variation_list in variations_info_dict.items():
                for i, parameter_variation in enumerate(variation_list):
                    logging.info(f"Variation {variation_type} {i}: {parameter_variation}")
                    d_meson_names = self.d_meson_types[d_meson_type]
                    for d_meson_name in d_meson_names:
                        logging.info(f"Processing Channel: {d_meson_name}")

                        for decay_file in self.get_base_llsw_decay_files_for_meson(d_meson_name=d_meson_name):
                            variation_name = self.get_llsw_variation_name(var_index=i, direction=variation_type)
                            variation_dir = os.path.join(self.decay_file_dir, variation_name, d_meson_name)
                            if not os.path.exists(variation_dir):
                                os.makedirs(variation_dir)

                            decay_file_manipulator = DecFileManipulator(
                                original_file=decay_file,
                                keyword=self.model_name
                            )
                            decay_file_manipulator.replace_parameters(params=parameter_variation)
                            decay_file_manipulator.save_as(
                                name_extension=self.get_variation_str(var_index=i, direction=variation_type),
                                new_dir=variation_dir,
                                force_overwrite=force_overwrite
                            )

    def get_future_output(self) -> Dict[str, Union[AnyStr, os.PathLike]]:
        output_dict = {}
        for d_meson_name, decay_files in self.get_base_llsw_decay_file_dict().items():
            for decay_file in decay_files:
                number_of_variations = self.get_number_of_variations_for_meson(d_meson_name=d_meson_name)
                for var_index, var_type in itertools.product(range(number_of_variations), ("up", "down")):
                    name_extension = self.get_variation_str(var_index=var_index, direction=var_type)
                    f = DecFileManipulator.get_result_file_name(original_file=decay_file, name_extension=name_extension)
                    variation_name = self.get_llsw_variation_name(var_index=var_index, direction=var_type)
                    output_key = f"{variation_name}_for_{os.path.splitext(decay_file)[0]}"
                    variation_dir = os.path.join(self.decay_file_dir, variation_name, d_meson_name)
                    variation_decay_file_path = os.path.join(variation_dir, f)
                    assert output_key not in output_dict.keys(), (output_key, output_dict)
                    assert variation_decay_file_path not in output_dict.values(), \
                        (variation_decay_file_path, output_dict)
                    output_dict.update({output_key: variation_decay_file_path})

        return output_dict


if __name__ == "__main__":
    dec_file_creator = DecayFileWithLLSWVariationsCreator(
        decay_file_directory=DecayFileWithLLSWVariationsCreator.decay_files_dir_path
    )
    dec_file_creator.create_llsw_variations()
