"""
Provides convenience functions for the simulation of D** events for the reweighting
of the ISGW2_nominal to the LLSW_nominal decay model.

Felix Metzner and Maximilian Welsch (2018-2020)
"""

import os
import basf2
import generators
import modularAnalysis

from variables import variables
from typing import Union, List, Dict, AnyStr

from analysistools.processing.observables_utils import makeROOTCompatible


__all__ = [
    "DststReweightingEventGenerator"
]


class DststReweightingEventGenerator:
    def __init__(
            self,
            number_of_events: int,
            decay_file_path: Union[os.PathLike, AnyStr],
            output_path: Union[os.PathLike, AnyStr]
    ) -> None:
        self._number_of_events = number_of_events
        self._decay_file_path = decay_file_path
        self._output_path = output_path

    def get_path(self) -> basf2.Path:
        main_path = basf2.create_path()

        modularAnalysis.setupEventInfo(noEvents=self._number_of_events, path=main_path)
        generators.add_evtgen_generator(path=main_path, finalstate='signal', signaldecfile=self._decay_file_path)
        modularAnalysis.loadGearbox(path=main_path)

        modularAnalysis.fillParticleListFromMC(decayString='Upsilon(4S)', cut='', addDaughters=True, path=main_path)
        modularAnalysis.matchMCTruth('Upsilon(4S)', path=main_path)

        modularAnalysis.variablesToNtuple(
            decayString='Upsilon(4S)',
            variables=self._get_variable_list(),
            filename=self._output_path,
            path=main_path
        )

        return main_path

    def _get_variable_list(self) -> List[str]:
        four_vectors = self._get_four_vectors()
        predefined_variables = self._get_predefined_variables()
        return four_vectors + predefined_variables

    @staticmethod
    def _get_four_vectors() -> List[str]:
        four_vector_components = ['E', 'px', 'py', 'pz']
        alias_list = []

        for particle in ['B', 'Dstarstar', 'l', 'nu', 'D', 'pi']:
            new_vars = DststReweightingEventGenerator._add_alias(particle=particle, observables=four_vector_components)
            alias_list.extend(new_vars)

        return alias_list

    @staticmethod
    def _get_predefined_variables() -> List[str]:
        return list(DststReweightingEventGenerator._get_predefined_variables_dict().values())

    @staticmethod
    def _get_predefined_variables_dict() -> Dict[str, str]:
        return {
            "wMC": "genUpsilon4S(mcDaughter(0, wMC))",
            "costhetalMC": "genUpsilon4S(mcDaughter(0, costhetalMC))",
            "costhetalWOPhotonsMC": "genUpsilon4S(mcDaughter(0, costhetalWOPhotonsMC))",
            "costhetavMC": "genUpsilon4S(mcDaughter(0, costhetavMC))",
            "chi_primeMC": "genUpsilon4S(mcDaughter(0, chi_primeMC))",
            "chi_primeWOPhotonsMC": "genUpsilon4S(mcDaughter(0, chi_primeWOPhotonsMC))",
            "chi_MC": "genUpsilon4S(mcDaughter(0, chi_MC))",
            "chiWOPhotons_MC": "genUpsilon4S(mcDaughter(0, chiWOPhotons_MC))",
            "chi_primeMC_inRF": "genUpsilon4S(mcDaughter(0, useRestFrame(chi_primeMC)))",
            "chi_primeWOPhotonsMC_inRF": "genUpsilon4S(mcDaughter(0, useRestFrame(chi_primeWOPhotonsMC)))",
            "chi_MC_inRF": "genUpsilon4S(mcDaughter(0, useRestFrame(chi_MC)))",
            "chiWOPhotons_MC_inRF": "genUpsilon4S(mcDaughter(0, useRestFrame(chiWOPhotons_MC)))",
            "chi_primeMC_inCMSF": "genUpsilon4S(mcDaughter(0, useCMSFrame(chi_primeMC)))",
            "chi_primeWOPhotonsMC_inCMSF": "genUpsilon4S(mcDaughter(0, useCMSFrame(chi_primeWOPhotonsMC)))",
            "chi_MC_inCMSF": "genUpsilon4S(mcDaughter(0, useCMSFrame(chi_MC)))",
            "chiWOPhotons_MC_inCMSF": "genUpsilon4S(mcDaughter(0, useCMSFrame(chiWOPhotons_MC)))",
            "d_meson_mass": "genUpsilon4S(mcDaughter(0, mcDaughter(0, M)))",
            "d_meson_delta_mass": "genUpsilon4S(mcDaughter(0, mcDaughter(0, dM)))",
            "p_lepton_star": "genUpsilon4S(mcDaughter(0, useRestFrame(mcDaughter(0, p))))",
        }

    @staticmethod
    def get_predefined_root_compatible_variables_dict() -> Dict[str, str]:
        return {
            key: makeROOTCompatible([value])[0]
            for key, value in DststReweightingEventGenerator._get_predefined_variables_dict().items()
        }

    @staticmethod
    def get_inverted_predefined_root_compatible_variables_dict() -> Dict[str, str]:
        return {v: k for k, v in DststReweightingEventGenerator.get_predefined_root_compatible_variables_dict().items()}

    @staticmethod
    def _add_alias(particle: str, observables: List[str]) -> List[str]:
        alias_list = []

        for observable in observables:
            alias = particle + '_' + observable

            if particle == 'B':
                variable_string = f'daughter(0, {observable})'
            elif particle == 'Dstarstar':
                variable_string = f'daughter(0, daughter(0, {observable}))'
            elif particle == 'l':
                variable_string = f'daughter(0, daughter(1, {observable}))'
            elif particle == 'nu':
                variable_string = f'daughter(0, daughter(2, {observable}))'
            elif particle == 'D':
                variable_string = f'daughter(0, daughter(0, daughter(0, {observable})))'
            elif particle == 'pi':
                variable_string = f'daughter(0, daughter(0, daughter(1, {observable})))'
            else:
                raise ValueError(f"The particle type {particle} is not valid.")

            variables.addAlias(alias, variable_string)
            alias_list.append(alias)

        return alias_list
