from abc import ABC, abstractmethod

import os
import numpy as np
import pandas as pd

from typing import Optional, Union, List, Tuple

from analysistools.batch_tools.file_handler import Belle_list_types


class FFReweightingBase(ABC):

    def __init__(self, info_target_path: Optional[Union[str, os.fspath]] = None) -> None:
        self.nominal_weight_key = None
        self.variation_weight_keys = list()

        self.b1_w = "genUpsilon4S__bomcDaughter__bo0__cm__spwMC__bc__bc"
        self.b1_costhetal = "genUpsilon4S__bomcDaughter__bo0__cm__spcosthetalMC__bc__bc"
        self.b1_costhetav = "genUpsilon4S__bomcDaughter__bo0__cm__spcosthetavMC__bc__bc"
        self.b1_chi = "genUpsilon4S__bomcDaughter__bo0__cm__spchi_MC__bc__bc"
        self.b1_chi_prime = "genUpsilon4S__bomcDaughter__bo0__cm__spchi_primeMC__bc__bc"

        self.b2_w = "genUpsilon4S__bomcDaughter__bo1__cm__spwMC__bc__bc"
        self.b2_costhetal = "genUpsilon4S__bomcDaughter__bo1__cm__spcosthetalMC__bc__bc"
        self.b2_costhetav = "genUpsilon4S__bomcDaughter__bo1__cm__spcosthetavMC__bc__bc"
        self.b2_chi = "genUpsilon4S__bomcDaughter__bo1__cm__spchi_MC__bc__bc"
        self.b2_chi_prime = "genUpsilon4S__bomcDaughter__bo1__cm__spchi_primeMC__bc__bc"

        self.b_index_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
        self.b_via_tau_index_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"

        self.info_target_path = info_target_path

        self.weight_name = None
        self.weight_info = None

        self.b1_info_vars = {
            self.b1_w: "w",
            self.b1_costhetal: "costhetal",
            self.b1_costhetav: "costhetav",
            self.b1_chi: "chi",
            self.b1_chi_prime: "chi_prime",
        }
        self.b2_info_vars = {
            self.b2_w: "w",
            self.b2_costhetal: "costhetal",
            self.b2_costhetav: "costhetav",
            self.b2_chi: "chi",
            self.b2_chi_prime: "chi_prime",
        }

    def release(self) -> None:
        if self.info_target_path is not None:
            dir_path, file_name = os.path.split(self.info_target_path)
            temp_target = os.path.join(dir_path, f"temp_{file_name}")

            self.weight_info.to_hdf(temp_target, "weight_info")
            os.rename(temp_target, self.info_target_path)

    @abstractmethod
    def add_weights(self, df: pd.DataFrame) -> None:
        pass

    @property
    def weight_keys(self) -> List[str]:
        weight_keys = [self.nominal_weight_key]
        weight_keys.extend(self.variation_weight_keys)
        return weight_keys

    def _write_weight_info(self, sample: pd.DataFrame, conditions: Tuple[pd.Series, pd.Series]) -> None:
        if self.info_target_path is None:
            return

        b1_condition, b2_condition = conditions

        common_cols = [c for c in sample.columns if c.startswith(self.weight_name)]
        common_cols.extend(["unique_event_id", "unique_candidate_id"])

        b1_vars = self.b1_info_vars
        b2_vars = self.b2_info_vars

        b1_sample = sample.loc[b1_condition, list(b1_vars.keys()) + common_cols]
        b2_sample = sample.loc[b2_condition, list(b2_vars.keys()) + common_cols]

        b1_sample.rename(columns=b1_vars, inplace=True)
        b2_sample.rename(columns=b2_vars, inplace=True)

        b1_sample["reweighted_B"] = 1
        b2_sample["reweighted_B"] = 2

        info = pd.concat([b1_sample, b2_sample], ignore_index=True, sort=False)

        if self.weight_info is None:
            self.weight_info = info
        else:
            self.weight_info = self.weight_info.append(info, ignore_index=True, sort=False)

    @staticmethod
    def reset_problematic_special_mc(
            df: pd.DataFrame,
            weight_base_name: str,
            gen_ids: List[int],
            reset_special_to: float = 0.
    ) -> None:
        nom_weight_col = f"{weight_base_name}nom"
        assert nom_weight_col in df.columns

        wcs = [c for c in df.columns if c.startswith(weight_base_name)]

        is_info_df = all(c in df.columns for c in ["w", "reweighted_B", "costhetal"])

        special_type_mask = df.data_type.isin(list(Belle_list_types.keys()))
        if is_info_df:
            special_exp_mask = df.unique_event_id.str.split("_").str[1].astype("float") < 30
        else:
            special_exp_mask = df["expNum"] < 30

        special_mask = special_type_mask & special_exp_mask

        if is_info_df:
            update_mask = (special_mask & ((df["w"] > 1.6) | (df["w"] < 0.)))
            if update_mask.any():
                print(f"Resetting the {weight_base_name} FF weights for {len(update_mask.index)} "
                      f"SpecialMC candidates to {reset_special_to}")
                if reset_special_to == 0.:
                    df.drop(index=df.loc[update_mask].index, inplace=True)
                else:
                    for col in [c for c in wcs if c.endswith("up") or c.endswith("down")]:
                        df.loc[update_mask, col] = df.loc[update_mask, col] / df.loc[update_mask, nom_weight_col]
                    df.loc[update_mask, nom_weight_col] = reset_special_to
                    df.loc[update_mask, "w"] = 1.7

            assert (df.loc[~special_mask, "w"] < 1.6).all()
            assert (df.loc[~special_mask, "w"] > 0.).all()
        else:
            b_i_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
            b_via_tau_i_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__" \
                              "bc__bc__bc__bc"

            b1_w = "genUpsilon4S__bomcDaughter__bo0__cm__spwMC__bc__bc"
            b2_w = "genUpsilon4S__bomcDaughter__bo1__cm__spwMC__bc__bc"
            b1_q = "genUpsilon4S__bomcDaughter__bo0__cm__spqSquaredMC__bc__bc"
            b2_q = "genUpsilon4S__bomcDaughter__bo1__cm__spqSquaredMC__bc__bc"

            b1_is_btodl = ((df.GEN_B1_ID.isin(gen_ids)) & ((df[b_i_var] == 1) | (df[b_via_tau_i_var] == 1)))
            b2_is_btodl = ((df.GEN_B2_ID.isin(gen_ids)) & ((df[b_i_var] == 2) | (df[b_via_tau_i_var] == 2)))

            update_mask_1 = (b1_is_btodl & special_mask & (df[b1_q] < 0.))
            update_mask_2 = (b2_is_btodl & special_mask & (df[b2_q] < 0.))

            if update_mask_1.any() or update_mask_2.any():
                print(f"Resetting the {weight_base_name} FF weights for "
                      f"{len(df.loc[update_mask_1].index) + len(df.loc[update_mask_2].index)} "
                      f"SpecialMC candidates to {reset_special_to}")

            if update_mask_1.any():
                for col in [c for c in wcs if c.endswith("up") or c.endswith("down")]:
                    if reset_special_to == 0.:
                        df.loc[update_mask_1, col] = reset_special_to
                    else:
                        df.loc[update_mask_1, col] = df.loc[update_mask_1, col] / df.loc[update_mask_1, nom_weight_col]
                df.loc[update_mask_1, nom_weight_col] = reset_special_to
            if update_mask_2.any():
                for col in [c for c in wcs if c.endswith("up") or c.endswith("down")]:
                    if reset_special_to == 0.:
                        df.loc[update_mask_2, col] = reset_special_to
                    else:
                        df.loc[update_mask_2, col] = df.loc[update_mask_2, col] / df.loc[update_mask_2, nom_weight_col]
                df.loc[update_mask_2, nom_weight_col] = reset_special_to

            assert (df.loc[(b1_is_btodl & (~special_mask)), b1_w] < 1.6).all()
            assert (df.loc[(b2_is_btodl & (~special_mask)), b2_w] < 1.6).all()
            assert (df.loc[(b1_is_btodl & (~special_mask)), b1_w] > 0.).all()
            assert (df.loc[(b2_is_btodl & (~special_mask)), b2_w] > 0.).all()
            assert (df.loc[(b1_is_btodl & (~special_mask)), b1_q] >= 0.).all()
            assert (df.loc[(b2_is_btodl & (~special_mask)), b2_q] >= 0.).all()


def get_varied_central_values(
        central: np.ndarray,
        covariance_matrix: np.ndarray
) -> List[Tuple[np.ndarray, np.ndarray]]:
    eigenvalues, eigenvectors = np.linalg.eig(covariance_matrix)

    eigenvalue_variations = []
    for index, eigenvalue in enumerate(eigenvalues):
        variation = np.sqrt(eigenvalue)

        up = central +  variation * eigenvectors[:, index]
        down = central - variation * eigenvectors[:, index]

        eigenvalue_variations.append((up, down))

    return eigenvalue_variations


if __name__ == "__main__":

    from rdstar.mc_reweighting.BtoDlnu import BtoDlnu_weights
    from rdstar.mc_reweighting.BtoDstlnu import BtoDstlnu_weights

    varied_central_values = get_varied_central_values(
        BtoDlnu_weights.central, BtoDlnu_weights.covariance_matrix
    )

    print("Varied parameter values for B->Dlv decays")
    for i, pair in enumerate(varied_central_values):
        print(f"variation {i}")
        print("\tvariation up:")
        print(f"\t{pair[0]}")
        print("\tvariation down:")
        print(f"\t{pair[1]}")

    varied_central_values = get_varied_central_values(
        BtoDstlnu_weights.central, BtoDstlnu_weights.covariance_matrix
    )
    print()
    print("Varied parameter values for B->Dstlv decays")
    for i, pair in enumerate(varied_central_values):
        print(f"variation {i}")
        print("\tvariation up:")
        print(f"\t{pair[0]}")
        print("\tvariation down:")
        print(f"\t{pair[1]}")
