import os
import pickle
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from PyFastBDT import FastBDT
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split

from typing import Union, Optional, List, Tuple, Dict, Generator, NamedTuple

from templatefitter.plotter.plot_utilities import export

from analysistools.plotting.mva_plots import plot_roc_curve

__all__ = [
    "MVAVarInfo",
    "MVABaseClass",
    "FeatureImportanceInfo"
]


class MVAVarInfo(NamedTuple):
    var: str  # The column name of the variable as available in the n-tuple
    latex: str  # Latex name for the variable to be used when plotting the variable
    bins: int  # Number of bins to be used when plotting the variable
    min: float  # Lower limit of range to be used when plotting the variable
    max: float  # Upper limit of range to be used when plotting the variable
    unit: Optional[str] = None  # Unit of the variable to be used when plotting the variable
    log: bool = False  # Whether to use a log scale for the y-axis when plotting the variable


class FeatureImportanceInfo(NamedTuple):
    feature_importance_type: str
    feature_importance: Dict[str, Union[int, float]]


class MVABaseClass:
    def __init__(
            self,
            variables: List[str],
            target_variable: str,
            weight: Optional[str] = None
    ) -> None:
        """
        Interface for the training and application a mva method.

        :param variables: List of variables to use for the
        mva.
        :param target_variable: Name of the variable for the
        'signal' when training the classifier.
        :param weight: Name of the weight to use in the DataFrame.
        """

        self._variables = variables
        self._target_variable = target_variable
        self._weight = weight
        self._has_trained_classifier = False

        self._classifier = None

    def add_classifier(
            self,
            classifier=FastBDT.Classifier(),
            is_trained: bool = False
    ) -> None:
        """
        Adds a mva classifier.
        Can be trained or untrained. The classifier should
        have a sklearn interface.
        :param classifier: A mva classifier with sklearn interface.
        :param is_trained: Flag if the classifier is already trained.
        :return:
        """
        assert self._classifier is None
        self._classifier = classifier
        self._has_trained_classifier = is_trained

    def load_classifier_weights(self, weight_file_path: Union[str, os.fspath]) -> None:
        """
        Loads the classifier weights from an already prepared weight file.
        :param weight_file_path: Path to the weight file.
        :return: None
        """
        if not os.path.exists(weight_file_path):
            raise ValueError(f"The provided path to the weight file does not exist:\n{weight_file_path}")

        self._classifier.load(weightfile=weight_file_path)
        self._has_trained_classifier = True

    def train(self, data: pd.DataFrame) -> None:
        """
        Trains the added classifier with data present
        in the DataFrame.

        :param data: Training data. DataFrame has to hold the
        mva as columns.
        :return: None.
        """
        assert all(v in data.columns for v in self._variables), [v for v in self._variables if v not in data.columns]
        assert self._target_variable in data.columns, self._target_variable

        x_train = data[self._variables].values
        y_train = data[self._target_variable].values.reshape(-1)

        if self._weight is None:
            w_train = np.ones(y_train.shape)
        else:
            w_train = data[self._weight].values

        w_train = self.normalize_weights(target=y_train, weights=w_train)

        self._classifier.fit(x_train, y_train, w_train)
        self._has_trained_classifier = True

    @staticmethod
    def normalize_weights(
            target: np.ndarray,
            weights: np.ndarray
    ) -> np.ndarray:
        """
        Normalizes event weights so that the sum of
        signal and background weights is equal.

        :param target: Target labels. Expected to be in {0, 1}.
        :param weights: Array of event weights.
        :return: Normalized event weights
        """
        n_sig = np.sum(target == 1)
        n_bkg = np.sum(target == 0)

        weights[np.where(target == 1)] *= 1 / n_sig
        weights[np.where(target == 0)] *= 1 / n_bkg

        return weights

    def apply(
            self,
            data: pd.DataFrame,
            output_column: str = 'mva_output'
    ) -> None:
        """
        Applies a train mva classifier to a
        pd.DataFrame and adds a new column with the classifier
        output to the pd.DataFrame.

        :param data: DataFrame where the classifier should
        be applied on.
        :param output_column: Name of the column containing the
        classifier output which will be appended to the given
        DataFrame.
        :return: None
        """
        if not self._has_trained_classifier:
            raise RuntimeError("Trying to apply an untrained classifier.")

        assert all(v in data.columns for v in self._variables), [v for v in self._variables if v not in data.columns]
        x_predict = data[self._variables].values

        pcs = self._classifier.predict(x_predict)

        data.loc[:, output_column] = pcs

    def save_classifier_weights(self, weight_file_path: Union[str, os.fspath], overwrite: bool = False) -> None:
        """
        Saves the weight file for a trained classifier.
        :param weight_file_path: Path under which the weight file shall be saved.
        :param overwrite: Whether to overwrite an existing weight file.
        :return: None
        """
        if os.path.isfile(weight_file_path) and not overwrite:
            raise RuntimeError("The specified weight file already exists and the overwrite option is set to False.")
        self._classifier.save(weightfile=weight_file_path)

        var_file_path = self.get_var_file_from_weight_file(weight_file_path=weight_file_path)
        self.dump_variables(var_file_path, overwrite=overwrite)

    def dump_variables(self, pickle_file_path: Union[str, os.fspath], overwrite: bool = False) -> None:
        """
        Dump variables used by the mva as pickled object.
        :param pickle_file_path: Path under which the variables shall be dumped as pickled object.
        :param overwrite:  Whether to overwrite an existing file.
        :return: None
        """
        MVABaseClass.dump_list_of_variables(
            variable_list=self._variables,
            pickle_file_path=pickle_file_path,
            overwrite=overwrite
        )

    def dump_classifier(self, pickle_file_path: Union[str, os.fspath], overwrite: bool = False) -> None:
        """
        Dump classifier object as pickled object.
        :param pickle_file_path: Path under which the classifier shall be dumped as pickled object.
        :param overwrite: Whether to overwrite an existing file.
        :return: None
        """
        if os.path.isfile(pickle_file_path) and not overwrite:
            raise RuntimeError("The specified pickle file already exists and the overwrite option is set to False.")
        with open(pickle_file_path, "wb") as pkl_file:
            pickle.dump(self, pkl_file)

    @staticmethod
    def get_var_file_from_weight_file(weight_file_path: Union[str, os.fspath]) -> str:
        """
        Convenience function to get the file name of the variable file to a weight file.
        :param weight_file_path: Path to the weight file.
        :return: Path to respective variable file.
        """
        return os.fspath(f"{os.path.splitext(weight_file_path)[0]}_vars.pkl")

    @staticmethod
    def load_classifier(
            weight_file_path: Union[str, os.fspath],
            var_file_path: Union[None, str, os.fspath] = None,
            target_var: Optional[str] = None
    ) -> "MVABaseClass":
        """
        Returns a classifier with the weights from the specified weight file.
        :param weight_file_path: Path to the weight file for the classifier.
        :param var_file_path: Path to the variable file for the classifier.
                              Default: None -> Trying to get it from weight file name.
        :param target_var: Target variable column string. Only necessary for validation.
        :return: A MVA classifier.
        """
        if not os.path.isfile(weight_file_path):
            raise ValueError(f"The specified weight file\n {weight_file_path} \ndoes not exists.")

        if var_file_path is None:
            var_file_path = MVABaseClass.get_var_file_from_weight_file(weight_file_path=weight_file_path)
            if not os.path.isfile(var_file_path):
                raise RuntimeError(f"Automatic extraction of variables file path from weight file failed!"
                                   f"Got variables file path\n{var_file_path}\nfrom weight file\n{weight_file_path}")
        variables = MVABaseClass.load_pickled_variables(pickle_file_path=var_file_path)

        classifier = MVABaseClass(
            variables=variables,
            target_variable=target_var if target_var is not None else "",
            weight=None
        )
        classifier.add_classifier()
        classifier.load_classifier_weights(weight_file_path=weight_file_path)
        return classifier

    @staticmethod
    def load_pickled_classifier(pickle_file_path: Union[str, os.fspath]) -> "MVABaseClass":
        """
        Loads a pickled classifier from the specified pickle file.
        :param pickle_file_path: Path to the pickle file containing a classifier object.
        :return: A MVA classifier.
        """
        if not os.path.isfile(pickle_file_path):
            raise ValueError(f"The specified pickle file\n {pickle_file_path} \ndoes not exists.")
        with open(pickle_file_path, "rb") as pkl_file:
            classifier = pickle.load(pkl_file)
        return classifier

    @staticmethod
    def load_pickled_variables(pickle_file_path: Union[str, os.fspath]) -> List[str]:
        """
        Loads a pickled variable list for a classifier from the specified pickle file.
        :param pickle_file_path: Path to the pickle file containing the variables.
        :return: A list of variables.
        """
        if not os.path.isfile(pickle_file_path):
            raise ValueError(f"The specified pickle file\n {pickle_file_path} \ndoes not exists.")
        with open(pickle_file_path, "rb") as pkl_file:
            variables = pickle.load(pkl_file)
        return variables

    @staticmethod
    def dump_list_of_variables(
            variable_list: List[str],
            pickle_file_path: Union[str, os.fspath],
            overwrite: bool = False
    ) -> None:
        """
        Dump given list of variables.

        :param variable_list: List of variables to be saved in pickle file.
        :param pickle_file_path: Path under which the variables shall be dumped as pickled object.
        :param overwrite: Whether to overwrite an existing file.
        :return: None
        """
        if os.path.isfile(pickle_file_path) and not overwrite:
            raise RuntimeError("The specified pickle file already exists and the overwrite option is set to False.")
        with open(pickle_file_path, "wb") as pkl_file:
            pickle.dump(variable_list, pkl_file)

    @staticmethod
    def prepare_data_samples(
            class1_dfs: List[pd.DataFrame],
            class2_dfs: List[pd.DataFrame],
            train_test_split_ratio: float = 0.5
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Splits data frames for category 1 and 2 into subsets for training and evaluation of the
        mva and returns a tuple of four data frames: Training set category 1, training set category 2,
        test set category 1 and test set category 2.

        :param class1_dfs: Data frame containing data of the first class.
        :param class2_dfs: Data frame containing data of the second class.
        :param train_test_split_ratio: Ratio by which the data frames will be split into train
        and test sample. Default = 0.5.
        :return: Tuple (class 1 train df, class 2 train df, class 1 test df, class 2 test df)
        """
        train1 = []
        test1 = []
        for df in class1_dfs:
            train, test = train_test_split(df, train_size=train_test_split_ratio)
            train1.append(train)
            test1.append(test)

        train2 = []
        test2 = []
        for df in class2_dfs:
            train, test = train_test_split(df, train_size=train_test_split_ratio)
            train2.append(train)
            test2.append(test)

        return (
            pd.concat(train1, ignore_index=True, sort=False),
            pd.concat(train2, ignore_index=True, sort=False),
            pd.concat(test1, ignore_index=True, sort=False),
            pd.concat(test2, ignore_index=True, sort=False)
        )

    @staticmethod
    def get_uncorrelated_variables(
            data_sample: pd.DataFrame,
            variables: List[str],
            max_correlation: float,
            signal_var: str
    ) -> List[str]:
        """
        Method to get a set of variables from the provided list "variables" which are
        uncorrelated to a specified other variable "signal_var".

        The given data_sample is used to evaluate the correlation. As correlated to
        signal_var are considered such variables for which abs(correlation to signal_var)
        is larger than max_correlation.

        The reduced list of variable strings is returned.

        :param data_sample: Set used to evaluate the correlation.
        :param variables: Variables to be checked for their correlation to signal_var.
        :param max_correlation: The maximal correlation allowed. Variables with values above are removed.
        :param signal_var: The signal variable against which the correlation is checked.
        :return: List of variable strings of uncorrelated variables.
        """
        corr = data_sample[variables + [signal_var]].corr()
        final_var_list = list(corr.query(f'abs({signal_var}) < {max_correlation}').index)

        print(f"The following variables have been omitted due to their correlation to the signal variable {signal_var}")
        print("\n".join([var for var in variables if var not in final_var_list]))

        return final_var_list

    def test_evaluation(self, test_data: pd.DataFrame, output_dir: Union[str, os.fspath]) -> float:
        """
        A test application of this MVA, yielding some performance values and plots.

        The MVA is applied to the provided test data sample provided via the parameter test_data
        based on which the performance of the MVA is evaluated.
        The results are written into the directory specified via output_dir.
        The return value is the area under the ROC curve for this MVA.

        :param test_data: Pandas DataFrame containing a labeled test data set.
        :param output_dir: The path to a directory into which the results will be written.
        :return: Returns the area under the roc curve.
        """
        self.apply(data=test_data, output_column="mva_output")

        # TODO: Add distribution plots of input pdfs and correlation plots using test_input...
        # test_input = test_data[self._variables].values
        test_truth = test_data[self._target_variable].values
        test_output = test_data["mva_output"].values

        roc_auc = roc_auc_score(test_truth, test_output)

        fig = plt.figure(num=None, figsize=(12, 9), dpi=100, facecolor='w', edgecolor='k')
        ax = plt.gca()
        fig, ax = plot_roc_curve(y_true=test_truth, y_predict=test_output, fig_ax_tuple=(fig, ax))
        export(fig=fig, target_dir=output_dir, filename="mva_roc_plot")

        return roc_auc

    def map_variables(self, variable_indices_list: List[int]) -> List[str]:
        variable_map = {i: var for i, var in enumerate(self._variables)}  # type: Dict[int, str]
        assert all(isinstance(i, int) and isinstance(v, str) for i, v in variable_map.items()), variable_map
        return [variable_map[index] for index in variable_indices_list]

    def _generate_feature_importance_info(
            self,
            importance: Dict[int, Union[int, float]],
            name: str
    ) -> FeatureImportanceInfo:
        assert isinstance(importance, dict), type(importance)
        assert all(isinstance(k, int) for k in importance.keys()), importance
        assert all(isinstance(v, (int, float)) for v in importance.values()), importance
        assert all(k < len(self._variables) for k in importance.keys()), (len(self._variables), list(importance.keys()))
        return FeatureImportanceInfo(
            feature_importance_type=name,
            feature_importance={v: r for v, r in zip(self.map_variables(list(importance.keys())), importance.values())}
        )

    @staticmethod
    def get_feature_importance_types() -> List[str]:
        return ["internal", "individual", "external"]

    @staticmethod
    def get_empty_feature_importance_infos() -> List[FeatureImportanceInfo]:
        imp_types = MVABaseClass.get_feature_importance_types()
        return [FeatureImportanceInfo(feature_importance_type=t, feature_importance={}) for t in imp_types]

    def get_feature_importance(
            self,
            train_data: pd.DataFrame,
            test_data: Optional[pd.DataFrame] = None
    ) -> Generator[FeatureImportanceInfo, None, None]:
        assert isinstance(self._classifier, FastBDT.Classifier), type(self._classifier)
        internal_feature_importance = self._classifier.internFeatureImportance()  # type: Dict[int, Union[int, float]]
        yield self._generate_feature_importance_info(importance=internal_feature_importance, name="internal")

        x_train = train_data[self._variables].values
        y_train = train_data[self._target_variable].values.reshape(-1)
        if self._weight is None:
            w_train = np.ones(y_train.shape)
        else:
            w_train = train_data[self._weight].values
        w_train = self.normalize_weights(target=y_train, weights=w_train)

        individual_feat_imp = self._classifier.individualFeatureImportance(  # type: Dict[int, Union[int, float]]
            X=x_train
        )
        yield self._generate_feature_importance_info(importance=individual_feat_imp, name="individual")

        if test_data is not None:
            x_test = test_data[self._variables].values
            y_test = test_data[self._target_variable].values.reshape(-1)
            if self._weight is None:
                w_test = np.ones(y_test.shape)
            else:
                w_test = test_data[self._weight].values
            w_test = self.normalize_weights(target=y_test, weights=w_test)
        else:
            x_test, y_test, w_test = None, None, None

        external_feat_importance = self._classifier.externFeatureImportance(  # type: Dict[int, Union[int, float]]
            X=x_train,
            y=y_train,
            weights=w_train,
            X_test=x_test,
            y_test=y_test,
            weights_test=w_test
        )
        yield self._generate_feature_importance_info(importance=external_feat_importance, name="external")
