"""
Definition of MVA Selection Setups

Patrick Ecker und Felix Metzner 2020
"""

import copy
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Tuple, List, Dict, Any

from rdstar.mva_methods.mva_methods import MVAVarInfo
from rdstar.offline_analysis.rdstar_samples import mva_samples

from rdstar.continuum_suppression.continuum_suppression import get_cs_variable_infos

from templatefitter.plotter.histogram_variable import HistVariable

__all__ = [
    "MVASelectorSetupBase",
    "MVASelectorSetups"
]


class MVASelectorSetupBase(ABC):
    selection_mva_name = ""  # type: str

    all_mva_variable_infos = ()  # type: Tuple[MVAVarInfo, ...]

    _training_sample_size = 0  # type: int
    _training_data_stream_key_dict = {}  # type: Dict[int, Tuple[str, ...]]

    _required_variables = ()  # type:  Tuple[str, ...]
    _modified_variables = {}  # type:  Dict[str, str] # Contains new_col_name: original_col_name

    _validation_variables = (  # type:  Tuple[str, ...]
        "m2RecoilSignalSide",
        "daughter__bo1__cm__spuseRestFrame__bodaughter__bo1__cm__spp__bc__bc__bc",
        "unique_event_id",
        "isSignalAcceptMissingNeutrino",
        "qSquaredReco",
        "daughter__bo0__cm__spMbc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
        "daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
        "daughter__bo1__cm__spchiProb__bc",
        "ROE_neextra__boCleanROE__bc",
        "daughter__bo0__cm__spdeltaE__bc",
        "daughter__bo1__cm__spROE_neextra__boCleanROE__bc__bc",
        "ROE_deltae__boCleanROE__bc",
        "daughter__bo1__cm__spROE_deltae__boCleanROE__bc__bc",
        "daughter__bo1__cm__spROE_eextra__boCleanROE__bc__bc",
        "daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc"
    )

    _fast_bdt_settings = {  # type: Dict[str, Any]
        "nTrees": 200,
    }

    @classmethod
    def get_selection_variable_dict(cls) -> Dict[str, HistVariable]:
        assert isinstance(cls.all_mva_variable_infos, tuple), type(cls.all_mva_variable_infos)
        assert len(cls.all_mva_variable_infos) > 0, f"No variable infos defined for MVASelector class {cls.__name__}!"
        return {
            var_info.var: HistVariable(
                df_label=var_info.var,
                n_bins=var_info.bins,
                scope=(var_info.min, var_info.max),
                var_name=var_info.latex,
                unit=var_info.unit,
                use_log_scale=var_info.log
            )
            for var_info in cls.all_mva_variable_infos
        }

    @classmethod
    def get_selection_mva_output_column_name(cls) -> str:
        assert isinstance(cls.selection_mva_name, str), type(cls.selection_mva_name)
        assert cls.selection_mva_name, f"selection_mva_name is not defined for MVASelector class {cls.__name__}!"
        return f"{cls.selection_mva_name}_output"

    @classmethod
    def get_target_column_name(cls) -> str:
        assert cls.selection_mva_name, \
            f"selection_mva_name is not defined for MVASelector class {cls.__name__}!"
        return f"{cls.selection_mva_name}_truth_target"

    @classmethod
    def get_training_sample_size(cls) -> int:
        assert isinstance(cls._training_sample_size, int), type(cls._training_sample_size)
        assert cls._training_sample_size > 0, cls._training_sample_size
        return cls._training_sample_size

    @classmethod
    def get_training_data_info_tuples(cls) -> List[Tuple[str, int]]:
        assert isinstance(cls._training_data_stream_key_dict, dict), type(cls._training_data_stream_key_dict)
        assert len(cls._training_data_stream_key_dict) > 0, \
            f"No Training data defined for MVASelector class {cls.__name__}!"
        return [
            (data_key, stream)
            for stream, data_keys in cls._training_data_stream_key_dict.items()
            for data_key in data_keys
        ]

    @classmethod
    def get_required_columns(cls) -> Tuple[str, ...]:
        train_vars = [
            mva_v.var if mva_v.var not in cls._modified_variables else cls._modified_variables[mva_v.var]
            for mva_v in cls.all_mva_variable_infos
        ]
        validation_vars = list(set(list(MVASelectorSetupBase._validation_variables) + list(cls._validation_variables)))
        return tuple(set(train_vars + list(cls._required_variables) + validation_vars))

    @classmethod
    def get_fast_bdt_settings(cls) -> Dict[str, Any]:
        return cls._fast_bdt_settings

    @classmethod
    @abstractmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _data_pre_processing(cls, df: pd.DataFrame) -> None:
        raise NotImplementedError()

    @classmethod
    def data_preparation(cls, df: pd.DataFrame, for_application_only: bool) -> None:
        if not for_application_only:
            cls._add_target_variable(df=df)
        cls._data_pre_processing(df=df)

        if for_application_only:
            assert all(c.var in df.columns for c in cls.all_mva_variable_infos), \
                [c.var for c in cls.all_mva_variable_infos if c.var not in df.columns]
        else:
            assert all(c in df.columns for c in cls.get_required_columns()), \
                [c for c in cls.get_required_columns() if c not in df.columns]

    @classmethod
    def data_preparation_for_validation(cls, df: pd.DataFrame) -> None:
        cls._add_target_variable(df=df)


class FirstSelectionMVASetup(MVASelectorSetupBase):
    selection_mva_name = "first_sel_mva"  # type: str

    e_extra = MVAVarInfo(
        var=f"ROE_eextra__boCleanROE__bc_for_{selection_mva_name}",
        latex=r"$\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
        bins=30, min=0., max=5., unit="GeV"
    )
    d_mass = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo0__cm__spM__bc__bc",
        latex=r"$\mathrm{M}_{D^{(*)}}$",
        bins=25, min=1.8, max=2.1, unit="GeV"
    )
    d_star_daughter_mass = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spM__bc__bc__bc",
        latex=r"$\mathrm{M}_\mathrm{D}$",
        bins=25, min=1.8, max=01.95, unit="GeV"
    )
    cs_mva = MVAVarInfo(
        var="cs_mva_output",
        latex=r"$\mathcal{P}_\mathrm{CS}$",
        bins=25, min=0., max=1.0
    )
    sig_b_decay_id = MVAVarInfo(
        var="daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc",
        latex=r"B Decay Mode ID",
        bins=70, min=10., max=80.
    )
    upsilon_decay_id = MVAVarInfo(
        var="extraInfo__bodecayModeID__bc",
        latex=r"$\Upsilon(\mathrm{4S})$ Decay Mode ID",
        bins=25, min=1000., max=1024.
    )

    all_mva_variable_infos = (  # type: Tuple[MVAVarInfo]
        e_extra,
        d_mass,
        d_star_daughter_mass,
        cs_mva,
        sig_b_decay_id,
        upsilon_decay_id
    )

    _required_variables = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc",
        "GEN_B1_ID",
        "GEN_B2_ID",
        "SIG_ID",
    )  # type:  Tuple[str, ...]

    # Contains information about modified variables in the form: {new_col_name: original_col_name}
    _modified_variables = {e_extra.var: "ROE_eextra__boCleanROE__bc"}  # type:  Dict[str, str]

    _training_sample_size = 50000  # type: int
    _training_data_stream_key_dict = {
        -999: tuple(copy.copy(mva_samples)),
        0: ("Charm", "UDS"),  # TODO: Should be Generic (Mixed, Charged) MC?!
    }  # type: Dict[int, Tuple[str, ...]]

    sig_id_to_gen_id_mapping = {
        1: (1,),
        2: (2,),
        3: (3,),
        4: (4,),
        5: (51, 52, 53, 54, 55, 56, 57, 58),
        6: (51, 52, 53, 54, 55, 56, 57, 58),
    }  # type: Dict[int, Tuple[int, ...]]

    @classmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        b_mdst_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
        b_mdst_tau_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
        b1_gen_id_col = "GEN_B1_ID"
        b2_gen_id_col = "GEN_B2_ID"
        sig_id_col = "SIG_ID"
        required_cols = [b_mdst_col, b_mdst_tau_col, b1_gen_id_col, b2_gen_id_col, sig_id_col]
        assert all(c in cls._required_variables for c in required_cols), \
            [c for c in required_cols if c not in cls._required_variables]
        assert all([c in df.columns for c in required_cols]), [c for c in required_cols if c not in df.columns]
        assert all([c in cls.get_required_columns() for c in required_cols]), \
            [c for c in required_cols if c not in cls.get_required_columns()]
        original_n_cols = len(df.columns)

        temp_sig_gen_id_col = "temp_mva_target_sig_gen_id"
        temp_sig_b_mdst_index_col = "temp_mva_target_sig_b_mdst_index"
        temp_columns = [temp_sig_gen_id_col, temp_sig_b_mdst_index_col]

        target_column = cls.get_target_column_name()
        new_columns = [target_column] + temp_columns
        assert all([c not in df.columns for c in new_columns]), [c for c in new_columns if c in df.columns]

        cond = df[b_mdst_col].isin([1, 2])
        cond_tau = df[b_mdst_tau_col].isin([1, 2])

        df.loc[:, temp_sig_b_mdst_index_col] = np.nan
        df.loc[cond, temp_sig_b_mdst_index_col] = df[b_mdst_col]
        df.loc[cond_tau, temp_sig_b_mdst_index_col] = df[b_mdst_tau_col]
        df.loc[(~cond) & (~cond_tau), temp_sig_b_mdst_index_col] = 0
        assert not pd.isnull(df[temp_sig_b_mdst_index_col]).any(), temp_sig_b_mdst_index_col

        cond_b1 = (df[temp_sig_b_mdst_index_col] == 1)
        cond_b2 = (df[temp_sig_b_mdst_index_col] == 2)

        df.loc[:, temp_sig_gen_id_col] = np.nan
        df.loc[cond_b1, temp_sig_gen_id_col] = df[b1_gen_id_col]
        df.loc[cond_b2, temp_sig_gen_id_col] = df[b2_gen_id_col]
        df.loc[(~cond_b1) & (~cond_b2), temp_sig_gen_id_col] = 0
        assert not pd.isnull(df[temp_sig_gen_id_col]).any(), temp_sig_gen_id_col

        target_condition = pd.Series(np.full(shape=len(df.index), fill_value=False, dtype=bool))
        for sig_id, gen_ids in cls.sig_id_to_gen_id_mapping.items():
            target_condition |= (df[sig_id_col] == sig_id) & (df[temp_sig_gen_id_col].isin(gen_ids))

        assert len(target_condition.index) == len(df.index), (len(target_condition.index), len(df.index))
        df.loc[:, target_column] = np.nan
        df.loc[target_condition.values, target_column] = 1
        df.loc[~target_condition.values, target_column] = 0

        for col in temp_columns:
            del df[col]

        assert not pd.isnull(df[target_column]).any(), target_column
        assert len(df.columns) == (1 + original_n_cols), (original_n_cols, len(df.columns))

    @classmethod
    def _data_pre_processing(cls, df: pd.DataFrame) -> None:
        assert cls.e_extra.var not in df.columns, cls.e_extra.var
        assert cls.e_extra.var in cls._modified_variables.keys(), cls.e_extra.var

        original_e_extra_col = cls._modified_variables[cls.e_extra.var]
        assert original_e_extra_col in df.columns, original_e_extra_col

        df.loc[:, cls.e_extra.var] = df[original_e_extra_col]
        df.loc[df[original_e_extra_col] < 0.6, cls.e_extra.var] = 0.6


class DSSMVASetup(FirstSelectionMVASetup):
    selection_mva_name = "dss_first_sel_mva"  # type: str

    _fast_bdt_settings = {  # type: Dict[str, Any]
        "nTrees": 200,
        # "flatnessLoss": 2.0,
        # "numberOfFlatnessFeatures": 2
    }

    _all_mva_variable_infos = list(FirstSelectionMVASetup.all_mva_variable_infos)  # type: List[MVAVarInfo]

    # add the tag side cs vars
    _all_mva_variable_infos.extend(get_cs_variable_infos(signal_side=False))

    p_cms = MVAVarInfo(
        var="useCMSFrame__bop__bc",
        latex=r"$\mathrm{p}_\mathrm{CMS}$",
        bins=30, min=0., max=3., unit="GeV"
    )

    p_lep_cms = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo1__cm__spuseCMSFrame__bop__bc__bc__bc",
        latex=r"$\mathrm{p}^\ell_\mathrm{CMS}$",
        bins=30, min=0., max=3., unit="GeV"
    )

    pt_lep_cms = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo1__cm__spuseCMSFrame__bopt__bc__bc__bc",
        latex=r"$\mathrm{p}^\ell_\mathrm{t,CMS}$",
        bins=30, min=0., max=3., unit="GeV"
    )

    belle_ecl_energy = MVAVarInfo(
        var="belleECLEnergy",
        latex=r"$\mathrm{E}^\mathrm{ECL}_\mathrm{BELLE}$",
        bins=30, min=2., max=10., unit="GeV"
    )

    m_upsilon = MVAVarInfo(
        var="M",
        latex=r"$\mathrm{M}^\mathrm{\Upsilon(4S)}$",
        bins=30, min=7., max=11., unit="GeV"
    )

    q = MVAVarInfo(
        var="Q",
        latex=r"$\mathrm{Q}$",
        bins=30, min=0., max=1., unit="GeV"
    )

    q_b_sig = MVAVarInfo(
        var="daughter__bo1__cm__spQ__bc",
        latex=r"$\mathrm{Q}_\mathrm{B,sig}$",
        bins=30, min=0., max=4., unit="GeV"
    )

    cos_theta_l = MVAVarInfo(
        var="costhetalReco",
        latex=r"$\mathrm{cos(\Theta)}_\mathrm{\ell,reco}$",
        bins=30, min=-1., max=1., unit="GeV"
    )

    m2RecoilSignalSide_sig = MVAVarInfo(
        var="m2RecoilSignalSide_sig",
        latex=r"$\mathrm{M}_\mathrm{miss,sig}^2$",
        bins=30, min=-4., max=10., unit="GeV"
    )

    m2RecoilSignalSide_bkg = MVAVarInfo(
        var="m2RecoilSignalSide_bkg",
        latex=r"$\mathrm{M}_\mathrm{miss,bkg}^2$",
        bins=30, min=-4., max=10., unit="GeV"
    )

    additional_mva_infos = [  # type: List[MVAVarInfo]
        p_cms,
        p_lep_cms,
        pt_lep_cms,
        belle_ecl_energy,
        m_upsilon,
        q,
        q_b_sig,
        cos_theta_l,
        # m2RecoilSignalSide_sig,
        # m2RecoilSignalSide_bkg
    ]
    _all_mva_variable_infos.extend(additional_mva_infos)
    all_mva_variable_infos = tuple(_all_mva_variable_infos)

    # assert all_mva_variable_infos[-2] == m2RecoilSignalSide_sig, all_mva_variable_infos[-1]
    # assert all_mva_variable_infos[-1] == m2RecoilSignalSide_bkg, all_mva_variable_infos[-1]

    _required_variables = (
        "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc",
        "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
    )  # type: Tuple[str, ...]

    _modified_variables = {
        FirstSelectionMVASetup.e_extra.var: "ROE_eextra__boCleanROE__bc",
        # m2RecoilSignalSide_sig.var: "m2RecoilSignalSide",
        # m2RecoilSignalSide_bkg.var: "m2RecoilSignalSide"
    }  # type:  Dict[str, str]

    dss_pdgs = [  # type: List[int]
        10411,
        10421,
        10413,
        10423,
        20413,
        20423,
        415,
        425
    ]

    @classmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        b1_pdg_col = "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"
        b2_pdg_col = "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"
        b_mdst_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
        b_mdst_tau_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
        required_cols = [b_mdst_col, b_mdst_tau_col, b1_pdg_col, b2_pdg_col]
        assert all(c in cls._required_variables for c in required_cols), \
            [c for c in required_cols if c not in cls._required_variables]
        assert all([c in df.columns for c in required_cols]), [c for c in required_cols if c not in df.columns]
        assert all([c in cls.get_required_columns() for c in required_cols]), \
            [c for c in required_cols if c not in cls.get_required_columns()]
        original_n_cols = len(df.columns)

        temp_sig_b_mdst_index_col = "temp_mva_target_sig_b_mdst_index"
        temp_pdg_col = "temp_mva_target_sig_gen_id"
        temp_columns = [temp_sig_b_mdst_index_col, temp_pdg_col]

        target_column = cls.get_target_column_name()
        new_columns = [target_column] + temp_columns
        assert all([c not in df.columns for c in new_columns]), [c for c in new_columns if c in df.columns]

        cond = df[b_mdst_col].isin([1, 2])
        cond_tau = df[b_mdst_tau_col].isin([1, 2])

        df.loc[:, temp_sig_b_mdst_index_col] = np.nan
        df.loc[cond, temp_sig_b_mdst_index_col] = df[b_mdst_col]
        df.loc[cond_tau, temp_sig_b_mdst_index_col] = df[b_mdst_tau_col]
        df.loc[(~cond) & (~cond_tau), temp_sig_b_mdst_index_col] = 0
        assert not pd.isnull(df[temp_sig_b_mdst_index_col]).any(), temp_sig_b_mdst_index_col

        cond_b1 = (df[temp_sig_b_mdst_index_col] == 1)
        cond_b2 = (df[temp_sig_b_mdst_index_col] == 2)

        df.loc[:, temp_pdg_col] = np.nan
        df.loc[cond_b1, temp_pdg_col] = df[b1_pdg_col]
        df.loc[cond_b2, temp_pdg_col] = df[b2_pdg_col]
        df.loc[(~cond_b1) & (~cond_b2), temp_pdg_col] = 0
        assert not pd.isnull(df[temp_pdg_col]).any(), temp_pdg_col

        target_cond = abs(df[temp_pdg_col]).isin(cls.dss_pdgs)

        target_col = cls.get_target_column_name()
        df.loc[target_cond, target_col] = 1
        df.loc[(~target_cond), target_col] = 0

        for col in temp_columns:
            del df[col]

        assert not pd.isnull(df[target_column]).any(), target_column
        assert len(df.columns) == (1 + original_n_cols), (original_n_cols, len(df.columns))

    # @classmethod
    # def _data_pre_processing(cls, df: pd.DataFrame) -> None:
    #     assert cls.e_extra.var not in df.columns, cls.e_extra.var
    #     assert cls.e_extra.var in cls._modified_variables.keys(), cls.e_extra.var
    #
    #     original_e_extra_col = cls._modified_variables[cls.e_extra.var]
    #     assert original_e_extra_col in df.columns, original_e_extra_col
    #
    #     df.loc[:, cls.e_extra.var] = df[original_e_extra_col]
    #     df.loc[df[original_e_extra_col] < 0.6, cls.e_extra.var] = 0.6
    #
    #     assert cls.m2RecoilSignalSide_sig.var not in df.columns, cls.m2RecoilSignalSide_sig.var
    #     assert cls.m2RecoilSignalSide_sig.var in cls._modified_variables.keys(), cls.m2RecoilSignalSide_sig.var
    #     assert cls.m2RecoilSignalSide_bkg.var not in df.columns, cls.m2RecoilSignalSide_bkg.var
    #     assert cls.m2RecoilSignalSide_bkg.var in cls._modified_variables.keys(), cls.m2RecoilSignalSide_bkg.var
    #
    #     m2RecoilColumn = cls._modified_variables[cls.m2RecoilSignalSide_sig.var]
    #     assert m2RecoilColumn in df.columns, m2RecoilColumn
    #     df.loc[:, cls.m2RecoilSignalSide_sig.var] = df[m2RecoilColumn]
    #     df.loc[:, cls.m2RecoilSignalSide_bkg.var] = df[m2RecoilColumn]
    #
    #     df.loc[df[cls.get_target_column_name()] == 0, cls.m2RecoilSignalSide_sig.var] = np.nan
    #     df.loc[df[cls.get_target_column_name()] == 1, cls.m2RecoilSignalSide_bkg.var] = np.nan


class SecondSelectionMVASetup(FirstSelectionMVASetup):
    selection_mva_name = "sec_sel_mva"  # type: str

    _required_variables = (
        "SIG_ID",
    )  # type: Tuple[str, ...]

    _target_sig_ids = [1, 2, 3, 4]  # type: List[int]

    @classmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        sig_id_col = "SIG_ID"
        required_cols = [sig_id_col]
        assert all(c in cls._required_variables for c in required_cols), \
            [c for c in required_cols if c not in cls._required_variables]
        assert all([c in df.columns for c in required_cols]), [c for c in required_cols if c not in df.columns]
        assert all([c in cls.get_required_columns() for c in required_cols]), \
            [c for c in required_cols if c not in cls.get_required_columns()]
        original_n_cols = len(df.columns)

        target_cond = df[sig_id_col].isin(cls._target_sig_ids)

        target_col = cls.get_target_column_name()
        assert target_col not in df.columns, target_col
        df.loc[target_cond, target_col] = 1
        df.loc[(~target_cond), target_col] = 0

        assert not pd.isnull(df[target_col]).any(), target_col
        assert len(df.columns) == (1 + original_n_cols), (original_n_cols, len(df.columns))


class SecondSelectionDSSMVASetup(SecondSelectionMVASetup):
    selection_mva_name = "dss_sec_sel_mva"

    _target_sig_ids = [5, 6]  # type: List[int]


class MVASelectorSetups:
    available_mva_setups = {  # type: Dict[str, MVASelectorSetupBase]
        FirstSelectionMVASetup.selection_mva_name: FirstSelectionMVASetup,
        DSSMVASetup.selection_mva_name: DSSMVASetup,
        SecondSelectionMVASetup.selection_mva_name: SecondSelectionMVASetup,
        SecondSelectionDSSMVASetup.selection_mva_name: SecondSelectionDSSMVASetup
    }

    @staticmethod
    def get_available_selection_mva_setup_names() -> Tuple[str, ...]:
        return tuple([sel_mva_name for sel_mva_name in MVASelectorSetups.available_mva_setups.keys()])

    @staticmethod
    def get_selection_mva_setup(setup_name: str):
        if setup_name not in MVASelectorSetups.available_mva_setups:
            raise KeyError(f"A Selection MVA with the name {setup_name} is not available."
                           f"The options are:\n\t"
                           + "\n\t-".join(MVASelectorSetups.get_available_selection_mva_setup_names()))
        return MVASelectorSetups.available_mva_setups[setup_name]

# TODO: Other possible mva variables:
# TODO: Check if binning is correct for all variables.
# MVAVarInfo(var="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc", bins=25, min=0., max=1.,
#           latex=r"$\mathcal{P}_\mathrm{FEI}$")
# MVAVarInfo(var="daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc", bins=25, min=-0.2, max=0.8,
#           latex=r"$\Delta \mathrm{M}_{D^{(*)}}$")
# MVAVarInfo(var="daughter__bo1__cm__spchiProb__bc", bins=25, min=-1., max=1.,
#           latex=r"$\mathcal{P}_{\mathrm{B},\mathrm{sig},\mathrm{Vertex}}$")
# MVAVarInfo(var="ROE_neextra__boCleanROE__bc", bins=25, min=0., max=8.,
#           latex=r"$\mathrm{nE}^\mathrm{ECL}_\mathrm{extra}$")
# MVAVarInfo(var="daughter__bo0__cm__spdeltaE__bc", bins=25, min=-0.15, max=0.1,
#           latex=r"$\Delta\mathrm{E}_\mathrm{tag}$")
# MVAVarInfo(var="daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spdM__bc__bc__bc", bins=25, min=-0.2,
#           max=0.2, latex=r"$\Delta \mathrm{M}_\mathrm{D}$")
# MVAVarInfo(var="daughter__bo1__cm__spROE_neextra__boCleanROE__bc__bc", bins=25, min=0., max=10.0,
#           latex=r"$\mathrm{nE}^\mathrm{ECL}_\mathrm{extra,sig}$")
# MVAVarInfo(var="ROE_deltae__boCleanROE__bc", bins=25, min=0., max=10.0,
#           latex=r"$\Delta \mathrm{E}_\mathrm{\Upsilon(4S),ROE}$")
# MVAVarInfo(var="daughter__bo1__cm__spROE_deltae__boCleanROE__bc__bc", bins=25, min=0., max=10.0,
#           latex=r"$\Delta \mathrm{E}_\mathrm{sig,ROE}$")
# MVAVarInfo(var="daughter__bo1__cm__spROE_eextra__boCleanROE__bc__bc", bins=25, min=0., max=10.0,
#           latex=r"$\mathrm{E}^\mathrm{ECL}_\mathrm{extra,sig}$")
