"""
Methods to do MC cleansing, such as:
    - Removal of Goity Roberts processes from generic Belle MC
    - Removal of D(2S) and D*(2S) modes from generic Belle MC


Felix Metzner 2020
"""

import copy
import logging
import pandas as pd

from typing import Tuple

from rdstar.offline_analysis.mc_ids import lepton_pdg_ids, tau_pdg_id, valid_forth_daughter, pizero_pdg_id, \
    picharged_pdg_id, d_pdg_ids, dst_pdg_ids

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "MCSampleCompositionCorrection"
]


class MCSampleCompositionCorrection:
    d2s_pdg_ids = (  # type: Tuple[int]
        100411,  # D(2S)+
        100421,  # D(2S)0
        100413,  # D*(2S)+
        100423,  # D*(2S)0
    )
    lepton_pdg_ids = tuple(copy.copy(lepton_pdg_ids) + [tau_pdg_id])  # type: Tuple[int]
    neutrino_pdg_ids = tuple([lep_id + 1 for lep_id in lepton_pdg_ids])  # type: Tuple[int]
    d_and_ds_pdg_ids = tuple(d_pdg_ids + dst_pdg_ids)
    pion_pdg_ids = tuple([pizero_pdg_id, picharged_pdg_id])

    def __init__(
            self,
            remove_goity_roberts: bool = True,
            remove_d2s: bool = True
    ):
        self._remove_goity_roberts = remove_goity_roberts
        self._remove_d2s = remove_d2s

    def apply_sample_correction(self, df: pd.DataFrame, sample_name: str) -> int:
        if not self._remove_goity_roberts and not self._remove_d2s:
            return

        self._check_columns(df=df, sample_name=sample_name)

        original_n_events = len(df.index)
        n_goity_roberts_removed = 0
        n_d2s_removed = 0

        if self._remove_goity_roberts:
            n_goity_roberts_removed = self.remove_goity_roberts_modes(df=df)
            logging.info(f"Removed {n_goity_roberts_removed} goity robert modes from sample {sample_name}.")

        if self._remove_d2s:
            n_d2s_removed = self.remove_d2s_modes(df=df)
            logging.info(f"Removed {n_d2s_removed} D(*)(2S) modes from sample {sample_name}.")

        assert len(df.index) == original_n_events - n_goity_roberts_removed - n_d2s_removed, \
            (len(df.index), original_n_events - n_goity_roberts_removed - n_d2s_removed, original_n_events,
             n_goity_roberts_removed, n_d2s_removed)

        return n_d2s_removed + n_goity_roberts_removed

    def remove_goity_roberts_modes(self, df: pd.DataFrame) -> int:
        b1_mask = self.check_b_meson_for_goity_roberts(df=df, b_index=0)
        b2_mask = self.check_b_meson_for_goity_roberts(df=df, b_index=1)

        n_dropped = len(df[b1_mask | b2_mask].index)
        df.drop(index=df[b1_mask | b2_mask].index, inplace=True)
        return n_dropped

    def remove_d2s_modes(self, df: pd.DataFrame) -> int:
        b1_mask = self.check_b_meson_for_d2s(df=df, b_index=0)
        b2_mask = self.check_b_meson_for_d2s(df=df, b_index=1)

        n_dropped = len(df[b1_mask | b2_mask].index)
        df.drop(index=df[b1_mask | b2_mask].index, inplace=True)
        return n_dropped

    @staticmethod
    def get_nth_daughter_of_mth_b_pdg(daughter_index: int, b_index: int) -> str:
        return f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo{daughter_index}__cmPDG__bc__bc__bc"

    def check_b_meson_for_goity_roberts(self, df: pd.DataFrame, b_index: int) -> pd.Series:
        d_meson_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=0, b_index=b_index)
        has_d_or_ds_as_1st = abs(df[d_meson_pdg_col_name]).isin(self.d_and_ds_pdg_ids)

        pion_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=1, b_index=b_index)
        has_pion_as_2nd = abs(df[pion_pdg_col_name]).isin(self.pion_pdg_ids)

        lepton_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=2, b_index=b_index)
        has_lepton_as_3rd = abs(df[lepton_pdg_col_name]).isin(self.lepton_pdg_ids)

        neutrino_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=3, b_index=b_index)
        has_neutrino_as_4th = abs(df[neutrino_pdg_col_name]).isin(self.neutrino_pdg_ids)

        fifth_daughter_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=4, b_index=b_index)
        valid_fifth_daughter = valid_forth_daughter
        has_only_4_daughters = df[fifth_daughter_pdg_col_name].isin(valid_fifth_daughter)

        return has_d_or_ds_as_1st & has_pion_as_2nd & has_lepton_as_3rd & has_neutrino_as_4th & has_only_4_daughters

    def check_b_meson_for_d2s(self, df: pd.DataFrame, b_index: int) -> pd.Series:
        d_meson_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=0, b_index=b_index)
        has_d2s_as_1st = abs(df[d_meson_pdg_col_name]).isin(self.d2s_pdg_ids)

        lepton_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=1, b_index=b_index)
        has_lepton_as_2nd = abs(df[lepton_pdg_col_name]).isin(self.lepton_pdg_ids)

        neutrino_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=2, b_index=b_index)
        has_neutrino_as_3rd = abs(df[neutrino_pdg_col_name]).isin(self.neutrino_pdg_ids)

        forth_daughter_pdg_col_name = self.get_nth_daughter_of_mth_b_pdg(daughter_index=3, b_index=b_index)
        has_only_3_daughters = df[forth_daughter_pdg_col_name].isin(valid_forth_daughter)

        return has_d2s_as_1st & has_lepton_as_2nd & has_neutrino_as_3rd & has_only_3_daughters

    def _check_columns(self, df: pd.DataFrame, sample_name: str) -> None:
        required_column_name_list = [
            self.get_nth_daughter_of_mth_b_pdg(daughter_index=b_daughter_index, b_index=b_index)
            for b_index in range(2)
            for b_daughter_index in range(5)
        ]

        if not all(column_name in df.columns for column_name in required_column_name_list):
            raise KeyError(f"Missing columns in sample {sample_name}:\n\t"
                           f"{[c for c in required_column_name_list if c not in df.columns]}")
