"""
Convenience functions to add particle id correction weights.

Felix Metzner 2019
"""

import os
import copy
import pandas as pd
import numpy as np

from collections import defaultdict

from analysistools.pid_corrections import pid_corrector

from typing import Union, Optional, List, Tuple


class ParticleIDCorrection:
    pid_l_corr_col_name = "__weight_pid_corr_lepton_ratio__"
    pid_l_corr_var_prefix = "__weight_pid_corr_lepton_variation_"
    pid_pi_corr_col_name = "__weight_pid_corr_pi_ratio__"
    pid_pi_corr_var_prefix = "__weight_pid_corr_pi_variation_"
    pid_piFromK_corr_col_name = "__weight_pid_corr_piFromK_ratio__"
    pid_piFromK_corr_var_prefix = "__weight_pid_corr_piFromK_variation_"
    pid_k_corr_col_name = "__weight_pid_corr_k_ratio__"
    pid_k_corr_var_prefix = "__weight_pid_corr_k_variation_"
    pid_kFromPi_corr_col_name = "__weight_pid_corr_kFromPi_ratio__"
    pid_kFromPi_corr_var_prefix = "__weight_pid_corr_kFromPi_variation_"

    def __init__(self, number_of_variations: Optional[int] = None):
        self.n_variations = number_of_variations

        self.e_cut = 0.6
        self.mu_cut = 0.9
        self.k_cut = 0.1
        self.pi_cut = 0.1

        self.e_pdg = 11
        self.mu_pdg = 13
        self.k_pdg = 321
        self.pi_pdg = 211

        self.d_pdgs = [411, 421]
        self.dst_pdgs = [413, 423]
        self.dstst_pdgs = [415, 425, 10413, 10423, 10411, 10421, 20413, 20423]

        self.exp_num_name = "expNum"

        self.mc_pdg_name = "matchedMC__boPDG__bc"
        self.pdg_name = "PDG"
        self.momentum_name = "p"
        self.theta_name = "theta"

        self.collect_information = False
        self.information_table = None

    def apply_pid_correction(
            self,
            sample: pd.DataFrame,
            apply_for_lep_fakes: bool = True,
            apply_for_had_fakes: bool = True,
            info_output_path: Union[None, str, os.fspath] = None,
    ):
        if info_output_path is not None:
            self.collect_information = True

        self._apply_lepton_pid_corrections(sample=sample, apply_for_fakes=apply_for_lep_fakes)
        self._apply_hadron_pid_corrections(sample=sample, apply_for_fakes=apply_for_had_fakes)

        if info_output_path is not None:
            self._write_info_table(target_path=info_output_path)

    @staticmethod
    def _get_pid_weight_col_name(correction_name: str):
        return pid_corrector.PIDCorrection.get_weight_column_name(column_name=correction_name)

    def _add_correction(self, sample, condition, corrections, values, names, base_column_name):
        assert len(values) == 3, "values is expected to be a tuple containing (ratio, stat, sys) values."
        assert len(names) == 3, "names is expected to be a tuple containing the name strings for (ratio, stat, sys)."

        ratio, stat, sys = values
        ratio_name, stat_name, sys_name = names
        new_columns = [col for col in names]

        assert pd.isnull(sample.loc[condition, ratio_name]).all()
        sample.loc[condition, ratio_name] = ratio
        assert pd.isnull(sample.loc[condition, stat_name]).all()
        sample.loc[condition, stat_name] = stat
        assert pd.isnull(sample.loc[condition, sys_name]).all()
        sample.loc[condition, sys_name] = sys

        if corrections is None:
            if self.n_variations is not None:
                for var in np.arange(self.n_variations):
                    new_var_col_name = ParticleIDCorrection._get_pid_weight_col_name(
                        f"{base_column_name}_{pid_corrector.PIDCorrection.variation_name}_{var}"
                    )
                    new_columns.append(new_var_col_name)
                    if new_var_col_name not in sample.columns:
                        sample.loc[:, new_var_col_name] = np.nan
                    assert pd.isnull(sample.loc[condition, new_var_col_name]).all()
                    sample.loc[condition, new_var_col_name] = 1.0
                new_var_uncert_col_name = ParticleIDCorrection._get_pid_weight_col_name(
                    f"{base_column_name}_{pid_corrector.PIDCorrection.uncertainty_from_var_name}"
                )
                new_columns.append(new_var_uncert_col_name)
                if new_var_uncert_col_name not in sample.columns:
                    sample.loc[:, new_var_uncert_col_name] = np.nan
                assert pd.isnull(sample.loc[condition, new_var_uncert_col_name]).all()
                sample.loc[condition, new_var_uncert_col_name] = 0.0
        else:
            variation_cols_and_names = pid_corrector.PIDCorrection.get_variation_columns(columns=corrections.columns)
            if self.n_variations is None:
                assert len(variation_cols_and_names) == 0
            else:
                assert len(variation_cols_and_names) == (self.n_variations + 1)
            for var_col, name in variation_cols_and_names:
                new_var_col_name = ParticleIDCorrection._get_pid_weight_col_name(f"{base_column_name}_{name}")
                new_columns.append(new_var_col_name)
                if new_var_col_name not in sample.columns:
                    sample.loc[:, new_var_col_name] = np.nan

                assert pd.isnull(sample.loc[condition, new_var_col_name]).all()
                sample.loc[condition, new_var_col_name] = corrections.loc[:, var_col]

        assert all(not pd.isnull(sample.loc[condition, col_name]).any() for col_name in new_columns)

    @staticmethod
    def reset_pid_weights_with_error_flag(df, ratio_columns, flag_uncert_dict):
        for wc in ratio_columns:
            sys_col = wc.replace("ratio", "sys_uncertainty")
            stat_col = wc.replace("ratio", "stat_uncertainty")
            assert sys_col in df.columns
            assert stat_col in df.columns

            pid_type = wc.split(pid_corrector.PIDCorrection.get_weight_column_name_start())[1].split("_ratio")[0]

            hadron_flag_cond = df[wc] < 0
            lepton_fake_cond = (df[stat_col] < 0) & (df[wc] >= 0)

            # Handling uncertainties for Kaon and Pion pID Correction with flag != 0
            for type_hint in df.loc[hadron_flag_cond, stat_col].unique():
                df.loc[hadron_flag_cond & (df[stat_col] == type_hint), sys_col] = flag_uncert_dict[pid_type][type_hint]
            df.loc[hadron_flag_cond, wc] = 1.0
            df.loc[hadron_flag_cond, stat_col] = 0.

            # Handling uncert. for lepton from hadron fake: Uncertainty fully in sys for cases where stat = -1 or -2.
            df.loc[lepton_fake_cond, stat_col] = 0.

    @staticmethod
    def apply_pid_weight_finalization(df: pd.DataFrame, flag_uncert_dict) -> Tuple[str, List[str]]:
        wcs = [c for c in df.columns if c.startswith(pid_corrector.PIDCorrection.get_weight_column_name_start())]
        assert not pd.isnull(df[wcs]).any().any()

        weight_col = ParticleIDCorrection._get_pid_weight_col_name("nominal")
        assert weight_col not in df.columns
        df.loc[:, weight_col] = np.nan

        pid_weight_cols = [
            c for c in df.columns
            if c.startswith(pid_corrector.PIDCorrection.get_weight_column_name_start()) and c.endswith("ratio__")
        ]

        # Resetting weights with error flag information:
        ParticleIDCorrection.reset_pid_weights_with_error_flag(
            df=df,
            ratio_columns=pid_weight_cols,
            flag_uncert_dict=flag_uncert_dict
        )

        # Separating lepton types [e, mu, eFromPi, muFromPi, eFromK, muFromK] into several columns, to be able to vary
        # them independently from each other.
        lepton_types = ParticleIDCorrection.separate_lepton_weights(df=df)
        orig_l_col = ParticleIDCorrection._get_pid_weight_col_name(f"lepton_ratio")
        new_l_cols = [ParticleIDCorrection._get_pid_weight_col_name(f"{t}_ratio") for t in lepton_types]
        assert np.allclose(df[orig_l_col], df[new_l_cols].product(axis=1), rtol=1e-08, atol=1e-10)

        # Calculating the weight products for all hadron types.
        hadron_types = ParticleIDCorrection.combine_all_hadron_weights(df=df)
        orig_had_cols = [c for c in pid_weight_cols if c.startswith("__weight_pid_corr_dst_pi_")
                         or c.startswith("__weight_pid_corr_dd_")]
        new_had_cols = [ParticleIDCorrection._get_pid_weight_col_name(f"{t}_ratio") for t in hadron_types]
        assert np.allclose(df[new_had_cols].product(axis=1), df[orig_had_cols].product(axis=1), rtol=1e-08, atol=1e-10)

        # Calculating the product of all nominal weights (ratio):
        nom_types = lepton_types + hadron_types
        pid_nominal_weight_cols = [ParticleIDCorrection._get_pid_weight_col_name(f"{t}_ratio") for t in nom_types]
        df.loc[:, weight_col] = df[pid_nominal_weight_cols].product(axis=1)

        assert np.allclose(df[weight_col], df[pid_weight_cols].product(axis=1), rtol=1e-08, atol=1e-10), f"" \
            f"{df.loc[df[weight_col] != df[pid_weight_cols].product(axis=1), pid_weight_cols + [weight_col]][:10]}\n" \
            f"{df[pid_weight_cols].product(axis=1)[:10]}"

        # Checking all weight values again for NaNs:
        wcs = [c for c in df.columns if c.startswith(pid_corrector.PIDCorrection.get_weight_column_name_start())]
        assert not pd.isnull(df[wcs]).any().any()

        return weight_col, pid_nominal_weight_cols

    @staticmethod
    def separate_lepton_weights(df: pd.DataFrame, variations: int = 20) -> List[str]:
        e_pdg_code = 11
        mu_pdg_code = 13
        k_pdg_code = 321
        pi_pdg_code = 211

        lepton_type_dict = {
            "e": {"PDG": e_pdg_code, "mcPDG": e_pdg_code},
            "mu": {"PDG": mu_pdg_code, "mcPDG": mu_pdg_code},
            "eFromPi": {"PDG": e_pdg_code, "mcPDG": pi_pdg_code},
            "muFromPi": {"PDG": mu_pdg_code, "mcPDG": pi_pdg_code},
            "eFromK": {"PDG": e_pdg_code, "mcPDG": k_pdg_code},
            "muFromK": {"PDG": mu_pdg_code, "mcPDG": k_pdg_code}
        }

        pdg_col = ParticleIDCorrection.get_lepton_var(variable="PDG")
        mc_pdg_col = ParticleIDCorrection.get_lepton_var(variable="matchedMC__boPDG__bc")

        assert pdg_col in df.columns
        assert mc_pdg_col in df.columns

        for l_type, l_info in lepton_type_dict.items():
            df_dict = {"ratio": np.ones(len(df.index))}
            df_dict.update({f"variation_{i}": np.ones(len(df.index)) for i in range(variations)})
            separated_weight = pd.DataFrame(df_dict, index=df.index)

            cond = ((abs(df[pdg_col]) == l_info["PDG"]) & (abs(df[mc_pdg_col]) == l_info["mcPDG"]))

            for col in separated_weight.columns:
                col_name = ParticleIDCorrection._get_pid_weight_col_name(f"lepton_{col}")
                separated_weight.loc[df.loc[cond].index, col] = df.loc[cond, col_name]

            assert pd.notnull(separated_weight).all().all()

            for col in separated_weight.columns:
                new_col = ParticleIDCorrection._get_pid_weight_col_name(f"{l_type}_{col}")
                df.loc[:, new_col] = separated_weight[col]

        return list(lepton_type_dict.keys())

    @staticmethod
    def combine_all_hadron_weights(df: pd.DataFrame) -> List[str]:
        hadron_types = ["k", "kFromPi", "pi", "piFromK"]
        for hadron_type in hadron_types:
            hadron_pid_corr = ParticleIDCorrection.get_hadron_weight_product_for_type(
                df=df,
                hadron_type=hadron_type,
                variations=20
            )
            for col in hadron_pid_corr.columns:
                new_col = ParticleIDCorrection._get_pid_weight_col_name(f"{hadron_type}_{col}")
                df.loc[:, new_col] = hadron_pid_corr[col]

        return hadron_types

    @staticmethod
    def get_hadron_weight_product_for_type(df: pd.DataFrame, hadron_type: str, variations=20) -> pd.DataFrame:
        k_pdg_code = 321
        pi_pdg_code = 211
        d_pdg_codes = [411, 421]
        dst_pdg_codes = [413, 423]

        hadron_type_dict = {
            "k": {"PDG": k_pdg_code, "mcPDG": k_pdg_code, "range": [0, 1]},
            "kFromPi": {"PDG": k_pdg_code, "mcPDG": pi_pdg_code, "range": [0, 1]},
            "pi": {"PDG": pi_pdg_code, "mcPDG": pi_pdg_code, "range": [1, 2, 3, 4]},
            "piFromK": {"PDG": pi_pdg_code, "mcPDG": k_pdg_code, "range": [1, 2, 3, 4]},
        }
        assert hadron_type in hadron_type_dict.keys(), hadron_type

        d_pdg_name = f"daughter__bo1__cm__spdaughter__bo0__cm__spPDG__bc__bc"
        is_d_cond = abs(df[d_pdg_name]).isin(d_pdg_codes)
        is_dst_cond = abs(df[d_pdg_name]).isin(dst_pdg_codes)

        df_dict = {"ratio": np.ones(len(df.index))}
        df_dict.update({f"variation_{i}": np.ones(len(df.index)) for i in range(variations)})
        weight_product = pd.DataFrame(df_dict, index=df.index)

        for index in hadron_type_dict[hadron_type]["range"]:
            d_pdg_col = ParticleIDCorrection.get_d_daughter_var(daughter=index, variable="PDG")
            d_mc_pdg_col = ParticleIDCorrection.get_d_daughter_var(daughter=index, variable="matchedMC__boPDG__bc")

            dd_pdg_col = ParticleIDCorrection.get_dd_daughter_var(daughter=index, variable="PDG")
            dd_mc_pdg_col = ParticleIDCorrection.get_dd_daughter_var(daughter=index, variable="matchedMC__boPDG__bc")

            d_cond = (is_d_cond & (abs(df[d_pdg_col]) == hadron_type_dict[hadron_type]["PDG"])
                      & (abs(df[d_mc_pdg_col]) == hadron_type_dict[hadron_type]["mcPDG"]))
            dst_cond = (is_dst_cond & (abs(df[dd_pdg_col]) == hadron_type_dict[hadron_type]["PDG"])
                        & (abs(df[dd_mc_pdg_col]) == hadron_type_dict[hadron_type]["mcPDG"]))

            assert not (d_cond & dst_cond).any()

            for col in weight_product.columns:
                col_name = ParticleIDCorrection._get_pid_weight_col_name(f"dd_{index}_{col}")
                weight_product.loc[df.loc[d_cond].index, col] *= df.loc[d_cond, col_name]
                weight_product.loc[df.loc[dst_cond].index, col] *= df.loc[dst_cond, col_name]

        dst_pi_pdg_col = ParticleIDCorrection.get_dst_pi_var(variable="PDG")
        dst_pi_mc_pdg_col = ParticleIDCorrection.get_dst_pi_var(variable="matchedMC__boPDG__bc")

        dst_pi_cond = (is_dst_cond & (abs(df[dst_pi_pdg_col]) == hadron_type_dict[hadron_type]["PDG"])
                       & (abs(df[dst_pi_mc_pdg_col]) == hadron_type_dict[hadron_type]["mcPDG"]))

        for col in weight_product.columns:
            col_name = ParticleIDCorrection._get_pid_weight_col_name(f"dst_pi_{col}")
            weight_product.loc[df.loc[dst_pi_cond].index, col] *= df.loc[dst_pi_cond, col_name]

        assert pd.notnull(weight_product).all().all()
        return weight_product

    @staticmethod
    def get_hadron_pid_flag_uncert_dict(df):
        assert "weight_type" in df.columns
        assert "ratio" in df.columns
        assert "stat" in df.columns

        flag_uncertainty_dict = defaultdict(dict)

        weight_types = [w for w in df.weight_type.unique() if w.startswith("piID") or w.startswith("kID")]
        type_hints = [-1, -2, -3, -4]

        for weight_type in weight_types:
            type_hint_from_name = -1 if weight_type.startswith("k") else -2
            if "From" in weight_type:
                type_hint_from_name += -2
            if not weight_type.split("From")[0].split("ID")[1]:
                assert type_hint_from_name % 2 == 0
                key = "dst_pi"
            else:
                assert weight_type.split("From")[0].split("ID")[1].isdigit()
                key = f"dd_{weight_type.split('From')[0].split('ID')[1]}"

            mask = (df.weight_type == weight_type) & (df.ratio > 0)
            check_mask = (df.weight_type == weight_type) & (df.ratio < 0)
            type_hints_from_df = df.loc[check_mask, "stat"].unique()
            assert len(type_hints_from_df) == 1
            assert type_hints_from_df[0] == type_hint_from_name
            type_hint = type_hint_from_name
            assert type_hint in type_hints

            lower_quantile = df.loc[mask, "ratio"].quantile(0.16)
            upper_quantile = df.loc[mask, "ratio"].quantile(0.84)
            median = df.loc[mask, "ratio"].median()
            flag_uncertainty_dict[key][type_hint] = max(abs(median - lower_quantile), abs(median - upper_quantile))

        assert len(flag_uncertainty_dict) == 6
        for expected_key in ["dst_pi", "dd_0", "dd_1", "dd_2", "dd_3", "dd_4"]:
            assert expected_key in flag_uncertainty_dict.keys()

        return flag_uncertainty_dict

    @staticmethod
    def get_lepton_var(variable: str):
        return f"daughter__bo1__cm__spdaughter__bo1__cm__sp{variable}__bc__bc"

    @staticmethod
    def get_d_daughter_var(daughter: int, variable: str):
        return f"daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo{daughter}__cm__sp{variable}__bc__bc__bc"

    @staticmethod
    def get_dd_daughter_var(daughter: int, variable: str):
        return f"daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spdaughter" \
            f"__bo{daughter}__cm__sp{variable}__bc__bc__bc__bc"

    @staticmethod
    def get_dst_pi_var(variable: str):
        return ParticleIDCorrection.get_dd_daughter_var(daughter=1, variable=variable)

    def _add_to_info_table(self, weighted_sample, name, column_dict, base_column_name):
        if not self.collect_information:
            return

        assert "momentum" in column_dict.values()
        assert "theta" in column_dict.values()
        assert "ratio" in column_dict.values()
        assert "sys" in column_dict.values()
        assert "stat" in column_dict.values()

        rename_col_dict = copy.deepcopy(column_dict)

        if self.n_variations is not None:
            for i in range(self.n_variations):
                col_name = ParticleIDCorrection._get_pid_weight_col_name(
                    f"{base_column_name}_{pid_corrector.PIDCorrection.variation_name}_{i}"
                )
                rename_col_dict[col_name] = f"var_{i}"

            col_name = ParticleIDCorrection._get_pid_weight_col_name(
                f"{base_column_name}_{pid_corrector.PIDCorrection.uncertainty_from_var_name}"
            )
            rename_col_dict[col_name] = f"uncert_from_var"

        columns = list(rename_col_dict.keys()) + [self.exp_num_name, "unique_event_id", "unique_candidate_id"]

        info = weighted_sample[columns].copy(deep=True)
        info["weight_type"] = name
        info.rename(columns=rename_col_dict, inplace=True)

        assert not pd.isnull(info).any().any()

        if self.information_table is None:
            self.information_table = info
        else:
            assert len(self.information_table.columns) == len(info.columns)
            len_before = len(self.information_table.index) + len(info.index)
            self.information_table = self.information_table.append(info, ignore_index=True, sort=False)
            assert len_before == len(self.information_table.index)

    def _write_info_table(self, target_path):
        if self.information_table is None:
            self.information_table = pd.DataFrame()
        dir_path, file_name = os.path.split(target_path)
        temp_target = os.path.join(dir_path, f"temp_{file_name}")

        self.information_table.to_hdf(temp_target, "weight_info")
        os.rename(temp_target, target_path)

    def _apply_lepton_pid_corrections(self, sample: pd.DataFrame, apply_for_fakes: bool = True):
        base_column_name = "lepton"
        new_ratio_name = self._get_pid_weight_col_name(f"{base_column_name}_ratio")
        new_stat_name = self._get_pid_weight_col_name(f"{base_column_name}_stat_uncertainty")
        new_sys_name = self._get_pid_weight_col_name(f"{base_column_name}_sys_uncertainty")

        assert new_ratio_name not in sample.columns
        assert new_stat_name not in sample.columns
        assert new_sys_name not in sample.columns

        sample.loc[:, new_ratio_name] = np.nan
        sample.loc[:, new_stat_name] = np.nan
        sample.loc[:, new_sys_name] = np.nan

        column_dict = {
            self.get_lepton_var(self.momentum_name): "momentum",
            self.get_lepton_var(self.theta_name): "theta",
            new_ratio_name: "ratio",
            new_stat_name: "stat",
            new_sys_name: "sys"
        }

        ratio_name = pid_corrector.PIDCorrection.get_weight_column_name("ratio")
        sys_name = pid_corrector.PIDCorrection.get_weight_column_name("sys")
        sys1_name = pid_corrector.PIDCorrection.get_weight_column_name("sys1")
        sys2_name = pid_corrector.PIDCorrection.get_weight_column_name("sys2")
        stat_name = pid_corrector.PIDCorrection.get_weight_column_name("stat")
        uncert_name = pid_corrector.PIDCorrection.get_weight_column_name("uncertainty")

        e_condition = abs(sample[self.get_lepton_var(self.pdg_name)]) == self.e_pdg
        mu_condition = abs(sample[self.get_lepton_var(self.pdg_name)]) == self.mu_pdg

        e_sample = sample.loc[e_condition]
        mu_sample = sample.loc[mu_condition]

        assert len(sample.index) == (len(e_sample.index) + len(mu_sample.index))

        e_from_e_cond = abs(e_sample[self.get_lepton_var(self.mc_pdg_name)]) == self.e_pdg
        e_from_k_cond = abs(e_sample[self.get_lepton_var(self.mc_pdg_name)]) == self.k_pdg
        e_from_pi_cond = abs(e_sample[self.get_lepton_var(self.mc_pdg_name)]) == self.pi_pdg
        e_from_o_cond = ~(abs(e_sample[self.get_lepton_var(self.mc_pdg_name)]).isin(
            [self.e_pdg, self.pi_pdg, self.k_pdg]))

        mu_from_mu_cond = abs(mu_sample[self.get_lepton_var(self.mc_pdg_name)]) == self.mu_pdg
        mu_from_k_cond = abs(mu_sample[self.get_lepton_var(self.mc_pdg_name)]) == self.k_pdg
        mu_from_pi_cond = abs(mu_sample[self.get_lepton_var(self.mc_pdg_name)]) == self.pi_pdg
        mu_f_o_cond = ~(abs(mu_sample[self.get_lepton_var(self.mc_pdg_name)]).isin(
            [self.mu_pdg, self.pi_pdg, self.k_pdg]))

        e_from_e_sample = e_sample.loc[e_from_e_cond]
        e_from_k_sample = e_sample.loc[e_from_k_cond]
        e_from_pi_sample = e_sample.loc[e_from_pi_cond]
        e_from_other_sample = e_sample.loc[e_from_o_cond]

        assert len(e_sample.index) == (len(e_from_e_sample.index) + len(e_from_k_sample.index)
                                       + len(e_from_pi_sample.index) + len(e_from_other_sample.index))

        if len(e_from_e_sample.index) > 0:
            e_pid_corrector = pid_corrector.ElectronIDCorrection(pid_cut=0.6, variations=self.n_variations)
            e_pid_corr = e_pid_corrector.get_pid_correction(
                df=e_from_e_sample,
                df_mom_col=self.get_lepton_var(self.momentum_name),
                df_theta_col=self.get_lepton_var(self.theta_name),
                df_exp_col=self.exp_num_name,
                convert_rad_to_deg=True
            )

            e_pid_corr[sys_name] = np.sqrt(
                e_pid_corr[sys1_name] * e_pid_corr[sys1_name] + e_pid_corr[sys2_name] * e_pid_corr[sys2_name]
            )

            self._add_correction(
                sample=sample, condition=(e_condition & e_from_e_cond),
                corrections=e_pid_corr,
                values=(e_pid_corr[ratio_name], e_pid_corr[stat_name], e_pid_corr[sys_name]),
                names=(new_ratio_name, new_stat_name, new_sys_name),
                base_column_name=base_column_name
            )
            self._add_to_info_table(
                weighted_sample=sample.loc[e_condition & e_from_e_cond],
                name="eID",
                column_dict=column_dict,
                base_column_name=base_column_name
            )

        if apply_for_fakes:
            if len(e_from_k_sample.index) > 0:
                e_k_fake_corrector = pid_corrector.ElectronFromHadronFakeIDCorrection(
                    pid_cut=self.e_cut,
                    hadron_type='kaon',
                    variations=self.n_variations
                )
                e_k_fake_corr = e_k_fake_corrector.get_pid_correction(
                    df=e_from_k_sample,
                    df_mom_col=self.get_lepton_var(self.momentum_name),
                    df_theta_col=self.get_lepton_var(self.theta_name),
                    df_exp_col=self.exp_num_name,
                    convert_rad_to_deg=True
                )

                self._add_correction(
                    sample=sample, condition=(e_condition & e_from_k_cond),
                    corrections=e_k_fake_corr,
                    values=(e_k_fake_corr[ratio_name], -1., e_k_fake_corr[uncert_name]),
                    names=(new_ratio_name, new_stat_name, new_sys_name),
                    base_column_name=base_column_name
                )
                self._add_to_info_table(
                    weighted_sample=sample.loc[e_condition & e_from_k_cond],
                    name="eIDFromK",
                    column_dict=column_dict,
                    base_column_name=base_column_name
                )

            if len(e_from_pi_sample.index) > 0:
                e_pi_fake_corrector = pid_corrector.ElectronFromHadronFakeIDCorrection(
                    pid_cut=self.e_cut,
                    hadron_type='pion',
                    variations=self.n_variations
                )
                e_pi_fake_corr = e_pi_fake_corrector.get_pid_correction(
                    df=e_from_pi_sample,
                    df_mom_col=self.get_lepton_var(self.momentum_name),
                    df_theta_col=self.get_lepton_var(self.theta_name),
                    df_exp_col=self.exp_num_name,
                    convert_rad_to_deg=True
                )

                self._add_correction(
                    sample=sample, condition=(e_condition & e_from_pi_cond),
                    corrections=e_pi_fake_corr,
                    values=(e_pi_fake_corr[ratio_name], -2., e_pi_fake_corr[uncert_name]),
                    names=(new_ratio_name, new_stat_name, new_sys_name),
                    base_column_name=base_column_name
                )
                self._add_to_info_table(
                    weighted_sample=sample.loc[e_condition & e_from_pi_cond],
                    name="eIDFromPi",
                    column_dict=column_dict,
                    base_column_name=base_column_name
                )
        else:
            e_from_o_cond = e_from_o_cond | e_from_k_cond | e_from_pi_cond

        self._add_correction(
            sample=sample, condition=(e_condition & e_from_o_cond),
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

        mu_from_mu_sample = mu_sample.loc[mu_from_mu_cond]
        mu_from_k_sample = mu_sample.loc[mu_from_k_cond]
        mu_from_pi_sample = mu_sample.loc[mu_from_pi_cond]
        mu_from_other_sample = mu_sample.loc[mu_f_o_cond]

        assert len(mu_sample.index) == (len(mu_from_mu_sample.index) + len(mu_from_k_sample.index)
                                        + len(mu_from_pi_sample.index) + len(mu_from_other_sample.index))

        if len(mu_from_mu_sample.index) > 0:
            mu_pid_corrector = pid_corrector.MuonIDCorrection(pid_cut=0.9, variations=self.n_variations)
            mu_pid_corr = mu_pid_corrector.get_pid_correction(
                df=mu_from_mu_sample,
                df_mom_col=self.get_lepton_var(self.momentum_name),
                df_theta_col=self.get_lepton_var(self.theta_name),
                df_exp_col=self.exp_num_name,
                convert_rad_to_deg=True
            )

            mu_pid_corr[sys_name] = np.sqrt(
                mu_pid_corr[sys1_name] * mu_pid_corr[sys1_name] + mu_pid_corr[sys2_name] * mu_pid_corr[sys2_name]
            )

            self._add_correction(
                sample=sample, condition=(mu_condition & mu_from_mu_cond),
                corrections=mu_pid_corr,
                values=(mu_pid_corr[ratio_name], mu_pid_corr[stat_name], mu_pid_corr[sys_name]),
                names=(new_ratio_name, new_stat_name, new_sys_name),
                base_column_name=base_column_name
            )
            self._add_to_info_table(
                weighted_sample=sample.loc[mu_condition & mu_from_mu_cond],
                name="muID",
                column_dict=column_dict,
                base_column_name=base_column_name
            )

        if apply_for_fakes:
            if len(mu_from_k_sample.index) > 0:
                mu_k_fake_corrector = pid_corrector.MuonFromHadronFakeIDCorrection(
                    pid_cut=self.mu_cut,
                    hadron_type='kaon',
                    variations=self.n_variations
                )
                mu_k_fake_corr = mu_k_fake_corrector.get_pid_correction(
                    df=mu_from_k_sample,
                    df_mom_col=self.get_lepton_var(self.momentum_name),
                    df_theta_col=self.get_lepton_var(self.theta_name),
                    df_exp_col=self.exp_num_name,
                    convert_rad_to_deg=True
                )

                self._add_correction(
                    sample=sample, condition=(mu_condition & mu_from_k_cond),
                    corrections=mu_k_fake_corr,
                    values=(mu_k_fake_corr[ratio_name], -1., mu_k_fake_corr[uncert_name]),
                    names=(new_ratio_name, new_stat_name, new_sys_name),
                    base_column_name=base_column_name
                )
                self._add_to_info_table(
                    weighted_sample=sample.loc[mu_condition & mu_from_k_cond],
                    name="muIDFromK",
                    column_dict=column_dict,
                    base_column_name=base_column_name
                )

            if len(mu_from_pi_sample.index) > 0:
                mu_pi_fake_corrector = pid_corrector.MuonFromHadronFakeIDCorrection(
                    pid_cut=self.mu_cut,
                    hadron_type='pion',
                    variations=self.n_variations
                )
                mu_pi_fake_corr = mu_pi_fake_corrector.get_pid_correction(
                    df=mu_from_pi_sample,
                    df_mom_col=self.get_lepton_var(self.momentum_name),
                    df_theta_col=self.get_lepton_var(self.theta_name),
                    df_exp_col=self.exp_num_name,
                    convert_rad_to_deg=True
                )

                self._add_correction(
                    sample=sample, condition=(mu_condition & mu_from_pi_cond),
                    corrections=mu_pi_fake_corr,
                    values=(mu_pi_fake_corr[ratio_name], -2., mu_pi_fake_corr[uncert_name]),
                    names=(new_ratio_name, new_stat_name, new_sys_name),
                    base_column_name=base_column_name
                )
                self._add_to_info_table(
                    weighted_sample=sample.loc[mu_condition & mu_from_pi_cond],
                    name="muIDFromPi",
                    column_dict=column_dict,
                    base_column_name=base_column_name
                )
        else:
            mu_f_o_cond = mu_f_o_cond | mu_from_k_cond | mu_from_pi_cond

        self._add_correction(
            sample=sample, condition=(mu_condition & mu_f_o_cond),
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

        assert not pd.isnull(sample[new_ratio_name]).any()
        assert not pd.isnull(sample[new_stat_name]).any()
        assert not pd.isnull(sample[new_sys_name]).any()

    def _apply_hadron_pid_corrections(self, sample: pd.DataFrame, apply_for_fakes: bool = True):
        d_pdg_name = f"daughter__bo1__cm__spdaughter__bo0__cm__sp{self.pdg_name}__bc__bc"

        d_condition = abs(sample[d_pdg_name]).isin(self.d_pdgs)
        dst_condition = abs(sample[d_pdg_name]).isin(self.dst_pdgs)

        assert len(sample.index) == (len(sample.loc[d_condition].index) + len(sample.loc[dst_condition].index))

        for d_index in range(0, 5):
            base_column_name = f"dd_{d_index}"
            new_ratio_name = self._get_pid_weight_col_name(f"{base_column_name}_ratio")
            new_stat_name = self._get_pid_weight_col_name(f"{base_column_name}_stat_uncertainty")
            new_sys_name = self._get_pid_weight_col_name(f"{base_column_name}_sys_uncertainty")

            assert new_ratio_name not in sample.columns
            assert new_stat_name not in sample.columns
            assert new_sys_name not in sample.columns

            sample.loc[:, new_ratio_name] = np.nan
            sample.loc[:, new_stat_name] = np.nan
            sample.loc[:, new_sys_name] = np.nan

            self._apply_d_daughters_pid_corrections(
                sample=sample,
                condition=d_condition,
                daughter_index=d_index,
                d_star=False,
                base_column_name=base_column_name,
                new_ratio_name=new_ratio_name,
                new_stat_name=new_stat_name,
                new_sys_name=new_sys_name,
                apply_for_fakes=apply_for_fakes
            )

            self._apply_d_daughters_pid_corrections(
                sample=sample,
                condition=dst_condition,
                daughter_index=d_index,
                d_star=True,
                base_column_name=base_column_name,
                new_ratio_name=new_ratio_name,
                new_stat_name=new_stat_name,
                new_sys_name=new_sys_name,
                apply_for_fakes=apply_for_fakes
            )

            assert not pd.isnull(sample[new_ratio_name]).any()
            assert not pd.isnull(sample[new_stat_name]).any()
            assert not pd.isnull(sample[new_sys_name]).any()

        self._apply_dst_pion_pid_corrections(sample=sample, apply_for_fakes=apply_for_fakes)

    def _apply_d_daughters_pid_corrections(
            self,
            sample: pd.DataFrame,
            condition: pd.Series(bool),
            daughter_index: int,
            d_star: bool,
            base_column_name: str,
            new_ratio_name: str,
            new_stat_name: str,
            new_sys_name: str,
            apply_for_fakes: bool
    ):
        ratio_name = pid_corrector.PIDCorrection.get_weight_column_name("ratio")
        sys_name = pid_corrector.PIDCorrection.get_weight_column_name("sys")
        stat_name = pid_corrector.PIDCorrection.get_weight_column_name("stat")

        if not d_star:
            pdg_col_name = self.get_d_daughter_var(daughter=daughter_index, variable=self.pdg_name)
            mc_pdg_col_name = self.get_d_daughter_var(daughter=daughter_index, variable=self.mc_pdg_name)
            mom_col_name = self.get_d_daughter_var(daughter=daughter_index, variable=self.momentum_name)
            theta_col_name = self.get_d_daughter_var(daughter=daughter_index, variable=self.theta_name)
        else:
            pdg_col_name = self.get_dd_daughter_var(daughter=daughter_index, variable=self.pdg_name)
            mc_pdg_col_name = self.get_dd_daughter_var(daughter=daughter_index, variable=self.mc_pdg_name)
            mom_col_name = self.get_dd_daughter_var(daughter=daughter_index, variable=self.momentum_name)
            theta_col_name = self.get_dd_daughter_var(daughter=daughter_index, variable=self.theta_name)

        assert pdg_col_name in sample.columns
        assert mc_pdg_col_name in sample.columns
        assert mom_col_name in sample.columns
        assert theta_col_name in sample.columns

        column_dict = {
            mom_col_name: "momentum",
            theta_col_name: "theta",
            new_ratio_name: "ratio",
            new_stat_name: "stat",
            new_sys_name: "sys"
        }

        kaon_condition = condition & (abs(sample[pdg_col_name]) == self.k_pdg)
        pion_condition = condition & (abs(sample[pdg_col_name]) == self.pi_pdg)
        others_condition = condition & ~(kaon_condition | pion_condition)

        kaon_sample = sample.loc[kaon_condition]
        pion_sample = sample.loc[pion_condition]
        others_sample = sample.loc[others_condition]

        assert len(sample.loc[condition].index) == (len(kaon_sample.index)
                                                    + len(pion_sample.index) + len(others_sample.index))

        # NOT Correcting none Kaon or Pion D decay daughters
        self._add_correction(
            sample=sample, condition=others_condition,
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

        # Correcting Kaons from D decay
        k_from_k_cond = abs(kaon_sample[mc_pdg_col_name]) == self.k_pdg
        k_from_pi_cond = abs(kaon_sample[mc_pdg_col_name]) == self.pi_pdg
        k_from_o_cond = ~(abs(kaon_sample[mc_pdg_col_name]).isin([self.k_pdg, self.pi_pdg]))

        k_f_k_sample = kaon_sample.loc[k_from_k_cond]
        k_f_pi_sample = kaon_sample.loc[k_from_pi_cond]
        k_f_o_sample = kaon_sample.loc[k_from_o_cond]

        assert len(kaon_sample.index) == (len(k_f_k_sample.index) + len(k_f_pi_sample.index) + len(k_f_o_sample.index))

        if len(k_f_k_sample.index) > 0:
            k_pid_corrector = pid_corrector.KaonIDCorrection(pid_cut=self.k_cut, variations=self.n_variations)
            k_pid_corr = k_pid_corrector.get_pid_correction(
                df=k_f_k_sample,
                df_mom_col=mom_col_name,
                df_theta_col=theta_col_name,
                df_exp_col=self.exp_num_name,
                convert_rad_to_deg=True
            )

            self._add_correction(
                sample=sample, condition=(kaon_condition & k_from_k_cond),
                corrections=k_pid_corr,
                values=(k_pid_corr[ratio_name], k_pid_corr[stat_name], k_pid_corr[sys_name]),
                names=(new_ratio_name, new_stat_name, new_sys_name),
                base_column_name=base_column_name
            )
            self._add_to_info_table(
                weighted_sample=sample.loc[kaon_condition & k_from_k_cond],
                name=f"kID{daughter_index}",
                column_dict=column_dict,
                base_column_name=base_column_name
            )

        if apply_for_fakes:
            if len(k_f_pi_sample.index) > 0:
                k_pi_fake_corrector = pid_corrector.KaonFromPionFakeIDCorrection(
                    pid_cut=self.k_cut,
                    variations=self.n_variations
                )
                k_pi_fake_corr = k_pi_fake_corrector.get_pid_correction(
                    df=k_f_pi_sample,
                    df_mom_col=mom_col_name,
                    df_theta_col=theta_col_name,
                    df_exp_col=self.exp_num_name,
                    convert_rad_to_deg=True
                )

                self._add_correction(
                    sample=sample, condition=(kaon_condition & k_from_pi_cond),
                    corrections=k_pi_fake_corr,
                    values=(k_pi_fake_corr[ratio_name], k_pi_fake_corr[stat_name], k_pi_fake_corr[sys_name]),
                    names=(new_ratio_name, new_stat_name, new_sys_name),
                    base_column_name=base_column_name
                )
                self._add_to_info_table(
                    weighted_sample=sample.loc[kaon_condition & k_from_pi_cond],
                    name=f"kID{daughter_index}FromPi",
                    column_dict=column_dict,
                    base_column_name=base_column_name
                )
        else:
            k_from_o_cond = k_from_o_cond | k_from_pi_cond

        self._add_correction(
            sample=sample, condition=(kaon_condition & k_from_o_cond),
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

        # Correcting Pions from D decay
        pi_from_pi_cond = abs(pion_sample[mc_pdg_col_name]) == self.pi_pdg
        pi_from_k_cond = abs(pion_sample[mc_pdg_col_name]) == self.k_pdg
        pi_from_o_cond = ~(abs(pion_sample[mc_pdg_col_name]).isin([self.k_pdg, self.pi_pdg]))

        pi_f_pi_sample = pion_sample.loc[pi_from_pi_cond]
        pi_f_k_sample = pion_sample.loc[pi_from_k_cond]
        pi_f_o_sample = pion_sample.loc[pi_from_o_cond]

        assert len(pion_sample.index) == (len(pi_f_pi_sample.index) + len(pi_f_k_sample.index)
                                          + len(pi_f_o_sample.index))

        if len(pi_f_pi_sample.index) > 0:
            pi_pid_corrector = pid_corrector.PionIDCorrection(pid_cut=self.pi_cut, variations=self.n_variations)
            pi_pid_corr = pi_pid_corrector.get_pid_correction(
                df=pi_f_pi_sample,
                df_mom_col=mom_col_name,
                df_theta_col=theta_col_name,
                df_exp_col=self.exp_num_name,
                convert_rad_to_deg=True
            )

            self._add_correction(
                sample=sample, condition=(pion_condition & pi_from_pi_cond),
                corrections=pi_pid_corr,
                values=(pi_pid_corr[ratio_name], pi_pid_corr[stat_name], pi_pid_corr[sys_name]),
                names=(new_ratio_name, new_stat_name, new_sys_name),
                base_column_name=base_column_name
            )
            self._add_to_info_table(
                weighted_sample=sample.loc[pion_condition & pi_from_pi_cond],
                name=f"piID{daughter_index}",
                column_dict=column_dict,
                base_column_name=base_column_name
            )

        if apply_for_fakes:
            if len(pi_f_k_sample.index) > 0:
                pi_k_fake_corrector = pid_corrector.PionFromKaonFakeIDCorrection(
                    pid_cut=self.pi_cut,
                    variations=self.n_variations
                )
                pi_k_fake_corr = pi_k_fake_corrector.get_pid_correction(
                    df=pi_f_k_sample,
                    df_mom_col=mom_col_name,
                    df_theta_col=theta_col_name,
                    df_exp_col=self.exp_num_name,
                    convert_rad_to_deg=True
                )

                self._add_correction(
                    sample=sample, condition=(pion_condition & pi_from_k_cond),
                    corrections=pi_k_fake_corr,
                    values=(pi_k_fake_corr[ratio_name], pi_k_fake_corr[stat_name], pi_k_fake_corr[sys_name]),
                    names=(new_ratio_name, new_stat_name, new_sys_name),
                    base_column_name=base_column_name
                )
                self._add_to_info_table(
                    weighted_sample=sample.loc[pion_condition & pi_from_k_cond],
                    name=f"piID{daughter_index}FromK",
                    column_dict=column_dict,
                    base_column_name=base_column_name
                )
        else:
            pi_from_o_cond = pi_from_o_cond | pi_from_k_cond

        self._add_correction(
            sample=sample, condition=(pion_condition & pi_from_o_cond),
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

    def _apply_dst_pion_pid_corrections(self, sample: pd.DataFrame, apply_for_fakes: bool):
        base_column_name = "dst_pi"
        new_ratio_name = self._get_pid_weight_col_name(f"{base_column_name}_ratio")
        new_stat_name = self._get_pid_weight_col_name(f"{base_column_name}_stat_uncertainty")
        new_sys_name = self._get_pid_weight_col_name(f"{base_column_name}_sys_uncertainty")

        assert new_ratio_name not in sample.columns
        assert new_stat_name not in sample.columns
        assert new_sys_name not in sample.columns

        pdg_col_name = self.get_dst_pi_var(variable=self.pdg_name)
        mc_pdg_col_name = self.get_dst_pi_var(variable=self.mc_pdg_name)
        mom_col_name = self.get_dst_pi_var(variable=self.momentum_name)
        theta_col_name = self.get_dst_pi_var(variable=self.theta_name)

        assert pdg_col_name in sample.columns
        assert mc_pdg_col_name in sample.columns
        assert mom_col_name in sample.columns
        assert theta_col_name in sample.columns
        assert self.exp_num_name in sample.columns

        column_dict = {
            mom_col_name: "momentum",
            theta_col_name: "theta",
            new_ratio_name: "ratio",
            new_stat_name: "stat",
            new_sys_name: "sys"
        }

        sample.loc[:, new_ratio_name] = np.nan
        sample.loc[:, new_stat_name] = np.nan
        sample.loc[:, new_sys_name] = np.nan

        d_pdg_name = f"daughter__bo1__cm__spdaughter__bo0__cm__sp{self.pdg_name}__bc__bc"

        dst_condition = abs(sample[d_pdg_name]).isin(self.dst_pdgs)
        dst_sample = sample.loc[dst_condition]
        other_sample = sample.loc[~dst_condition]

        assert len(sample.index) == (len(dst_sample.index) + len(other_sample.index))

        # NOT Correcting none D* decays
        self._add_correction(
            sample=sample, condition=~dst_condition,
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

        # Correcting Charged Pion in D* -> D pi+/-
        ratio_name = pid_corrector.PIDCorrection.get_weight_column_name("ratio")
        sys_name = pid_corrector.PIDCorrection.get_weight_column_name("sys")
        stat_name = pid_corrector.PIDCorrection.get_weight_column_name("stat")

        pion_condition = abs(dst_sample[pdg_col_name]) == self.pi_pdg
        others_condition = abs(dst_sample[pdg_col_name]) != self.pi_pdg

        pion_sample = dst_sample.loc[pion_condition]
        others_sample = dst_sample.loc[others_condition]

        assert len(dst_sample.index) == (len(pion_sample.index) + len(others_sample.index))

        pi_f_pi_cond = abs(pion_sample[mc_pdg_col_name]) == self.pi_pdg
        pi_f_k_cond = abs(pion_sample[mc_pdg_col_name]) == self.k_pdg
        pi_f_o_cond = ~(abs(pion_sample[mc_pdg_col_name]).isin([self.pi_pdg, self.k_pdg]))

        pi_f_pi_sample = pion_sample.loc[pi_f_pi_cond]
        pi_f_k_sample = pion_sample.loc[pi_f_k_cond]
        pi_f_o_sample = pion_sample.loc[pi_f_o_cond]

        assert len(pion_sample.index) == (len(pi_f_pi_sample.index) + len(pi_f_k_sample.index)
                                          + len(pi_f_o_sample.index))

        if len(pi_f_pi_sample.index) > 0:
            pi_pid_corrector = pid_corrector.PionIDCorrection(pid_cut=self.pi_cut, variations=self.n_variations)
            pi_pid_corr = pi_pid_corrector.get_pid_correction(
                df=pi_f_pi_sample,
                df_mom_col=mom_col_name,
                df_theta_col=theta_col_name,
                df_exp_col=self.exp_num_name,
                convert_rad_to_deg=True
            )

            self._add_correction(
                sample=sample, condition=(dst_condition & pion_condition & pi_f_pi_cond),
                corrections=pi_pid_corr,
                values=(pi_pid_corr[ratio_name], pi_pid_corr[stat_name], pi_pid_corr[sys_name]),
                names=(new_ratio_name, new_stat_name, new_sys_name),
                base_column_name=base_column_name
            )
            self._add_to_info_table(
                weighted_sample=sample.loc[dst_condition & pion_condition & pi_f_pi_cond],
                name=f"piID",
                column_dict=column_dict,
                base_column_name=base_column_name
            )

        if apply_for_fakes:
            if len(pi_f_k_sample.index) > 0:
                pi_k_fake_corrector = pid_corrector.PionFromKaonFakeIDCorrection(
                    pid_cut=self.pi_cut,
                    variations=self.n_variations
                )
                pi_k_fake_corr = pi_k_fake_corrector.get_pid_correction(
                    df=pi_f_k_sample,
                    df_mom_col=mom_col_name,
                    df_theta_col=theta_col_name,
                    df_exp_col=self.exp_num_name,
                    convert_rad_to_deg=True
                )

                self._add_correction(
                    sample=sample, condition=(dst_condition & pion_condition & pi_f_k_cond),
                    corrections=pi_k_fake_corr,
                    values=(pi_k_fake_corr[ratio_name], pi_k_fake_corr[stat_name], pi_k_fake_corr[sys_name]),
                    names=(new_ratio_name, new_stat_name, new_sys_name),
                    base_column_name=base_column_name
                )
                self._add_to_info_table(
                    weighted_sample=sample.loc[dst_condition & pion_condition & pi_f_k_cond],
                    name=f"piIDFromK",
                    column_dict=column_dict,
                    base_column_name=base_column_name
                )
        else:
            pi_f_o_cond = pi_f_o_cond | pi_f_k_cond

        self._add_correction(
            sample=sample, condition=(dst_condition & pion_condition & pi_f_o_cond),
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

        # NOT Correcting other D* decays: D* -> D gamma or D* -> D pi0
        self._add_correction(
            sample=sample, condition=(dst_condition & others_condition),
            corrections=None,
            values=(1.0, 0.0, 0.0),
            names=(new_ratio_name, new_stat_name, new_sys_name),
            base_column_name=base_column_name
        )

        assert not pd.isnull(sample[new_ratio_name]).any()
        assert not pd.isnull(sample[new_stat_name]).any()
        assert not pd.isnull(sample[new_sys_name]).any()
