"""
Provides fit models based on the TemplateFitter package and tailored to this analysis.

Felix Metzner 2019
"""

import copy
import numpy as np
import pandas as pd

from collections import defaultdict
from typing import Optional, Tuple, List, Dict, NamedTuple

from templatefitter.fit_model.template import Template
from templatefitter.fit_model.model_builder import FitModel
from templatefitter.fit_model.parameter_handler import ParameterHandler

from templatefitter.plotter.histogram_variable import HistVariable

import rdstar.offline_analysis.histogram_variables as rdstar_hist_vars
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask
from rdstar.offline_analysis.mc_ids import sig_id_to_label, rdstar_sig_queries
from rdstar.offline_analysis.systematics import full_systematics_cols, get_systematics
from rdstar.offline_analysis.rdstar_plots import rdstar_plot_order, rdstar_dec_modes, get_rdstar_color_dict

__all__ = [
    "FitCase",
    "rdstar_fit_cases",
    "FitModelCreator"
]


class FitCase(NamedTuple):
    name: str
    base_dir_name: str
    sample_key: str
    output_tag: str
    plot_info_text: str
    asimov_data: bool
    round_asimov_data: bool
    hist_variables: Tuple[Optional[HistVariable], ...]
    binning_by_mode_dicts: Tuple[Optional[Dict[str, int]], ...]
    scope_by_mode_dicts: Tuple[Optional[Dict[str, Tuple[float, float]]], ...]


signal_fit_case = FitCase(
    name="signal_fit",
    base_dir_name="Fitting_Signal",
    sample_key="",
    output_tag="Fit_Signal_FixedNuisanceParams",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr, rdstar_hist_vars.pl_star),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None)
)

signal_asimov_fit_case = FitCase(
    name="signal_asimov",
    base_dir_name="Fitting_AsimovSignal",
    sample_key="",
    output_tag="AsimovFit_Signal_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr, rdstar_hist_vars.pl_star),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None)
)

q2_sideband_asimov_fit_case = FitCase(
    name="q2_sideband_asimov",
    base_dir_name="Fitting_AsimovQ2Sideband",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb, rdstar_hist_vars.pl_star),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None)
)

q2_sideband_fit_case = FitCase(
    name="q2_sideband_fit",
    base_dir_name="Fitting_Q2Sideband",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_FixedNuisanceParams",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb, rdstar_hist_vars.pl_star),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None)
)

rdstar_fit_cases = {  # type: Dict[str, FitCase]
    # signal_fit_case.name: signal_fit_case,  # TODO: Include for Box Opening
    signal_asimov_fit_case.name: signal_asimov_fit_case,
    q2_sideband_asimov_fit_case.name: q2_sideband_asimov_fit_case,
    q2_sideband_fit_case.name: q2_sideband_fit_case,
}


class FitModelCreator:
    default_weight_column_name = WeightFinalizationTask.final_weight_col  # type: str

    decay_mode_column_name = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str
    mc_signal_id_column_name = "SIG_ID"  # type: str

    fit_cases = rdstar_fit_cases  # type: Dict[str, FitCase]

    def __init__(
            self,
            fit_case: str,
            weight_column_name: Optional[str] = None
    ) -> None:
        if not fit_case in self.fit_cases.keys():
            raise ValueError(f"The parameter 'fit_case' must be one of {list(self.fit_cases.keys())}!\n"
                             f"You provided '{fit_case}'...")

        self._fit_case = fit_case

        if weight_column_name is None:
            self._weight_column_name = self.default_weight_column_name
        else:
            self._weight_column_name = weight_column_name

        self.param_handler = ParameterHandler()
        self.model = FitModel(parameter_handler=self.param_handler)

        self.channel_templates = defaultdict(list)  # Dict[str, List[str]]
        self.channel_eff_parameters = defaultdict(list)  # Dict[str, List[str]]

        self.initial_yield_dict = {}  # type: Dict[str, float]
        self.initial_eff_dict = defaultdict(dict)  # type: Dict[str, Dict[str, float]]

        all_mode_ids = [d_id for d_ids in rdstar_dec_modes.values() for d_id in d_ids]
        mode_infos = [(k, v) for k, v in rdstar_dec_modes.items()]
        self.mode_infos = copy.copy(mode_infos)  # type: List[Tuple[str, List[int]]]

        mode_infos.append(("all_modes", all_mode_ids))
        self.all_mode_ids = all_mode_ids  # type: List[int]
        self.mode_infos_with_all = mode_infos

    def setup_model(
            self,
            mc_df: pd.DataFrame,
            data_df: Optional[pd.DataFrame] = None,
            use_asimov_data: bool = False,
            round_asimov_data_bin_counts: bool = False
    ) -> Tuple[ParameterHandler, FitModel]:
        if not use_asimov_data:
            assert data_df is not None
            assert isinstance(data_df, pd.DataFrame), type(data_df)
        else:
            assert data_df is None

        self._get_initial_mc_yields(mc_df=mc_df)
        self._get_initial_mc_efficiencies(mc_df=mc_df)

        self._add_yield_parameters()
        self._add_templates_and_channels(mc_df=mc_df)

        if use_asimov_data:
            self.model.add_asimov_data_from_templates(round_bin_counts=round_asimov_data_bin_counts)
        else:
            self._add_data(data_df=data_df)

        self.model.finalize_model()

        return self.param_handler, self.model

    def _get_initial_mc_yields(self, mc_df: pd.DataFrame) -> None:
        assert len(self.initial_yield_dict) == 0, self.initial_yield_dict

        for component in rdstar_plot_order:
            reco_id = rdstar_sig_queries[component]["id"]
            sig_id_constraint = mc_df.SIG_ID == reco_id
            assert isinstance(sig_id_constraint, pd.Series)

            initial_yield = 0.
            for mode_name, mode_ids in self.mode_infos_with_all:
                variable_range_constraint = pd.Series(np.ones(len(mc_df.index), dtype=bool))
                scope_for_mode = self.get_fit_variable_scopes_for_mode(mode=mode_name)
                for col, scope in zip(self.fit_variable_column_names, scope_for_mode):
                    variable_range_constraint &= ((mc_df[col] >= scope[0]) & (mc_df[col] <= scope[1]))

                mode_constraint = mc_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(self.all_mode_ids)
                all_constraints = sig_id_constraint & mode_constraint & variable_range_constraint

                initial_yield += mc_df.loc[all_constraints, self.weight_col].sum()

            self.initial_yield_dict[component] = initial_yield

    def _get_initial_mc_efficiencies(self, mc_df: pd.DataFrame) -> None:
        assert len(self.initial_yield_dict) == len(rdstar_plot_order), \
            (self.initial_yield_dict, len(self.initial_yield_dict), len(rdstar_plot_order))
        assert len(self.initial_eff_dict) == 0, self.initial_eff_dict

        for component in rdstar_plot_order:
            reco_id = rdstar_sig_queries[component]["id"]
            sig_id_constraint = mc_df.SIG_ID == reco_id
            assert isinstance(sig_id_constraint, pd.Series)

            for mode_name, mode_ids in self.mode_infos_with_all:
                variable_range_constraint = pd.Series(np.ones(len(mc_df.index), dtype=bool))
                scope_for_mode = self.get_fit_variable_scopes_for_mode(mode=mode_name)
                for col, scope in zip(self.fit_variable_column_names, scope_for_mode):
                    variable_range_constraint &= ((mc_df[col] >= scope[0]) & (mc_df[col] <= scope[1]))

                mode_constraint = mc_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids)

                all_constraints = sig_id_constraint & mode_constraint & variable_range_constraint
                count_in_mode = mc_df.loc[all_constraints, self.weight_col].sum()

                self.initial_eff_dict[component][mode_name] = count_in_mode * 1. / self.initial_yield_dict[component]

    def _add_yield_parameters(self) -> None:
        for component in rdstar_plot_order:
            self.model.add_model_parameter(
                name=f"yield_{component}",
                parameter_type=ParameterHandler.yield_parameter_type,
                floating=True,
                initial_value=self.initial_yield_dict[component],
                constrain_to_value=None,
                constraint_sigma=None
            )

    def _add_templates_and_channels(self, mc_df: pd.DataFrame) -> None:
        for mode_name, mode_ids in self.mode_infos:
            decay_mode_condition = self._get_decay_mode_condition(df=mc_df, mode_ids=mode_ids)

            for component in rdstar_plot_order:
                reco_id = rdstar_sig_queries[component]["id"]
                component_condition = mc_df[self.mc_signal_id_column_name] == reco_id
                assert isinstance(component_condition, pd.Series)
                template_condition = decay_mode_condition & component_condition

                template = Template(
                    name=f"{mode_name}_{component}",
                    latex_label=sig_id_to_label[reco_id],
                    color=get_rdstar_color_dict()[reco_id],
                    process_name=component,
                    dimensions=self.fit_dimensions,
                    bins=self.get_fit_variable_binnings_for_mode(mode=mode_name),
                    scope=self.get_fit_variable_scopes_for_mode(mode=mode_name),
                    params=self.param_handler,
                    data_column_names=self.fit_variable_column_names,
                    data=mc_df.loc[template_condition],
                    weights=self._weight_column_name,
                    systematics=get_systematics(df=mc_df, sys_cols_infos=full_systematics_cols),
                    log_scale_mask=False
                )

                template_id = self.model.add_template(
                    template=template,
                    yield_parameter=f"yield_{component}",
                    use_other_systematics=True
                )
                self.channel_templates[mode_name].append(f"temp_{template_id}")

                eff_param_name = f"eff_{mode_name}_{component}"
                self.model.add_model_parameter(
                    name=eff_param_name,
                    parameter_type="efficiency",
                    floating=False,
                    initial_value=self.initial_eff_dict[component][mode_name],
                    constrain_to_value=self.initial_eff_dict[component][mode_name],
                    constraint_sigma=0.1
                )
                self.channel_eff_parameters[mode_name].append(eff_param_name)

            self.model.add_channel(
                efficiency_parameters=self.channel_eff_parameters[mode_name],
                name=mode_name,
                components=self.channel_templates[mode_name],
            )

    def _add_data(self, data_df: pd.DataFrame, weight_column_name: Optional[str] = None) -> None:
        data_channel_infos = {}
        data_channel_weight_infos = {}
        for mode_name, mode_ids in self.mode_infos:
            decay_mode_condition = self._get_decay_mode_condition(df=data_df, mode_ids=mode_ids)
            data_channel_infos.update({f"{mode_name}": data_df.loc[decay_mode_condition]})
            if weight_column_name is not None:
                assert weight_column_name in data_df.columns, weight_column_name
                data_channel_weight_infos.update({f"{mode_name}": weight_column_name})

        self.model.add_data(
            channels=data_channel_infos,
            channel_weights=data_channel_weight_infos if weight_column_name is not None else None
        )

    def _get_decay_mode_condition(self, df: pd.DataFrame, mode_ids: List[int]) -> pd.Series:
        return df[self.decay_mode_column_name].isin(mode_ids)

    @property
    def fit_model(self) -> FitModel:
        return self.model

    @property
    def fit_case(self) -> str:
        return self._fit_case

    @property
    def weight_col(self) -> str:
        return self._weight_column_name

    @property
    def fit_variables(self) -> Tuple[HistVariable, ...]:
        return self.fit_cases[self.fit_case].hist_variables

    @property
    def fit_dimensions(self) -> int:
        return len(self.fit_variables)

    @property
    def fit_variable_column_names(self) -> List[str]:
        return [h_var.df_label for h_var in self.fit_variables]

    def get_fit_variable_binnings_for_mode(self, mode: str) -> Tuple[int, ...]:
        mode_dependent_bins = self.fit_cases[self.fit_case].binning_by_mode_dicts
        return tuple([
            binning[mode] if binning is not None else h_var.n_bins
            for binning, h_var in zip(mode_dependent_bins, self.fit_variables)
        ])

    def get_fit_variable_scopes_for_mode(self, mode: str) -> Tuple[Tuple[float, float], ...]:
        mode_dependent_scopes = self.fit_cases[self.fit_case].scope_by_mode_dicts
        return tuple([
            scope[mode] if scope is not None else h_var.scope
            for scope, h_var in zip(mode_dependent_scopes, self.fit_variables)
        ])
