"""
Histogram Variables

This file provides the histogram variables used in the analysis.

Felix Metzner 2020
"""

from typing import Dict

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.offline_analysis.corrections.resolution_correction import MissingMassResolutionCorrector

__all__ = [
    # For Fitting
    "missing_mass_res_corr",
    "missing_mass_res_corr_q2sb",
    "missing_mass_binning_dict",
    "pl_star",
    # For Sideband Validation:
    "neextra_fine",
    "eextra",
    "cs_mva_output",
    "loose_cs_mva_output",
    "pl_star_SidebandValidation",
    "missing_mass_q2_SidebandValidation",
    "missing_mass_res_corr_q2sbSidebandValidation",
    "mbc_SidebandValidation",
    "tag_delta_e",
    "mbc_in_mbcsb_SidebandValidation",
    "mbc_in_mbcsb_SidebandValidation_rough_binning",
    "sig_vertex_fit_prob",
]

# Missing Mass Squared Definitions

missing_mass_res_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_corrected_col_name(original_column="m2RecoilSignalSide"),
    n_bins=48,
    scope=(-2., 10.),
    var_name=r"$\mathrm{M}_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_mass_res_corr_q2sb = HistVariable(
    df_label=MissingMassResolutionCorrector.get_corrected_col_name(original_column="m2RecoilSignalSide"),
    n_bins=25,
    scope=(-2., 4.),
    var_name=r"$\mathrm{M}_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_mass_binning_dict = {'DZ': 22, 'DM': 20, 'DSZ': 22, 'DSM': 17, 'all_modes': 22}  # type: Dict[str, int]

# Lepton Momentum

pl_star = HistVariable(
    df_label="pStarL",
    n_bins=20,
    scope=(0.1, 2.2),
    var_name=r"$\mathrm{p}_\ell^\ast$",
    unit=r"$\mathrm{GeV}$"
)

# Sideband Validation Variables

neextra_fine = HistVariable(
    df_label="ROE_neextra__boCleanROE__bc",
    n_bins=40,
    scope=(0., 2.5),
    var_name=r"$\mathrm{nE}^\mathrm{ECL}_\mathrm{extra}$",
    unit=r"$\mathrm{GeV}$"
)

eextra = HistVariable(
    df_label="ROE_eextra__boCleanROE__bc",
    n_bins=20,
    scope=(0., 3.),
    var_name=r"$\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
    unit=r"$\mathrm{GeV}$"
)

cs_mva_output = HistVariable(
    df_label="cs_mva_output",
    n_bins=40,
    scope=(0.5, 1.),
    var_name=r"$\mathcal{P}_\mathrm{CS}$",
    unit=None
)

loose_cs_mva_output = HistVariable(
    df_label="cs_mva_output",
    n_bins=40,
    scope=(0.2, 1.),
    var_name=r"$\mathcal{P}_\mathrm{CS}$",
    unit=None
)

sig_vertex_fit_prob = HistVariable(
    df_label="daughter__bo1__cm__spchiProb__bc",
    n_bins=40,
    scope=(0., 1.),
    var_name=r"$\mathcal{P}_\mathrm{B sig Vertex}$",
    unit=None
)

pl_star_SidebandValidation = HistVariable(
    df_label="pStarL",
    n_bins=25,
    scope=(0.1, 2.2),
    var_name=r"$\mathrm{p}_\ell^\ast$",
    unit=r"$\mathrm{GeV}$"
)

missing_mass_q2_SidebandValidation = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=25,
    scope=(-2., 4.),
    var_name=r"$\mathrm{M}_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_mass_res_corr_q2sbSidebandValidation = HistVariable(
    df_label=MissingMassResolutionCorrector.get_corrected_col_name(original_column="m2RecoilSignalSide"),
    n_bins=25,
    scope=(-2., 4.),
    var_name=r"$\mathrm{M}_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

mbc_SidebandValidation = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=25,
    scope=(5.27, 5.29),
    var_name=r"$\mathrm{M}_\mathrm{bc}^\mathrm{tag}$",
    unit="GeV"
)

tag_delta_e = HistVariable(
    df_label="daughter__bo0__cm__spdeltaE__bc",
    n_bins=30,
    scope=(-0.05, 0.1),
    var_name=r"$\Delta\mathrm{E}^\mathrm{tag}$",
    unit="GeV"
)

mbc_in_mbcsb_SidebandValidation = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=25,
    scope=(5.22, 5.27),
    var_name=r"$M_\mathrm{bc}$",
    unit="GeV"
)

mbc_in_mbcsb_SidebandValidation_rough_binning = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=15,
    scope=(5.22, 5.27),
    var_name=r"$M_\mathrm{bc}$",
    unit="GeV"
)

# Unused Variables
# Dstarstar Variables
missing_mass_dstarstar = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=25,
    scope=(-2.5, 8.),
    var_name=r"$M_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_mass_dstarstar_rough_binning = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=15,
    scope=(-2.5, 8.),
    var_name=r"$M_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_dstarstar_mass_ge_1p5 = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=25,
    scope=(1.5, 8.),
    var_name=r"$M_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_mass_dstarstar_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_corrected_col_name(original_column="m2RecoilSignalSide"),
    n_bins=25,
    scope=(-2.5, 8.),
    var_name=r"$M_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_mass_dstarstar_rough_binning_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_corrected_col_name(original_column="m2RecoilSignalSide"),
    n_bins=15,
    scope=(-2.5, 8.),
    var_name=r"$M_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

missing_dstarstar_mass_ge_1p5_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_corrected_col_name(original_column="m2RecoilSignalSide"),
    n_bins=25,
    scope=(1.5, 8.),
    var_name=r"$M_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$"
)

# Other
diff_dstar_d_mass = HistVariable(
    df_label="daughter__bo1__cm__spformula__bodaughterInvM__bo0__cm__sp2__bc__sp__mi__spdaughter__bo0__cm__spInvM__bc__bc__bc",
    n_bins=30,
    scope=(0.1, 1.0),
    var_name=r"$M(D^{(\ast)}\pi^0) - M(D^{(\ast)})$",
    unit=r"$\mathrm{GeV}$"
)

missing_mom_cms = HistVariable(
    df_label="missingMomentumInCMSFrame",
    n_bins=30,
    scope=(0., 1.5),
    var_name=r"$p_\mathrm{miss}^\ast$",
    unit="GeV"
)
