"""
Analysis tasks to be executed on n-tuple level.

Felix Metzner 2019
"""

import os
import yaml
import logging
import b2luigi as luigi

from collections import Iterable
from typing import Any, Union, Optional, List, Tuple, Dict, Generator

from analysistools.batch_tools.file_handler import load_json_config, Belle_url_mc_types

from rdstar.mva_selection.mva_selector_setups import MVASelectorSetups
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.combiner_tasks import SampleCombinerTask
from rdstar.offline_analysis.tasks.manager_tasks import SelectionManagerTask
from rdstar.offline_analysis.tasks.selection_validation_tasks import SelectionEvaluationTask

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
mpl_logger = logging.getLogger('matplotlib')
mpl_logger.setLevel(logging.WARNING)

__all__ = [
    "OfflineAnalysisMaster"
]


class OfflineAnalysisMaster(luigi.WrapperTask):
    """
    Entry function to run the whole offline analysis.
    """
    offline_config_paths = luigi.ListParameter(positional=False, hashed=True)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        base_offline_config = None  # type: Optional[Dict[str, Any]]
        self.base_offline_config_path = None  # type: Optional[Union[str, os.fspath]]
        self.offline_configs = []  # type: List[Tuple[Dict[str, Any], os.fspath]]

        assert isinstance(self.offline_config_paths, Iterable)
        for offline_config_path in self.offline_config_paths:
            with open(offline_config_path) as f:
                offline_config = yaml.load(f)

            self.offline_configs.append((offline_config, offline_config_path))

            if isinstance(offline_config["cs_config"], dict) and base_offline_config is None:
                base_offline_config = offline_config
                self.base_offline_config_path = offline_config_path
            elif isinstance(offline_config["cs_config"], dict) and base_offline_config is not None:
                raise RuntimeError("Exactly one offline config with a defined cs_config is required!")

        if base_offline_config is None:
            raise RuntimeError("Exactly one offline config with a defined cs_config is required!")

        self.working_dir = base_offline_config["working_dir"]

        self.cs_tag = base_offline_config["cs_config"]["cs_tag"]
        self.cs_fastbdt_settings = base_offline_config["cs_config"]["cs_fastbdt_settings"]

        self.selection_names = SelectionSetups.get_available_selection_names()
        self.tuple_type = "Y4S_ntuple"

    def get_log_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def get_task_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    @staticmethod
    def get_batch_configs(offline_config: Dict[str, Any]) -> List[Dict[str, Any]]:
        config_dir = utilities.get_config_directory()  # type: os.fspath
        base_path = os.path.join(config_dir, offline_config["production_tag"])  # type: os.fspath
        config_files = [os.path.join(base_path, c) for c in offline_config["batch_configs"]]  # type: List[os.fspath]
        return [load_json_config(c) for c in config_files]

    def get_params_from_batch_config(self, batch_config):
        assert self.tuple_type in batch_config["output"]

        for data_type in batch_config["data"]:
            if batch_config["data"][data_type]["streams"] is None:
                streams = [None]
            else:
                streams = batch_config["data"][data_type]["streams"]

            n_subsets = 1
            if data_type in Belle_url_mc_types:
                if data_type == "Mixed":
                    n_subsets = utilities.GlobalParameters().n_subsets_mixed
                else:
                    n_subsets = utilities.GlobalParameters().n_subsets_generic

            for stream in streams:
                yield {
                    "data_type": data_type,
                    "stream": stream if stream is not None else -999,
                    "n_subsets": n_subsets
                }

    def get_all_defined_configs(self) -> Generator[Dict[str, Any], None, None]:
        for offline_config, offline_config_path in self.offline_configs:
            for batch_config in self.get_batch_configs(offline_config):
                for batch_params in self.get_params_from_batch_config(batch_config):
                    yield {
                        "offline_config_path": offline_config_path,
                        "tuple_type": self.tuple_type,
                        "cs_tag": self.cs_tag,
                        "cs_fastbdt_settings": self.cs_fastbdt_settings,
                        **batch_params
                    }

    def get_data_type_config_combinations(self) -> List[Tuple[str, List[str], List[Dict[str, Any]]]]:
        data_type_combos = SampleCombinerTask.get_data_combo_dict()
        all_config_params = list(self.get_all_defined_configs())

        output = []
        for combo_name, data_types in data_type_combos.items():
            relevant_configs = []
            for data_type in data_types:
                relevant_configs.extend([c for c in all_config_params if c["data_type"] == data_type])

            if len(relevant_configs) > 0:
                output.append((combo_name, data_types, relevant_configs))

        return output

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield SelectionManagerTask(
            selection_setup_names=self.selection_names,
            config_combinations=self.get_data_type_config_combinations(),
            working_dir=self.working_dir,
            cs_tag=self.cs_tag
        )

        if len(self.selection_names) > 0:
            yield SelectionEvaluationTask(
                selection_setup_names=self.selection_names,
                mva_setup_name_for_truth=MVASelectorSetups.get_available_selection_mva_setup_names()[0],
                reference_selection=None,
                config_combinations=self.get_data_type_config_combinations(),
                working_dir=self.working_dir,
                cs_tag=self.cs_tag
            )


if __name__ == "__main__":
    config_base_path = utilities.get_config_directory()  # type: os.fspath

    offline_config_paths = [  # type: List[os.fspath]
        os.path.join(config_base_path, "prod_30thMar2020_stream0", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream1", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream2", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream3", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream4", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream5", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream6", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream7", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream8", "offline_ana_conf.yaml"),
        os.path.join(config_base_path, "prod_30thMar2020_stream9", "offline_ana_conf.yaml")
    ]

    num_workers = 10
    luigi.process(
        OfflineAnalysisMaster(offline_config_paths=offline_config_paths),
        workers=num_workers
    )
