"""
This module provides the plotting methods for the rdstar analysis.

Felix Metzner 2019
"""

import os
import numpy as np
import pandas as pd
from scipy.stats import chi2
import matplotlib.pyplot as plt

from typing import Any, Union, Optional, Tuple, List, Dict, Generator

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import StackedHistogramPlot
from templatefitter.plotter.histogram_plot_base import FigureType, AxesType

from rdstar.offline_analysis.mc_ids import sig_id_to_label

plot_style.set_matplotlibrc_params()

__all__ = [
    "rdstar_plot_order",
    "dss_dec_modes",
    "rdstar_dec_modes",
    "dss_dec_titles",
    "rdstar_dec_titles",
    "dss_mc_titles",
    "decay_mode_id_dict",
    "u4s_decay_mode_id_dict",
    "get_rdstar_color_dict",
    "plot_chi_squared_toys",
    "DecayModeDictType",
    "DecayTitleDictType"
]


DecayModeDictType = Dict[str, List[int]]
DecayTitleDictType = Dict[str, str]


rdstar_plot_order = ["CBKG", "GBKG", "HBKG", "DSSlbkg", "DSSl", "DSl", "Dl", "DStau", "Dtau"]  # type: List[str]


dss_dec_modes = {  # type: DecayModeDictType
    "DZPi": [63, 64],
    "DMPi": [61, 62],
    "DSZPi": [67, 68],
    "DSMPi": [65, 66],
}

rdstar_dec_modes = {  # type: DecayModeDictType
    "DZ": [13, 14],
    "DM": [11, 12],
    "DSZ": [17, 18],
    "DSM": [15, 16],
}

dss_dec_titles = {  # type: DecayTitleDictType
    "DZPi": r"$B^- \rightarrow D^0 \pi^0 \ell^-$",
    "DMPi": r"$B^0 \rightarrow D^- \pi^0 \ell^+$",
    "DSZPi": r"$B^- \rightarrow D^{*0} \pi^0 \ell^-$",
    "DSMPi": r"$B^0 \rightarrow D^{*-} \pi^0 \ell^+$",
}

rdstar_dec_titles = {  # type: DecayTitleDictType
    "DZ": r"$B^- \rightarrow D^0 \ell^-$",
    "DM": r"$B^0 \rightarrow D^- \ell^+$",
    "DSZ": r"$B^- \rightarrow D^{*0} \ell^-$",
    "DSM": r"$B^0 \rightarrow D^{*-} \ell^+$",
    "all_modes": "All Reconstruction Modes",
}

dss_mc_titles = {
    "Done": r"$B\rightarrow D_1 \ell \nu$",
    "Dtwostar": r"$B\rightarrow D_2^* \ell \nu$",
    "Dprimeone": r"$B\rightarrow D_1^\prime \ell \nu$",
    "Dzerostar": r"$B\rightarrow D_0^\ast \ell \nu$",
}


def decay_mode_id_dict() -> Dict[int, str]:
    return {  # type: Dict[int, str]
        11: r"$B^0_\mathrm{sig} \rightarrow D^- e^+$",
        12: r"$B^0_\mathrm{sig} \rightarrow D^- \mu^+$",
        13: r"$B^-_\mathrm{sig} \rightarrow D^0 e^-$",
        14: r"$B^-_\mathrm{sig} \rightarrow D^0 \mu^-$",
        15: r"$B^0_\mathrm{sig} \rightarrow D^{*-} e^+$",
        16: r"$B^0_\mathrm{sig} \rightarrow D^{*-} \mu^+$",
        17: r"$B^-_\mathrm{sig} \rightarrow D^{*0} e^-$",
        18: r"$B^-_\mathrm{sig} \rightarrow D^{*0} \mu^-$",

        61: r"$B^0_\mathrm{sig} \rightarrow D^- e^+ \pi^0$",
        62: r"$B^0_\mathrm{sig} \rightarrow D^- \mu^+ \pi^0$",
        63: r"$B^-_\mathrm{sig} \rightarrow D^0 e^- \pi^0$",
        64: r"$B^-_\mathrm{sig} \rightarrow D^0 \mu^- \pi^0$",
        65: r"$B^0_\mathrm{sig} \rightarrow D^{*-} e^+ \pi^0$",
        66: r"$B^0_\mathrm{sig} \rightarrow D^{*-} \mu^+ \pi^0$",
        67: r"$B^-_\mathrm{sig} \rightarrow D^{*0} e^- \pi^0$",
        68: r"$B^-_\mathrm{sig} \rightarrow D^{*0} \mu^- \pi^0$",

        71: r"$B^0_\mathrm{sig} \rightarrow \overline{D^0} e^+ \pi^-$",
        72: r"$B^0_\mathrm{sig} \rightarrow \overline{D^0} \mu^+ \pi^-$",
        73: r"$B^-_\mathrm{sig} \rightarrow D^+ e^- \pi^-$",
        74: r"$B^-_\mathrm{sig} \rightarrow D^+ \mu^- \pi^-$",
        75: r"$B^0_\mathrm{sig} \rightarrow \overline{D^{*0}} e^+ \pi^-$",
        76: r"$B^0_\mathrm{sig} \rightarrow \overline{D^{*0}} \mu^+ \pi^-$",
        77: r"$B^-_\mathrm{sig} \rightarrow D^{*+} e^- \pi^-$",
        78: r"$B^-_\mathrm{sig} \rightarrow D^{*+} \mu^- \pi^-$",
    }


def u4s_decay_mode_id_dict() -> Dict[int, str]:
    return {  # type: Dict[int, str]
        1001: r"$\Upsilon$(4S)$ \rightarrow B^+_\mathrm{tag} B^-_\mathrm{sig}$",
        1000: r"$\Upsilon$(4S)$ \rightarrow B^0_\mathrm{tag} \bar{B}^0_\mathrm{sig}$",
        1011: r"$\Upsilon$(4S)$ \rightarrow B^0_\mathrm{tag} B^-_\mathrm{sig}$",
        1021: r"$\Upsilon$(4S)$ \rightarrow B^0_\mathrm{tag} B^+_\mathrm{sig}$",
        1010: r"$\Upsilon$(4S)$ \rightarrow B^+_\mathrm{tag} B^0_\mathrm{sig}$",
        1020: r"$\Upsilon$(4S)$ \rightarrow B^-_\mathrm{tag} B^0_\mathrm{sig}$",
    }


# Define reconstructed decay modes to be combined:
def get_mode_dict() -> Tuple[Dict[str, List[int]], Dict[str, str]]:
    rdstar_mode_dict = {  # type: Dict[str, List[int]]
        "DZ": [13, 14],
        "DM": [11, 12],
        "DSZ": [17, 18],
        "DSM": [15, 16],
        "DZPi": [63, 64],
        "DMPi": [61, 62],
        "DSZPi": [67, 68],
        "DSMPi": [65, 66],
        "DZPiPlus": [71, 72],
        "DMPiPlus": [73, 74],
        "DSZPiPlus": [75, 76],
        "DSMPiPlus": [77, 78],
    }

    mode_to_title = {  # type: Dict[str, str]
        "DZ": r"$B^- \rightarrow D^0 \ell^-$",
        "DM": r"$B^0 \rightarrow D^- \ell^+$",
        "DSZ": r"$B^- \rightarrow D^{*0} \ell^-$",
        "DSM": r"$B^0 \rightarrow D^{*-} \ell^+$",
        "DZPi": r"$B^- \rightarrow D^0 \pi^0 \ell^-$",
        "DMPi": r"$B^0 \rightarrow D^- \pi^0 \ell^+$",
        "DSZPi": r"$B^- \rightarrow D^{*0} \pi^0 \ell^-$",
        "DSMPi": r"$B^0 \rightarrow D^{*-} \pi^0 \ell^+$",
        "DZPiPlus": r"$B^0 \rightarrow \overline{D^0} \ell^+ \pi^-$",
        "DMPiPlus": r"$B^- \rightarrow D^+ \ell^- \pi^-$",
        "DSZPiPlus": r"$B^0 \rightarrow \overline{D^{*0}} \ell^+ \pi^-$",
        "DSMPiPlus": r"$B^- \rightarrow D^{*+} \ell^- \pi^-$",
    }

    return rdstar_mode_dict, mode_to_title


def get_rdstar_color_dict() -> Dict[int, str]:
    color_dict = dict()
    color_dict[1] = plot_style.KITColors.kit_green
    color_dict[2] = plot_style.KITColors.kit_maygreen
    color_dict[3] = plot_style.KITColors.kit_blue
    color_dict[4] = plot_style.KITColors.kit_cyan
    color_dict[5] = plot_style.KITColors.kit_red
    color_dict[6] = plot_style.KITColors.kit_orange
    color_dict[7] = plot_style.KITColors.kit_brown
    color_dict[8] = plot_style.KITColors.grey
    color_dict[9] = plot_style.KITColors.light_grey

    assert len(color_dict.keys()) == len(sig_id_to_label.keys()), (len(color_dict.keys()), len(sig_id_to_label.keys()))
    assert set(color_dict.keys()) == set(sig_id_to_label.keys()), (color_dict.keys(), sig_id_to_label.keys())

    return color_dict


def plot_stacked_for_all_modes(
        df: pd.DataFrame,
        variable: HistVariable,
        weight_column: Optional[str] = None,
        cut_string: Optional[str] = None,
        y_lim_percentage: Optional[float] = None,
        output_name: Optional[str] = None,
        output_path: Optional[Union[str, os.fspath]] = None
) -> Generator[Tuple[FigureType, AxesType, Dict[str, Any]], None, None]:
    if weight_column is not None:
        assert isinstance(weight_column, str)
        assert weight_column in df.columns
    if (output_name is not None) and (output_path is None):
        raise ValueError("If a output file name is provided, the output_path must also be set!")

    mode_dict, mode_to_title = get_mode_dict()
    color_dict = get_rdstar_color_dict()

    for mode, mode_id in mode_dict.items():
        shp = StackedHistogramPlot(variable)
        for ID in sorted(list(df.SIG_ID.unique()), reverse=True):
            if cut_string:
                shp.add_component(
                    sig_id_to_label[ID],
                    df[df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_id)].query(
                        f"(SIG_ID == {ID}) and {cut_string}"),
                    weights=weight_column,
                    color=color_dict[ID]
                )
            else:
                shp.add_component(
                    sig_id_to_label[ID],
                    df[df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_id)].query(
                        f"SIG_ID == {ID}"),
                    weights=weight_column,
                    color=color_dict[ID]
                )

        fig = plt.figure(num=None, figsize=(6, 4.5), dpi=100, facecolor='w', edgecolor='k')
        ax = plt.gca()

        ax.set_title(mode_to_title[mode])

        shp.plot_on(ax=ax)

        if y_lim_percentage is not None:
            assert 0. < y_lim_percentage < 1.
            y_min, y_max = ax.get_ylim()
            ax.set_ylim(bottom=y_min, top=y_lim_percentage*y_max)

        if output_name:
            export(fig, f"{output_name}_{mode}", output_path)

        yield fig, ax, {"filename": f"{output_name}_{mode}", "target_dir": output_path}


def plot_chi_squared_toys(
        toys: np.ndarray,
        bin_edges: np.ndarray,
        ndf: int,
        observed_chi2: float,
        fig_size: Tuple[Union[int, float], Union[int, float]] = (5.5, 5.),
        full_output_path: Optional[Union[os.fspath, str]] = None,
        tag: Optional[str] = None
) -> Tuple[FigureType, AxesType]:
    fig, ax = plt.subplots(1, 1, figsize=fig_size, dpi=200)

    ax.hist(
        toys,
        bins=bin_edges,
        color=plot_style.KITColors.kit_blue,
        label=r"Toy $\chi^2$ Distribution",
        density=True
    )

    x_vals = np.linspace(start=bin_edges[0], stop=np.ceil(bin_edges[-1]), num=100)
    ax.plot(
        x_vals,
        chi2.pdf(x_vals, ndf),
        c=plot_style.KITColors.kit_red,
        ls="-",
        lw=2.,
        label=r"$\chi^2$ Distribution"
    )

    ax.axvline(
        x=observed_chi2,
        c=plot_style.KITColors.kit_orange,
        ls="--",
        lw=2.,
        label=rf"Observed $\chi^2$ = {observed_chi2:0.2f}"
    )

    ax.set_xlabel(r"$\chi^2$")
    ax.legend(frameon=False)

    if full_output_path is not None:
        output_path, base_name = os.path.split(full_output_path)
        output_name, _ = os.path.splitext(base_name)
        if tag is not None:
            output_name = f"{output_name}_{tag}"
        export(fig, output_name, output_path)

    return fig, ax
