"""
Weights for RDStar analysis samples

Felix Metzner 2019
"""

import math
import pandas as pd

from typing import Union, Optional, Dict

from analysistools.batch_tools.file_handler import Belle_url_mc_types, Belle_list_types, Belle_data_types
from analysistools.offline_analysis import sample_weights as sw


__all__ = [
    "scale_weight_column_name",
    "signal_mc_bfs",
    "signal_mc_evts",
    "add_scaling_weights"
]


BFSDType = Dict[str, Dict[str, Union[float, str, bool]]]
EvtNumDictType = Dict[str, Dict[str, int]]

scale_weight_column_name = sw.scale_weight_column_name  # type: str

# Some background information:
# - The D** branching fraction estimates are taken from Table V in https://arxiv.org/pdf/1606.09300.pdf
# - The D'_1 is also called D^*_1 sometimes.
# - The branching fraction (bf) of the modes involving `D_1` are smaller, because the sub modes `D_1 -> D pi pi`
#   where the D_1 decays into two pions is excluded, as this decay is listed separately in the gab modes GapD1Mixed
#   and GapD1Charged. Because the gab case combines the electron and muon decay, one has to consider the relation
#       BF(B0 -> D_1 l nu) == 0.5 * (BF(BZD1E) + BF(BZD1Mu) + BF(GapD1Mixed))
sig_mc_bfs = {
    # B0
    "DSME": {"bf": 0.05110, "charged_B": False, "decay_descr": "B0 -> D*- e+ nu_e"},
    "DME": {"bf": 0.02140, "charged_B": False, "decay_descr": "B0 -> D- e+ nu_e"},
    "BZD1E": {"bf": 0.00422, "charged_B": False, "decay_descr": "B0 -> D_1- e+ nu_e"},
    "BZD0sE": {"bf": 0.00362, "charged_B": False, "decay_descr": "B0 -> D_0*- e+ nu_e"},
    "BZDp1E": {"bf": 0.00401, "charged_B": False, "decay_descr": "B0 -> D'_1- e+ nu_e"},
    "BZD2sE": {"bf": 0.00347, "charged_B": False, "decay_descr": "B0 -> D_2*- e+ nu_e"},
    "DSMMu": {"bf": 0.05110, "charged_B": False, "decay_descr": "B0 -> D*- mu+ nu_mu"},
    "DMMu": {"bf": 0.02140, "charged_B": False, "decay_descr": "B0 -> D- mu+ nu_mu"},
    "BZD1Mu": {"bf": 0.00422, "charged_B": False, "decay_descr": "B0 -> D_1- mu+ nu_mu"},
    "BZD0sMu": {"bf": 0.00362, "charged_B": False, "decay_descr": "B0 -> D_0*- mu+ nu_mu"},
    "BZDp1Mu": {"bf": 0.00401, "charged_B": False, "decay_descr": "B0 -> D'_1- mu+ nu_mu"},
    "BZD2sMu": {"bf": 0.00347, "charged_B": False, "decay_descr": "B0 -> D_2*- mu+ nu_mu"},
    "DSMTau": {"bf": 0.0125416, "charged_B": False, "decay_descr": "B0 -> D*- tau+ nu_tau"},
    "DMTau": {"bf": 0.0063687, "charged_B": False, "decay_descr": "B0 -> D- tau+ nu_tau"},
    # "BZD1Tau": {"bf": 0.0013, "charged_B": False, "decay_descr": "B0 -> D_1- tau+ nu_tau"},  # Not considered
    # "BZD0sTau": {"bf": 0.0013, "charged_B": False, "decay_descr": "B0 -> D_0*- tau+ nu_tau"},  # Not considered
    # "BZDp1Tau": {"bf": 0.0020, "charged_B": False, "decay_descr": "B0 -> D'_1- tau+ nu_tau"},  # Not considered
    # "BZD2sTau": {"bf": 0.0020, "charged_B": False, "decay_descr": "B0 -> D_2*- tau+ nu_tau"},  # Not considered
    "GapD1Mixed": {"bf": 0.00564, "charged_B": False, "decay_descr": "B0 -> D_1- (-> D pi pi) l+ nu_l"},  # TODO: Update?
    "GapDEtaMixed": {"bf": 0.00868, "charged_B": False, "decay_descr": "B0 -> D- eta l+ nu_l"},  # TODO: Update!
    "GapDstEtaMixed": {"bf": 0.00868, "charged_B": False, "decay_descr": "B0 -> D*- eta l+ nu_l"},  # TODO: Update!
    "GapDpipiMixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D pi pi l+ nu_l"},  # TODO: Update!
    "GapDstpipiMixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D* pi pi l+ nu_l"},  # TODO: Update!
    "GapDEtaViaD0sMixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D_0* (-> D eta) l+ nu_l"},  # TODO: Update!
    "GapDstEtaViaDp1Mixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D'1 (-> D* eta) l+ nu_l"},  # TODO: Update!
    "GapDPiPiViaDp1Mixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D'1 (-> D pi pi) l+ nu_l"},  # TODO: Update!
    "GapDPiPiViaD0sMixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D_0s (-> D pi pi) l+ nu_l"},  # TODO: Update!
    "GapDstPiPiViaDp1Mixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D'1 (-> D* pi pi) l+ nu_l"},  # TODO: Update!
    "GapDstPiPiViaD0sMixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D_0s (-> D* pi pi) l+ nu_l"},  # TODO: Update!
    # B+
    "DSZE": {"bf": 0.05490, "charged_B": True, "decay_descr": "B+ -> anti-D*0 e+ nu_e"},
    "DZE": {"bf": 0.02310, "charged_B": True, "decay_descr": "B+ -> anti-D0 e+ nu_e"},
    "BMD1E": {"bf": 0.00454, "charged_B": True,  "decay_descr": "B+ -> anti-D_10 e+ nu_e"},
    "BMD0sE": {"bf": 0.00389, "charged_B": True,  "decay_descr": "B+ -> anti-D_0*0 e+ nu_e"},
    "BMDp1E": {"bf": 0.00431, "charged_B": True,  "decay_descr": "B+ -> anti-D'_10 e+ nu_e"},
    "BMD2sE": {"bf": 0.00373, "charged_B": True,  "decay_descr": "B+ -> anti-D_2*0 e+ nu_e"},
    "DSZMu": {"bf": 0.05490, "charged_B": True, "decay_descr": "B+ -> anti-D*0 mu+ nu_mu"},
    "DZMu": {"bf": 0.02310, "charged_B": True, "decay_descr": "B+ -> anti-D0 mu+ nu_mu"},
    "BMD1Mu": {"bf": 0.00454, "charged_B": True,  "decay_descr": "B+ -> anti-D_10 mu+ nu_mu"},
    "BMD0sMu": {"bf": 0.00389, "charged_B": True,  "decay_descr": "B+ -> anti-D_0*0 mu+ nu_mu"},
    "BMDp1Mu": {"bf": 0.00431, "charged_B": True,  "decay_descr": "B+ -> anti-D'_10 mu+ nu_mu"},
    "BMD2sMu": {"bf": 0.00373, "charged_B": True,  "decay_descr": "B+ -> anti-D_2*0 mu+ nu_mu"},
    "DSZTau": {"bf": 0.0134976, "charged_B": True, "decay_descr": "B+ -> anti-D*0 tau+ nu_tau"},
    "DZTau": {"bf": 0.0068527, "charged_B": True, "decay_descr": "B+ -> anti-D0 tau+ nu_tau"},
    # "BMD1Tau": {"bf": 0.0013, "charged_B": True, "decay_descr": "B+ -> anti-D_10 tau+ nu_tau"},  # Not considered
    # "BMD0sTau": {"bf": 0.0013, "charged_B": True, "decay_descr": "B+ -> anti-D_0*0 tau+ nu_tau"},  # Not considered
    # "BMDp1Tau": {"bf": 0.0020, "charged_B": True, "decay_descr": "B+ -> anti-D'_10 tau+ nu_tau"},  # Not considered
    # "BMD2sTau": {"bf": 0.0020, "charged_B": True, "decay_descr": "B+ -> anti-D_2*0 tau+ nu_tau"},  # Not considered
    "GapD1Charged": {"bf": 0.00607, "charged_B": True, "decay_descr": "B+ -> anti-D_10 (-> D pi pi) l+ nu_l"},  # TODO: Update?
    "GapDEtaCharged": {"bf": 0.00804, "charged_B": True, "decay_descr": "B+ -> anti-D0 eta l+ nu_l"},  # TODO: Update!
    "GapDstEtaCharged": {"bf": 0.00804, "charged_B": True, "decay_descr": "B+ -> anti-D*0 eta l+ nu_l"},  # TODO: Update!
    "GapDpipiCharged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D pi pi l+ nu_l"},  # TODO: Update!
    "GapDstpipiCharged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D* pi pi l+ nu_l"},  # TODO: Update!
    "GapDEtaViaD0sCharged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D_0* (-> D eta) l+ nu_l"},  # TODO: Update!
    "GapDstEtaViaDp1Charged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D'1 (-> D* eta) l+ nu_l"},  # TODO: Update!
    "GapDPiPiViaDp1Charged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D'1 (-> D pi pi) l+ nu_l"},  # TODO: Update!
    "GapDPiPiViaD0sCharged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D_0s (-> D pi pi) l+ nu_l"},  # TODO: Update!
    "GapDstPiPiViaDp1Charged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D'1 (-> D* pi pi) l+ nu_l"},  # TODO: Update!
    "GapDstPiPiViaD0sCharged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D_0s (-> D* pi pi) l+ nu_l"},  # TODO: Update!
}  # type: BFSDType


# Adding branching fraction infos also for efficiency and MVA training samples.
# For the efficiency samples of the B -> D** decays these are not correct, as we have separate
# efficiency samples for the charged and neutral pion cases, but this does not affect the
# efficiency calculation performed in the EfficiencyEvaluationTask!
Sig_MVA_Info = {f"{k}MVA": v for k, v in sig_mc_bfs.items() if k.startswith("D")}  # type: BFSDType

Dss_PiZero_Eff_Tmp = {f"{k}Eff": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_PiZero_MVA_Tmp = {f"{k}MVA": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_PiPlus_Eff_Tmp = {f"{k}ChargedPiEff": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_PiPlus_MVA_Tmp = {f"{k}ChargedPiMVA": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType


# Adapting branching fractions of charged and neutral pion D** decays:
def update_branching_fraction(bf_dict: BFSDType, bf_factor: float) -> BFSDType:
    assert all(isinstance(v["bf"], float) for v in bf_dict.values()), \
        [(k, v["bf"], type(v["bf"])) for k, v in bf_dict.items() if not isinstance(v["bf"], float)]
    assert all("bf" in v.keys() for v in bf_dict.values()), \
        [(k, v.keys()) for k, v in bf_dict.items() if "bf" not in v.keys()]
    new_dict = {k: {kk: vv * bf_factor if kk is "bf" else vv for kk, vv in v.items()} for k, v in bf_dict.items()}
    assert len(new_dict.keys()) == len(bf_dict.keys()), (len(new_dict.keys()), len(bf_dict.keys()))
    assert len(new_dict.keys()) == len(bf_dict.keys()), (len(new_dict.keys()), len(bf_dict.keys()))
    assert all((new_dict[k][vk] == v[vk] or vk is "bf") for k, v in bf_dict.items() for vk in v.keys())
    assert all(set(v.keys()) == set(bf_dict[list(bf_dict.keys())[0]].keys()) for v in new_dict.values())
    return new_dict


Dss_PiZero_Eff_Info = update_branching_fraction(bf_dict=Dss_PiZero_Eff_Tmp, bf_factor=float(1./3.))  # type: BFSDType
Dss_PiZero_MVA_Info = update_branching_fraction(bf_dict=Dss_PiZero_MVA_Tmp, bf_factor=float(1./3.))  # type: BFSDType
Dss_PiPlus_Eff_Info = update_branching_fraction(bf_dict=Dss_PiPlus_Eff_Tmp, bf_factor=float(2./3.))  # type: BFSDType
Dss_PiPlus_MVA_Info = update_branching_fraction(bf_dict=Dss_PiPlus_MVA_Tmp, bf_factor=float(2./3.))  # type: BFSDType
assert all(math.isclose(Dss_PiZero_Eff_Info[f"{k}Eff"]["bf"] + Dss_PiPlus_Eff_Info[f"{k}ChargedPiEff"]["bf"], v["bf"])
           for k, v in sig_mc_bfs.items() if k.startswith("B"))
assert all(math.isclose(Dss_PiZero_MVA_Info[f"{k}MVA"]["bf"] + Dss_PiPlus_MVA_Info[f"{k}ChargedPiMVA"]["bf"], v["bf"])
           for k, v in sig_mc_bfs.items() if k.startswith("B"))


assert all(k not in sig_mc_bfs for k in Sig_MVA_Info), [k for k in Sig_MVA_Info if k in sig_mc_bfs]
sig_mc_bfs.update(Sig_MVA_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiZero_Eff_Info), [k for k in Dss_PiZero_Eff_Info if k in sig_mc_bfs]
sig_mc_bfs.update(Dss_PiZero_Eff_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiZero_MVA_Info), [k for k in Dss_PiZero_MVA_Info if k in sig_mc_bfs]
sig_mc_bfs.update(Dss_PiZero_MVA_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiPlus_Eff_Info), [k for k in Sig_MVA_Info if k in Dss_PiPlus_Eff_Info]
sig_mc_bfs.update(Dss_PiPlus_Eff_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiPlus_MVA_Info), [k for k in Sig_MVA_Info if k in Dss_PiPlus_MVA_Info]
sig_mc_bfs.update(Dss_PiPlus_MVA_Info)
signal_mc_bfs = sig_mc_bfs  # type: BFSDType

#  Number of simulated events in private MC samples
signal_mc_evts = {
    # B0
    "DSME": {"prod": 2500000, "meta": 2500000, "summary": 2498371},
    "DSMEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498475},
    "DME": {"prod": 2500000, "meta": 2500000, "summary": 2498270},
    "DMEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498281},
    "BZD1E": {"prod": 72999988, "meta": 72999988, "summary": 72969720},
    "BZD0sE": {"prod": 72999988, "meta": 72999988, "summary": 72949473},
    "BZDp1E": {"prod": 72999988, "meta": 72999988, "summary": 72969660},
    "BZD2sE": {"prod": 72999988, "meta": 72999988, "summary": 72973031},
    "BZD1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2498797},
    "BZD0sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2498778},
    "BZDp1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2498744},
    "BZD2sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2498819},
    "BZD1EMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498796},
    "BZD0sEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498668},
    "BZDp1EMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498833},
    "BZD2sEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498786},
    "BZD1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499035},
    "BZDp1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499014},
    "BZD2sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499165},
    "BZD0sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499104},
    "BZD1EChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499074},
    "BZDp1EChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498957},
    "BZD2sEChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499218},
    "BZD0sEChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499121},
    "DSMMu": {"prod": 2500000, "meta": 2500000, "summary": 2495590},
    "DSMMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2495605},
    "DMMu": {"prod": 2500000, "meta": 2500000, "summary": 2496307},
    "DMMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2496123},
    "BZD1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72934855},
    "BZD0sMu": {"prod": 72999988, "meta": 72999988, "summary": 72939202},
    "BZDp1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72932851},
    "BZD2sMu": {"prod": 72999988, "meta": 72999988, "summary": 72943431},
    "BZD1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497267},
    "BZD0sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497331},
    "BZDp1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497329},
    "BZD2sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497514},
    "BZD1MuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497218},
    "BZD0sMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497377},
    "BZDp1MuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497272},
    "BZD2sMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497468},
    "BZD1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498031},
    "BZDp1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497905},
    "BZD2sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498408},
    "BZD0sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498136},
    "BZD1MuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497991},
    "BZDp1MuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497922},
    "BZD2sMuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498369},
    "BZD0sMuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498137},
    "DSMTau": {"prod": 2500000, "meta": 2500000, "summary": 2495585},
    "DSMTauMVA": {"prod": 2500000, "meta": 2500000, "summary": 2495535},
    "DMTau": {"prod": 2500000, "meta": 2500000, "summary": 2495428},
    "DMTauMVA": {"prod": 2500000, "meta": 2500000, "summary": 2495498},
    "GapD1Mixed": {"prod": 49999999, "meta": 49999999, "summary": 49954056},
    "GapDEtaMixed": {"prod": 39999994, "meta": 39999994, "summary": 39978590},
    "GapDstEtaMixed": {"prod": 39999994, "meta": 39999994, "summary": 39980161},
    "GapDpipiMixed": {"prod": 9499997, "meta": 9499997, "summary": 9497078},
    "GapDstpipiMixed": {"prod": 45999994, "meta": 45999994, "summary": 45986056},
    "GapDEtaViaD0sMixed": {"prod": 39999994, "meta": 39999994, "summary": 39970050},
    "GapDstEtaViaDp1Mixed": {"prod": 39999994, "meta": 39999994, "summary": 39973790},
    "GapDPiPiViaDp1Mixed": {"prod": 5999997, "meta": 5999997, "summary": 5995108},
    "GapDPiPiViaD0sMixed": {"prod": 5999997, "meta": 5999997, "summary": 5994354},
    "GapDstPiPiViaDp1Mixed": {"prod": 25000002, "meta": 25000002, "summary": 24977791},
    "GapDstPiPiViaD0sMixed": {"prod": 25000002, "meta": 25000002, "summary": 24974270},
    # B+
    "DSZE": {"prod": 2500000, "meta": 2500000, "summary": 2498902},
    "DSZEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499045},
    "DZE": {"prod": 2500000, "meta": 2500000, "summary": 2498891},
    "DZEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498861},
    "BMD1E": {"prod": 72999988, "meta": 72999988, "summary": 72972061},
    "BMD0sE": {"prod": 72999988, "meta": 72999988, "summary": 72970858},
    "BMDp1E": {"prod": 72999988, "meta": 72999988, "summary": 72965689},
    "BMD2sE": {"prod": 72999988, "meta": 72999988, "summary": 72975847},
    "BMD1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2499211},
    "BMD0sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2499129},
    "BMDp1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2499198},
    "BMD2sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2499248},
    "BMD1EMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499194},
    "BMD0sEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499146},
    "BMDp1EMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499263},
    "BMD2sEMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499264},
    "BMD1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499030},
    "BMDp1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498962},
    "BMD2sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499062},
    "BMD0sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498916},
    "BMD1EChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498913},
    "BMDp1EChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498921},
    "BMD2sEChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2499091},
    "BMD0sEChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498894},
    "DSZMu": {"prod": 2500000, "meta": 2500000, "summary": 2497236},
    "DSZMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497272},
    "DZMu": {"prod": 2500000, "meta": 2500000, "summary": 2497364},
    "DZMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497305},
    "BMD1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72939807},
    "BMD0sMu": {"prod": 72999988, "meta": 72999988, "summary": 72939605},
    "BMDp1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72889052},
    "BMD2sMu": {"prod": 72999988, "meta": 72999988, "summary": 72949175},
    "BMD1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498314},
    "BMD0sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498148},
    "BMDp1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498328},
    "BMD2sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498388},
    "BMD1MuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498340},
    "BMD0sMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498184},
    "BMDp1MuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498359},
    "BMD2sMuMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498384},
    "BMD1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497710},
    "BMDp1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497695},
    "BMD2sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498204},
    "BMD0sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497833},
    "BMD1MuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497759},
    "BMDp1MuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497635},
    "BMD2sMuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2498219},
    "BMD0sMuChargedPiMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497807},
    "DSZTau": {"prod": 2500000, "meta": 2500000, "summary": 2497350},
    "DSZTauMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497181},
    "DZTau": {"prod": 2500000, "meta": 2500000, "summary": 2497049},
    "DZTauMVA": {"prod": 2500000, "meta": 2500000, "summary": 2497120},
    "GapD1Charged": {"prod": 49999999, "meta": 49999999, "summary": 49965108},
    "GapDEtaCharged": {"prod": 39999994, "meta": 39999994, "summary": 39985913},
    "GapDstEtaCharged": {"prod": 39999994, "meta": 39999994, "summary": 39987282},
    "GapDpipiCharged": {"prod": 9499997, "meta": 9499997, "summary": 9497676},
    "GapDstpipiCharged": {"prod": 45999994, "meta": 45999994, "summary": 45989597},
    "GapDEtaViaD0sCharged": {"prod": 39999994, "meta": 39999994, "summary": 39980542},
    "GapDstEtaViaDp1Charged": {"prod": 39999994, "meta": 39999994, "summary": 39983032},
    "GapDPiPiViaDp1Charged": {"prod": 5999997, "meta": 5999997, "summary": 5996234},
    "GapDPiPiViaD0sCharged": {"prod": 5999997, "meta": 5999997, "summary": 5995530},
    "GapDstPiPiViaDp1Charged": {"prod": 25000002, "meta": 25000002, "summary": 24983424},
    "GapDstPiPiViaD0sCharged": {"prod": 25000002, "meta": 25000002, "summary": 24982046},
}  # type: EvtNumDictType


# Checking consistency of the event numbers in the signal_mc_evts dictionary:
assert all(v["prod"] == v["meta"] for v in signal_mc_evts.values()), \
    [(k, v["prod"], v["meta"]) for k, v in signal_mc_evts.items() if not v["prod"] == v["meta"]]
assert all(v["prod"] >= v["summary"] for v in signal_mc_evts.values()), \
    [(k, v["prod"], v["summary"]) for k, v in signal_mc_evts.items() if v["prod"] < v["summary"]]


# Checking consistency of the dictionaries signal_mc_evts and signal_mc_bfs:
assert set(signal_mc_evts.keys()) == set(signal_mc_bfs.keys())


def add_scaling_weights(
        df: pd.DataFrame,
        data_type: str,
        generic_stream_dict: Optional[Dict[str, int]] = None
) -> None:
    if data_type in Belle_url_mc_types:
        applied_weight = sw.add_public_mc_weight(
            df=df,
            streams=generic_stream_dict[data_type] if generic_stream_dict is not None else 1.
        )
    elif data_type in Belle_list_types:
        applied_weight = sw.add_public_mc_weight(
            df=df,
            streams=sw.FIXED_SPECIAL_WEIGHTS[data_type]
        )
    elif data_type in signal_mc_evts:
        n_events_source = "prod"
        applied_weight = sw.add_private_mc_weight(
            df=df,
            produced_events=signal_mc_evts[data_type][n_events_source],
            charged_B=signal_mc_bfs[data_type]["charged_B"],
            branching_fraction=signal_mc_bfs[data_type]["bf"]
        )
    elif data_type.startswith("B") and (data_type.endswith("Test") or data_type.endswith("Eff")):
        # No weighting for additional D** sample:
        applied_weight = 1.0
        df.loc[:, sw.scale_weight_column_name] = applied_weight
    elif data_type in Belle_data_types:
        # Also adding no scaling weight for actual data.
        applied_weight = 1.0
        df.loc[:, sw.scale_weight_column_name] = applied_weight
    else:
        raise RuntimeError(f"Data type {data_type} is not covered by add_scaling_weights!")

    assert applied_weight > 0.

    print(f"Applied scaling factor for {data_type:>10}: {applied_weight:2.4f}")


old_sig_mc_bfs = {
    # B0
    "OldDSME": {"bf": 0.05110, "charged_B": False, "decay_descr": "B0 -> D*- e+ nu_e"},
    "OldDME": {"bf": 0.02140, "charged_B": False, "decay_descr": "B0 -> D- e+ nu_e"},
    "OldBZD1E": {"bf": 0.00422, "charged_B": False, "decay_descr": "B0 -> D_1- e+ nu_e"},
    "OldBZD0sE": {"bf": 0.00362, "charged_B": False, "decay_descr": "B0 -> D_0*- e+ nu_e"},
    "OldBZDp1E": {"bf": 0.00401, "charged_B": False, "decay_descr": "B0 -> D'_1- e+ nu_e"},
    "OldBZD2sE": {"bf": 0.00347, "charged_B": False, "decay_descr": "B0 -> D_2*- e+ nu_e"},
    "OldDSMMu": {"bf": 0.05110, "charged_B": False, "decay_descr": "B0 -> D*- mu+ nu_mu"},
    "OldDMMu": {"bf": 0.02140, "charged_B": False, "decay_descr": "B0 -> D- mu+ nu_mu"},
    "OldBZD1Mu": {"bf": 0.00422, "charged_B": False, "decay_descr": "B0 -> D_1- mu+ nu_mu"},
    "OldBZD0sMu": {"bf": 0.00362, "charged_B": False, "decay_descr": "B0 -> D_0*- mu+ nu_mu"},
    "OldBZDp1Mu": {"bf": 0.00401, "charged_B": False, "decay_descr": "B0 -> D'_1- mu+ nu_mu"},
    "OldBZD2sMu": {"bf": 0.00347, "charged_B": False, "decay_descr": "B0 -> D_2*- mu+ nu_mu"},
    "OldDSMTau": {"bf": 0.0125416, "charged_B": False, "decay_descr": "B0 -> D*- tau+ nu_tau"},
    "OldDMTau": {"bf": 0.0063687, "charged_B": False, "decay_descr": "B0 -> D- tau+ nu_tau"},
    "OldGapD1Mixed": {"bf": 0.00564, "charged_B": False, "decay_descr": "B0 -> D_1- (-> D pi pi) l+ nu_l"},
    "OldGapDetaMixed": {"bf": 0.00868, "charged_B": False, "decay_descr": "B0 -> D(*)- eta l+ nu_l"},
    "OldGapDpipiMixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D(*) pi pi l+ nu_l"},
    # B+
    "OldDSZE": {"bf": 0.05490, "charged_B": True, "decay_descr": "B+ -> anti-D*0 e+ nu_e"},
    "OldDZE": {"bf": 0.02310, "charged_B": True, "decay_descr": "B+ -> anti-D0 e+ nu_e"},
    "OldBMD1E": {"bf": 0.00454, "charged_B": True,  "decay_descr": "B+ -> anti-D_10 e+ nu_e"},
    "OldBMD0sE": {"bf": 0.00389, "charged_B": True,  "decay_descr": "B+ -> anti-D_0*0 e+ nu_e"},
    "OldBMDp1E": {"bf": 0.00431, "charged_B": True,  "decay_descr": "B+ -> anti-D'_10 e+ nu_e"},
    "OldBMD2sE": {"bf": 0.00373, "charged_B": True,  "decay_descr": "B+ -> anti-D_2*0 e+ nu_e"},
    "OldDSZMu": {"bf": 0.05490, "charged_B": True, "decay_descr": "B+ -> anti-D*0 mu+ nu_mu"},
    "OldDZMu": {"bf": 0.02310, "charged_B": True, "decay_descr": "B+ -> anti-D0 mu+ nu_mu"},
    "OldBMD1Mu": {"bf": 0.00454, "charged_B": True,  "decay_descr": "B+ -> anti-D_10 mu+ nu_mu"},
    "OldBMD0sMu": {"bf": 0.00389, "charged_B": True,  "decay_descr": "B+ -> anti-D_0*0 mu+ nu_mu"},
    "OldBMDp1Mu": {"bf": 0.00431, "charged_B": True,  "decay_descr": "B+ -> anti-D'_10 mu+ nu_mu"},
    "OldBMD2sMu": {"bf": 0.00373, "charged_B": True,  "decay_descr": "B+ -> anti-D_2*0 mu+ nu_mu"},
    "OldDSZTau": {"bf": 0.0134976, "charged_B": True, "decay_descr": "B+ -> anti-D*0 tau+ nu_tau"},
    "OldDZTau": {"bf": 0.0068527, "charged_B": True, "decay_descr": "B+ -> anti-D0 tau+ nu_tau"},
    "OldGapD1Charged": {"bf": 0.00607, "charged_B": True, "decay_descr": "B+ -> anti-D_10 (-> D pi pi) l+ nu_l"},
    "OldGapDetaCharged": {"bf": 0.00804, "charged_B": True, "decay_descr": "B+ -> anti-D(*)0 eta l+ nu_l"},
    "OldGapDpipiCharged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D(*) pi pi l+ nu_l"},
}  # type: BFSDType

# Adding branching fraction infos also for test and efficiency samples.
# For the efficiency samples of the B -> D** decays these are not correct, as we have separate
# efficiency samples for the charged and neutral pion cases, but this does not affect the
# efficiency calculation performed in the EfficiencyEvaluationTask!
OldDss_Test_Info = {f"{k}Test": v for k, v in old_sig_mc_bfs.items() if k.startswith("OldB")}  # type: BFSDType
OldDss_Eff_Info = {f"{k}Eff": v for k, v in old_sig_mc_bfs.items() if k.startswith("OldB")}  # type: BFSDType
ODss_CPiEff_Info = {f"{k}ChargedPiEff": v for k, v in old_sig_mc_bfs.items() if k.startswith("OldB")}  # type: BFSDType
old_sig_mc_bfs.update(OldDss_Test_Info)
old_sig_mc_bfs.update(OldDss_Eff_Info)
old_sig_mc_bfs.update(ODss_CPiEff_Info)
old_signal_mc_bfs = old_sig_mc_bfs

old_signal_mc_evts = {
    # B0
    "OldDSME": {"prod": 2500000, "meta": 2500000, "summary": 2498381},
    "OldDME": {"prod": 2500000, "meta": 2500000, "summary": 2498259},
    "OldBZD1E": {"prod": 72999988, "meta": 72955090, "summary": 72970488},
    "OldBZD0sE": {"prod": 72999988, "meta": 72999988, "summary": 72972981},
    "OldBZDp1E": {"prod": 72999988, "meta": 72999988, "summary": 72971014},
    "OldBZD2sE": {"prod": 72899988, "meta": 72999988, "summary": 72972825},
    "OldBZD1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499020},
    "OldBZD0sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499081},
    "OldBZDp1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499004},
    "OldBZD2sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499058},
    "OldBZD1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2498775},
    "OldBZD0sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2498849},
    "OldBZDp1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2498816},
    "OldBZD2sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2498877},
    "OldBZD1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499099},
    "OldBZDp1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499081},
    "OldBZD2sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499205},
    "OldBZD0sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499172},
    "OldDSMMu": {"prod": 2500000, "meta": 2500000, "summary": 2495652},
    "OldDMMu": {"prod": 2500000, "meta": 2500000, "summary": 2496318},
    "OldBZD1Mu": {"prod": 72999988, "meta": 72928778, "summary": 72935520},
    "OldBZD0sMu": {"prod": 72999988, "meta": 72999988, "summary": 72945960},
    "OldBZDp1Mu": {"prod": 72899988, "meta": 72999988, "summary": 72937478},
    "OldBZD2sMu": {"prod": 72999988, "meta": 72999988, "summary": 72885254},
    "OldBZD1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497746},
    "OldBZD0sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498095},
    "OldBZDp1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497821},
    "OldBZD2sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498115},
    "OldBZD1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497335},
    "OldBZD0sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497482},
    "OldBZDp1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497318},
    "OldBZD2sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497486},
    "OldBZD1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497982},
    "OldBZDp1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498097},
    "OldBZD2sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498273},
    "OldBZD0sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498428},
    "OldDSMTau": {"prod": 2500000, "meta": 2500000, "summary": 2495647},
    "OldDMTau": {"prod": 2500000, "meta": 2500000, "summary": 2495568},
    "OldGapD1Mixed": {"prod": 9999998, "meta": 9999998, "summary": 9990965},
    "OldGapDetaMixed": {"prod": 12059997, "meta": 12059997, "summary": 12053835},
    "OldGapDpipiMixed": {"prod": 9479999, "meta": 9479999, "summary": 9477142},
    # B+
    "OldDSZE": {"prod": 2500000, "meta": 2500000, "summary": 2499051},
    "OldDZE": {"prod": 2500000, "meta": 2500000, "summary": 2498851},
    "OldBMD1E": {"prod": 72999988, "meta": 72999988, "summary": 72972451},
    "OldBMD0sE": {"prod": 72999988, "meta": 72999988, "summary": 72976112},
    "OldBMDp1E": {"prod": 72999988, "meta": 72999988, "summary": 72973658},
    "OldBMD2sE": {"prod": 72999988, "meta": 72986680, "summary": 72976436},
    "OldBMD1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499020},
    "OldBMD0sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499081},
    "OldBMDp1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499004},
    "OldBMD2sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499195},
    "OldBMD1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2499285},
    "OldBMD0sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2499248},
    "OldBMDp1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2499208},
    "OldBMD2sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2499272},
    "OldBMD1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498942},
    "OldBMDp1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499014},
    "OldBMD2sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499063},
    "OldBMD0sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499165},
    "OldDSZMu": {"prod": 2500000, "meta": 2500000, "summary": 2497271},
    "OldDZMu": {"prod": 2500000, "meta": 2500000, "summary": 2497540},
    "OldBMD1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72941370},
    "OldBMD0sMu": {"prod": 72899988, "meta": 72999988, "summary": 72951794},
    "OldBMDp1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72942764},
    "OldBMD2sMu": {"prod": 72999988, "meta": 72999988, "summary": 72950054},
    "OldBMD1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497746},
    "OldBMD0sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498095},
    "OldBMDp1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497821},
    "OldBMD2sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498308},
    "OldBMD1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498342},
    "OldBMD0sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498401},
    "OldBMDp1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498384},
    "OldBMD2sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498407},
    "OldBMD1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497809},
    "OldBMDp1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497867},
    "OldBMD2sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498191},
    "OldBMD0sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498300},
    "OldDSZTau": {"prod": 2500000, "meta": 2500000, "summary": 2497354},
    "OldDZTau": {"prod": 2500000, "meta": 2500000, "summary": 2497181},
    "OldGapD1Charged": {"prod": 9999998, "meta": 9928198, "summary": 9993029},
    "OldGapDetaCharged": {"prod": 12759999, "meta": 12759999, "summary": 12755786},
    "OldGapDpipiCharged": {"prod": 10039998, "meta": 10039998, "summary": 10037734},
}  # type: EvtNumDictType

# Checking consistency of the dictionaries signal_mc_evts and signal_mc_bfs:
assert set(old_signal_mc_evts.keys()) == set(old_signal_mc_bfs.keys())
