"""
Collection of data samples used in the RDStar analysis

Felix Metzner 2019
"""

from typing import Dict, List

from rdstar.offline_analysis import rdstar_sample_weights
from analysistools.batch_tools.file_handler import Belle_list_types


__all__ = [
    "data_type_combinations",
    "recorded_data_samples",
    "main_sample_keys",
    "main_samples",
    "alternative_gap_samples",
    "efficiency_samples",
    "mva_samples",
    "off_resonance_mc_samples",
    "unused_samples"
]

private_data_types = list(rdstar_sample_weights.signal_mc_bfs.keys())

data_type_combinations = {
    "Charged": ["Charged"],
    "Mixed": ["Mixed"],
    "UDS": ["UDS"],
    "Charm": ["Charm"],
    "Special": [c for c in Belle_list_types if "psi" not in c],
    "SpecialPsi": [c for c in Belle_list_types if "psi" in c],
    "Dss": [c for c in private_data_types if c.startswith("B") and (c.endswith("Charged") or c.endswith("Mixed"))],
    "Gap": [c for c in private_data_types if c.startswith("Gap") and not "ViaD" in c],
    "GapAlternative": [c for c in private_data_types if c.startswith("Gap") and ("ViaD" in c or c.startswith("GapD1"))],
    "SignalEff": [c for c in private_data_types if c.startswith("D") and not c.endswith("MVA")],
    "SignalMVA": [c for c in private_data_types if c.startswith("D") and c.endswith("MVA")],
    "DssPiZeroEff": [c for c in private_data_types if c.startswith("B") and c.endswith("Eff") and "Charged" not in c],
    "DssPiPlusEff": [c for c in private_data_types if c.startswith("B") and c.endswith("ChargedPiEff")],
    "DssPiZeroMVA": [c for c in private_data_types if c.startswith("B") and c.endswith("MVA") and "Charged" not in c],
    "DssPiPlusMVA": [c for c in private_data_types if c.startswith("B") and c.endswith("ChargedPiMVA")],
    "OffResonanceUDS": ["OffResonanceUDS"],
    "OffResonanceCharm": ["OffResonanceCharm"],
    "DataY4S": ["DataY4S"],
    "DataContinuum": ["DataContinuum"]
}  # type: Dict[str, List[str]]

recorded_data_samples = [
    *data_type_combinations["DataY4S"],
    *data_type_combinations["DataContinuum"],
]  # type: List[str]

main_sample_keys = [
    "Charged",
    "Mixed",
    "UDS",
    "Charm",
    "Special",
    "Dss",
    "Gap",
]  # type: List[str]

main_samples = [
    *data_type_combinations["Charged"],
    *data_type_combinations["Mixed"],
    *data_type_combinations["UDS"],
    *data_type_combinations["Charm"],
    *data_type_combinations["Special"],
    *data_type_combinations["Dss"],
    *data_type_combinations["Gap"],
]  # type: List[str]

alternative_gap_samples = [
    *data_type_combinations["GapAlternative"],
]  # type: List[str]

efficiency_samples = [
    *data_type_combinations["SignalEff"],
    *data_type_combinations["DssPiZeroEff"],
    *data_type_combinations["DssPiPlusEff"],
]  # type: List[str]

mva_samples = [
    *data_type_combinations["SignalMVA"],
    *data_type_combinations["DssPiZeroMVA"],
    *data_type_combinations["DssPiPlusMVA"],
]  # type: List[str]

off_resonance_mc_samples = [
    *data_type_combinations["OffResonanceUDS"],
    *data_type_combinations["OffResonanceCharm"],
]  # type: List[str]

unused_samples = [
    *data_type_combinations["SpecialPsi"],
]  # type: List[str]

all_samples = recorded_data_samples + main_samples + alternative_gap_samples + efficiency_samples + mva_samples \
              + off_resonance_mc_samples + unused_samples  # type: List[str]

assert len([s for v in data_type_combinations.values() for s in v]) == len(all_samples)
assert len(set([s for v in data_type_combinations.values() for s in v])) == len(set(all_samples))
