"""
Definition of Classes implementing functionalities related to Event Selections.

Felix Metzner 2019
"""

import os
import itertools

from abc import ABC, abstractmethod
from typing import Union, Optional, List, Tuple, NamedTuple, Dict, Generator

from analysistools.processing.observables_utils import makeROOTCompatible, invert_make_ROOT_compatible

from rdstar.mva_selection.mva_selector_setups import MVASelectorSetupBase, MVASelectorSetups

__all__ = [
    "BCSVariableInfo",
    "CutInfo",
    "MVAInfo",
    "SelectionSpecifics",
    "SubSelectionInfo",
    "SelectionInfo",
    "SelectionSetupBase"
]


class BCSVariableInfo(NamedTuple):
    """
    Definition of a information set used for the BCS:
    Contains:
      - bcs_variable: variable name of variable to be used for bcs
                      or "random" if selection shall be done random
                      or None e.g. if now secondary bcs variable shall be used or no BCS shall be applied at all.
      - absolute_value: Boolean indicating if absolute value of variable should be taken for the bcs
      - max: Boolean whether to select the candidate based on the maximum (True) or minimum (False)
    """
    bcs_variable: Optional[str]
    absolute_value: bool = False
    max: bool = True


class CutInfo:
    """
    Definition of a information set for cuts used for the Selection:
    Contains:
      - cut_name: Name of the cut.
      - cut_string: The corresponding cut string.
      - cut_label: The latex label for the cut.
      - cut_var: The corresponding variable.
      - cut_type: The cut type, e.g. smaller, greater, equals, abs etc.
      - cut_limits: The limits where the cut is done.
      - cut_abs: Boolean indicating whether to use the absolute value of the variable for the cut. Default = False.
      - cut_online: Whether cut is performed online or not. Default = False.
    """
    possible_cut_types = [
        "",
        "<",
        ">",
        "==",
        "<=",
        ">=",
        ("<", "<"),
        ("<", "<="),
        ("<=", "<"),
        ("<=", "<=")
    ]  # type: List[Union[str, Tuple[str, str]]]

    def __init__(
            self,
            cut_str: str,
            cut_label: str,
            cut_name: str,
            cut_var: str,
            cut_type: Union[str, Tuple[str, str]],
            cut_limits: Union[float, Tuple[float, float]],
            cut_abs: bool = False,
            cut_online: bool = False
    ) -> None:

        self.__cut_str = cut_str
        self.__cut_label = cut_label
        self.__cut_name = cut_name
        self.__cut_var = cut_var
        self.__cut_type = cut_type
        self.__cut_limits = cut_limits
        self.__cut_abs = cut_abs
        self.__cut_online = cut_online

        self.check_valid()

    @property
    def cut_str(self) -> str:
        return self.__cut_str

    @property
    def cut_label(self) -> str:
        return self.__cut_label

    @property
    def cut_name(self) -> str:
        return self.__cut_name

    @property
    def cut_var(self) -> str:
        return self.__cut_var

    @property
    def cut_type(self) -> Union[str, Tuple[str, str]]:
        return self.__cut_type

    @property
    def cut_limits(self) -> Union[float, Tuple[float, float]]:
        return self.__cut_limits

    @property
    def cut_abs(self) -> bool:
        return self.__cut_abs

    @property
    def cut_online(self) -> bool:
        return self.__cut_online

    def check_valid(self) -> None:
        assert self.cut_type in self.possible_cut_types
        if isinstance(self.cut_limits, Tuple):
            assert isinstance(self.cut_type, type(self.cut_limits))
            assert self.cut_limits[0] < self.cut_limits[1]
        assert self.cut_str in self.generate_cut_str()

    def generate_cut_str(self) -> List[str]:
        var = self.cut_var
        if self.cut_type == "":
            return [""]
        if isinstance(self.cut_type, str):
            if self.cut_abs:
                var = f"abs({var})"
            cut_strings = [f"{var} {self.cut_type} {limit}" for limit in self.float_str_conversions(self.cut_limits)]
        else:
            if self.cut_abs:
                var = f"abs({var})"
            cut_strings = [
                f"{item[0]} {self.cut_type[0]} {var} {self.cut_type[1]} {item[1]}"
                for item in itertools.product(
                    self.float_str_conversions(self.cut_limits[0]),
                    self.float_str_conversions(self.cut_limits[1])
                )]

        return cut_strings

    @staticmethod
    def float_str_conversions(f: float) -> List[str]:
        if f % 1 == 0:
            return [str(float(f)), str(float(f)).strip("0"), str(float(f)).strip(".0")]
        else:
            return [str(float(f))]

    def scale_cut(self, percentage: float = 0.1) -> "CutInfo":
        assert 0. <= percentage <= 1., percentage
        if self.cut_type == "==":
            cut_str = ""
            cut_type = ""
            cut_limits = 0
        elif self.cut_type in ["<", "<="]:
            temp_cut_limits = self.cut_limits
            scale_factor = (1. + percentage) if self.cut_limits > 0. else (1. - percentage)
            self.__cut_limits = scale_factor * self.cut_limits
            cut_str = self.generate_cut_str()[0]
            cut_type = self.cut_type
            cut_limits = self.cut_limits
            self.__cut_limits = temp_cut_limits
        elif self.cut_type in [">", ">="]:
            temp_cut_limits = self.cut_limits
            scale_factor = (1. - percentage) if self.cut_limits > 0. else (1. + percentage)
            self.__cut_limits = scale_factor * self.cut_limits
            cut_str = self.generate_cut_str()[0]
            cut_type = self.cut_type
            cut_limits = self.cut_limits
            self.__cut_limits = temp_cut_limits
        else:
            temp_cut_limits = self.cut_limits
            left_scale_factor = (1. - percentage) if self.cut_limits[0] > 0. else (1. + percentage)
            right_scale_factor = (1. + percentage) if self.cut_limits[1] > 0. else (1. - percentage)
            self.__cut_limits = (left_scale_factor * self.cut_limits[0], right_scale_factor * self.cut_limits[1])
            cut_str = self.generate_cut_str()[0]
            cut_type = self.cut_type
            cut_limits = self.cut_limits
            self.__cut_limits = temp_cut_limits

        return CutInfo(
            cut_str=cut_str,
            cut_label=self.cut_label,
            cut_name=self.cut_name,
            cut_var=self.cut_var,
            cut_type=cut_type,
            cut_limits=cut_limits,
            cut_abs=self.cut_abs,
            cut_online=self.cut_online
        )

    def make_root_compatible(self) -> "CutInfo":
        temp_cut_var = self.cut_var
        self.__cut_var = makeROOTCompatible(variables=[self.__cut_var])[0]
        cut_var = self.cut_var
        cut_str = self.generate_cut_str()[0]
        self.__cut_var = temp_cut_var
        return CutInfo(
            cut_str=cut_str,
            cut_label=self.cut_label,
            cut_name=self.cut_name,
            cut_var=cut_var,
            cut_type=self.cut_type,
            cut_limits=self.cut_limits,
            cut_abs=self.cut_abs,
            cut_online=self.cut_online
        )

    def invert_make_root_compatible(self) -> "CutInfo":
        temp_cut_var = self.cut_var
        self.__cut_var = invert_make_ROOT_compatible(variables=[self.__cut_var])[0]
        cut_var = self.cut_var
        cut_str = self.generate_cut_str()[0]
        self.__cut_var = temp_cut_var
        return CutInfo(
            cut_str=cut_str,
            cut_label=self.cut_label,
            cut_name=self.cut_name,
            cut_var=cut_var,
            cut_type=self.cut_type,
            cut_limits=self.cut_limits,
            cut_abs=self.cut_abs,
            cut_online=self.cut_online
        )

    def __eq__(self, other: "CutInfo") -> bool:
        if self.cut_str not in other.generate_cut_str():
            return False
        if self.cut_label != other.cut_label:
            return False
        if self.cut_name != other.cut_name:
            return False
        if self.cut_var != other.cut_var:
            return False
        if self.cut_abs != other.cut_abs:
            return False
        if self.cut_online != other.cut_online:
            return False
        return True

    def __str__(self) -> str:
        return (
                self.cut_name + " :\n" +
                "Cut String: " + self.cut_str + "\n" +
                "Cut Label: " + self.cut_label + "\n" +
                "Cut Variable: " + self.cut_var + "\n" +
                "Cut Type: " + str(self.cut_type) + "\n" +
                "Cut Limits: " + str(self.cut_limits) + "\n" +
                "Cut Abs: " + str(self.cut_abs) + "\n" +
                "Cut Online: " + str(self.cut_online) + "\n"
        )


class MVAInfo(NamedTuple):
    mva_setup_name: str
    mva_post_cut: float
    mva_setup: MVASelectorSetupBase


class SelectionSpecifics(NamedTuple):
    name: str
    cuts: Tuple[CutInfo, ...]
    primary_bcs_var: BCSVariableInfo
    secondary_bcs_var: BCSVariableInfo
    check_cuts: Tuple[CutInfo, ...]
    check_bcs: bool
    mva_setup_name: Optional[str] = None
    post_mva_cut_info: Optional[CutInfo] = None


class SubSelectionInfo(NamedTuple):
    selection_tag: str
    cut_list: Tuple[CutInfo, ...]
    check_cuts: Tuple[CutInfo, ...]
    check_bcs: bool


class SelectionInfo(NamedTuple):
    selection_tag: str
    selection: SelectionSpecifics
    target_path: Union[str, os.fspath, os.PathLike]
    output_key: str
    info_target_path: Union[None, str, os.fspath, os.PathLike] = None
    info_output_key: Optional[str] = None


class SelectionSetupBase(ABC):
    selection_name = ""  # type: str
    _mva_setup_name = None  # type: Optional[str]
    _post_mva_cut = None  # type: Optional[float]
    all_cut_infos = []  # type: List[CutInfo]

    @staticmethod
    @abstractmethod
    def get_bcs_info() -> Tuple[Optional[BCSVariableInfo], Optional[BCSVariableInfo]]:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_selection(cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        raise NotImplementedError()

    @classmethod
    def get_cut_label_dict(cls) -> Dict[str, str]:
        cut_label_dict = {cut.cut_str: cut.cut_label for cut in cls.all_cut_infos}
        mva_cut_info = cls.get_post_mva_cut_info()
        if mva_cut_info is not None:
            assert mva_cut_info.cut_str not in cut_label_dict, (mva_cut_info.cut_str, cut_label_dict)
            cut_label_dict.update({mva_cut_info.cut_str: mva_cut_info.cut_label})
        return cut_label_dict

    @classmethod
    def get_cut(cls, cut_name: str) -> str:
        cuts = [cut.cut_str for cut in cls.all_cut_infos if cut.cut_name == cut_name]
        if len(cuts) == 1:
            return cuts[0]
        elif len(cuts) == 0:
            raise ValueError(f"No cut named {cut_name} available in the Selection class {cls.__name__}!")
        else:
            raise ValueError(f"Name {cut_name} not unique in the Selection class {cls.__name__}!")

    @classmethod
    def get_mva_setup_name(cls) -> Optional[str]:
        return cls._mva_setup_name

    @classmethod
    def get_mva_setup(cls) -> Optional[MVASelectorSetupBase]:
        if cls._mva_setup_name is None:
            return None
        return MVASelectorSetups.get_selection_mva_setup(setup_name=cls._mva_setup_name)

    @classmethod
    def get_post_mva_cut(cls) -> float:
        assert cls.get_mva_setup_name() is not None
        assert cls._post_mva_cut is not None
        assert isinstance(cls._post_mva_cut, float), type(cls._post_mva_cut)
        assert 0 <= cls._post_mva_cut <= 1.0, cls._post_mva_cut
        return cls._post_mva_cut

    @classmethod
    def get_post_mva_cut_info(cls) -> Optional[CutInfo]:
        if cls.get_mva_setup_name() is None:
            return None
        mva_setup = cls.get_mva_setup()
        mva_output_column_name = mva_setup.get_selection_mva_output_column_name()
        return CutInfo(
            cut_str=f"{mva_output_column_name} > {cls._post_mva_cut}",
            cut_label=r"$\mathcal{P}_\mathrm{Sel MVA} > $" + f"{cls._post_mva_cut}",
            cut_name=f"{mva_output_column_name}_cut",
            cut_var=mva_output_column_name,
            cut_type=">",
            cut_limits=cls._post_mva_cut
        )

    @classmethod
    def get_mva_info(cls) -> Optional[MVAInfo]:
        if cls.get_mva_setup_name() is None:
            return None
        return MVAInfo(
            mva_setup_name=cls.get_mva_setup_name(),
            mva_post_cut=cls.get_post_mva_cut(),
            mva_setup=cls.get_mva_setup()
        )

    @classmethod
    def get_selection_tags(cls, cs_tag: str) -> Tuple[str, ...]:
        return tuple([info.selection_tag for info in cls.get_selection(cs_tag=cs_tag)])

    @classmethod
    def get_main_selection_tag(cls, cs_tag: str) -> str:
        # Getting first element returned by the generator get_selection.
        # This first item is supposed to be the main selection definition of the analysis.
        return next(cls.get_selection(cs_tag=cs_tag)).selection_tag

    @classmethod
    def get_selection_specifics(cls, cs_tag: str) -> Dict[str, SelectionSpecifics]:
        return {infos.selection_tag: SelectionSpecifics(
            name=infos.selection_tag,
            cuts=infos.cut_list,
            primary_bcs_var=cls.get_bcs_info()[0],
            secondary_bcs_var=cls.get_bcs_info()[1],
            check_cuts=infos.check_cuts,
            check_bcs=infos.check_bcs,
            mva_setup_name=cls.get_mva_setup_name(),
            post_mva_cut_info=cls.get_post_mva_cut_info()
        ) for infos in cls.get_selection(cs_tag=cs_tag)}

    @classmethod
    def print_selection_info(cls, cs_tag: str) -> None:
        print(f"Selection name: {cls.selection_name}\n")

        print("\tAll defined selections:\n\t\t- "
              + "\n\t\t- ".join(cls.get_selection_tags(cs_tag=cs_tag)) + "\n")

        print(f"\tMain selection tag: {cls.get_main_selection_tag(cs_tag=cs_tag)}\n")

        if cls.get_post_mva_cut_info() is not None:
            print(
                f"\tMVA Selection Info:\n\t\t"
                f"MVA Selection Name: {cls.get_mva_setup_name()}\n\t\t"
                f"MVA Selection Cut: {cls.get_post_mva_cut_info().cut_str}\n"
            )

        for key, value in cls.get_selection_specifics(cs_tag=cs_tag).items():
            cut_string_list = [f"C {c.cut_str}" if c in value.check_cuts else f"  {c.cut_str}" for c in value.cuts]
            print(f"\t{key}:\n\t\t" + "\n\t\t".join(cut_string_list) + "\n")


if __name__ == "__main__":
    c = CutInfo(
        cut_str="3. <= useCMSFrame(pt) < 4.",
        cut_label="test",
        cut_name="test",
        cut_type=("<=", "<"),
        cut_var="useCMSFrame(pt)",
        cut_abs=False,
        cut_limits=(3, 04.0),
        cut_online=True
    )
    print(c)
