"""
Combination of Data Samples

Felix Metzner 2019
"""

import os
import json
import copy
import numpy as np
import pandas as pd
import multiprocessing
import b2luigi as luigi

from functools import partial
from collections import defaultdict, Iterable, abc
from typing import Any, Union, Optional, Tuple, List, Dict, Generator

from analysistools.batch_tools.file_handler import Belle_url_mc_types

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.rdstar_sample_weights import scale_weight_column_name as scale_col_name
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask, DataPreparationMergerTask

__all__ = [
    "SampleCombinerTask",
    "SelectionInfoCombinerTask"
]


class SampleCombinerTask(utilities.BaseOfflineTask):
    combo_name = luigi.Parameter(positional=False)
    working_dir = luigi.Parameter(positional=False)
    selection_name = luigi.Parameter(positional=False)  # type: str
    config_params = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name = "CombinedSamples"

    data_type_column_name = "data_type"

    @property
    def resources(self) -> Dict[str, int]:
        if self.combo_name in ["Charm", "UDS", "Charged", "Mixed"]:
            return {"offline_memory": 10}
        else:
            return {"offline_memory": 1}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.config_params, abc.Sequence)
        assert isinstance(self.working_dir, str), type(self.working_dir)

        self.base_path = os.path.join(os.fspath(self.working_dir), self.base_dir_name, self.selection_name)

        self.cs_tag = self.config_params[0]["cs_tag"]

        self.pid_weight_info_key = "pid_weight_info"
        self.ff_d_weight_info_key = "ff_d_weight_info"
        self.ff_ds_weight_info_key = "ff_ds_weight_info"
        self.ff_dss_weight_info_key = "ff_dss_weight_info"

        self.evt_id_col = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()
        self.c_id_col = "unique_candidate_id"
        self.id_col = "combined_candidate_id"

        self.pid_sample_random_seed = 47471337

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    @staticmethod
    def get_data_combo_dict() -> Dict[str, List[str]]:
        return rdstar_samples.data_type_combinations

    def get_input_file_info(self) -> Generator[Tuple[str, str, int], None, None]:
        assert isinstance(self.config_params, abc.Sequence)
        for params in self.config_params:
            stream = params["stream"]
            stream_tag = f"_{stream}"
            if stream == -999:
                stream_tag = ""
            yield f"{params['data_type']}{stream_tag}_{params['tuple_type']}", params['data_type'], stream

    def get_paths_for_key(self, input_key: str) -> Generator[Tuple[str, str, int, os.fspath], None, None]:
        for base_key, data_type, stream in self.get_input_file_info():
            for key, path in self.input_dict.items():
                is_hdf5_file = os.path.splitext(path)[1] == ".h5"
                if key.endswith(input_key) and key.startswith(base_key) and is_hdf5_file:
                    yield key, data_type, stream, path

    def load_and_combine(self, input_key: str) -> pd.DataFrame:
        assert not str(self.combo_name).startswith("Data")
        initial_length = 0
        dfs = []
        scale_weight_dict = defaultdict(list)
        for key, data_type, stream, path in self.get_paths_for_key(input_key=input_key):
            df = pd.read_hdf(path)
            assert isinstance(df, pd.DataFrame)
            if len(df.index) > 0:
                initial_length += len(df.index)

                df.loc[:, "stream"] = stream
                df.loc[:, self.data_type_column_name] = data_type

                if input_key == self.pid_weight_info_key:
                    if not len(df.index) == len(np.unique(df.index)):
                        df.reset_index(inplace=True)
                    self.add_unique_id(df=df)
                    selection_conds = self.get_selection_conditions(data_type=data_type, stream=stream)
                    evt_cond = df[self.evt_id_col].isin(selection_conds[0])
                    id_cond = df.loc[evt_cond, self.id_col].isin(selection_conds[1])
                    selected_indices = df.loc[evt_cond].loc[id_cond].index  # type: pd.Index
                    assert len(selected_indices) == len(set(selected_indices)), \
                        (len(selected_indices), len(set(selected_indices)))

                    sampled_indices = df.loc[~df.index.isin(selected_indices)].sample(
                        frac=0.1,
                        random_state=self.pid_sample_random_seed
                    ).index  # type: pd.Index
                    assert len(sampled_indices) == len(set(sampled_indices)), \
                        (len(sampled_indices), len(set(sampled_indices)))

                    n_combined_indices = len(set(selected_indices) | set(sampled_indices))
                    assert n_combined_indices == len(selected_indices) + len(sampled_indices), \
                        (n_combined_indices, len(selected_indices) + len(sampled_indices))

                    chosen_indices = selected_indices.union(sampled_indices)  # type: pd.Index

                    assert len(chosen_indices) == len(selected_indices) + len(sampled_indices), \
                        (len(chosen_indices), len(selected_indices) + len(sampled_indices))

                    df = df.iloc[chosen_indices]
                    del df[self.id_col]
                    utilities.decompose_event_id(df=df, remove_original_column=True)

                if data_type in Belle_url_mc_types.keys() and input_key not in self.get_info_keys():
                    scale_weight = df[scale_col_name].unique()
                    scale_correction = df[DataPreparationTask.scale_corr_column].unique()
                    if len(scale_weight) == 1 and len(scale_correction) == 1:
                        scale_weight_dict[data_type].append(scale_weight[0] - scale_correction[0])
                    else:
                        raise RuntimeError(
                            f"Unexpected scaling_weight = {scale_weight}, scale_correction = {scale_correction} pair."
                        )

            dfs.append(df)

        combined_df = pd.concat(dfs, axis=0, ignore_index=True, sort=False)

        for data_type in scale_weight_dict.keys():
            scaling_weight = 1. / sum(scale_weight_dict[data_type])
            print(f"Extracted the scaling weights {scale_weight_dict[data_type]} for {data_type} and applying the "
                  f"weight factor 1. / {sum(scale_weight_dict[data_type])} = {scaling_weight}")

            combined_df.loc[combined_df.data_type == data_type, scale_col_name] = scaling_weight

        if input_key != self.pid_weight_info_key:
            assert len(combined_df.index) == initial_length

        if input_key == self.pid_weight_info_key:
            combined_df.loc[:, self.data_type_column_name] = combined_df[self.data_type_column_name].astype('category')
            combined_df.loc[:, "weight_type"] = combined_df.weight_type.astype('category')

        return combined_df

    def add_unique_id(self, df: pd.DataFrame, check_if_unique: bool = False) -> np.ndarray:
        id_values = utilities.add_unique_candidate_id(
            df=df,
            id_column_name=self.id_col,
            event_id_name=self.evt_id_col,
            candidate_id_name=self.c_id_col,
            check_if_unique=check_if_unique
        )
        return id_values

    def get_selection_conditions(self, data_type: str, stream: Union[int, float]) -> Tuple[np.ndarray, np.ndarray]:
        selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
        type_info = f"{data_type}" if stream == -999 else f"{data_type}_{stream}"
        df_paths = [  # type: List[os.fspath]
            path for selection_tag in selection_tags for key, path in self.input_dict.items()
            if key.endswith(selection_tag) and key.startswith(type_info) and os.path.splitext(path)[1] == ".h5"
        ]
        assert len(df_paths) == len(selection_tags), (len(selection_tags), len(df_paths), selection_tags, df_paths)

        cols = [self.evt_id_col, self.c_id_col]
        df = pd.concat(
            [df.loc[:, cols] for path in df_paths for df in [pd.read_hdf(path)] if isinstance(df, pd.DataFrame)],
            axis=0,
            ignore_index=True,
            sort=False
        )
        unique_candidate_id = self.add_unique_id(df=df)

        return df[self.evt_id_col].unique(), np.unique(unique_candidate_id)

    def get_data_path(self, input_key: str) -> os.fspath:
        assert str(self.combo_name).startswith("Data")
        paths = []
        for key, data_type, stream, path in self.get_paths_for_key(input_key=input_key):
            paths.append(path)
        assert len(paths) == 1
        return paths[0]

    def get_info_keys(self) -> List[str]:
        return [
            self.pid_weight_info_key,
            self.ff_d_weight_info_key,
            self.ff_ds_weight_info_key,
            self.ff_dss_weight_info_key,
        ]

    def get_subselection_keys(self, selection_keys: List[str]) -> List[str]:
        subselection_keys = []
        possible_extensions = DataSelectionTask.get_possible_subsample_extensions(input_dict=self.input_dict)
        for selection_key in selection_keys:
            for ext in possible_extensions:
                tag = f"{selection_key}{ext}"
                if any(key.endswith(tag) for key in self.input_dict.keys()):
                    subselection_keys.append(tag)

        return subselection_keys

    def get_keys(self) -> List[str]:
        _selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
        selection_keys = list(_selection_tags)
        selection_keys.extend(self.get_subselection_keys(selection_keys))

        if str(self.combo_name).startswith("Data"):
            return selection_keys

        keys = self.get_info_keys()
        keys.extend(selection_keys)

        if str(self.combo_name).startswith("Special"):
            keys.remove(self.ff_dss_weight_info_key)
        return keys

    def combine_sample_with_key(self, key: str, dry_run: bool) -> Tuple[str, os.fspath]:
        output_key = f"{self.combo_name}_{key}"
        output_path = os.path.join(self.base_path, f"{output_key}.h5")
        temp_output_path = os.path.join(self.base_path, f"temp_{output_key}.h5")

        if not dry_run:
            if str(self.combo_name).startswith("Data"):
                if os.path.exists(output_path):
                    os.remove(output_path)
                os.symlink(self.get_data_path(input_key=key), output_path)
            else:
                df = self.load_and_combine(key)
                df.to_hdf(
                    temp_output_path,
                    key=output_key,
                    format="table" if key == self.pid_weight_info_key else "fixed"
                )
                os.rename(temp_output_path, output_path)

        return output_key, output_path

    def combine_samples(self, dry_run: bool = False) -> List[Tuple[str, os.fspath]]:
        if not dry_run:
            os.makedirs(self.base_path, exist_ok=True)

        with multiprocessing.Pool(processes=4) as pool:
            output_tuples = pool.map(
                partial(SampleCombinerTask.combine_sample_with_key, self, dry_run=dry_run), self.get_keys()
            )

        return output_tuples

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.combine_samples()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.combine_samples(dry_run=True):
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    @staticmethod
    def get_reduced_params_for_data_preparation_task(params_dict: Dict[str, Any]) -> Dict[str, Any]:
        exclude_list = [
            "cs_tag",
            "cs_fastbdt_settings",
        ]

        return {k: v for k, v in params_dict.items() if k not in exclude_list}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_params, Iterable)
        for params in self.config_params:
            yield DataSelectionTask(selection_name=self.selection_name, **params)
            yield DataPreparationMergerTask(**self.get_reduced_params_for_data_preparation_task(params))


class SelectionInfoCombinerTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)
    selection_name = luigi.Parameter(positional=False)  # type: str
    config_combinations = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name = "CombinedSelectionInfo"

    _info_combos = {
        "main": rdstar_samples.main_samples,
        "eff": rdstar_samples.efficiency_samples,
        "data": rdstar_samples.recorded_data_samples,
        "allMC": rdstar_samples.main_samples + rdstar_samples.efficiency_samples,
    }
    _info_combos.update({k: [k] for k in rdstar_samples.recorded_data_samples})
    info_combos = _info_combos
    info_names = [k for k in info_combos.keys()]
    main_info_name = "main"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.working_dir, str), type(self.working_dir)
        self.base_path = os.path.join(os.fspath(self.working_dir), self.base_dir_name, self.selection_name)

        assert isinstance(self.config_combinations, abc.Sequence)
        self.cs_tag = self.config_combinations[0][2][0]["cs_tag"]
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_selection_info_files_per_tag(self) -> Dict[str, List[str]]:
        file_dict = defaultdict(list)
        type_dict = defaultdict(list)
        selection_tags = []
        for key, path in self.input_dict.items():
            has_prefix = os.path.basename(path).startswith(DataSelectionTask.evaluation_info_prefix)
            is_json = os.path.splitext(path)[1] == ".json"
            if not (has_prefix and is_json):
                continue

            data_type = key.split("_", 1)[0]
            selection_tag = key.split(f"{DataSelectionTask.evaluation_info_prefix}_", 1)[1]
            if selection_tag not in selection_tags:
                selection_tags.append(selection_tag)

            for combo_name, combo_keys in self.info_combos.items():
                if data_type in combo_keys:
                    type_dict[f"{combo_name}_{selection_tag}"].append(data_type)
                    file_dict[f"{combo_name}_{selection_tag}"].append(path)

        for combo_name in self.info_combos:
            keys = [f"{combo_name}_{st}" for st in selection_tags]
            assert all(len(type_dict[t]) == len(type_dict[keys[0]]) for t in keys)
        assert all(len(file_dict[t]) == len(type_dict[t]) for t in file_dict.keys())

        for data_type in rdstar_samples.recorded_data_samples:
            assert all(len(v) == 1 for k, v in file_dict.items() if k.startswith(data_type))

        assert all(len(file_dict[t]) == len(set(file_dict[t])) for t in file_dict.keys())
        assert all(os.path.exists(p) for t in file_dict.keys() for p in file_dict[t])

        return file_dict

    @staticmethod
    def update_selection_dicts(
            dict1: Dict[str, Any],
            dict2: Dict[str, Any],
            exclude_keys: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        if exclude_keys is None:
            exclude_keys = []
        result_dict = {}
        d1_keys = [k for k in dict1.keys() if k not in exclude_keys]
        d2_keys = [k for k in dict2.keys() if k not in exclude_keys]

        for key in d1_keys:
            if key not in d2_keys:
                result_dict.update({key: copy.copy(dict1[key])})
            else:
                if isinstance(dict1[key], int) and isinstance(dict2[key], int):
                    result_dict.update({key: dict1[key] + dict2[key]})
                elif isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
                    # Calling function recursively, if value is of type dict:
                    result_dict.update({
                        key: SelectionInfoCombinerTask.update_selection_dicts(dict1[key], dict2[key], exclude_keys)
                    })
                else:
                    raise ValueError(
                        f"Dictionaries can not be merged, because values are of different type:\n"
                        f"for key {key}:\n"
                        f"\tdict1 has value {dict1[key]} of type {type(dict1[key])}\n"
                        f"\tdict2 has value {dict2[key]} of type {type(dict2[key])}..."
                    )

        for key in [k for k in d2_keys if k not in d1_keys]:
            result_dict.update({key: copy.copy(dict2[key])})

        return result_dict

    @staticmethod
    def merge_selection_dicts(
            dict1: Dict[str, Dict[str, Any]],
            dict2: Dict[str, Dict[str, Any]]
    ) -> Dict[str, Dict[str, Any]]:
        assert set(dict1.keys()) == set(dict2.keys()), (dict1.keys(), dict2.keys())
        merged_dict = {}
        for key in dict1.keys():
            assert "bcs" in dict1[key].keys() and "bcs" in dict2[key].keys()
            assert dict1[key]["bcs"] == dict2[key]["bcs"]
            assert "cuts" in dict1[key].keys() and "cuts" in dict2[key].keys()
            if not (dict2[key]["cuts"] is None and dict1[key]["cuts"] is None):
                assert set(dict1[key]["cuts"]) == set(dict2[key]["cuts"])
            assert set(dict1[key].keys()) == set(dict2[key].keys())
            assert len(dict1[key].keys()) == len(dict2[key].keys())
            merged_dict[key] = {
                "bcs": copy.copy(dict1[key]["bcs"]),
                "cuts": copy.copy(dict1[key]["cuts"])
            }
            merged_dict[key].update(SelectionInfoCombinerTask.update_selection_dicts(
                dict1=dict1[key],
                dict2=dict2[key],
                exclude_keys=["bcs", "cuts"]
            ))
        assert set(merged_dict.keys()) == set(dict1.keys())
        return merged_dict

    def merge_selection_infos(self, file_paths: List[str]) -> Optional[Dict[str, Dict[str, Any]]]:
        combined_dict = None
        for path in file_paths:
            info_dict = utilities.load_dict_from_json(path)
            if combined_dict is None:
                combined_dict = copy.copy(info_dict)
            else:
                combined_dict = self.merge_selection_dicts(combined_dict, info_dict)

        return combined_dict

    def combine_selection_infos(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        output_dict = {}

        selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
        result_keys = [f"{combo}_{sel_tag}" for sel_tag in selection_tags for combo in self.info_combos]

        for key in result_keys:
            output_key = f"{DataSelectionTask.evaluation_info_prefix}_{key}"
            output_path = os.path.join(self.base_path, f"{output_key}.json")
            output_dict.update({output_key: output_path})

        if not dry_run:
            os.makedirs(self.base_path, exist_ok=True)

            selection_info_files_per_tag_dict = self.get_selection_info_files_per_tag()
            assert len(result_keys) == len(selection_info_files_per_tag_dict)
            assert set(result_keys) == set(selection_info_files_per_tag_dict.keys())

            for result_key, in_files in selection_info_files_per_tag_dict.items():
                assert result_key in result_keys, result_key
                assert len(in_files) > 0, (result_key, in_files)
                output_key = f"{DataSelectionTask.evaluation_info_prefix}_{result_key}"
                output_path = output_dict[output_key]
                tmp_path = os.path.join(self.base_path, f"tmp_{output_key}.json")

                if not dry_run:
                    if len(in_files) == 1:
                        if os.path.exists(output_path):
                            os.remove(output_path)
                        os.symlink(in_files[0], output_path)
                    else:
                        selection_info = self.merge_selection_infos(file_paths=in_files)
                        with open(tmp_path, 'w') as out_file:
                            json.dump(selection_info, out_file, indent=2)
                        os.rename(src=tmp_path, dst=output_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.combine_selection_infos()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.combine_selection_infos(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for params in [c for _, _, rcs in self.config_combinations for c in rcs]:
            yield DataSelectionTask(selection_name=self.selection_name, **params)
