"""
Data Reduction Tasks

    These tasks have the purpose of reducing the size of the data frames by
        - Application of Continuum Suppression and a light cut on the output
        - Removing unneeded columns from the data frames

Felix Metzner 2020
"""

import os
import yaml
import logging
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Dict, Generator

from analysistools.dataframe_tools import hdf5_utils

from rdstar import continuum_suppression as cs_tools
from rdstar.offline_analysis import offline_observables
from rdstar.offline_analysis import rdstar_sample_weights

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.cs_tasks import CSTestTask
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationSplitterTask
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask, DataPreparationMergerTask

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "DataReductionTask",
]


@inherits(CSTestTask)
@inherits(DataPreparationSplitterTask)
class DataReductionTask(utilities.BaseOfflineTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.base_dir = os.path.join(self.working_dir, self.data_type)

        if isinstance(offline_config["cs_config"], dict):
            self.cs_offline_config_path = self.offline_config_path
            cs_config = offline_config["cs_config"]
        else:
            configs_dir = utilities.get_config_directory()
            self.cs_offline_config_path = os.path.join(configs_dir, offline_config["cs_config"])

            if not os.path.isfile(self.cs_offline_config_path):
                raise RuntimeError(f"Encountered invalid offline_config path for cs training:\n"
                                   f"\t{self.cs_offline_config_path}")

            with open(self.cs_offline_config_path) as cs_f:
                cs_offline_config = yaml.load(cs_f)
                cs_config = cs_offline_config["cs_config"]

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.file_name = f'{self.data_type}{self.stream_tag}_{self.tuple_type}'
        self.input_key = f"{self.file_name}_{DataPreparationTask.base_file_tag}"

        self.target_path = os.path.join(self.base_dir, f"{self.file_name}_with_CS_{self.cs_tag}.h5")
        self.tmp_target_path = os.path.join(self.base_dir, f"tmp_{self.file_name}_with_CS_{self.cs_tag}.h5")
        self.output_key = f"{self.file_name}_with_CS_{self.cs_tag}"

        self.cs_column_name = cs_config["cs_column_name"]
        self.cs_cut_value = cs_config["cs_cut_value"] if "cs_cut_value" in cs_config else None

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def apply_cs_mva(self, df: pd.DataFrame) -> None:
        cs_classifier = cs_tools.ContinuumSuppression.load_classifier(
            weight_file_path=self.input_dict["cs_output_weight_file"],
            var_file_path=self.input_dict["cs_var_list_file"]
        )
        logging.info("  Applying CS classifier to data frame...")

        cs_classifier.apply(data=df, output_column=self.cs_column_name)

        if self.cs_cut_value is not None:
            logging.info("  Applying cut on CS output to data frame...")
            df.query(f"{self.cs_column_name} > {self.cs_cut_value}", inplace=True)

    @staticmethod
    def discard_unneeded_columns(df: pd.DataFrame) -> None:
        discard_cols = [c for c in offline_observables.observables_to_discard if c in df.columns]
        df.drop(discard_cols, axis=1, inplace=True)

    def add_scaling_weights(self, df: pd.DataFrame) -> None:
        rdstar_sample_weights.add_scaling_weights(
            df=df,
            data_type=self.data_type,
            generic_stream_dict=None
        )

    def data_reduction(self) -> None:
        df = hdf5_utils.load_hdf5(hdf5_file_path=self.input_dict[self.input_key], hdf5_key=self.file_name)

        if self.data_type.startswith("Data"):
            utilities.add_unique_event_id(df=df, data_type=f"{self.data_type}{self.stream_tag}")

        self.apply_cs_mva(df=df)

        self.discard_unneeded_columns(df=df)
        self.add_scaling_weights(df=df)

        df.to_hdf(path_or_buf=self.tmp_target_path, key=self.file_name)
        os.rename(src=self.tmp_target_path, dst=self.target_path)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.data_reduction()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {self.output_key: luigi.LocalTarget(self.target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CSTestTask, offline_config_path=self.cs_offline_config_path)
        yield self.clone(DataPreparationMergerTask)
