"""
Fitting Tasks

These tasks will
    - perform the actual fit of the analysis (TODO)
    - perform sideband fits on data to validate the model and the data-MC agreement (TODO)
    - run asimov fits on sideband samples (TODO)
    - run asimov fits on signal region samples (TODO)
    - perform toy studies (TODO)
    - perform profile scans (TODO)
    - plot the template shapes (TODO)
    - plot the distributions before and after the fits (TODO)

Felix Metzner 2020
"""

import os
import logging
import pandas as pd
import b2luigi as luigi

from collections import abc
from luigi.util import inherits
from PyPDF2 import PdfFileMerger
from typing import Tuple, List, Dict, Union, Generator

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.rdstar_plots import rdstar_dec_titles
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask, WeightAnalysisTask
from rdstar.offline_analysis.fitting.rdstar_fit_models import FitModelCreator, FitCase, rdstar_fit_cases

from templatefitter.fitter import TemplateFitter
from templatefitter.plotter.fit_result_plots import FitResultPlotter

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "FitTask",
    "FittingTasksWrapperTask",
]


# TODO: Maybe combine sample for fit beforehand.
# TODO: Maybe move fit samples to Deepthought SSDs

@inherits(WeightAnalysisTask)
class FitTask(utilities.BaseOfflineTask):
    fit_base_dir_name = "FitResults"  # type: str

    fit_case_key = luigi.Parameter(positional=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert isinstance(self.working_dir, (str, os.PathLike))
        assert isinstance(self.cs_tag, str), self.cs_tag
        assert isinstance(self.fit_case_key, str), self.fit_case_key
        assert isinstance(self.selection_name, str), self.selection_name

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag
        )

        self.fit_case_name = self.fit_case_key  # type: str
        self.fit_case = FitModelCreator.fit_cases[self.fit_case_key]  # type: FitCase
        self.base_dir_name = self.fit_case.base_dir_name
        self.plot_info_text = self.fit_case.plot_info_text
        self.fit_output_tag = self.fit_case.output_tag

        input_sample_key = FitModelCreator.fit_cases[self.fit_case_key].sample_key
        if input_sample_key is not "":
            input_sample_key = f"_{input_sample_key}"

        self.base_path = os.path.join(self.working_dir, self.fit_base_dir_name, self.main_sel_tag, self.base_dir_name)

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.mc_sample_keys = rdstar_samples.main_sample_keys

        self.full_weight_column_name = "__weight_final_wFEI_corr__"  # TODO: Remove once provided by weight finalization

        self.key = f"_{self.selection_name}_with_CS_{self.cs_tag}{input_sample_key}"
        self.input_appendix = f"{WeightFinalizationTask.appendix}_{ResolutionCorrectionTask.appendix}"

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def load_mc_sample(self) -> pd.DataFrame:
        mc_paths = [  # type: List[os.fspath]
            p for mc_key in self.mc_sample_keys for k, p in self.input_dict.items()
            if (k.startswith(f"{mc_key}_{self.input_appendix}_")
                and k.endswith(self.key)
                and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, self.key)

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        self.add_final_weight_col(mc_sample=mc_sample)  # TODO: Remove once provided by weight finalization

        return mc_sample

    def load_recorded_data_sample(self) -> pd.DataFrame:
        sample_key = "DataY4S"
        data_paths = [  # type: List[os.fspath]
            p for k, p in self.input_dict.items()
            if (k.startswith(f"{sample_key}_{self.input_appendix}_")
                and k.endswith(self.key)
                and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(data_paths) == 1, (len(data_paths), data_paths, self.key)
        data_path = data_paths[0]
        assert os.path.exists(data_path), data_path
        data_df = pd.read_hdf(data_path)
        assert isinstance(data_df, pd.DataFrame)
        return data_df

    def add_final_weight_col(self,
                             mc_sample: pd.DataFrame) -> None:  # TODO: Remove once provided by weight finalization
        assert self.full_weight_column_name not in mc_sample.columns
        assert WeightFinalizationTask.final_weight_col in mc_sample.columns
        assert TagCorrection.tagging_correction_name in mc_sample.columns

        product_cols = [
            WeightFinalizationTask.final_weight_col,
            TagCorrection.tagging_correction_name
        ]
        mc_sample.loc[:, self.full_weight_column_name] = mc_sample[product_cols].product(axis=1)

    def get_fit_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}"
        return key, f"{key}.json"

    def get_fit_plot_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}"
        return key, f"{key}.pdf"

    @staticmethod
    def merge_plot_pdf_files(
            input_paths: List[Union[str, os.fspath, os.PathLike]],
            output_path: Union[str, os.fspath, os.PathLike],
            overwrite: bool = False
    ) -> None:
        pdf_merger = PdfFileMerger()
        assert all(os.path.exists(path) for path in input_paths), ([p for p in input_paths if not os.path.exists(p)])
        assert all(os.path.splitext(path)[1] == ".pdf" for path in input_paths), \
            ([p for p in input_paths if not os.path.splitext(p)[1] == ".pdf"])
        assert os.path.splitext(output_path)[1] == ".pdf", output_path
        if not overwrite:
            assert not os.path.exists(output_path), output_path

        for pdf_path in input_paths:
            pdf_merger.append(pdf_path)

        pdf_merger.write(fileobj=output_path)
        pdf_merger.close()

    def do_fit(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        output_dict = dict()  # type: Dict[str, os.fspath]
        fit_result_key, fit_result_file_name = self.get_fit_output_key_file_name_pair(file_tag="FitResult")
        temp_fit_result_path = os.path.join(self.base_path, f"temp_{fit_result_file_name}")
        fit_result_path = os.path.join(self.base_path, fit_result_file_name)
        output_dict.update({fit_result_key: fit_result_path})

        pre_fit_plot_key, pre_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PreFitPlot")
        temp_pre_fit_plot_path = os.path.join(self.base_path, f"temp_{pre_fit_plot_file_name}")
        pre_fit_plot_path = os.path.join(self.base_path, pre_fit_plot_file_name)
        output_dict.update({pre_fit_plot_key: pre_fit_plot_path})

        post_fit_plot_key, post_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PostFitPlot")
        temp_post_fit_plot_path = os.path.join(self.base_path, f"temp_{post_fit_plot_file_name}")
        post_fit_plot_path = os.path.join(self.base_path, post_fit_plot_file_name)
        output_dict.update({post_fit_plot_key: post_fit_plot_path})

        if not dry_run:
            mc_sample = self.load_mc_sample()

            fit_model_creator = FitModelCreator(
                fit_case=self.fit_case_name,
                weight_column_name=self.full_weight_column_name
            )
            fit_model_creator.setup_model(
                mc_df=mc_sample,
                data_df=None if self.fit_case.asimov_data else self.load_recorded_data_sample(),
                use_asimov_data=self.fit_case.asimov_data,
                round_asimov_data_bin_counts=self.fit_case.round_asimov_data
            )
            fit_model = fit_model_creator.fit_model

            fit_result_plotter = FitResultPlotter(
                variables=fit_model_creator.fit_variables,
                fit_model=fit_model,
                reference_dimension=0,
                channel_label_dict=rdstar_dec_titles,
                data_label=self.plot_info_text
            )

            pre_fit_plot_paths = fit_result_plotter.plot_fit_result(
                use_initial_values=True,
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag
            )
            assert len(pre_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["png"]), \
                (len(pre_fit_plot_paths["pdf"]), len(pre_fit_plot_paths["png"]))
            self.merge_plot_pdf_files(input_paths=pre_fit_plot_paths["pdf"], output_path=temp_pre_fit_plot_path)

            fit_result_plotter.plot_fit_result_projections(
                project_to=0,
                use_initial_values=True,
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag
            )
            fit_result_plotter.plot_fit_result_projections(
                project_to=1,
                use_initial_values=True,
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag
            )
            os.rename(src=temp_pre_fit_plot_path, dst=pre_fit_plot_path)

            fitter = TemplateFitter(fit_model=fit_model, minimizer_id="iminuit")
            fit_result = fitter.do_fit(update_templates=True, get_hesse=True, verbose=True, fix_nui_params=True)
            fit_result.save_to(path=temp_fit_result_path, overwrite=True)
            os.rename(src=temp_fit_result_path, dst=fit_result_path)

            post_fit_plot_paths = fit_result_plotter.plot_fit_result(
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag
            )
            assert len(post_fit_plot_paths["pdf"]) == len(post_fit_plot_paths["png"]), \
                (len(post_fit_plot_paths["pdf"]), len(post_fit_plot_paths["png"]))
            assert len(post_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["pdf"]), \
                (len(post_fit_plot_paths["pdf"]), len(pre_fit_plot_paths["pdf"]))
            self.merge_plot_pdf_files(input_paths=post_fit_plot_paths["pdf"], output_path=temp_post_fit_plot_path)

            fit_result_plotter.plot_fit_result_projections(
                project_to=0,
                use_initial_values=False,
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag
            )
            fit_result_plotter.plot_fit_result_projections(
                project_to=1,
                use_initial_values=False,
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag
            )
            os.rename(src=temp_post_fit_plot_path, dst=post_fit_plot_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_fit()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_fit(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations
            )


@inherits(WeightAnalysisTask)
class FittingTasksWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def get_task_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for fit_case_key in rdstar_fit_cases.keys():
            yield self.clone(FitTask, fit_case_key=fit_case_key)
