"""
Tasks producing the Validation Results for the Analysis

Felix Metzner 2019
"""

import os
import b2luigi as luigi
from luigi.util import inherits

from typing import Tuple, List, Generator, Any

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.result_tasks import EfficiencyEvaluationTask
from rdstar.offline_analysis.tasks.fitting_tasks import FittingTasksWrapperTask
from rdstar.offline_analysis.tasks.flow_validation_tasks import AnalysisEvaluationTask
from rdstar.offline_analysis.tasks.sideband_validation_tasks import SidebandEvaluationWrapperTask
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionValidationTask

from rdstar.offline_analysis.selection.selection_setups import DSSSelection

__all__ = [
    "ValidationWrapperTask",
    "SelectionManagerTask"
]


@inherits(
    AnalysisEvaluationTask,
    SidebandEvaluationWrapperTask,
    ResolutionCorrectionValidationTask
)
class ValidationWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def get_task_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(AnalysisEvaluationTask)
        yield self.clone(SidebandEvaluationWrapperTask)
        if self.selection_name != DSSSelection.selection_name:
            yield self.clone(ResolutionCorrectionValidationTask)


class SelectionManagerTask(luigi.WrapperTask):
    selection_setup_names = luigi.TupleParameter(positional=False, hashed=True)  # type: Tuple[str, ...]
    config_combinations = luigi.ListParameter(positional=False, hashed=True)  # type: List[Any]
    working_dir = luigi.Parameter(positional=False)  # type: str
    cs_tag = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def get_task_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for selection_name in self.selection_setup_names:
            yield self.clone(FittingTasksWrapperTask, selection_name=selection_name)
            yield self.clone(ValidationWrapperTask, selection_name=selection_name)
            yield self.clone(EfficiencyEvaluationTask, selection_name=selection_name)
