"""
Signal Selection Tasks

    - Best candidate selection
    - Signal selection cuts
    - Creation of side band samples

Felix Metzner 2019
"""

import os
import json
import yaml
import logging
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Union, List, Dict, Any, Generator

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.data_reduction_tasks import DataReductionTask
from rdstar.offline_analysis.selection.selection import Selection, SelectionSetupInterpreter

from rdstar.offline_analysis.tasks.mva_selector_tasks import MVASelectorTrainTask, MVASelectorTestTask

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "DataSelectionTask"
]


@inherits(DataReductionTask)
class DataSelectionTask(utilities.BaseOfflineTask):
    selection_name = luigi.Parameter(positional=False)  # type: str

    evaluation_info_prefix = "selection_info"

    # Column name information:
    u_eid_col = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.base_dir = os.path.join(self.working_dir, self.data_type)

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.bcs_random_seed = 1337

        self.file_name = f'{self.data_type}{self.stream_tag}_{self.tuple_type}'
        self.input_key = f"{self.file_name}_with_CS_{self.cs_tag}"

        self.selection_setup_handler = SelectionSetupInterpreter(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
            output_directory_path=self.base_dir,
            output_file_name=self.file_name,
            evaluation_output_prefix=self.evaluation_info_prefix
        )

        self.event_selection = Selection(
            selection_name=self.selection_name,
            unique_event_id_column_name=self.u_eid_col,
            bcs_random_seed=self.bcs_random_seed
        )
        self.mva_info = self.event_selection.get_mva_setup_info()

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    @staticmethod
    def get_main_config_path(config_path: Union[str, os.PathLike]) -> Union[str, os.PathLike]:
        with open(config_path) as f:
            offline_config = yaml.load(f)

        if isinstance(offline_config["cs_config"], dict):
            return config_path
        else:
            assert isinstance(offline_config["cs_config"], str), offline_config["cs_config"]
            config_base_path = utilities.get_config_directory()  # type: os.fspath
            new_config_path = os.path.join(config_base_path, offline_config["cs_config"])
            assert os.path.exists(new_config_path), new_config_path
            return new_config_path

    @staticmethod
    def get_main_working_dir(config_path: Union[str, os.PathLike]) -> Union[str, os.PathLike]:
        main_config_path = DataSelectionTask.get_main_config_path(config_path=config_path)
        with open(main_config_path) as f:
            offline_config = yaml.load(f)
        return offline_config["working_dir"]

    @staticmethod
    def get_possible_subsample_extensions(input_dict: Dict[str, Any]) -> List[str]:
        return SelectionSetupInterpreter.get_possible_subsample_extensions(input_dict=input_dict)

    def data_selection(self) -> None:
        df = pd.read_hdf(path_or_buf=self.input_dict[self.input_key], key=self.file_name)
        assert isinstance(df, pd.DataFrame), type(df)

        if self.data_type.startswith("Data") \
                and utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name() not in df.columns:
            utilities.add_unique_event_id(df=df, data_type=f"{self.data_type}{self.stream_tag}")

        if self.mva_info is not None:
            mva_setup_name = self.mva_info.mva_setup_name
            weight_file_path = self.input_dict[MVASelectorTrainTask.get_weight_output_key(setup_name=mva_setup_name)]
            var_file_path = self.input_dict[MVASelectorTrainTask.get_variables_output_key(setup_name=mva_setup_name)]
            self.event_selection.apply_selection_mva(  # TODO: Order of application?
                df=df,
                mva_selection_weight_file_path=weight_file_path,
                mva_selection_var_list_file_path=var_file_path
            )

        for selection_info in self.selection_setup_handler.get_selection_infos():
            selected_df, eval_info = self.event_selection.apply_selection(
                df=df,
                selection_info=selection_info,
                evaluate_selection=True,
                data_type=self.data_type
            )

            # TODO: MVA Application after preselection?
            # TODO: Write out dataset for MVA Training after preselection
            # TODO: Cut on MVA output? Where? Currently it is done in self.event_selection.apply_selection!

            selected_df.to_hdf(path_or_buf=self.get_temp_file_path(path=selection_info.target_path), key=self.file_name)
            os.rename(src=self.get_temp_file_path(path=selection_info.target_path), dst=selection_info.target_path)

            assert eval_info is not None
            with open(self.get_temp_file_path(path=selection_info.info_target_path), 'w') as out_file:
                json.dump(eval_info, out_file, indent=2)
            os.rename(
                src=self.get_temp_file_path(path=selection_info.info_target_path),
                dst=selection_info.info_target_path
            )

            # TODO: How to extend Selection info if Selection is split into multiple tasks?

            for subselection_info in self.selection_setup_handler.get_subselection_info(selection_info=selection_info):
                # TODO: Decide for which sub selections the MVA method should be applied in multiple steps
                selected_df, _ = self.event_selection.apply_selection(df=df, selection_info=subselection_info)
                tmp_sub_sel_target_path = self.get_temp_file_path(path=subselection_info.target_path)
                selected_df.to_hdf(path_or_buf=tmp_sub_sel_target_path, key=self.file_name)
                os.rename(src=tmp_sub_sel_target_path, dst=subselection_info.target_path)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.data_selection()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {
            output_key: luigi.LocalTarget(target_path)
            for output_key, target_path in self.selection_setup_handler.get_selection_output_dict().items()
        }

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(DataReductionTask)
        if self.mva_info is not None:
            yield self.clone(
                MVASelectorTrainTask,
                mva_selector_setup_name=self.mva_info.mva_setup_name,
                working_dir=self.get_main_working_dir(config_path=self.offline_config_path),
                main_offline_config_path=self.get_main_config_path(config_path=self.offline_config_path)
            )
            yield self.clone(
                MVASelectorTestTask,
                mva_selector_setup_name=self.mva_info.mva_setup_name,
                mva_selector_working_point=self.mva_info.mva_post_cut,
                working_dir=self.get_main_working_dir(config_path=self.offline_config_path),
                main_offline_config_path=self.get_main_config_path(config_path=self.offline_config_path)
            )
