"""
Tasks for evaluating different selections

Patrick Ecker 2020
"""

import os
import json
import logging
import pandas as pd
import numpy as np
import b2luigi as luigi
import matplotlib.pyplot as plt

from collections import abc
from luigi.util import inherits
from typing import Union, Optional, Tuple, Dict, Generator

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_style import set_matplotlibrc_params

from rdstar.mva_selection.mva_selector_setups import MVASelectorSetups

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.combiner_tasks import SampleCombinerTask
from rdstar.offline_analysis.tasks.manager_tasks import SelectionManagerTask
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "SelectionEvaluationTask"
]


@inherits(SelectionManagerTask)
class SelectionEvaluationTask(utilities.BaseOfflineTask):
    mva_setup_name_for_truth = luigi.Parameter(positional=False)  # type: str
    reference_selection = luigi.OptionalParameter(positional=False, default="")  # type: Optional[str]
    base_dir_name = "SelectionEvaluation"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.working_dir, str), self.working_dir
        self.base_path = os.path.join(os.fspath(self.working_dir), self.base_dir_name)

        self._number_of_selections = len(self.selection_names)  # type: int
        assert self._number_of_selections > 0, "Only one selection name provided. Nothing to compare..."

        self._truth_mva_setup_class = MVASelectorSetups.get_selection_mva_setup(
            setup_name=self.mva_setup_name_for_truth)
        self._truth_col_name = self._truth_mva_setup_class.get_target_column_name()

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    @property
    def selection_names(self) -> Tuple[str, ...]:
        assert isinstance(self.selection_setup_names, tuple), type(self.selection_setup_names)
        assert len(self.selection_setup_names) > 0, self.selection_setup_names
        assert all(isinstance(sel_name, str) for sel_name in self.selection_setup_names), self.selection_setup_names
        assert len(set(self.selection_setup_names)) == len(self.selection_setup_names), self.selection_setup_names
        return self.selection_setup_names

    @property
    def number_of_selections(self) -> int:
        return self._number_of_selections

    @property
    def combo_names(self) -> Tuple[str, ...]:
        assert isinstance(self.config_combinations, abc.Sequence), type(self.config_combinations)
        assert all(isinstance(combo_info, tuple) for combo_info in self.config_combinations)
        assert all(len(combo_info) == 3 for combo_info in self.config_combinations)
        assert all(isinstance(combo_name, str) for combo_name, _, _ in self.config_combinations)
        return tuple([combo_name for combo_name, _, _ in self.config_combinations])

    @property
    def reference_selection_name(self) -> str:
        if self.reference_selection is None:
            return self.selection_names[0]
        elif isinstance(self.reference_selection, str) and self.reference_selection in self.selection_names:
            return self.reference_selection
        else:
            raise ValueError(f"The value {self.reference_selection} for the parameter reference_selection is invalid!"
                             f"\nValid options are\n\t- " + "\n\t- ".join(self.selection_names))

    @staticmethod
    def get_data_dict_key(combination_name: str, selection_name: str) -> str:
        return f"{combination_name}_{selection_name}"

    def get_data_frames(self) -> Dict[str, pd.DataFrame]:
        df_dict = {}
        for selection_name in self.selection_names:
            for combo_name in self.combo_names:
                logging.info(f"Loading combo {combo_name} for SelectionEvaluationTask.")
                comb_sel_key = self.get_data_dict_key(combo_name, selection_name)
                assert comb_sel_key not in df_dict, (comb_sel_key, df_dict.keys())

                if combo_name in rdstar_samples.recorded_data_samples:
                    continue
                sel_tag = SelectionSetups.get_main_selection_tag(selection_name=selection_name, cs_tag=self.cs_tag)
                df_path = self.input_dict[f"{combo_name}_{sel_tag}"]
                df = pd.read_hdf(df_path)
                assert isinstance(df, pd.DataFrame)

                if self._truth_col_name not in df.columns:
                    self._truth_mva_setup_class.data_preparation_for_validation(df=df)

                df_dict[comb_sel_key] = df

        return df_dict

    def get_candidate_numbers(self) -> Dict[str, Dict[str, int]]:
        numbers_dict = {}  # type: Dict[str, Dict[str, int]]
        data = self.get_data_frames()  # type: Dict[str, pd.DataFrame]

        for combo_name in self.combo_names:
            if combo_name in rdstar_samples.recorded_data_samples:
                continue
            numbers = {}  # type: Dict[str, int]
            for selection_name in self.selection_names:
                comb_sel_key = self.get_data_dict_key(combo_name, selection_name)
                assert comb_sel_key in data.keys(), (comb_sel_key, data.keys())

                df = data[comb_sel_key]
                n_corr = len(df.loc[df[self._truth_col_name] == 1].index)
                n_incorr = len(df.loc[df[self._truth_col_name] == 0].index)
                assert len(df.index) == n_corr + n_incorr, (len(df.index), n_corr + n_incorr, n_corr, n_incorr)

                numbers[f"{selection_name}_corr"] = n_corr
                numbers[f"{selection_name}_incorr"] = n_incorr
            numbers_dict[combo_name] = numbers

        return numbers_dict

    @staticmethod
    def safe_division(numerator: int, denominator: int) -> float:
        if numerator == 0:
            # Returns 1. if division by zero!
            return 1.
        return numerator / denominator if denominator != 0 else 0.

    def plot_candidate_numbers(
            self,
            numbers: Dict[str, Dict[str, int]],
            output_path: Optional[str] = None,
            plot_name_tag: Optional[str] = None,
            dry_run: bool = False
    ) -> Dict[str, Union[str, os.fspath, os.PathLike]]:
        file_suffix = "" if plot_name_tag is None else f"_{plot_name_tag}"
        file_name = f"selection_comparison_plot{file_suffix}"
        if output_path is not None:
            output_info = {file_name: os.path.join(output_path, f"{file_name}.pdf")}
        else:
            output_info = {}

        if dry_run:
            return output_info

        set_matplotlibrc_params()

        fig, axes = plt.subplots(nrows=2, ncols=2, sharey="all", gridspec_kw={'hspace': 0.2, 'wspace': 0.1})
        (ax1, ax2), (ax3, ax4) = axes

        i = 0
        ref_sel_name = self.reference_selection_name
        for sel in self.selection_names:
            ax1.barh(
                y=np.arange(len(numbers.keys())) + i,
                width=[numbers[combo_name][f"{sel}_corr"] for combo_name in numbers.keys()],
                height=0.8 / self.number_of_selections,
                label=sel
            )
            ax2.barh(
                y=np.arange(len(numbers.keys())) + i,
                width=[numbers[combo_name][f"{sel}_incorr"] for combo_name in numbers.keys()],
                height=0.8 / self.number_of_selections,
                label=sel
            )

            ratios_corr = np.array([
                self.safe_division(numbers[combo_name][f"{sel}_corr"], numbers[combo_name][f"{ref_sel_name}_corr"])
                for combo_name in numbers.keys()
            ])
            ratios_incorr = np.array([
                self.safe_division(numbers[combo_name][f"{sel}_incorr"], numbers[combo_name][f"{ref_sel_name}_incorr"])
                for combo_name in numbers.keys()
            ])

            # Plotting ratios minus 1.0 to show improvement relative to reference selection.
            # Adding 0.005 * max absolute value to visualize also if no improvement is achieved and to show reference!
            ax3.barh(
                y=np.arange(len(numbers.keys())) + i,
                width=ratios_corr - 1. + 0.005 * max(np.abs(ratios_corr)),
                height=0.8 / self.number_of_selections,
                label=sel
            )
            ax4.barh(
                y=np.arange(len(numbers.keys())) + i,
                width=ratios_incorr - 1. + 0.005 * max(np.abs(ratios_incorr)),
                height=0.8 / self.number_of_selections,
                label=sel
            )

            i += (1 / self.number_of_selections - 0.8 / (4 * self.number_of_selections))

        for ax in fig.get_axes():
            ax.set_xscale("log")

        ax1.set_yticks(np.arange(len(numbers.keys())) + 0.8 / self.number_of_selections)
        ax1.set_yticklabels(numbers.keys(), rotation=0)

        for ax in [ax1, ax2]:
            ax.tick_params(axis='both', which="both", labelsize=7)

        ax2.legend(loc="best")
        ax1.set_title("Correct Reconstructed")
        ax2.set_title("Incorrect Reconstructed")

        for ax in [ax3, ax4]:
            ax.set_xticks([])
            ax.set_xscale("linear")

        if output_path:
            export(
                fig=fig,
                filename=file_name,
                target_dir=output_path,
                close_figure=True
            )
        return output_info

    def run_selection_evaluation(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        numbers_dict_path = os.path.join(self.base_path, "numbers_dict.json")
        output_dict = {"numbers_dict": numbers_dict_path}
        plot_output_info_dict = self.plot_candidate_numbers(numbers={}, output_path=self.base_path, dry_run=True)
        output_dict.update(plot_output_info_dict)

        if not dry_run:
            numbers_dict = self.get_candidate_numbers()
            tmp_path = utilities.BaseOfflineTask.get_temp_file_path(path=numbers_dict_path)
            with open(file=tmp_path, mode="w") as f:
                json.dump(numbers_dict, fp=f)
            os.rename(src=tmp_path, dst=numbers_dict_path)

            self.plot_candidate_numbers(numbers=numbers_dict, output_path=self.base_path, dry_run=False)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.run_selection_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.run_selection_evaluation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence), type(self.config_combinations)
        assert all(isinstance(combo_info, tuple) for combo_info in self.config_combinations), self.config_combinations
        assert all(len(combo_info) == 3 for combo_info in self.config_combinations), self.config_combinations

        for selection_name in self.selection_names:
            for combo_name, _, relevant_configs in self.config_combinations:
                yield self.clone(
                    SampleCombinerTask,
                    combo_name=combo_name,
                    selection_name=selection_name,
                    config_params=relevant_configs
                )
