"""
Tasks performing the Sideband Validation

Felix Metzner 2019
"""

import os
import copy
import logging
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from collections import abc, defaultdict

from typing import Optional, List, Tuple, Dict, Generator

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection
from rdstar.offline_analysis.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.validation.sideband_validation import plot_data_mc_comparison, sideband_info_collection, \
    SidebandInfo, create_sideband_infos_with_systematics

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask, WeightAnalysisTask

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "SidebandEvaluationWrapperTask",
    "SidebandEvaluationTask"
]


@inherits(WeightAnalysisTask)
class SidebandEvaluationWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    def get_task_file_dir(self) -> os.fspath:
        return os.fspath(self.working_dir)

    @staticmethod
    def get_sorted_sideband_info_dict() -> Dict[str, List[SidebandInfo]]:
        sideband_infos = copy.copy(sideband_info_collection)  # type: List[SidebandInfo]
        sorted_sideband_info_dict = defaultdict(list)
        for sideband_info in sideband_infos:
            sorted_sideband_info_dict[sideband_info.key].append(sideband_info)

        return sorted_sideband_info_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for side_band_key in self.get_sorted_sideband_info_dict().keys():
            yield self.clone(SidebandEvaluationTask, side_band_key=side_band_key)


@inherits(WeightAnalysisTask)
class SidebandEvaluationTask(utilities.BaseOfflineTask):
    side_band_key = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "SidebandValidation"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert isinstance(self.working_dir, (str, os.PathLike))
        assert isinstance(self.selection_name, str), self.selection_name
        assert isinstance(self.cs_tag, str), self.cs_tag

        self.base_path = os.path.join(self.working_dir, self.base_dir_name, self.selection_name)  # type: os.fspath

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.mc_sample_keys = rdstar_samples.main_sample_keys
        self.data_sample_key = "DataY4S"

        self.full_weight_column_name = "__weight_final_wFEI_corr__"  # TODO: Remove once provided by weight finalization

        self.key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}_"

        side_band_info_dict = SidebandEvaluationWrapperTask.get_sorted_sideband_info_dict()
        self.side_band_infos = side_band_info_dict[self.side_band_key]  # type: List[SidebandInfo]
        self.side_band_infos_with_sys = create_sideband_infos_with_systematics(  # type: List[SidebandInfo]
            sideband_info_list=self.side_band_infos,
            systematics_collection=full_systematics_cols,
            sys_name_tag=full_systematics_name
        )

        self.input_appendix = f"{WeightFinalizationTask.appendix}_{ResolutionCorrectionTask.appendix}"
        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_mc_and_data_file_paths(self) -> Optional[Tuple[List[os.PathLike], os.PathLike]]:
        full_key = f"{self.key_prefix}{self.side_band_key}"

        mc_paths = [  # type: List[os.fspath]
            p for mc_key in self.mc_sample_keys for k, p in self.input_dict.items()
            if (k.startswith(f"{mc_key}_{self.input_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5")
        ]

        if len(mc_paths) == 0:
            return None

        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p for k, p in self.input_dict.items()
            if (k.startswith(f"{self.data_sample_key}_{self.input_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        return mc_paths, data_path

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        mc_paths, data_path = self.get_mc_and_data_file_paths()

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame)

        return mc_sample, data_sample

    @property
    def sideband_samples_are_available(self) -> bool:
        paths = self.get_mc_and_data_file_paths()
        return False if paths is None else True

    def produce_sideband_plot(
            self,
            mc_df: Optional[pd.DataFrame],
            data_df: Optional[pd.DataFrame],
            input_tuple: Tuple[SidebandInfo, str],
            dry_run: bool
    ) -> Dict[str, os.fspath]:
        sideband_info, style = input_tuple
        assert sideband_info.key == self.side_band_key, (sideband_info.key, self.side_band_key)

        if not dry_run:
            assert isinstance(mc_df, pd.DataFrame), type(mc_df)
            assert isinstance(data_df, pd.DataFrame), type(data_df)

        plot_output_dict = plot_data_mc_comparison(
            mc_df=mc_df,
            data_df=data_df,
            dry_run=dry_run,
            weight_col=self.full_weight_column_name,
            sideband_info=sideband_info,
            style=style,
            normalize_mc_to_data=True,
            ratio_type="vs_uncert",
            gof_check_method="toys",
            adaptive_binning=True,
            export_path=self.base_path
        )

        return plot_output_dict

    def do_sideband_validation(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        output_dict = {}

        if not self.sideband_samples_are_available:
            return output_dict

        mc_df, data_df = (None, None)
        if not dry_run:
            mc_df, data_df = self.get_mc_and_data_sample()
            assert self.full_weight_column_name not in mc_df.columns
            assert WeightFinalizationTask.final_weight_col in mc_df.columns
            assert TagCorrection.tagging_correction_name in mc_df.columns

            product_cols = [
                WeightFinalizationTask.final_weight_col,
                TagCorrection.tagging_correction_name
            ]
            mc_df.loc[:, self.full_weight_column_name] = mc_df[product_cols].product(axis=1)

        sb_plot_infos = [(sb_info, style) for sb_info in self.side_band_infos_with_sys for style in sb_info.styles]
        assert all(sb_info.key == self.side_band_key for sb_info, _ in sb_plot_infos), \
            (self.side_band_key, [sbi.key for sbi, _ in sb_plot_infos])

        for sb_plot_info_tuple in sb_plot_infos:
            eval_output = self.produce_sideband_plot(
                mc_df=mc_df,
                data_df=data_df,
                input_tuple=sb_plot_info_tuple,
                dry_run=dry_run
            )

            assert all(k not in output_dict.keys() for k in eval_output.keys()), \
                [k for k in eval_output.keys() if k in output_dict.keys()]
            assert all(v not in output_dict.values() for v in eval_output.values()), \
                [v for v in eval_output.values() if v in output_dict.values()]
            output_dict.update(eval_output)

        assert len(output_dict) % len(sb_plot_infos) == 0, (len(output_dict), len(sb_plot_infos))
        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_sideband_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_sideband_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def complete(self):
        # Catching special case of a SidebandEvaluationTask which cannot produce output.
        if isinstance(self.output(), dict) and len(self.output()) == 0:
            logging.warning(f"SidebandEvaluationTask cannot produce output for "
                            f"selection name {self.selection_name} and sideband key {self.side_band_key}!")
            return True
        return super().complete()

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations
            )
