"""
Utility Functions for Analysis Tasks

Felix Metzner 2019
"""

import os
import json
import rdstar
import socket
import inspect

import numpy as np
import pandas as pd
import b2luigi as luigi

from typing import Any, Union, Optional, List, Dict, NamedTuple

__all__ = [
    "TaskType",
    "InputDictType",
    "GlobalParameters",
    "BaseOfflineTask",
    "get_config_directory",
    "host_check",
    "load_dict_from_json",
    "to_float_if_float",
    "fix_json_dict",
    "decompose_event_id",
    "recreate_event_id",
    "UniqueEventIDColumnInfos",
    "add_unique_candidate_id"
]

TaskType = luigi.Task
InputDictType = Dict[str, os.fspath]


class GlobalParameters(luigi.Config):
    work_host_list = luigi.ListParameter(default=[])
    copy_host_list = luigi.ListParameter(default=[])

    n_subsets_mixed = luigi.IntParameter(default=1)
    n_subsets_generic = luigi.IntParameter(default=1)


class BaseOfflineTask(luigi.Task):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_task_file_dir(self) -> os.fspath:
        assert self.working_dir is not None
        return os.fspath(self.working_dir)

    def get_log_file_dir(self) -> os.fspath:
        assert self.get_task_file_dir() is not None
        return self.get_task_file_dir()

    @staticmethod
    def get_temp_file_path(
            path: Union[str, os.PathLike, os.fspath],
            prefix: Optional[str] = None
    ) -> Union[str, os.PathLike, os.fspath]:
        if prefix is None:
            prefix = "temp"
        dir_path, file_name = os.path.split(path)
        return os.path.join(dir_path, f"{prefix}_{file_name}")


def get_config_directory() -> os.fspath:
    return os.path.join(os.path.dirname(inspect.getfile(rdstar)), "configs")


def host_check(hosts: List[str], message: Optional[str] = None) -> bool:
    if not hosts:
        return True

    if message is None:
        message = "The task must be run on a suitable machine...\n"
        message += f"Switch to any of the following hosts:\n {hosts}"

    if not any(host in socket.gethostname() for host in hosts):
        raise RuntimeError(message)

    return True


def load_dict_from_json(path: os.fspath) -> Dict[Any, Any]:
    with open(path, 'r') as in_file:
        eval_dict = json.load(in_file)
    return fix_json_dict(eval_dict)


def to_float_if_float(input_string: str) -> Union[float, str]:
    try:
        num = float(input_string)
    except ValueError:
        return input_string
    return num


def fix_json_dict(json_dict: Dict[str, Any]) -> Dict[Union[float, str], Any]:
    if not isinstance(json_dict, dict):
        return json_dict
    return {to_float_if_float(k): fix_json_dict(v) for k, v in json_dict.items()}


class UniqueEventIDColumnInfos(NamedTuple):
    event_id_col: str = "unique_event_id"
    data_type_col: str = "temp_data_type_info"
    orig_exp_col: str = "expNum"
    orig_run_col: str = "runNum"
    orig_evt_col: str = "evtNum"
    orig_other_col: str = "totalEnergyOfParticlesInList__bogamma__clfsp__bc"
    exp_col: str = "temp_exp_info"
    run_col: str = "temp_run_info"
    evt_col: str = "temp_evt_info"
    other_col: str = "temp_gamma_energy_info"
    split_str: str = "_"
    split_n: int = 4
    round_to: int = 6

    @staticmethod
    def get_unique_event_id_col_name() -> str:
        return UniqueEventIDColumnInfos().event_id_col


def add_unique_event_id(
        df: pd.DataFrame,
        data_type: str,
) -> None:
    num_cols = len(df.columns)
    col_info = UniqueEventIDColumnInfos()
    assert col_info.event_id_col not in df.columns, col_info.event_id_col

    assert col_info.orig_exp_col in df.columns, col_info.orig_exp_col
    assert col_info.orig_run_col in df.columns, col_info.orig_run_col
    assert col_info.orig_evt_col in df.columns, col_info.orig_evt_col
    assert col_info.orig_other_col in df.columns, col_info.orig_other_col

    assert col_info.data_type_col not in df.columns, col_info.data_type_col
    df.loc[:, col_info.data_type_col] = data_type

    temp_cols = [col_info.data_type_col]
    info_cols = [col_info.data_type_col]
    int_info_cols = [col_info.orig_exp_col, col_info.orig_run_col, col_info.orig_evt_col]
    float_info_cols = [col_info.orig_other_col]

    for col in int_info_cols:
        converted_col_name = f"temp_{col}"
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, converted_col_name] = df[col].astype(float).astype(str)
        temp_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    for col in float_info_cols:
        rounded_col_name = f"temp_{col}_rounded"
        assert rounded_col_name not in df.columns, rounded_col_name
        converted_col_name = f"temp_{col}_str"
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, rounded_col_name] = df[col].round(decimals=col_info.round_to)
        temp_cols.append(rounded_col_name)
        df.loc[:, converted_col_name] = df[rounded_col_name].astype(str)
        temp_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    df.loc[:, col_info.event_id_col] = pd.Series(df.loc[:, info_cols].values.tolist()).str.join(col_info.split_str)

    for column in temp_cols:
        del df[column]

    assert not pd.isnull(df[col_info.event_id_col]).any()
    # Checking if there is only the one new column in the data frame now: The event_id_col.
    assert len(df.columns) == num_cols + 1, (len(df.columns), num_cols + 1)


def decompose_event_id(df: pd.DataFrame, remove_original_column: bool = False) -> None:
    col_info = UniqueEventIDColumnInfos()
    num_cols = len(df.columns)

    assert col_info.event_id_col in df.columns, col_info.data_type_col

    assert col_info.data_type_col not in df.columns, col_info.data_type_col
    assert col_info.exp_col not in df.columns, col_info.exp_col
    assert col_info.run_col not in df.columns, col_info.run_col
    assert col_info.evt_col not in df.columns, col_info.evt_col
    assert col_info.other_col not in df.columns, col_info.other_col

    split_id = df.loc[:, col_info.event_id_col].str.rsplit(pat=col_info.split_str, n=col_info.split_n, expand=True)
    assert not pd.isnull(split_id).any().any()

    df.loc[:, col_info.data_type_col] = split_id[0].astype('category')
    df.loc[:, col_info.exp_col] = split_id[1].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.run_col] = split_id[2].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.evt_col] = split_id[3].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.other_col] = split_id[4].astype("float")

    if remove_original_column:
        del df[col_info.event_id_col]
        assert len(df.columns) == num_cols + 4, (len(df.columns), num_cols)
    else:
        assert len(df.columns) == num_cols + 5, (len(df.columns), num_cols)


def recreate_event_id(df: pd.DataFrame, remove_original_columns: bool = False) -> None:
    col_info = UniqueEventIDColumnInfos()

    assert col_info.data_type_col in df.columns, col_info.data_type_col
    assert col_info.exp_col in df.columns, col_info.exp_col
    assert col_info.run_col in df.columns, col_info.run_col
    assert col_info.evt_col in df.columns, col_info.evt_col
    assert col_info.other_col in df.columns, col_info.other_col

    assert col_info.event_id_col not in df.columns, col_info.data_type_col

    info_cols = [col_info.data_type_col]
    int_info_cols = [col_info.exp_col, col_info.run_col, col_info.evt_col]
    float_info_cols = [col_info.other_col]
    original_info_cols = info_cols + int_info_cols + float_info_cols
    converted_cols = []

    for col in int_info_cols:
        converted_col_name = f"temp_{col}"
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, converted_col_name] = df[col].astype(float).astype(str)
        converted_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    for col in float_info_cols:
        rounded_col_name = f"temp_{col}_rounded"
        converted_col_name = f"temp_{col}_str"
        assert rounded_col_name not in df.columns, rounded_col_name
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, rounded_col_name] = df[col].round(decimals=col_info.round_to)
        converted_cols.append(rounded_col_name)
        df.loc[:, converted_col_name] = df[rounded_col_name].astype(str)
        converted_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    df.loc[:, col_info.event_id_col] = pd.Series(df.loc[:, info_cols].values.tolist()).str.join(col_info.split_str)
    assert not pd.isnull(df[col_info.event_id_col]).any()

    for column in converted_cols:
        del df[column]
    if remove_original_columns:
        for column in original_info_cols:
            del df[column]


def add_unique_candidate_id(
        df: pd.DataFrame,
        id_column_name: str,
        event_id_name: str,
        candidate_id_name: str,
        check_if_unique: bool = True,
        mask: Optional[pd.Series] = None
) -> np.ndarray:
    assert event_id_name in df.columns, event_id_name
    assert candidate_id_name in df.columns, candidate_id_name
    assert id_column_name not in df.columns, id_column_name

    active_mask = False
    if mask is None:
        mask = pd.Series(np.full(shape=len(df.index), fill_value=True))
    else:
        active_mask = True
        assert len(mask) == len(df.index), (len(mask), len(df.index))

    cand_id_col_name = candidate_id_name
    if df[candidate_id_name].dtype == np.float or df[candidate_id_name].dtype == np.int:
        temp_cand_id_as_str_name = f"temp_{candidate_id_name}"
        assert temp_cand_id_as_str_name not in df.columns, temp_cand_id_as_str_name

        df.loc[mask, temp_cand_id_as_str_name] = df[candidate_id_name].astype(str)
        if active_mask:
            df.loc[~mask, temp_cand_id_as_str_name] = 'not_in_mask'

        cand_id_col_name = temp_cand_id_as_str_name

    cols_to_combine = [event_id_name, cand_id_col_name]
    df.loc[mask, id_column_name] = pd.Series(df.loc[mask, cols_to_combine].values.tolist()).str.join("_")
    if active_mask:
        df.loc[~mask, id_column_name] = 'not_in_mask'

    assert not pd.isnull(df.loc[mask, id_column_name]).any()

    assert id_column_name in df.columns, id_column_name
    if check_if_unique:
        assert len(df.loc[mask].index) == len(df.loc[mask, id_column_name].unique()), \
            (len(df.loc[mask].index), len(df.loc[mask, id_column_name].unique()))

    if cand_id_col_name != candidate_id_name:
        del df[cand_id_col_name]

    return df.loc[mask, id_column_name].values
