"""
Weight Evaluation and Finalization Tasks

Felix Metzner 2019
"""

import os
import json
import numpy as np
import pandas as pd
import multiprocessing
import b2luigi as luigi

from functools import partial
from luigi.util import inherits
from collections import defaultdict, Iterable

from typing import Any, Union, Optional, Tuple, List, Dict, Generator

from analysistools.pid_corrections.pid_weight_plots import plot_pid_weights

from rdstar.mc_reweighting.BtoDlnu.BtoDlnu_weights import BtoDFFReweighting
from rdstar.mc_reweighting.BtoDstlnu.BtoDstlnu_weights import BtoDstarFFReweighting
from rdstar.mc_reweighting.mc_reweighting_plots import plot_reweighted as plot_ff_reweighted

from rdstar.offline_analysis.rdstar_samples import data_type_combinations
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.corrections.k_short_corrections import KShortEffCorrector
from rdstar.offline_analysis.corrections.tracking_corrections import TrackingEffCorrector
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection
from rdstar.offline_analysis.rdstar_sample_weights import scale_weight_column_name as scale_weight_col

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.combiner_tasks import SampleCombinerTask
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask

__all__ = [
    "WeightAnalysisTask",
    "WeightFinalizationTask"
]


class WeightAnalysisTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)
    selection_name = luigi.Parameter(positional=False)
    cs_tag = luigi.Parameter(positional=False)
    config_combinations = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name = "Weight_Analysis"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.working_dir, (str, os.PathLike))

        assert isinstance(self.selection_name, str), self.selection_name
        assert isinstance(self.cs_tag, str), self.cs_tag
        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag
        )

        self.base_path = os.path.join(os.fspath(self.working_dir), self.base_dir_name, self.main_sel_tag)

        self.pid_weight_info_key = "pid_weight_info"
        self.ff_d_weight_info_key = "ff_d_weight_info"
        self.ff_ds_weight_info_key = "ff_ds_weight_info"
        self.ff_dss_weight_info_key = "ff_dss_weight_info"

        self.weight_info_keys = [
            self.pid_weight_info_key,
            self.ff_d_weight_info_key,
            self.ff_ds_weight_info_key,
            self.ff_dss_weight_info_key
        ]

        self.evt_id_col = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()
        self.c_id_col = "unique_candidate_id"
        self.id_col = "combined_candidate_id"

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_weight_info(self, weight_key: str) -> pd.DataFrame:
        dfs = [pd.read_hdf(path) for key, path in self.input_dict.items()
               if key.endswith(weight_key) and not key.startswith("SpecialPsi")]

        df = pd.concat(dfs, axis=0, ignore_index=True, sort=False)
        if "index" in df.columns:
            del df["index"]

        del dfs
        if weight_key == self.pid_weight_info_key:
            if self.evt_id_col not in df.columns:
                utilities.recreate_event_id(df=df, remove_original_columns=True)
        return df

    def add_unique_id(
            self,
            df: pd.DataFrame,
            check_if_unique: bool = False,
            mask: Optional[pd.Series] = None
    ) -> np.ndarray:
        id_values = utilities.add_unique_candidate_id(
            df=df,
            id_column_name=self.id_col,
            event_id_name=self.evt_id_col,
            candidate_id_name=self.c_id_col,
            check_if_unique=check_if_unique,
            mask=mask
        )
        return id_values

    def get_selection_conditions(self) -> Tuple[np.ndarray, np.ndarray]:
        dfs = [pd.read_hdf(path) for key, path in self.input_dict.items() if key.endswith(self.main_sel_tag)
               and not key.startswith("Data")]  # type: List[Union[pd.DataFrame, pd.Series, Any]]
        df = pd.concat(
            [sdf.loc[:, [self.evt_id_col, self.c_id_col]] for sdf in dfs],
            axis=0,
            ignore_index=True,
            sort=False
        )
        if "index" in df.columns:
            del df["index"]

        unique_candidate_id = self.add_unique_id(df=df, check_if_unique=True)

        return df[self.evt_id_col].unique(), np.unique(unique_candidate_id)

    def produce_weight_plots(
            self,
            df: pd.DataFrame,
            weight_key: str,
            output_path: os.fspath,
            selection_conditions: Optional[Tuple[np.ndarray, np.ndarray]] = None
    ) -> None:
        if weight_key == self.pid_weight_info_key:
            plot_pid_weights(df, output_path, convert_rad_to_degree=True)
            if selection_conditions is not None:
                evt_cond = df[self.evt_id_col].isin(selection_conditions[0])
                self.add_unique_id(df=df)
                cand_cond = df.loc[evt_cond, self.id_col].isin(selection_conditions[1])
                plot_pid_weights(
                    df=df.loc[evt_cond].loc[cand_cond],
                    output_path=output_path,
                    convert_rad_to_degree=False,  # Conversion not necessary, because it is already done above.
                    file_name_extension="withCuts"
                )
        else:
            plot_ff_reweighted(df, output_path)
            evt_cond = None
            cand_cond = None
            if selection_conditions is not None:
                evt_cond = df[self.evt_id_col].isin(selection_conditions[0])
                self.add_unique_id(df=df)
                cand_cond = df.loc[evt_cond, self.id_col].isin(selection_conditions[1])
                plot_ff_reweighted(
                    df=df.loc[evt_cond].loc[cand_cond],
                    output_path=output_path,
                    file_name_extension="withCuts"
                )

            if weight_key == self.ff_d_weight_info_key:
                BtoDFFReweighting.finalize_weights(df=df)
            elif weight_key == self.ff_ds_weight_info_key:
                BtoDstarFFReweighting.finalize_weights(df=df)
            else:
                return

            plot_ff_reweighted(
                df=df,
                output_path=output_path,
                file_name_extension="withWeightReset",
                normal_condition=True
            )
            if selection_conditions is not None:
                plot_ff_reweighted(
                    df=df.loc[evt_cond].loc[cand_cond],
                    output_path=output_path,
                    file_name_extension="withCuts_withWeightReset",
                    normal_condition=True
                )

    def analyse_weight_info(self, dry_run: bool = False) -> Generator[Tuple[str, os.fspath], None, None]:
        selection_conds = None
        if not dry_run:
            selection_conds = self.get_selection_conditions()

        for weight_key in self.weight_info_keys:
            output_key = f"combined_{weight_key}"
            output_path = os.path.join(self.base_path, f"{output_key}.h5")
            temp_output_path = os.path.join(self.base_path, f"temp_{output_key}.h5")

            plot_output_key = f"combined_{weight_key}_plot"
            plot_output_path = os.path.join(self.base_path, f"{output_key}_plot.pdf")

            dict_output_key = "pid_hadron_for_flag_dict"
            dict_output_path = os.path.join(self.base_path, f"{dict_output_key}.json")

            if not dry_run:
                df = self.get_weight_info(weight_key)

                if weight_key == self.pid_weight_info_key:
                    flag_dict = ParticleIDCorrection.get_hadron_pid_flag_uncert_dict(df)
                    with open(dict_output_path, 'w') as out_file:
                        json.dump(flag_dict, out_file, indent=2)
                else:
                    df.to_hdf(temp_output_path, key=output_key)
                    os.rename(temp_output_path, output_path)

                self.produce_weight_plots(
                    df=df,
                    weight_key=weight_key,
                    output_path=plot_output_path,
                    selection_conditions=selection_conds,
                )

            yield plot_output_key, plot_output_path
            if weight_key == self.pid_weight_info_key:
                yield dict_output_key, dict_output_path
            else:
                yield output_key, output_path

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        list(self.analyse_weight_info())

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.analyse_weight_info(dry_run=True):
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, Iterable)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield SampleCombinerTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                config_params=relevant_configs
            )


@inherits(WeightAnalysisTask)
@inherits(SampleCombinerTask)
class WeightFinalizationTask(utilities.BaseOfflineTask):
    appendix = "withFinalWeights"
    final_weight_col = "__weight_full__"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_input_files(self) -> Dict[str, os.fspath]:
        input_dict = {}
        for selection_tag in SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag):
            input_dict.update({k: v for k, v in self.input_dict.items() if k.endswith(selection_tag)})
            for ext in DataSelectionTask.get_possible_subsample_extensions(input_dict=self.input_dict):
                input_dict.update({k: v for k, v in self.input_dict.items() if k.endswith(selection_tag + ext)})
        return input_dict

    def get_flag_info_dict(self) -> Dict[str, Any]:
        dict_file_path = self.input_dict["pid_hadron_for_flag_dict"]
        return utilities.load_dict_from_json(path=dict_file_path)

    def add_finalized_weights(self, df: pd.DataFrame, key_hint: Optional[str] = None) -> None:
        combo_hint = self.combo_name
        if key_hint is not None:
            combo_hint += f" with {key_hint} selection"

        # Particle ID Weights
        nom_pid_col, nom_pid_cols = ParticleIDCorrection.apply_pid_weight_finalization(
            df=df,
            flag_uncert_dict=self.get_flag_info_dict()
        )

        # Form Factor Weights for B -> D l nu
        BtoDFFReweighting.finalize_weights(df=df, threshold=3., reset_special_to=0.)

        # Form Factor Weights for B -> D* l nu
        BtoDstarFFReweighting.finalize_weights(
            df=df, threshold=5., set_to=1.0, reset_special_to=0., combo_name=combo_hint
        )

        ff_cols = [c for c in df.columns if c.startswith("weight_ff_") and c.endswith("nom")]
        assert len(ff_cols) == 6, (len(ff_cols), ff_cols)

        for c in ff_cols:
            df.loc[np.isnan(df[c]) & df.data_type.isin(data_type_combinations["Special"]), c] = 1.0
            for cc in [col for col in df.columns if col.startswith(c.replace("_nom", "_var_"))]:
                df.loc[np.isnan(df[cc]) & df.data_type.isin(data_type_combinations["Special"]), cc] = 1.0

        df.loc[:, "__weight_ff_nom__"] = df[ff_cols].product(axis=1)

        # nom_tracking_col = self.get_tracking_variation(df=df)
        #
        # column_combinations = self.get_weight_combinations(
        #     ("__weight_ff_nom__", ff_cols),
        #     (nom_pid_col, nom_pid_cols),
        #     nom_tracking_col,
        #     "__weight_slow_pi_ratio__",
        #     "__weight_k_short_ratio__",
        #     df=df
        # )
        # for varied_weight_name, column_list in column_combinations:
        #     df.loc[:, varied_weight_name] = df[column_list].product(axis=1)
        #     assert pd.notnull(df[varied_weight_name]).all()

        # Combining all weights
        weight_col_list = [
            "__weight_ff_nom__",
            nom_pid_col,
            TrackingEffCorrector.slow_pi_corr_col_name,
            KShortEffCorrector.k_short_corr_col_name,
            scale_weight_col
        ]
        df.loc[:, self.final_weight_col] = df[weight_col_list].product(axis=1)
        assert pd.notnull(df[self.final_weight_col]).all()

    @staticmethod
    def get_tracking_variation(
            df: pd.DataFrame,
            tracking_uncert_col: str = TrackingEffCorrector.tracking_systematic_name
    ) -> str:
        tracking_weight_name = "__weight_tracking_eff_ratio__"
        tracking_var_up_name = "__weight_tracking_eff_var_1__"
        tracking_var_down_name = "__weight_tracking_eff_var_2__"
        df.loc[:, tracking_weight_name] = np.ones(len(df.index))
        df.loc[:, tracking_var_up_name] = np.ones(len(df.index)) + df[tracking_uncert_col]
        df.loc[:, tracking_var_down_name] = np.ones(len(df.index)) - df[tracking_uncert_col]

        return tracking_weight_name

    @staticmethod
    def get_weight_combinations(
            *args: Union[str, List[str], Tuple[str, List[str]]],
            df: pd.DataFrame
    ) -> List[Tuple[str, List[str]]]:
        assert len(args) > 0
        info_dict = {
            "__weight_slow_pi_": ("ratio", "var"),
            "__weight_pid_corr_": ("ratio", "variation"),
            "__weight_k_short_": ("ratio", "var"),
            "__weight_tracking_eff_": ("ratio", "var"),
            "weight_ff_": ("nom", "var"),
        }

        input_args = defaultdict(list)
        columns = []

        for arg in args:
            if isinstance(arg, str):
                assert arg not in columns
                assert arg not in input_args.keys()
                assert "ratio" in arg or "nom" in arg
                assert arg in df.columns
                input_args[arg].append(arg)
                columns.append(arg)
            elif isinstance(arg, list):
                for col in arg:
                    assert col not in columns
                    assert col not in input_args.keys()
                    assert "ratio" in col or "nom" in col
                    assert col in df.columns
                    input_args[col].append(col)
                    columns.append(col)
            elif isinstance(arg, tuple):
                assert all(c not in columns for c in arg[1])
                assert arg[0] not in input_args.keys()
                assert all(c in df.columns for c in arg[1])
                assert arg[0] in df.columns
                assert all(("ratio" in c or "nom" in c) for c in arg[1])
                assert "ratio" in arg[0] or "nom" in arg[0]
                input_args[arg[0]].extend(arg[1])
                columns.extend(arg[1])
            else:
                raise TypeError(f"The input arguments of the function get_weight_combinations must be one of\n"
                                f" - str\n - list(str)\n - tuple(str, list(str))\n"
                                f"However, the argument {arg} of type {type(arg)} is not.")

        combinations = defaultdict(list)

        for col in columns:
            col_type = [c for c in info_dict.keys() if col.startswith(c)][0]
            variation_cols = [c for c in df.columns if c.startswith(col.split(info_dict[col_type][0], 1)[0])
                              and info_dict[col_type][1] in c]
            for var_col in variation_cols:
                new_col_name = f"__weight_full_var_{var_col.replace('weight_', '').replace('__', '')}__"
                combinations[new_col_name].append(var_col)
                combinations[new_col_name].extend([k for k, v in input_args.items() if col not in v])
                combinations[new_col_name].extend([v for cols in input_args.values() if col in cols
                                                   for v in [c for c in cols if c != col]])

        return [(k, v) for k, v in combinations.items()]

    def apply_weight_finalization_for_key(
            self,
            key_path_tuple: Tuple[str, os.fspath],
            dry_run: bool
    ) -> Tuple[str, os.fspath]:
        input_key, input_path = key_path_tuple
        dir_path, file_name = os.path.split(input_path)
        output_name_1, output_name_2 = file_name.split("_", 1)
        output_name = f"{output_name_1}_{self.appendix}_{output_name_2}"
        output_path = os.path.join(dir_path, output_name)
        temp_output_path = os.path.join(dir_path, f"temp_{output_name}")
        output_key_1, output_key_2 = input_key.split("_", 1)
        output_key = f"{output_key_1}_{self.appendix}_{output_key_2}"

        if not dry_run:
            if input_key.startswith("Data"):
                if os.path.exists(output_path):
                    os.remove(output_path)
                os.symlink(input_path, output_path)
            else:
                df = pd.read_hdf(input_path)
                assert isinstance(df, pd.DataFrame)
                self.add_finalized_weights(df, key_hint=output_key_2)
                df.to_hdf(temp_output_path, key=output_key)
                os.rename(temp_output_path, output_path)

        return output_key, output_path

    def apply_weight_finalization(self, dry_run: bool = False) -> Generator[Tuple[str, os.fspath], None, None]:
        with multiprocessing.Pool(processes=10) as pool:
            output_tuples = pool.map(
                partial(WeightFinalizationTask.apply_weight_finalization_for_key, self, dry_run=dry_run),
                self.get_input_files().items()
            )

        for output_key, output_path in output_tuples:
            yield output_key, output_path

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        list(self.apply_weight_finalization())

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.apply_weight_finalization(dry_run=True):
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(
            SampleCombinerTask,
            combo_name=self.combo_name,
            working_dir=self.working_dir,
            selection_name=self.selection_name,
            config_params=self.config_params
        )

        yield self.clone(
            WeightAnalysisTask,
            working_dir=self.working_dir,
            selection_name=self.selection_name,
            config_combinations=self.config_combinations
        )
