"""
Definition of convenience functions for the analysis-specific decay channel reconstruction.

it originates from:
Felix Metzner 2018
"""

import basf2

from analysistools.processing import roe_masks_utils
from analysistools.processing import decay_channels_utils
from analysistools.processing import reco_utilities as reco_utils
from analysistools.processing import best_cand_selection_utils as bcs

from rdstar.online_analysis import roe_masks
from rdstar.online_analysis import skim_cuts
from rdstar.online_analysis import decay_channels
from rdstar.online_analysis import best_cand_selection
from rdstar.online_analysis import final_state_particles as fsp


__all__ = [
    "fsp",
    "Reconstructor"
]


class Reconstructor:
    def __init__(self, path, config, decay_channels_module, do_gen_level_reco=False) -> None:
        self.path = path
        self.config = config
        self.decay_channels = decay_channels_module
        self.do_gen_level_reco = do_gen_level_reco

    def reconstruct_Kshorts(self):
        for channel in self.decay_channels.KS0_decay_channels:
            reco_utils.reconstruct_decay_channel(
                path=self.path,
                decay_label=channel,
                decay_chl_dict=self.decay_channels.decay_channel_dict,
                use_mc_info=self.config["use_mc_info"],
                use_b2bii=self.config["use_b2bii"],
                do_gen_reco=self.do_gen_level_reco
            )

        reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=self.decay_channels.KS0_decay_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="reco"
        )

    def reconstruct_pi0(self):
        for channel in self.decay_channels.pi0_decay_channels:
            reco_utils.reconstruct_decay_channel(
                path=self.path,
                decay_label=channel,
                decay_chl_dict=self.decay_channels.decay_channel_dict,
                use_mc_info=self.config["use_mc_info"],
                use_b2bii=self.config["use_b2bii"],
                do_gen_reco=self.do_gen_level_reco
            )

        reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=self.decay_channels.pi0_decay_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="reco"
        )

    def reconstruct_tau_for_gen(self):
        if self.config["use_mc_info"] is False:
            return
        if not self.do_gen_level_reco:
            return

        for channel in self.decay_channels.tau_decay_channels:
            reco_utils.reconstruct_decay_channel(
                path=self.path,
                decay_label=channel,
                decay_chl_dict=self.decay_channels.decay_channel_dict,
                use_mc_info=self.config["use_mc_info"],
                use_b2bii=self.config["use_b2bii"],
                do_gen_reco=self.do_gen_level_reco
            )

        reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=self.decay_channels.tau_decay_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="all"
        )

    def reconstruct_D(self):
        for channel in self.decay_channels.D_decay_channels:
            reco_utils.reconstruct_decay_channel(
                path=self.path,
                decay_label=channel,
                decay_chl_dict=self.decay_channels.decay_channel_dict,
                use_mc_info=self.config["use_mc_info"],
                use_b2bii=self.config["use_b2bii"],
                do_gen_reco=self.do_gen_level_reco
            )

        particle_lists, _ = reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=self.decay_channels.D_decay_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="all"
        )

        bcs.apply_bcs_from_dict(self.path, best_cand_selection.D_meson_by_dM_bcs, particle_lists)

    def reconstruct_Dstar(self):
        for channel in self.decay_channels.Dstar_decay_channels:
            reco_utils.reconstruct_decay_channel(
                path=self.path,
                decay_label=channel,
                decay_chl_dict=self.decay_channels.decay_channel_dict,
                use_mc_info=self.config["use_mc_info"],
                use_b2bii=self.config["use_b2bii"],
                do_gen_reco=self.do_gen_level_reco
            )

        reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=self.decay_channels.Dstar_decay_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="all"
        )

    def reconstruct_D2star_for_gen(self):
        if self.config["use_mc_info"] is False:
            return
        if not self.do_gen_level_reco:
            return

        for channel in self.decay_channels.D2star_decay_channels:
            reco_utils.fillParticleListsFromMC([(f"{channel}_gen", "")], path=self.path, addDaughters=True)

        # for channel in self.decay_channels.D2star_decay_channels:
        #     reco_utils.reconstruct_decay_channel(
        #         path=self.path,
        #         decay_label=channel,
        #         decay_chl_dict=self.decay_channels.decay_channel_dict,
        #         use_mc_info=self.config["use_mc_info"],
        #         use_b2bii=self.config["use_b2bii"]
        #     )

        # for channel in self.decay_channels.D2star_decay_channels:
        #     reco_utils.findMCDecay(
        #         list_name=f"{channel}_gen",
        #         decay=self.decay_channels.decay_channel_dict["decay_label_to_chain"][channel].format(suf="", neutrino=""),
        #         writeOut=True,
        #         path=path
        #     )

        reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=self.decay_channels.D2star_decay_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="all"
        )

    def reconstruct_sig_B(self):
        sig_b_channels = {
            **self.decay_channels.B_decay_channels,
            **self.decay_channels.B_Dlpi0_decay_channels,
            **self.decay_channels.B_Dlpim_decay_channels,
            **self.decay_channels.Btau_decay_channels,
            **self.decay_channels.B_D2star_decay_channels,
        }
        for channel in sig_b_channels:
            reco_utils.reconstruct_decay_channel(
                path=self.path,
                decay_label=channel,
                decay_chl_dict=self.decay_channels.decay_channel_dict,
                use_mc_info=self.config["use_mc_info"],
                use_b2bii=self.config["use_b2bii"],
                do_gen_reco=self.do_gen_level_reco,
                add_neutrinos=True
            )

        reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=sig_b_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="sig"
        )

    def reconstruct_upsilon4s(self):
        for channel in self.decay_channels.U4S_decay_channels:
            reco_utils.reconstruct_decay_channel(
                path=self.path,
                decay_label=channel,
                decay_chl_dict=self.decay_channels.decay_channel_dict,
                use_mc_info=self.config["use_mc_info"],
                use_b2bii=self.config["use_b2bii"],
                do_gen_reco=self.do_gen_level_reco
            )

        particle_lists, _ = reco_utils.merge_particle_lists(
            path=self.path,
            merge_channel_dict=self.decay_channels.U4S_decay_channels,
            decay_chl_dict=self.decay_channels.decay_channel_dict,
            use_mc_info=self.config["use_mc_info"],
            do_gen_reco=self.do_gen_level_reco,
            label="all"
        )

        bcs_dicts = tuple([
            best_cand_selection.upsilon4s_with_fei_probability_bcs,
            best_cand_selection.upsilon4s_with_vertex_fit_probability_bcs,
            best_cand_selection.upsilon4s_with_roe_eextra_bcs,
            best_cand_selection.upsilon4s_with_roe_neextra_bcs,
            best_cand_selection.upsilon4s_with_deltaE_bcs,
            best_cand_selection.upsilon4s_with_DMeson_deltaM_bcs,
        ])
        for bcs_dict in bcs_dicts:
            bcs.apply_bcs_from_dict(path=self.path, bcs_dict=bcs_dict, particle_lists=particle_lists)

    def print_decay_summary(
            self,
            fsp_collection=None,
            list_of_skims=None,
            bcs_dict=None,
            roe_mask_list=None
    ):
        if fsp_collection:
            print("Final State Particles Selection Criteria:")
            print(fsp_collection)

        if self.decay_channels.decay_channel_dict:
            print("\n\nDecay Channels Reconstruction Criteria:")
            decay_channels_utils.print_decay_channels_table(decay_chl_dict=self.decay_channels.decay_channel_dict)

        if list_of_skims:
            print("\n\nPreselection and Skim Criteria:")
            skim_cuts.print_skim_cuts(skim_dict_list=list_of_skims)

        if bcs_dict:
            print("\n\nBest Candidate Selection Criteria:")
            bcs.print_bcs_settings(bcs_settings_dict=bcs_dict)

        if roe_mask_list:
            b2bii = True
            print(f"\n\nAll Defined ROE Masks for use_b2bii = {b2bii}:")
            roe_masks_utils.print_roe_masks(mask_list=roe_mask_list, use_b2bii=b2bii)


if __name__ == "__main__":
    reco = Reconstructor(path=basf2.create_path(), config=None, decay_channels_module=decay_channels)
    reco.print_decay_summary(
        fsp_collection=fsp.fsp_collection,
        list_of_skims=skim_cuts.list_of_skims,
        bcs_dict=best_cand_selection.my_bcs_settings,
        roe_mask_list=roe_masks.roe_mask_list
    )
