"""
Convenience functions to retrieve some common sets of variables.

It is important to note, that the final Y(4S) particle is always
reconstructed in such a way, that the first daughter (index 0) is
the tag B meson, and the second (index 1) is the signal B meson.

Felix Metzner 2018
"""

from typing import List

from analysistools.processing.observables_utils import VariableCollection, VariableListBase, VariableCollectionFactory
from analysistools.processing import roe_masks_utils

from rdstar.online_analysis import roe_masks


########################
#  Final Variable Set  #
########################

def get_upsilon4s_collection(configuration):
    """
    Retrieve complete set of Upsilon(4S) variables for the analysis.

    The returned list is defined relative to an Y(4S) particle list.
    This Y(4S) list must be defined so that daughter 1 is the signal B
    and daughter 0 is the tag side B meson.

    :param configuration: Analysis configuration.
    :return: VariableCollection of basf2 variables
    """
    ups_coll = UpsilonVariableCollectionFactory(configuration).create("reco")

    tag_b = TagSideVariableCollectionFactory(configuration).create()
    sig_b = get_signal_side_collection(configuration)

    ups_coll.add_daughter_variables(tag_b.as_list(), 0)
    ups_coll.add_daughter_variables(sig_b.as_list(), 1)

    return ups_coll


###################
# Generator Level #
###################

def get_gen_level_upsilon4s_vars(configuration):
    """
    Combines Upsilon(4S), Signal B and TagSide B collections
    of generator level variables.

    :param configuration: Analysis config dictionary.
    :return: Combined VariableCollection.
    """

    ups_coll = UpsilonVariableCollectionFactory(configuration).create("gen")

    tag_b = TagSideVariableCollectionFactory(configuration).create()
    gen_level_b = get_gen_level_b_collection(configuration)

    ups_coll.add_daughter_variables(tag_b.as_list(), 0)
    ups_coll.add_daughter_variables(gen_level_b.as_list(), 1)

    return ups_coll


def get_gen_level_b_collection(configuration):
    """
    Combines Signal B, D/D*, lepton, pi0 and neutrino collections
    for generator level info.

    :param configuration: Analysis config dictionary.
    :return: Combined VariableCollection.
    """
    b_vars = SignalBVariableCollectionFactory(configuration).create("gen")
    d_vars = DVariableCollectionFactory(configuration).create("gen")
    lep_vars = LeptonVariableCollectionFactory(configuration).create("gen")

    # also neutrino variables
    pi0_vars = Pi0VariableCollectionFactory(configuration).create("gen")

    b_vars.add_daughter_variables(d_vars.as_list(), 0)
    b_vars.add_daughter_variables(d_vars.as_list(), 0, 0)
    b_vars.add_daughter_variables(lep_vars.as_list(), 1)
    b_vars.add_daughter_variables(pi0_vars.as_list(), 2)
    b_vars.add_daughter_variables(pi0_vars.as_list(), 3)

    return b_vars


def get_direct_gen_level_b_collection(configuration):
    """
    Combines Signal B, D/D*, lepton, pi0 and neutrino collections
    for generator level info.

    :param configuration: Analysis config dictionary.
    :return: Combined VariableCollection.
    """
    b_vars = SignalBVariableCollectionFactory(configuration).create("gen")
    d_vars = DVariableCollectionFactory(configuration).create("gen")
    lep_vars = LeptonVariableCollectionFactory(configuration).create("gen")

    # These are also the neutrino variables!
    pi0_vars = Pi0VariableCollectionFactory(configuration).create("gen")

    b_vars.add_mc_daughter_variables(d_vars.as_list(), 0)
    b_vars.add_mc_daughter_variables(d_vars.as_list(), 0, 0)
    b_vars.add_mc_daughter_variables(lep_vars.as_list(), 1)
    b_vars.add_mc_daughter_variables(pi0_vars.as_list(), 2)
    b_vars.add_mc_daughter_variables(pi0_vars.as_list(), 3)

    return b_vars


def get_signal_side_collection(configuration):
    """
    Combines Signal B, D/D*, lepton, pi0 and neutrino collections
    for reco level info.

    :param configuration: Analysis config dictionary.
    :return: Combined VariableCollection.
    """
    b_vars = SignalBVariableCollectionFactory(configuration).create("reco")

    dst_vars = DstarVariableCollectionFactory(configuration).create()
    d_vars = DVariableCollectionFactory(configuration).create("reco")
    lep_vars = LeptonVariableCollectionFactory(configuration).create("reco")
    pi0_vars = Pi0VariableCollectionFactory(configuration).create("reco")

    b_vars.add_daughter_variables(dst_vars.as_list(), 0)
    b_vars.add_daughter_variables(d_vars.as_list(), 0, 0)
    b_vars.add_daughter_variables(lep_vars.as_list(), 1)
    b_vars.add_daughter_variables(pi0_vars.as_list(), 2)

    return b_vars


################################
# VariableCollection Factories #
################################

class UpsilonVariableCollectionFactory(VariableCollectionFactory):
    """
    Creates a VariableCollection for the Upsilon(4S) either for
    reco or gen level.
    """

    def _create_reco_level_collection(self) -> VariableCollection:
        u4s_vars = UpsilonVariables(use_decay_hash=False, use_mc_info=self._use_mc_info)
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)
        loose_mc_vars = LooseMCVariables(use_mc_info=self._use_mc_info)
        event_vars = EventVariables(use_mc_info=self._use_mc_info, use_b2bii=self._use_b2bii)
        kin_vars = KinematicVariables()
        roe_vars = ROEVariables(roe_mask=self._roe_mask, use_b2bii=self._use_b2bii)

        return VariableCollection(
            "Upsilon Reco Variables",
            extra=u4s_vars.get(),
            mc=mc_vars.get(),
            loose_mc=loose_mc_vars.get(),
            event=event_vars.get(),
            kinematic=kin_vars.get(),
            roe=roe_vars.get()
        )

    def _create_gen_level_collection(self) -> VariableCollection:
        u4s_vars = UpsilonVariables(use_decay_hash=False, use_mc_info=self._use_mc_info)
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)
        event_vars = EventVariables(use_mc_info=self._use_mc_info, use_b2bii=self._use_b2bii)
        roe_vars = ROEVariables(roe_mask=self._roe_mask,
                                use_b2bii=self._use_b2bii)

        return VariableCollection(
            "Upsilon Gen Variables",
            extra=u4s_vars.get(),
            mc=mc_vars.get(),
            event=event_vars.get(),
            roe=roe_vars.get()
        )

    def create(self, level: str) -> VariableCollection:
        """
        Creates a reco or gen level Pi0 VariableCollection.

        :param level: Either "reco" or "gen".
        :return: VariableCollection based on the level argument.
        """

        if level == "reco":
            return self._create_reco_level_collection()
        elif level == "gen":
            return self._create_gen_level_collection()
        else:
            raise ValueError(f"Keyword level:{level} not known.")


class SignalBVariableCollectionFactory(VariableCollectionFactory):
    """
    Creates a VariableCollection for the Signal Side B meson
    either for reco or gen level.
    """

    def _create_reco_level_collection(self) -> VariableCollection:
        b_vars = SignalSideBVariables(use_mc_info=self._use_mc_info)
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)
        loose_mc_vars = LooseMCVariables(use_mc_info=self._use_mc_info)
        kin_vars = KinematicVariables()
        roe_vars = ROEVariables(roe_mask=self._roe_mask, use_b2bii=self._use_b2bii)
        cs_vars = ContinuumSuppressionVariables()

        return VariableCollection(
            "Signal B Reco Variables",
            extra=b_vars.get(),
            mc=mc_vars.get(),
            loose_mc=loose_mc_vars.get(),
            kinematic=kin_vars.get(),
            roe=roe_vars.get(),
            cs=cs_vars.get()
        )

    def _create_gen_level_collection(self) -> VariableCollection:
        b_vars = SignalSideBGenLevelVariables()
        event_vars = EventVariables(use_mc_info=self._use_mc_info, use_b2bii=self._use_b2bii)
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)
        kin_vars = KinematicVariables()
        mc_kin_vars = MCKinematicVariables(use_mc_info=self._use_mc_info)

        return VariableCollection(
            "Signal B Gen Variables",
            extra=b_vars.get(),
            event=event_vars.get(),
            mc=mc_vars.get(),
            kinematic=kin_vars.get(),
            mc_kinematic=mc_kin_vars.get()
        )

    def create(self, level: str) -> VariableCollection:
        """
        Creates a reco or gen level signal-side B meson VariableCollection.

        :param level: Either "reco" or "gen".
        :return: VariableCollection based on the level argument.
        """

        if level == "reco":
            return self._create_reco_level_collection()
        elif level == "gen":
            return self._create_gen_level_collection()
        else:
            raise ValueError(f"Keyword level:{level} not known.")


class TagSideVariableCollectionFactory(VariableCollectionFactory):
    """
    Factory for tag side B meson variables. This list of
    variables includes some features of the daughters 0, 1 and 2
    of the B_tag meson.  The actual number of daughters may vary
    of course and they themselves might have children which are
    not considered here.
    """

    def create(self):
        tag_side_vars = TagSideVariables(use_mc_info=self._use_mc_info)
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)
        loose_mc_vars = LooseMCVariables(use_mc_info=self._use_mc_info)
        cs_vars = ContinuumSuppressionVariables()

        return VariableCollection(
            "Tag Side Variables",
            extra=tag_side_vars.get(),
            mc=mc_vars.get(),
            loose_mc=loose_mc_vars.get(),
            cs=cs_vars.get()
        )


class LeptonVariableCollectionFactory(VariableCollectionFactory):
    """
    Creates a collection of lepton variables for either
    reco or gen level.
    """

    def _create_reco_level_collection(self) -> VariableCollection:
        lep_vars = LeptonVariables(use_mc_info=self._use_mc_info)
        kin_vars = KinematicVariables()
        track_vars = TrackVariables(use_b2bii=self._use_b2bii)
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)

        return VariableCollection(
            "Lepton Variables Reco Level",
            extra=lep_vars.get(),
            kinematic=kin_vars.get(),
            track=track_vars.get(),
            mc=mc_vars.get()
        )

    def _create_gen_level_collection(self) -> VariableCollection:
        lep_vars = LeptonGenLevelVariables()
        mc_kin_vars = MCKinematicVariables(use_mc_info=self._use_mc_info)

        return VariableCollection(
            "Lepton Variables Gen Level",
            extra=lep_vars.get(),
            mc_kinematic=mc_kin_vars.get()
        )

    def create(self, level: str) -> VariableCollection:
        """
        Creates a reco or gen level Pi0 VariableCollection.

        :param level: Either "reco" or "gen".
        :return: VariableCollection based on the level argument.
        """

        if level == "reco":
            return self._create_reco_level_collection()
        elif level == "gen":
            return self._create_gen_level_collection()
        else:
            raise ValueError(f"Keyword level:{level} not known.")


class Pi0VariableCollectionFactory(VariableCollectionFactory):
    """
    Creates a VariableCollection for pi0 particles.
    """

    def _create_gen_level_collection(self):
        pi0_vars = Pi0GenLevelVariables()
        mc_kin_vars = MCKinematicVariables(use_mc_info=self._use_mc_info)

        return VariableCollection(
            "Pi0 Gen Level",
            extra=pi0_vars.get(),
            mc_kinematic=mc_kin_vars.get()
        )

    def _create_reco_level_colleciton(self):
        pi0_vars = Pi0Variables(use_mc_info=self._use_mc_info)
        kin_vars = KinematicVariables()
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)

        return VariableCollection(
            "Pi0 Reco Level",
            extra=pi0_vars.get(),
            kin_vars=kin_vars.get(),
            mc=mc_vars.get()
        )

    def create(self, level: str) -> VariableCollection:
        """
        Creates a reco or gen level Pi0 VariableCollection.

        :param level: Either "reco" or "gen".
        :return: VariableCollection based on the level argument.
        """

        if level == "reco":
            return self._create_reco_level_colleciton()
        elif level == "gen":
            return self._create_gen_level_collection()
        else:
            raise ValueError(f"Keyword level:{level} not known.")


class DstarVariableCollectionFactory(VariableCollectionFactory):
    """
    Collection of variables associated with D* mesons.
    """

    def create(self):
        dst_vars = DstarVariables(use_mc_info=self._use_mc_info)
        kin_vars = KinematicVariables()
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)

        dstar_meson_collection = VariableCollection(
            "Dstar Variables",
            extra=dst_vars.get(),
            kinematic=kin_vars.get(),
            mc=mc_vars.get()
        )

        daughter_variables = DDaughterVariables(use_mc_info=self._use_mc_info).get()

        for daughter_index in range(0, 5):
            dstar_meson_collection.add_daughter_variables(daughter_variables, daughter_index)

        return dstar_meson_collection


class DVariableCollectionFactory(VariableCollectionFactory):
    """
    Generates a list of variables for D mesons.
    """

    def _create_gen_level_collection(self) -> VariableCollection:
        """
        Creates a gen level D meson VariableCollection

        :return: Gen level D meson VariableCollection
        """
        extra_vars = DGenLevelVariables()
        mc_kin_vars = MCKinematicVariables(self._use_mc_info)

        return VariableCollection(
            "D Variables (Gen Level)",
            extra=extra_vars.get(),
            mc_kinematic=mc_kin_vars.get()
        )

    def _create_reco_level_collection(self) -> VariableCollection:
        """
        Creates a standard D meson VariableCollection.

        :return: Standard D meson VariableCollection.
        """
        d_extra_vars = DVariables(use_mc_info=self._use_mc_info)
        kin_vars = KinematicVariables()
        mc_vars = MCVariables(use_mc_info=self._use_mc_info)

        d_meson_collection = VariableCollection(
            "D Variables",
            extra=d_extra_vars.get(),
            kinematic=kin_vars.get(),
            mc=mc_vars.get()
        )

        daughter_variables = DDaughterVariables(use_mc_info=self._use_mc_info).get()

        for daughter_index in range(0, 5):
            d_meson_collection.add_daughter_variables(daughter_variables, daughter_index)

        return d_meson_collection

    def create(self, level: str):
        """
        Creates standard or gen level D meson VariableCollections.

        :param level: Either "standard" or "gen"
        :return: D meson VariableCollection.
        """
        if level == "reco":
            return self._create_reco_level_collection()
        elif level == "gen":
            return self._create_gen_level_collection()
        else:
            raise ValueError(f"Keyword level:{level} not known.")


########################
# Variable Lists #
########################

class UpsilonVariables(VariableListBase):
    """
    List of variables associated with the Upsilon(4S).
    """

    def __init__(self, use_decay_hash=False, **kwargs):
        super().__init__(**kwargs)
        self._use_dec_hash = use_decay_hash

    def get(self):
        extra_vars = [
            "extraInfo(decayModeID)",
            "m2RecoilSignalSide",
            "m2RecoilSignalSideRescaledTag",
            "m2RecoilSignalSideNoScaling",
            "m2RecoilSignalSideNoFSRCorrection",
            "pStarL",
            "pStarLNoFSR",
            "qSquaredReco",
            "qSquaredReco2",
            "qSquaredReco3",
            "qSquaredReco4",
            "qSquaredRecoDStarStar",
            "wReco",
            "costhetalReco",
            "costhetavReco",
            "chi_primeReco",
            "chi_Reco",
            "missingMomentum",
            "missingMomentumRescaledTag",
            "missingMomentumInCMSFrame",
            "missingMomentumRescaledTagInCMSFrame",
            "missingMomentumInBSigRestFrame",
            "missingMomentumRescaledTagInBSigRestFrame",
            "angleWithMissingMomentum(1,0)",
            "angleWithMissingMomentum(1,1)",
            "angleWithMissingMomentum(1,2)",
            "missingEnergyCMS",
            "missingEnergyNoScalingCMS",
            "missingEnergyLab",
            "missingEnergyNoScalingLab",
            "missingEnergyRescaledTagLab",
            "chiProb",
            "extraInfo(feiTagSideSignalProbabilityRank)",
            "extraInfo(vertexFitChiProbabilityRank)",
            "extraInfo(roeEExtraRank)",
            "extraInfo(roeNEExtraRank)",
            "extraInfo(deltaERank)",
            "extraInfo(U4SDMesonDeltaMRank)",
            "useCMSFrame(fsrMomentumX)",
            "useCMSFrame(fsrMomentumY)",
            "useCMSFrame(fsrMomentumZ)",
            "useCMSFrame(fsrE)",
            "useLabFrame(fsrMomentumX)",
            "useLabFrame(fsrMomentumY)",
            "useLabFrame(fsrMomentumZ)",
            "useLabFrame(fsrE)",
            "useRestFrame(fsrMomentumX)",
            "useRestFrame(fsrMomentumY)",
            "useRestFrame(fsrMomentumZ)",
            "useRestFrame(fsrE)",
            "wasFSRCorrected"
        ]

        if self._use_mc_info:
            extra_vars.extend([
                "mcMissingMassSquared",
                "mcMissingMomentum",
                "mcMissingMomentumInBSigRestFrame",
                "genUpsilon4S(PDG)",
                "genUpsilon4S(mcDaughter(0, PDG))",
                "genUpsilon4S(mcDaughter(0, mdstIndex))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0,PDG)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0,mcDaughter(0,PDG))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0,mcDaughter(1,PDG))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0,mcDaughter(2,PDG))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,PDG)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,mcDaughter(0,PDG))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,mcDaughter(1,PDG))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,mcDaughter(2,PDG))))",
                "genUpsilon4S(mcDaughter(1, PDG))",
                "genUpsilon4S(mcDaughter(1, mdstIndex))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0,PDG)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0,mcDaughter(0,PDG))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0,mcDaughter(1,PDG))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0,mcDaughter(2,PDG))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,PDG)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,mcDaughter(0,PDG))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,mcDaughter(1,PDG))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,mcDaughter(2,PDG))))",
                "genUpsilon4S(mcDaughter(0, charge))",
                "genUpsilon4S(mcDaughter(0, E))",
                "genUpsilon4S(mcDaughter(0, qSquaredMC))",
                "genUpsilon4S(mcDaughter(0, wMC))",
                "genUpsilon4S(mcDaughter(0, costhetalMC))",
                "genUpsilon4S(mcDaughter(0, costhetalWOPhotonsMC))",
                "genUpsilon4S(mcDaughter(0, costhetavMC))",
                "genUpsilon4S(mcDaughter(0, chi_primeMC))",
                "genUpsilon4S(mcDaughter(0, chi_primeWOPhotonsMC))",
                "genUpsilon4S(mcDaughter(0, chi_MC))",
                "genUpsilon4S(mcDaughter(0, chiWOPhotons_MC))",
                "genUpsilon4S(mcDaughter(0, useRestFrame(chi_primeMC)))",
                "genUpsilon4S(mcDaughter(0, useRestFrame(chi_primeWOPhotonsMC)))",
                "genUpsilon4S(mcDaughter(0, useRestFrame(chi_MC)))",
                "genUpsilon4S(mcDaughter(0, useRestFrame(chiWOPhotons_MC)))",
                "genUpsilon4S(mcDaughter(0, useCMSFrame(chi_primeMC)))",
                "genUpsilon4S(mcDaughter(0, useCMSFrame(chi_primeWOPhotonsMC)))",
                "genUpsilon4S(mcDaughter(0, useCMSFrame(chi_MC)))",
                "genUpsilon4S(mcDaughter(0, useCMSFrame(chiWOPhotons_MC)))",
                "genUpsilon4S(mcDaughter(1, charge))",
                "genUpsilon4S(mcDaughter(1, E))",
                "genUpsilon4S(mcDaughter(1, qSquaredMC))",
                "genUpsilon4S(mcDaughter(1, wMC))",
                "genUpsilon4S(mcDaughter(1, costhetalMC))",
                "genUpsilon4S(mcDaughter(1, costhetalWOPhotonsMC))",
                "genUpsilon4S(mcDaughter(1, costhetavMC))",
                "genUpsilon4S(mcDaughter(1, chi_primeMC))",
                "genUpsilon4S(mcDaughter(1, chi_primeWOPhotonsMC))",
                "genUpsilon4S(mcDaughter(1, chi_MC))",
                "genUpsilon4S(mcDaughter(1, chiWOPhotons_MC))",
                "genUpsilon4S(mcDaughter(1, useRestFrame(chi_primeMC)))",
                "genUpsilon4S(mcDaughter(1, useRestFrame(chi_primeWOPhotonsMC)))",
                "genUpsilon4S(mcDaughter(1, useRestFrame(chi_MC)))",
                "genUpsilon4S(mcDaughter(1, useRestFrame(chiWOPhotons_MC)))",
                "genUpsilon4S(mcDaughter(1, useCMSFrame(chi_primeMC)))",
                "genUpsilon4S(mcDaughter(1, useCMSFrame(chi_primeWOPhotonsMC)))",
                "genUpsilon4S(mcDaughter(1, useCMSFrame(chi_MC)))",
                "genUpsilon4S(mcDaughter(1, useCMSFrame(chiWOPhotons_MC)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(2,PDG)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(3,PDG)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(4,PDG)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(2,PDG)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(3,PDG)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(4,PDG)))",

                # Variables for Hammer:
                "genUpsilon4S(mcDaughter(0, px))",
                "genUpsilon4S(mcDaughter(0, py))",
                "genUpsilon4S(mcDaughter(0, pz))",
                "genUpsilon4S(mcDaughter(0, E))",

                "genUpsilon4S(mcDaughter(1, px))",
                "genUpsilon4S(mcDaughter(1, py))",
                "genUpsilon4S(mcDaughter(1, pz))",
                "genUpsilon4S(mcDaughter(1, E))",

                "genUpsilon4S(mcDaughter(0, mcDaughter(0,px)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0,py)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0,pz)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0,E)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(0,px))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(0,py))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(0,pz))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(0,E))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(1,px))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(1,py))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(1,pz))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(0, mcDaughter(1,E))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,px)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,py)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,pz)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1,E)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(0,px))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(0,py))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(0,pz))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(0,E))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(1,px))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(1,py))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(1,pz))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(1,E))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(2,px))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(2,py))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(2,pz))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(1, mcDaughter(2,E))))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(2,px)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(2,py)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(2,pz)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(2,E)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(3,px)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(3,py)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(3,pz)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(3,E)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(4,px)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(4,py)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(4,pz)))",
                "genUpsilon4S(mcDaughter(0, mcDaughter(4,E)))",

                "genUpsilon4S(mcDaughter(1, mcDaughter(0,px)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0,py)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0,pz)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0,E)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(0,px))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(0,py))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(0,pz))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(0,E))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(1,px))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(1,py))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(1,pz))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(0, mcDaughter(1,E))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,px)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,py)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,pz)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1,E)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(0,px))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(0,py))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(0,pz))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(0,E))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(1,px))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(1,py))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(1,pz))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(1,E))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(2,px))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(2,py))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(2,pz))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(1, mcDaughter(2,E))))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(2,px)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(2,py)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(2,pz)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(2,E)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(3,px)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(3,py)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(3,pz)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(3,E)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(4,px)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(4,py)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(4,pz)))",
                "genUpsilon4S(mcDaughter(1, mcDaughter(4,E)))",
            ])

        if self._use_dec_hash:
            extra_vars.extend([
                "extraInfo(DecayHash)",
                "extraInfo(DecayHashExtended)",
            ])

        return extra_vars


class SignalSideBGenLevelVariables(VariableListBase):
    """
    List of variables associated with the signal side
    B meson on generator level.
    """

    def get(self) -> List[str]:
        b_gen_vars = [
            "matchedMC(PDG)",
            "matchedMC(charge)",
            "extraInfo(decayModeID)",
            "useRestFrame(daughter(0, p))",
            "useRestFrame(daughter(1, p))",
            "useRestFrame(daughter(2, p))",
            "daughterInvM(0,2)"
        ]
        reweighting_vars = [
            "qSquared",
            "w",
            "costhetal",
            "costhetav",
            "chi_prime",
            "alt_chi"
        ]

        return b_gen_vars + reweighting_vars


class SignalSideBVariables(VariableListBase):
    """
    List of variables associated with the signal side
    B meson on reco Level.
    """

    def get(self) -> List[str]:
        b_sig_vars = [
            "PDG",
            "extraInfo(decayModeID)",
            "Mbc",
            "E",
            "px", "py", "pz",
            "deltaE",
            "chiProb",
            "charge",
            "decayAngle(0)",
            "decayAngle(1)",
            "cosThetaBetweenParticleAndTrueB",
            "daughterInvM(0, 2)",
            "formula(daughterInvM(0, 2) - daughter(0, InvM))",
            "daughter(2, p)",
            "daughter(2, theta)",
            "useCMSFrame(daughter(0, p))",
            "useCMSFrame(daughter(1, p))",
            "useCMSFrame(daughter(2, p))",
            "useCMSFrame(daughterAngle(0, 1))",
            "useCMSFrame(daughterAngleInBetween(0, 1))",
            "useCMSFrame(daughterDiffOfPhi(0, 1))",
            "useCMSFrame(daughterAngle(0, 2))",
            "useCMSFrame(daughterAngleInBetween(0, 2))",
            "useCMSFrame(daughterDiffOfPhi(0, 2))",
            "useRestFrame(daughter(0, p))",
            "useRestFrame(daughter(1, p))",
            "useRestFrame(daughter(2, p))",
            "checkIfInfoIsInChain(extraInfo(fsrCorrected) > 0)",
            "checkIfInfoIsInChain(extraInfo(hasFSRCorrection) > 0)"
        ]

        if self._use_mc_info:
            b_sig_vars.extend([
                "matchedMC(PDG)",
                "matchedMC(charge)",
                "matchedMC(E)",
                "matchedMC(px)",
                "mcDaughter(1, PDG)",
                "mcDaughter(1, mcDaughter(0, PDG))"
            ])
        return b_sig_vars


class LeptonGenLevelVariables(VariableListBase):
    """
    List of variables associated with leptons on
    generator level.
    """

    def get(self):
        return [
            "matchedMC(PDG)",
            "mcDaughter(0, PDG)",
            "mcMother(PDG)",
            "mcMother(mcDaughter(0, PDG))"
        ]


class LeptonVariables(VariableListBase):
    """
    List of variables associated with leptons.
    """

    def get(self):
        lepton_extra_vars = [
            "PDG",
            "chiProb",
        ]

        if self._use_mc_info:
            lepton_extra_vars.extend([
                "matchedMC(PDG)",
                "mcMother(PDG)",
                "mcMother(mcMother(PDG))",
                "mcMother(mdstIndex)",
                "mcMother(mcMother(mdstIndex))",
                "mcMother(charge)",
                "mcMother(mcMother(charge))",
                "mcMother(mcDaughter(0, PDG))",
                "mcMother(mcMother(mcDaughter(0, PDG)))",
                "mcMother(mcDaughter(0, mcDaughter(0, PDG)))",
                "mcMother(mcDaughter(0, mcDaughter(1, PDG)))",
                "mcMother(mcDaughter(0, mcDaughter(2, PDG)))",
                "mcMother(mcMother(mcDaughter(0, mcDaughter(0, PDG))))",
                "mcMother(mcMother(mcDaughter(0, mcDaughter(1, PDG))))",
                "mcMother(mcMother(mcDaughter(0, mcDaughter(2, PDG))))",
                "mcMother(mcDaughter(1,PDG))",
                "mcMother(mcMother(mcDaughter(1,PDG)))",
                "mcMother(mcDaughter(2,PDG))",
                "mcMother(mcMother(mcDaughter(2,PDG)))",
                "mcMother(mcDaughter(3,PDG))",
                "mcMother(mcMother(mcDaughter(3,PDG)))"
            ])

        return lepton_extra_vars


class Pi0GenLevelVariables(VariableListBase):
    """
    List of variables associated with pi0 particles on
    generator level.
    """

    def get(self):
        return [
            "matchedMC(PDG)",
            "daughterAngle(0,1)"
        ]


class Pi0Variables(VariableListBase):
    """
    List of variables associated with pi0 particles on
    reco level.
    """

    def get(self):
        pi0_vars = [
            "PDG",
            "daughterInvariantMass(0, 1)",
            "chiProb",
            "abs(SigMBF)",
            "daughterAngle(0,1)",
            "useRestFrame(daughterAngle(0,1))",
            "useCMSFrame(daughterAngle(0,1))",
            "formula(daughterDiffOf(0, 1, E) / daughterSumOf(E))",
            "daughter(0, goodBelleGamma)",
            "daughter(0, clusterE9E21)",
            "daughter(1, goodBelleGamma)",
            "daughter(1, clusterE9E21)"
        ]

        if self._use_mc_info:
            pi0_vars.extend([
                "matchedMC(PDG)",
                "mcMother(PDG)",
            ])

        return pi0_vars


class DDaughterVariables(VariableListBase):
    """
    Collection of variables associated with D mesons Daughters
    in standard reconstruction.
    """

    def get(self):
        variables = [
            "p",
            "px",
            "py",
            "pz",
            "E",
            "theta",
            "PDG",
        ]
        if self._use_mc_info:
            variables.extend([
                "matchedMC(PDG)"
            ])

        return variables


class DVariables(VariableListBase):
    """
    Collection of variables associated with D mesons in
    standard reconstruction.
    """

    def get(self):
        variables = [
            "E",
            "p", "pt",
            "px", "py", "pz",
            "dr", "dz",
            "theta",
            "PDG",
            "extraInfo(decayModeID)",
            "abs(SigM)",
            "abs(SigMBF)",
            "chiProb"
        ]
        if self._use_mc_info:
            variables.extend([
                "matchedMC(PDG)",
                "matchedMC(E)"
            ])

        return variables


class DGenLevelVariables(VariableListBase):
    """
    Collection of variables associated with D mesons in
    gen level reconstruction.
    """

    def get(self):
        return [
            "matchedMC(PDG)",
            "extraInfo(decayModeID)",
            "mcDaughter(0, PDG)",
            "mcDaughter(1, PDG)",
            "mcMother(PDG)"
        ]


class DstarVariables(VariableListBase):
    """
    List of variables associated with D* mesons
    """

    def get(self):
        dst_vars = [
            "E",
            "p", "pt",
            "px", "py", "pz",
            "dr", "dz",
            "theta",
            "PDG",
            "extraInfo(decayModeID)",
            "abs(SigM)",
            "abs(SigMBF)",
            "daughterMotherDiffOf(0, M)",
            "daughterInvariantMass(0, 1)",
            "chiProb"
        ]

        if self._use_mc_info:
            dst_vars.extend([
                "matchedMC(PDG)",
                "matchedMC(E)"
            ])
        return dst_vars


class TagSideVariables(VariableListBase):
    """
    List of tag side B meson variables.
    """

    def get(self):
        variables = [
            "extraInfo(decayModeID)",
            "extraInfo(uniqueSignal)",
            "extraInfo(SignalProbability)",
            "extraInfo(FEIProbabilityRank)",
            "cosThetaBetweenParticleAndTrueB",
            "Mbc",
            "deltaE",
            "nRemainingTracksInEvent",
            "chiProb",
            "dx", "dy", "dz",
            "useCMSFrame(p)",
            "useCMSFrame(pt)",
            "PDG",
            "matchedMC(PDG)" if self._use_mc_info else "",
            "charge",
            "matchedMC(charge)" if self._use_mc_info else "",
            "daughter(0, M)",
            "daughter(0, chiProb)",
            "daughter(0, useCMSFrame(p))",
            "daughter(0, useCMSFrame(pt))",
            "daughter(1, M)",
            "daughter(1, chiProb)",
            "daughter(1, useCMSFrame(p))",
            "daughter(1, useCMSFrame(pt))",
            "daughter(2, M)",
            "daughter(2, chiProb)",
            "daughter(2, useCMSFrame(p))",
            "daughter(2, useCMSFrame(pt))",
        ]

        return list(filter(lambda x: x is not "", variables))


class MCVariables(VariableListBase):
    """
    This class stores variables generated by MC matching of
    the particles. These are only available on MC of course.
    If use_mc_info is False, an empty list will be returned.
    """

    def get(self):
        if not self._use_mc_info:
            return []

        return [
            "mcErrors",
            "isSignal",
            "isExtendedSignal",
            "isPrimarySignal",
            "isSignalAcceptMissingNeutrino"
        ]


class LooseMCVariables(VariableListBase):
    """
    This class stores extraInfo variables available when loose
    MC matching was performed. These are only available on MC
    of course. If use_mc_info is False, an empty list will be
    returned.
    """

    def get(self):
        if not self._use_mc_info:
            return []

        return [
            "extraInfo(looseMCMotherPDG)",
            "extraInfo(looseMCWrongDaughterN)",
            "extraInfo(looseMCWrongDaughterPDG)",
            "extraInfo(looseMCWrongDaughterBiB)"
        ]


class MCKinematicVariables(VariableListBase):
    """
    This class stores a list of MC kinematic variables.
    These are only available on MC of course.
    If use_mc_info is False, an empty list will be returned.
    """

    def get(self):
        if not self._use_mc_info:
            return []

        return [
            "matchedMC(px)",
            "matchedMC(py)",
            "matchedMC(pz)",
            "matchedMC(E)",
            "matchedMC(M)",
            "matchedMC(Q)",
        ]


class EventVariables(VariableListBase):
    """
    This class stores variables describing a single event.
    Returned values depend on the use_mc_info and use_b2bii
    flags.
    """

    def get(self):
        event_vars = [
            "productionIdentifier",
            "expNum",
            "runNum",
            "evtNum",
            "nTracks",
            "nCleanedTracks(dr < 2 and abs(dz) < 4)",
            "nCleanedTracks(dr < 2 and abs(dz) < 4 and pt > 0.2)",
            "nCleanedECLClusters(goodBelleGamma == 1 and clusterTrackMatch == 0)"
            if self._use_b2bii
            else "nCleanedECLClusters(goodGamma == 1 and clusterTrackMatch == 0)",
            "nECLClusters",
            "belleECLEnergy" if self._use_b2bii else "",
            "totalEnergyOfParticlesInList(gamma:fsp)",
            "nKLMClusters",
            "KLMEnergy",
            "Ecms",
            "beamE",
            "beamPx",
            "beamPy",
            "beamPz",
            "isContinuumEvent" if self._use_mc_info else "",
            "isNotContinuumEvent" if self._use_mc_info else "",
        ]

        return list(filter(lambda x: x is not "", event_vars))


class KinematicVariables(VariableListBase):
    """
    This class stores kinematic variables.
    """

    def get(self):
        return [
            "M",
            "dM",
            "InvM",
            "Q",
            "dQ",
            "useCMSFrame(p)",
            "useCMSFrame(pt)",
        ]


class PhotonVariables(VariableListBase):
    """
    This class stores photon variables. The list depends on
    the use_b2bii flag
    """

    def get(self):
        return [
            "E",
            "goodBelleGamma" if self._use_b2bii else "goodGamma",
            "clusterE9E25",
            "clusterE9E21",
            "clusterE1E9",
            "clusterTiming",
            "clusterHighestE",
            "clusterTrackMatch",
            "clusterNHits",
            "clusterReg",
            "clusterConnectedRegionID"
        ]


class KshortVariables(VariableListBase):
    """
    This class stores variables for Kshort particles.
    """

    def get(self):
        return [
            "extraInfo(goodKs)",
            "extraInfo(ksnbVLike)",
            "extraInfo(ksnbNoLam)",
            "extraInfo(ksnbStandard)",
        ]


class TrackVariables(VariableListBase):
    """
    This class stores track variables. The list depends on
    the use_b2bii flag.
    """

    def get(self):
        variables = [
            "E",
            "p", "pt",
            "px", "py", "pz",
            "dr", "dz",
            "theta",
            "pValue",
            "nCDCHits",
            "nSVDHits",
            "nPXDHits" if not self._use_b2bii else "",
            "nVXDHits" if not self._use_b2bii else "",
            "eIDBelle" if self._use_b2bii else "electronID",
            "muIDBelle" if self._use_b2bii else "muonID",
            "muIDBelleQuality" if self._use_b2bii else "",
            "atcPIDBelle(3,2)" if self._use_b2bii else "kaonID",
            "kIDBelle" if self._use_b2bii else "",
            "atcPIDBelle(2,3)" if self._use_b2bii else "pionID",
            "trackCurlerCloneIndicator",
        ]

        return list(filter(lambda x: x is not "", variables))


class ContinuumSuppressionVariables(VariableListBase):
    """
    This class stores continuum suppression variables.
    """

    def get(self):
        return [
            "R2",
            "cosTBTO",
            "cosTBz",
            "thrustBm",
            "thrustOm",
            "KSFWVariables(et)",
            "KSFWVariables(mm2)",
            "KSFWVariables(hso00)",
            "KSFWVariables(hso01)",
            "KSFWVariables(hso02)",
            "KSFWVariables(hso03)",
            "KSFWVariables(hso04)",
            "KSFWVariables(hso10)",
            "KSFWVariables(hso12)",
            "KSFWVariables(hso14)",
            "KSFWVariables(hso20)",
            "KSFWVariables(hso22)",
            "KSFWVariables(hso24)",
            "KSFWVariables(hoo0)",
            "KSFWVariables(hoo1)",
            "KSFWVariables(hoo2)",
            "KSFWVariables(hoo3)",
            "KSFWVariables(hoo4)",
            "CleoCone(1)",
            "CleoCone(2)",
            "CleoCone(3)",
            "CleoCone(4)",
            "CleoCone(5)",
            "CleoCone(6)",
            "CleoCone(7)",
            "CleoCone(8)",
            "CleoCone(9)",
        ]


class ROEVariables(VariableListBase):
    """
    This class stores ROE variables. Getting list of ROE
    variables for a certain ROE mask or all ROE masks
    defined in the analysis specific roe_mask module.
    To get the variables for all ROE masks, use "all"
    as roe_mask_name.
    """

    @staticmethod
    def _create_roe_variables(roe_mask):
        """
        Helper function to create a list of ROE variables.
        :param roe_mask: Name of the ROE mask used.
        :return:
        """
        return [
            f"ROE_deltae({roe_mask})",
            f"ROE_mbc({roe_mask})",
            f"ROE_charge({roe_mask})",
            f"nROE_Tracks({roe_mask})",
            f"ROE_eextra({roe_mask})",
            f"ROE_neextra({roe_mask})",
        ]

    def get(self):
        if self._roe_mask == "all":
            var_list = list()
            for mask in roe_masks_utils.get_roe_masks(roe_masks.roe_mask_list, self._use_b2bii):
                var_list.extend(self._create_roe_variables(mask))
            return var_list
        else:
            return self._create_roe_variables(self._roe_mask)


if __name__ == "__main__":
    config = {
        "use_mc_info": True,
        "use_b2bii": True,
        "roe_mask": "CleanROE"
    }

    ups4s_coll = get_upsilon4s_collection(config)
    gen_lvl_b_coll = get_gen_level_b_collection(config)
    gen_lvl_ups4s_coll = get_gen_level_upsilon4s_vars(config)

    print(ups4s_coll)
    print(gen_lvl_ups4s_coll)
    print(gen_lvl_b_coll)
