"""
Steering file for testing the n-tuple file generation for the B -> D(*) tau nu analysis.

Contains the reconstruction analog to reconstruction.py but produces additional output.

Felix Metzner 2018
"""

import basf2
import b2biiConversion

from modularAnalysis import *

from analysistools.processing import path_utilities
from analysistools.processing import reco_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis.decay_utilities import Reconstructor
from rdstar.online_analysis import observables, decay_utilities, skim_cuts, decay_channels

input_files, input_sequences, output_dict, config = path_utilities.get_commandline_args()

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()

b2biiConversion.convertBelleMdstToBelleIIMdst(
    inputBelleMDSTFile=input_files,
    applyHadronBJSkim=True,
    useBelleDBServer=os.environ["BELLE_POSTGRES_SERVER"],
    path=main_path,
    entrySequences=input_sequences
)

setAnalysisConfigParams({"mcMatchingVersion": "Belle"}, main_path)

main_path.add_module("CurlerCloneMarker")
decay_utilities.fsp.fsp_collection.get_final_state_particles(path=main_path, use_mc_info=config["use_mc_info"])

skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.pre_skim_cuts)

reconstructor = Reconstructor(
    path=main_path,
    config=config,
    decay_channels_module=decay_channels,
    do_gen_level_reco=True
)

reconstructor.reconstruct_tau_for_gen()

reconstructor.reconstruct_D()
reconstructor.reconstruct_Dstar()

reconstructor.reconstruct_D2star_for_gen()

reconstructor.reconstruct_sig_B()

if config["use_mc_info"] is True:
    b_gen_coll = observables.get_gen_level_b_collection(config)

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["B_minus_generator_info_ntuple"],
        particleList="B-:sig_gen",
        variables=b_gen_coll.as_list()
    )

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["B_zero_generator_info_ntuple"],
        particleList="B0:sig_gen",
        variables=b_gen_coll.as_list()
    )

    copyLists(
        outputListName="B0:starstar_gen",
        inputListNames=["B0:D1_e_gen", "B0:D1_mu_gen", "B0:D0*_e_gen", "B0:D0*_mu_gen", "B0:D'1_e_gen", "B0:D'1_mu_gen",
                        "B0:D2*_e_gen", "B0:D2*_mu_gen"],
        writeOut=True,
        path=main_path
    )

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["B_zero_starstar_generator_info_ntuple"],
        particleList="B0:starstar_gen",
        variables=b_gen_coll.as_list()
    )

    copyLists(
        outputListName="B-:starstar_gen",
        inputListNames=["B-:D10_e_gen", "B-:D10_mu_gen", "B-:D0*0_e_gen", "B-:D0*0_mu_gen", "B-:D'10_e_gen",
                        "B-:D'10_mu_gen", "B-:D2*0_e_gen", "B-:D2*0_mu_gen"],
        writeOut=True,
        path=main_path
    )

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["B_minus_starstar_generator_info_ntuple"],
        particleList="B-:starstar_gen",
        variables=b_gen_coll.as_list()
    )

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["D_2_starstar0_generator_info_ntuple"],
        particleList="D_2*0:all_gen",
        variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)",
                        "mcMother(PDG)"] +
                       observables.EventVariables(use_mc_info=True, use_b2bii=True).get() +
                       observables.MCVariables(use_mc_info=True).get())
    )

    fillParticleListsFromMC([("D_2*+:mc", "")], path=main_path, addDaughters=True)
    fillParticleListsFromMC([("D_2*0:mc", "")], path=main_path, addDaughters=True)

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["D_2_starstar0_generator_direct_info_ntuple"],
        particleList="D_2*0:mc",
        variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)",
                        "mcMother(PDG)", "mcDaughter(0, mcDaughter(0, PDG))", "mcDaughter(0, mcDaughter(1, PDG))",
                        "mcDaughter(0, mcDaughter(2, PDG))", "mcDaughter(0, mcDaughter(0, mcDaughter(0, PDG)))"])
    )

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["D_2_starstarPlus_generator_direct_info_ntuple"],
        particleList="D_2*+:mc",
        variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)",
                        "mcMother(PDG)", "mcDaughter(0, mcDaughter(0, PDG))", "mcDaughter(0, mcDaughter(1, PDG))",
                        "mcDaughter(0, mcDaughter(2, PDG))", "mcDaughter(0, mcDaughter(0, mcDaughter(0, PDG)))", "PDG"])
    )

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["Dstar0_generator_info_ntuple"],
        particleList="D*0:all_gen",
        variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)",
                        "mcMother(PDG)", "mcDaughter(0, mcDaughter(0, PDG))", "mcDaughter(0, mcDaughter(1, PDG))",
                        "mcDaughter(0, mcDaughter(2, PDG))", "mcDaughter(0, mcDaughter(0, mcDaughter(0, PDG)))", "PDG",
                        "extraInfo(decayModeID)"])
    )
    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["D0_generator_info_ntuple"],
        particleList="D0:all_gen",
        variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)",
                        "mcMother(PDG)", "mcDaughter(0, mcDaughter(0, PDG))", "mcDaughter(0, mcDaughter(1, PDG))",
                        "mcDaughter(0, mcDaughter(2, PDG))", "mcDaughter(0, mcDaughter(0, mcDaughter(0, PDG)))", "PDG",
                        "extraInfo(decayModeID)"])
    )

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["DstarMinus_generator_info_ntuple"],
        particleList="D*-:all_gen",
        variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)",
                        "mcMother(PDG)", "mcDaughter(0, mcDaughter(0, PDG))", "mcDaughter(0, mcDaughter(1, PDG))",
                        "mcDaughter(0, mcDaughter(2, PDG))", "mcDaughter(0, mcDaughter(0, mcDaughter(0, PDG)))", "PDG",
                        "extraInfo(decayModeID)"])
    )
    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["DMinus_generator_info_ntuple"],
        particleList="D-:all_gen",
        variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)",
                        "mcMother(PDG)", "mcDaughter(0, mcDaughter(0, PDG))", "mcDaughter(0, mcDaughter(1, PDG))",
                        "mcDaughter(0, mcDaughter(2, PDG))", "mcDaughter(0, mcDaughter(0, mcDaughter(0, PDG)))", "PDG",
                        "extraInfo(decayModeID)"])
    )

main_path.add_module(
    "VariablesToNtuple",
    fileName=output_dict["B_minus_ntuple"],
    particleList="B-:sig",
    variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)"] +
                   observables.get_signal_side_collection(config).as_list())
)

main_path.add_module(
    "VariablesToNtuple",
    fileName=output_dict["B_zero_ntuple"],
    particleList="B0:sig",
    variables=list(["mcDaughter(0, PDG)", "mcDaughter(1, PDG)", "mcDaughter(2, PDG)", "mcDaughter(3, PDG)"] +
                   observables.get_signal_side_collection(config).as_list())
)

skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.post_signal_skim_cuts)

reco_utilities.reconstruct_tag_B(
    path=main_path,
    config=config,
    b_tag_chl_dict=decay_channels.BTag_decay_channels,
    decay_chl_dict=decay_channels.decay_channel_dict
)
skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.post_tag_skim_cuts)

reconstructor.reconstruct_upsilon4s()

if config["use_mc_info"] is True:
    main_path.add_module(
        "ParticleMCDecayString",
        fileName=(output_dict["Y4S_decay_trees"]),
        listName="Upsilon(4S):all"
    )

up4s_coll = observables.get_upsilon4s_collection(config)

main_path.add_module(
    "VariablesToNtuple",
    fileName=output_dict["Y4S_ntuple"],
    particleList="Upsilon(4S):all",
    variables=up4s_coll.as_list()
)

if config["use_mc_info"] is True:
    ups4s_gen_coll = observables.get_gen_level_upsilon4s_vars(config)
    b_direct_gen_coll = observables.get_direct_gen_level_b_collection(config)

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["Y4S_generator_level_reco_ntuple"],
        particleList="Upsilon(4S):all_gen",
        variables=ups4s_gen_coll.as_list()
    )

    fillParticleListsFromMC([("B-:mc", "")], addDaughters=True, path=main_path)
    fillParticleListsFromMC([("B0:mc", "")], addDaughters=True, path=main_path)
    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["B_minus_generator_direct_info_ntuple"],
        particleList="B-:mc",
        variables=b_direct_gen_coll.as_list()
    )
    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["B_zero_generator_direct_info_ntuple"],
        particleList="B0:mc",
        variables=b_direct_gen_coll.as_list()
    )

basf2.print_path(main_path)
basf2.process(main_path)
print(statistics)
