"""
b2luigi setup to perform studies of the final state radiation and
Bremsstrahlungs correction.

Felix Metzner 2018
"""

import os
import numpy as np
import pandas as pd
import b2luigi as luigi

from enum import Enum

import matplotlib as mpl
mpl.use("Agg")

from b2luigi.core.utils import product_dict
from b2luigi.basf2_helper import Basf2PathTask, Basf2nTupleMergeTask

from templatefitter.plotter import plot_style

from analysistools.dataframe_tools.hdf5_utils import load_root_file
from analysistools.batch_tools.file_handler import get_file_set, load_json_config
from rdstar.studies.reconstruction_fsr_correction_study import create_fsr_test_path


class FSRCMethods(Enum):
    centeredCone = "centeredCone"
    offsetCone = "offsetCone"
    arc = "arc"


class FSRReconstructionTask(Basf2PathTask):
    method = luigi.EnumParameter(enum=FSRCMethods, positional=False)
    max_phi_angle = luigi.FloatParameter(positional=False)
    max_theta_angle = luigi.FloatParameter(positional=False)
    max_energy = luigi.FloatParameter(positional=False)
    momentum_factor = luigi.FloatParameter(positional=False)

    file_chunk = luigi.DictParameter(positional=False, hashed=True)
    config = luigi.DictParameter(positional=False, hashed=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = "l"
        if self.config["queue"]:
            self.queue = self.config["queue"]
        self.outputs = {
            "e_ntuple": "e_ntuple.root",
            "gamma_ntuple": "gamma_ntuple.root"
        }

    def create_path(self):
        file_list = list(self.file_chunk["files"])
        sequence_list = list(self.file_chunk["parts"])
        output_dict = {key: self.get_output_file_name(value) for key, value in self.outputs.items()}

        return create_fsr_test_path(
            input_files=file_list,
            input_sequences=sequence_list,
            output_dict=output_dict,
            config=self.config,
            method=self.method.value,
            max_phi_angle=self.max_phi_angle,
            max_theta_angle=self.max_theta_angle,
            max_energy=self.max_energy,
            momentum_factor=self.momentum_factor
        )

    def output(self):
        for file_name in self.outputs.values():
            yield self.add_to_output(file_name)


class FSRCProcessingTask(Basf2nTupleMergeTask):
    batch_config_file = luigi.Parameter(positional=False, hashed=True)

    method = luigi.EnumParameter(enum=FSRCMethods, positional=False)
    max_phi_angle = luigi.FloatParameter(positional=False)
    max_theta_angle = luigi.FloatParameter(positional=False)
    max_energy = luigi.FloatParameter(positional=False)
    momentum_factor = luigi.FloatParameter(positional=False)

    def requires(self):
        batch_config = load_json_config(self.batch_config_file, hint="Loading batch config file")

        global_config_file = batch_config["global_config"]
        global_config = load_json_config(global_config_file, hint="Loading global config file")

        config = batch_config.copy()
        config.update(global_config)
        config["events_per_chunk"] = 200000

        max_events = config["events_per_chunk"]
        file_dict_location = config["file_dict_location"]
        signal_mc_config = config["signal_MC_config"]

        streams = list(set([stream for stream in [d["streams"] for d in config["data"].values()]]))
        if len(streams) > 1:
            raise RuntimeError("The provided configuration is not suitable for this study. "
                               "Multiple streams are not allowed!")
        else:
            stream = streams[0]
        for chunk in get_file_set(
                list(config["data"].keys()),
                events_per_job=max_events,
                path_to_signal_mc_config=signal_mc_config,
                stream=stream,
                fraction=1.0,
                dict_directory=file_dict_location
        ):
            yield self.clone(FSRReconstructionTask, file_chunk=chunk, config=config)


@luigi.requires(FSRCProcessingTask)
class FSRCEvaluationTask(luigi.WrapperTask):
    @staticmethod
    def eval_and_plot(data, var, output, all=False):
        import matplotlib.pyplot as plt
        assert var in ["E", "P"]

        plot_style.set_matplotlibrc_params()
        colors = plot_style.KITColors()

        base_query = "abs(matchedMC__boPDG__bc) == 11 " \
                     "and dr < 2. and abs(dz) < 4. " \
                     "and useCMSFrame__bopt__bc > 0.1 " \
                     "and eIDBelle > 0.6 and E > 0.1"
        sig_query = base_query + " and extraInfo__bofsrCorrected__bc == 1"

        fig = plt.figure(num=None, figsize=(8, 5), dpi=100, facecolor="w", edgecolor="k")

        n_with_corr = data.query(sig_query).E.count()
        n_all = data.query(base_query).E.count()
        fraction_corrected = n_with_corr / n_all
        percent_corrected = 100.0 * fraction_corrected

        if all:
            eval_query = base_query
        else:
            eval_query = sig_query

        data.query(eval_query)[f"diff_{var}"].hist(bins=30, label="w/o Correction", color=colors.kit_blue)
        data.query(eval_query)[f"diff_{var}_corr"].hist(bins=30, label="w/ Correction", color=colors.kit_red, alpha=0.5)

        ax = plt.gca()

        ax.legend(loc=0, fancybox=True, shadow=True, frameon=False)

        all_label = ""
        if not all:
            all_label = " (only corrected Electrons)"

        if var == "P":
            plt.xlabel(r"$p_{Reco} - p_{MC}$ in GeV")
            plt.title(f"FSR-Corrected Electron Momentum Residuals{all_label}")
        if var == "E":
            plt.xlabel(r"$E_{Reco} - E_{MC}$ in GeV")
            plt.title(f"FSR-Corrected Electron Energy Residuals{all_label}")

        plt.text(
            x=plt.xlim()[0] + 0.1 * (plt.xlim()[1] - plt.xlim()[0]),
            y=plt.ylim()[0] + 0.9 * (plt.ylim()[1] - plt.ylim()[0]),
            s=f"Corrected {percent_corrected:.2f}% of {n_all} Electrons",
            fontsize=12,
            bbox=dict(facecolor=colors.grey, alpha=0.3, edgecolor=None)
        )

        plt.tight_layout()
        fig.savefig(output, bbox_inches="tight")

        cov = data.query(eval_query)[f"diff_{var}"].quantile(0.84) - data.query(eval_query)[f"diff_{var}"].quantile(0.16)
        cov_corr = (data.query(eval_query)[f"diff_{var}_corr"].quantile(0.84)
                    - data.query(eval_query)[f"diff_{var}_corr"].quantile(0.16))

        all_ident = ""
        if all:
            all_ident = "_all"

        return {
            f"coverage68{all_ident}_{var}": [cov],
            f"coverage68{all_ident}_{var}_corr": [cov_corr],
            f"max{all_ident}_{var}": [data.query(eval_query)[f"diff_{var}"].max()],
            f"max{all_ident}_{var}_corr": [data.query(eval_query)[f"diff_{var}"].max()],
            f"min{all_ident}_{var}": [data.query(eval_query)[f"diff_{var}"].min()],
            f"min{all_ident}_{var}_corr": [data.query(eval_query)[f"diff_{var}"].min()],
            "n_all": [n_all],
            "n_with_corr": [n_with_corr],
            "fraction_corrected": [fraction_corrected]
        }

    def run(self):
        data = load_root_file(self.get_input_file_names("e_ntuple.root")[0])

        var_tuples = [
            ("diff_E", "matchedMC__boE__bc", "E"),
            ("diff_P", "matchedMC__bop__bc", "p"),
        ]

        for name, mc, rec in var_tuples:
            data[name] = data[rec] - data[mc]
            rec_corr = "extraInfo__bofsrCorrectedE__bc" if rec == "E" else "extraInfo__bofsrCorrectedP__bc"
            data[f"{name}_corr"] = np.where(data[rec_corr] != -999, data[rec_corr] - data[mc], data[name])

        res = self.eval_and_plot(data=data, var="E", output=self.get_output_file_name("residuals_E.pdf"))
        res2 = self.eval_and_plot(data=data, var="P", output=self.get_output_file_name("residuals_p.pdf"))
        res3 = self.eval_and_plot(data=data, var="E", output=self.get_output_file_name("residuals_all_E.pdf"), all=True)
        res4 = self.eval_and_plot(data=data, var="P", output=self.get_output_file_name("residuals_all_p.pdf"), all=True)
        res.update(res2)
        res.update(res3)
        res.update(res4)
        res.update({
            "method": [self.method.value],
            "max_phi_angle": [self.max_phi_angle],
            "max_theta_angle": [self.max_theta_angle],
            "max_energy": [self.max_energy],
            "momentum_factor": [self.momentum_factor]
        })

        pd.DataFrame(res).to_hdf(self.get_output_file_name("results.h5"), key="results")

    def output(self):
        yield self.add_to_output("residuals_E.pdf")
        yield self.add_to_output("residuals_p.pdf")
        yield self.add_to_output("residuals_all_E.pdf")
        yield self.add_to_output("residuals_all_p.pdf")
        yield self.add_to_output("results.h5")


class FSRCStudyMasterTask(luigi.Task):
    batch_config_file = luigi.Parameter(positional=False, hashed=True)

    def requires(self):
        methods = [method for method in FSRCMethods]
        max_phi_angles = list(np.linspace(start=0.5, stop=3.0, num=6))
        max_theta_angles = list(np.linspace(start=0.5, stop=3.0, num=6))
        max_energies = list(np.linspace(start=0.4, stop=1.0, num=4))
        momentum_factors = list(np.linspace(start=0.4, stop=1.0, num=4))

        parameter_combinations = product_dict(
            method=methods,
            max_phi_angle=max_phi_angles,
            max_energy=max_energies,
            momentum_factor=momentum_factors
        )

        for parameters in parameter_combinations:
            if parameters["method"] == FSRCMethods.arc:
                for max_theta_angle in max_theta_angles:
                    yield self.clone(FSRCEvaluationTask, max_theta_angle=max_theta_angle, **parameters)
            else:
                yield self.clone(FSRCEvaluationTask, max_theta_angle=1.0, **parameters)

    def run(self):
        dfs = []
        for result in self.get_input_file_names("results.h5"):
            dfs.append(pd.read_hdf(result, key="results"))

        res = pd.concat(dfs, ignore_index=True, sort=False)
        res.to_hdf(self.get_output_file_name("results.h5"), key="results")

    def output(self):
        yield self.add_to_output("results.h5")


if __name__ == "__main__":
    num_workers = 1000

    config_base_dir = "/home/belle2/metzner/RDStar/rdstar/configs/"
    config_path = os.path.join(config_base_dir, "test_batch_config.json")

    luigi.process(FSRCStudyMasterTask(batch_config_file=config_path), workers=num_workers)
