"""
Plotting for testing the randomness of gsim.

Felix Metzner 2020
"""

import os
import uproot
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from templatefitter.plotter.plot_style import set_matplotlibrc_params


set_matplotlibrc_params()

base_dir_path = "/home/belle2/metzner/temp_gsim_test/"
output_dir = os.path.join(base_dir_path, "plots")
os.makedirs(output_dir, exist_ok=True)

input_dict = {
    "old_exp55_out": "/group/belle2/TMP/metzner/mc_production_august_2020/GapDPiPiViaDstst/gsim/mdst/GapDpipiViaDoneprimeCharged/evtgen_exp_55_GapDpipiViaDoneprimeCharged-1.mdst",
    "new1_exp55_out": "/home/belle2/rvantond/temp_resubmitting_one_mc_job/gsim/mdst/GapDpipiViaDoneprimeCharged/evtgen_exp_55_GapDpipiViaDoneprimeCharged-1.mdst",
    "new2_exp55_out": "/home/belle2/rvantond/temp_resubmitting_one_mc_job_take2/gsim/mdst/GapDpipiViaDoneprimeCharged/evtgen_exp_55_GapDpipiViaDoneprimeCharged-1.mdst",
    "new3_exp55_out": "/home/belle2/rvantond/temp_resubmitting_one_mc_job_take3/gsim/mdst/GapDpipiViaDoneprimeCharged/evtgen_exp_55_GapDpipiViaDoneprimeCharged-1.mdst",
    "new1_exp19a_out": "/home/belle2/rvantond/temp_resubmitting_one_mc_job/gsim/mdst/GapDpipiViaDoneprimeCharged/evtgen_exp_19a_GapDpipiViaDoneprimeCharged-1.mdst",
    "new2_exp19a_out": "/home/belle2/rvantond/temp_resubmitting_one_mc_job_take2/gsim/mdst/GapDpipiViaDoneprimeCharged/evtgen_exp_19a_GapDpipiViaDoneprimeCharged-1.mdst",
    "new3_exp19a_out": "/home/belle2/rvantond/temp_resubmitting_one_mc_job_take3/gsim/mdst/GapDpipiViaDoneprimeCharged/evtgen_exp_19a_GapDpipiViaDoneprimeCharged-1.mdst",
}

exp_keys = ["exp19a", "exp55"]


def get_n_tuple_paths(key):
    return {
        "piNeutral_reco_vertex": os.path.join(base_dir_path, f"{key}_piNeutral_reco_vertex_ntuple.root"),
        "piNeutral_reco": os.path.join(base_dir_path, f"{key}_piNeutral_reco_ntuple.root"),
        "piNeutral_from_list": os.path.join(base_dir_path, f"{key}_piNeutral_from_list_ntuple.root"),
        "e": os.path.join(base_dir_path, f"{key}_e_ntuple.root"),
        "mu": os.path.join(base_dir_path, f"{key}_mu_ntuple.root"),
        "piCharged": os.path.join(base_dir_path, f"{key}_piCharged_ntuple.root"),
        "B_zero_generator_direct_info": os.path.join(base_dir_path, f"{key}_B_zero_generator_direct_info_ntuple.root"),
        "B_minus_generator_direct_info": os.path.join(base_dir_path, f"{key}_B_minus_generator_direct_info_ntuple.root"),
        "gamma": os.path.join(base_dir_path, f"{key}_gamma_ntuple.root"),
        "K": os.path.join(base_dir_path, f"{key}_K_ntuple.root"),
        "K_S0": os.path.join(base_dir_path, f"{key}_K_S0_ntuple.root"),
    }


def load_df_for_key(n_tuple_key):
    loaded_dfs = {}
    for in_key in input_dict.keys():
        file_path = get_n_tuple_paths(key=in_key)[n_tuple_key]
        hdf_file_path = os.path.splitext(file_path)[0] + ".h5"
        if os.path.exists(hdf_file_path):
            loaded_dfs.update({in_key: pd.read_hdf(hdf_file_path, key=in_key)})
        else:
            data_file = uproot.open(file_path)
            tree = data_file[data_file.keys()[0]]
            loaded_dfs.update({in_key: tree.pandas.df()})
            loaded_dfs[in_key].to_hdf(hdf_file_path, key=in_key)
    assert len(loaded_dfs) == len(input_dict), (len(loaded_dfs), len(input_dict))
    return loaded_dfs


# Print Lepton Moments:
e_dfs = load_df_for_key(n_tuple_key="e")
gamma_dfs = load_df_for_key(n_tuple_key="gamma")
ks_dfs = load_df_for_key(n_tuple_key="K_S0")

assert all(isinstance(df, pd.DataFrame) for df in e_dfs.values())

for exp_k in exp_keys:
    best_range = [np.inf, -np.inf]
    for k, df in e_dfs.items():
        if df["p"].max() > best_range[1]:
            best_range[1] = df["p"].max()
        if df["p"].min() < best_range[0]:
            best_range[0] = df["p"].min()

    fig = plt.figure(figsize=(12., 8.), dpi=200)
    ax = plt.gca()
    for k, df in e_dfs.items():
        if exp_k not in k:
            continue
        df.hist(column="p", ax=ax, label=k, histtype="step", range=tuple(best_range), bins=30)
    ax.set_title(f"Electron Momentum for {exp_k}")
    ax.set_xlabel(r"$p_\ell$ in GeV")
    ax.legend(frameon=True, loc="best", ncol=2)
    fig.savefig(os.path.join(output_dir, f"e_moment_{exp_k}.png"), bbox_inches="tight")
    plt.close(fig)

evt_vars = [
    "nTracks",
    "nECLClusters",
    "belleECLEnergy",
    "nKLMClusters",
    "KLMEnergy",
    "nMCParticles",
    "IPX",
]

gamma_vars = [
    "mcMother__boPDG__bc",
    "goodBelleGamma",
    "clusterE9E25",
    "clusterNHits",
    "clusterHighestE",
]

ks_vars = [
    "extraInfo__boksnbStandard__bc",
]

for var in evt_vars:
    for exp_k in exp_keys:
        best_range = [np.inf, -np.inf]
        for k, df in e_dfs.items():
            if df[var].max() > best_range[1]:
                best_range[1] = df[var].max()
            if df[var].min() < best_range[0]:
                best_range[0] = df[var].min()

        fig = plt.figure(figsize=(12., 8.), dpi=200)
        ax = plt.gca()
        for k, df in e_dfs.items():
            if exp_k not in k:
                continue
            df.hist(column=var, ax=ax, label=k, histtype="step", range=tuple(best_range), bins=30)
        ax.set_title(f"Electron {var} for {exp_k}")
        ax.set_xlabel(f"{var}")
        ax.legend(frameon=True, loc="best", ncol=2)
        fig.savefig(os.path.join(output_dir, f"e_{var}_{exp_k}.png"), bbox_inches="tight")
        plt.close(fig)

for var in gamma_vars:
    for exp_k in exp_keys:
        best_range = [np.inf, -np.inf]
        for k, df in gamma_dfs.items():
            if df[var].max() > best_range[1]:
                best_range[1] = df[var].max()
            if df[var].min() < best_range[0]:
                best_range[0] = df[var].min()

        fig = plt.figure(figsize=(12., 8.), dpi=200)
        ax = plt.gca()
        for k, df in gamma_dfs.items():
            if exp_k not in k:
                continue
            df.hist(column=var, ax=ax, label=k, histtype="step", range=tuple(best_range), bins=30)
        ax.set_title(f"Gamma {var} for {exp_k}")
        ax.set_xlabel(f"{var}")
        ax.legend(frameon=True, loc="best", ncol=2)
        fig.savefig(os.path.join(output_dir, f"gamma_{var}_{exp_k}.png"), bbox_inches="tight")
        plt.close(fig)

for exp_k in exp_keys:
    best_range = [np.inf, -np.inf]
    var = "extraInfo__boksnbStandard__bc"
    for k, df in ks_dfs.items():
        if df[var].max() > best_range[1]:
            best_range[1] = df[var].max()
        if df[var].min() < best_range[0]:
            best_range[0] = df[var].min()

    fig = plt.figure(figsize=(12., 8.), dpi=200)
    ax = plt.gca()
    for k, df in ks_dfs.items():
        if exp_k not in k:
            continue
        df.hist(column="extraInfo__boksnbStandard__bc", ax=ax, label=k, histtype="step", range=tuple(best_range), bins=30)
    ax.set_title(f"ksnbStandard for {exp_k}")
    ax.set_xlabel("ksnbStandard")
    ax.legend(frameon=True, loc="best", ncol=2)
    fig.savefig(os.path.join(output_dir, f"ks_ksnbStandard_{exp_k}.png"), bbox_inches="tight")
    plt.close(fig)

