import os
import yaml
import logging
import b2luigi as luigi
import pandas as pd

from typing import List, Dict, Generator
import uproot

from analysistools.batch_tools.file_handler import load_json_config

from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask

from rdstar.offline_analysis.tasks import utilities
from rdstar.studies.low_level_cut_study.utilities.utilities import get_config_base_path
from rdstar.studies.low_level_cut_study.utilities.plotting import LowLevelCutPlotter

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
mpl_logger = logging.getLogger('matplotlib')
mpl_logger.setLevel(logging.WARNING)

__all__ = [
    "MergeNTuplesTask",
    "LowLevelCutsEvaluationTask",
    "LowLevelCutStudyMaster"
]


class MergeNTuplesTask(utilities.BaseOfflineTask):
    config_path = luigi.Parameter(positional=False)
    ntuple_samples_combination = luigi.TupleParameter(positional=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(os.fspath(self.config_path)) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.batch_configs = offline_config["batch_configs"]

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    @staticmethod
    def get_sample_ntuple_combinations(configs) -> Dict[str, List[str]]:
        config_files = [os.path.join(get_config_base_path(), file) for file in configs]
        configs = [load_json_config(c) for c in config_files]

        combinations = {}
        for config in configs:
            samples = config["data"].keys()
            n_tuples = config["output"]
            for tuple in n_tuples:
                combinations.update({tuple: [str(s) for s in samples]})

        return combinations

    def run_merging(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        n_tuple_dir = os.path.join(self.working_dir, "CombinedNTuples")
        output_key = f"{self.ntuple_samples_combination[0]}_total"
        output_path = os.path.join(n_tuple_dir, f"{self.ntuple_samples_combination[0]}_total.h5")

        combos = [f"{sample}_{self.ntuple_samples_combination[0]}" for sample in self.ntuple_samples_combination[1]]
        if not dry_run:
            dfs = []
            number_of_events = 0
            for combo in combos:
                assert os.path.isfile(path=self.input_dict[combo]), self.input_dict[combo]
                assert os.path.splitext(p=self.input_dict[combo])[1] == ".root", os.path.splitext(
                    p=self.input_dict[combo])
                data_file = uproot.open(self.input_dict[combo])
                tree = data_file[data_file.keys()[0]]
                df = tree.pandas.df()
                df = df.loc[:, ~df.columns.duplicated()]
                dfs.append(df)
                number_of_events += len(dfs[-1].index)

            total_n_tuple = pd.concat(dfs, ignore_index=True, sort=False)
            assert len(total_n_tuple.index) == number_of_events, (len(total_n_tuple.index), number_of_events)

            assert isinstance(total_n_tuple, pd.DataFrame)
            total_n_tuple.to_hdf(path_or_buf=output_path, key="low_level")

        return {output_key: output_path}

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_merging()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {
            output_key: luigi.LocalTarget(target_path)
            for output_key, target_path in self.run_merging(dry_run=True).items()
        }

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield DataCopyTask(offline_config_path=self.config_path)


class LowLevelCutsEvaluationTask(utilities.BaseOfflineTask):
    config_path = luigi.Parameter(positional=False)
    n_tuple = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(os.fspath(self.config_path)) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.batch_configs = offline_config["batch_configs"]

        self.cut_evaluation_dir = os.path.join(self.working_dir, "CutEvaluation")

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def run_evaluation(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        output_dir = os.path.join(self.cut_evaluation_dir, self.n_tuple)
        llcp = LowLevelCutPlotter(
            n_tuple=self.n_tuple,
            target_path=output_dir
        )
        output = llcp.generate_output_dict()
        if not dry_run:
            df = pd.read_hdf(self.input_dict[f"{self.n_tuple}_total"])
            assert isinstance(df, pd.DataFrame)
            llcp.get_df(df=df)
            plot_dict = llcp.plot_on()
            assert plot_dict == llcp.generate_output_dict()

        return output

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {
            output_key: luigi.LocalTarget(target_path)
            for output_key, target_path in self.run_evaluation(dry_run=True).items()
        }

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        ntuple_samples_combinations = MergeNTuplesTask.get_sample_ntuple_combinations(configs=self.batch_configs)
        yield MergeNTuplesTask(
            config_path=self.config_path,
            ntuple_samples_combination=(self.n_tuple, ntuple_samples_combinations[self.n_tuple])
        )


class LowLevelCutStudyMaster(luigi.WrapperTask):
    config_path = luigi.Parameter(positional=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(os.fspath(self.config_path)) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.batch_configs = offline_config["batch_configs"]

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for n_tuple in MergeNTuplesTask.get_sample_ntuple_combinations(configs=self.batch_configs).keys():
            yield LowLevelCutsEvaluationTask(config_path=self.config_path, n_tuple=n_tuple)


if __name__ == "__main__":
    config_base_path = get_config_base_path()

    offline_config_path = os.path.join(config_base_path, "low_level_cut_study_conf.yaml")  # type: os.fspath

    num_workers = 10
    luigi.process(
        LowLevelCutStudyMaster(config_path=offline_config_path),
        workers=num_workers
    )
