"""
Steering file for the n-tuple production for the D/D* low level cut study.

Patrick Ecker 2020
"""

import basf2
import b2biiConversion

from modularAnalysis import *

from analysistools.processing import path_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis import decay_utilities, skim_cuts
from rdstar.studies.low_level_cut_study.utilities import decay_channels_cut_study
from rdstar.online_analysis.decay_utilities import Reconstructor

input_files, input_sequences, output_dict, config = path_utilities.get_commandline_args()

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()

b2biiConversion.convertBelleMdstToBelleIIMdst(
    inputBelleMDSTFile=input_files,
    applyHadronBJSkim=True,
    useBelleDBServer=os.environ['BELLE_POSTGRES_SERVER'],
    path=main_path,
    entrySequences=input_sequences
)

setAnalysisConfigParams({'mcMatchingVersion': 'Belle'}, main_path)

# Get Final State Particles
main_path.add_module('CurlerCloneMarker')
decay_utilities.fsp.fsp_collection.get_final_state_particles(path=main_path, use_mc_info=config['use_mc_info'])
skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.pre_skim_cuts)

reconstructor = Reconstructor(path=main_path, config=config, decay_channels_module=decay_channels_cut_study)

# Build intermediate D mesons
reconstructor.reconstruct_D()
reconstructor.reconstruct_Dstar()

###########
# D meson #
###########

general_vars = [
    'PDG',
    'matchedMC(PDG)',

    'mcMother(PDG)',
    'mcMother(mcMother(PDG))',
    'mcMother(mcMother(mcMother(PDG)))',

    'mcMother(mcDaughter(0, PDG))',
    'mcMother(mcDaughter(1, PDG))',

    'mcMother(mcMother(mcDaughter(0, PDG)))',
    'mcMother(mcMother(mcDaughter(1, PDG)))',

    'mcMother(mcMother(mcMother(mcDaughter(0, PDG))))',
    'mcMother(mcMother(mcMother(mcDaughter(1, PDG))))',

    'mcDaughter(0, PDG)',
    'mcDaughter(1, PDG)',

    "isSignal",
    'extraInfo(decayModeID)',
    'mcDaughter(0, PDG)',
    'mcDaughter(1, PDG)',
    'matchedMC(E)',

    'dM', 'abs(dM)', 'M', 'InvM',
    'abs(SigMBF)',
    'SigMBF',

    'E',
    'Q', 'dQ',
    'useCMSFrame(p)',
    'useCMSFrame(pt)',
    'p', 'pt', 'px', 'py', 'pz',

    'isContinuumEvent',

    'useCMSFrame(daughter(0, p))',
    'useCMSFrame(daughter(1, p))',
    'useCMSFrame(daughter(2, p))',
    'useCMSFrame(daughter(3, p))',
    'useCMSFrame(daughter(4, p))',
    'useRestFrame(daughter(0, p))',
    'useRestFrame(daughter(1, p))',
    'useRestFrame(daughter(2, p))',
    'useRestFrame(daughter(3, p))',
    'useRestFrame(daughter(4, p))',
    'daughter(0, InvM)',
    'daughter(1, InvM)',
    'daughter(2, InvM)',
    'daughter(3, InvM)',
    'daughter(4, InvM)',

    'daughter(0, trackCurlerCloneIndicator)',
    'daughter(1, trackCurlerCloneIndicator)',
    'daughter(2, trackCurlerCloneIndicator)',
    'daughter(3, trackCurlerCloneIndicator)',
    'daughter(4, trackCurlerCloneIndicator)',

]

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['dzero_ntuple'],
    particleList='D0:all',
    variables=general_vars
)

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['dcharged_ntuple'],
    particleList='D+:all',
    variables=general_vars
)

############
# D* meson #
############

dst_vars = [
    'daughterMotherDiffOf(0, M)',
    'daughterMotherDiffOf(0, InvM)',
    'daughterInvariantMass(0, 1)',
    'abs(massDifferenceSignificance(0))',
    'daughter(1, p)',
    'daughter(1, InvM)',
    'daughter(1, SigMBF)',
    'daughter(1, M)',
    'daughter(1, E)',
    'useCMSFrame(daughter(1, p))',
    'useRestFrame(daughter(1, p))',
    'daughter(0, p)',
    'daughter(0, InvM)',
    'daughter(0, SigMBF)',
    'daughter(0, M)',
    'daughter(0, E)',
    'useCMSFrame(daughter(0, p))',
    'useRestFrame(daughter(0, p))',
    'useLabFrame(daughterAngleInBetween(0, 1))',
    'daughter(1, trackCurlerCloneIndicator)',
]

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['dstzero_ntuple'],
    particleList='D*0:all',
    variables=general_vars + dst_vars
)

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['dstcharged_ntuple'],
    particleList='D*+:all',
    variables=general_vars + dst_vars
)

basf2.print_path(main_path)
basf2.process(main_path)
print(statistics)
