"""
This file updates the final state particles cut infos for the low level cut study.

Patrick Ecker 2020
"""

from typing import List

from analysistools.processing.final_state_particles_utils import FinalStateParticleCollection

from rdstar.online_analysis.final_state_particles import *
from rdstar.offline_analysis.selection.selection_info_containers import CutInfo

__all__ = [
    "fsp_collection",
    "fsp_cut_sets"
]


default_scale_factor = 1.


def scale_cut_list(cut_list: List[CutInfo], scale_factor: float) -> List[CutInfo]:
    return [
        cut_info.scale_cut(percentage=scale_factor) for cut_info in cut_list
        if cut_info.scale_cut(percentage=scale_factor).cut_str
    ]


study_electron_cuts = scale_cut_list(cut_list=electron_cuts, scale_factor=default_scale_factor)
study_muon_cuts = scale_cut_list(cut_list=muon_cuts, scale_factor=default_scale_factor)
study_tau_cuts = scale_cut_list(cut_list=tau_cuts, scale_factor=default_scale_factor)
study_charged_pion_cuts = scale_cut_list(cut_list=charged_pion_cuts, scale_factor=default_scale_factor)
study_neutral_pion_cuts = scale_cut_list(cut_list=neutral_pion_cuts, scale_factor=default_scale_factor)
study_photon_cuts = scale_cut_list(cut_list=photon_cuts, scale_factor=default_scale_factor)
study_charged_kaon_cuts = scale_cut_list(cut_list=charged_kaon_cuts, scale_factor=default_scale_factor)
study_neutral_kaon_cuts = scale_cut_list(cut_list=neutral_kaon_cuts, scale_factor=default_scale_factor)
study_neutrino_cuts = scale_cut_list(cut_list=neutrino_cuts, scale_factor=default_scale_factor)
# electron_cuts = []
# muon_cuts = []
# tau_cuts = []
# charged_pion_cuts = []
# neutral_pion_cuts = []
# photon_cuts = []
# charged_kaon_cuts = []
# neutral_kaon_cuts = []
# neutrino_cuts = []

# Particle Definition
electron = generate_fsp(
    label="e",
    particle_list="e-:fsp",
    origin="reco",
    cut_infos=study_electron_cuts,
    electron_fsr_correction=electron_fsr_correction_settings
)

muon = generate_fsp(
    label="mu",
    particle_list="mu-:fsp",
    origin="reco",
    cut_infos=study_muon_cuts
)

tau = generate_fsp(
    label="tau",
    particle_list="tau-:mc",
    origin="mc",
    cut_infos=study_tau_cuts
)

charged_pion = generate_fsp(
    label="pi+",
    particle_list="pi+:fsp",
    origin="reco",
    cut_infos=study_charged_pion_cuts
)

neutral_pion = generate_fsp(
    label="pi0",
    particle_list="pi0:fsp",
    origin=["pi0:mdst"],
    cut_infos=study_neutral_pion_cuts
)

photon = generate_fsp(
    label="gamma",
    particle_list="gamma:fsp",
    origin=["gamma:mdst", "gamma:v0mdst"],
    cut_infos=study_photon_cuts
)

charged_kaon = generate_fsp(
    label="K+",
    particle_list="K+:fsp",
    origin="reco",
    cut_infos=study_charged_kaon_cuts
)

neutral_kaon = generate_fsp(
    label="K_S0",
    particle_list="K_S0:fsp",
    origin=["K_S0:mdst"],
    cut_infos=study_neutral_kaon_cuts
)

e_neutrino = generate_fsp(
    label="nu_e",
    particle_list="nu_e:mc",
    origin="mc",
    cut_infos=study_neutrino_cuts
)

mu_neutrino = generate_fsp(
    label="nu_mu",
    particle_list="nu_mu:mc",
    origin="mc",
    cut_infos=study_neutrino_cuts
)

tau_neutrino = generate_fsp(
    label="nu_tau",
    particle_list="nu_tau:mc",
    origin="mc",
    cut_infos=study_neutrino_cuts
)

# Add Particles to FSPCollection
fsp_collection = FinalStateParticleCollection()
fsp_collection.add_fsps(
    electron=electron,
    muon=muon,
    # tau=tau,
    charged_pion=charged_pion,
    neutral_pion=neutral_pion,
    photon=photon,
    charged_kaon=charged_kaon,
    neutral_kaon=neutral_kaon,
    e_neutrino=e_neutrino,
    mu_neutrino=mu_neutrino,
    tau_neutrino=tau_neutrino
)

fsp_cut_sets = {
    electron.label: study_electron_cuts,
    muon.label: study_muon_cuts,
    tau.label: study_tau_cuts,
    charged_pion.label: study_charged_pion_cuts,
    neutral_pion.label: study_neutral_pion_cuts,
    photon.label: study_photon_cuts,
    charged_kaon.label: study_charged_kaon_cuts,
    neutral_kaon.label: study_neutral_kaon_cuts,
    e_neutrino.label: study_neutrino_cuts,
    mu_neutrino.label: study_neutrino_cuts,
    tau_neutrino.label: study_neutrino_cuts,
}

if __name__ == '__main__':
    print(fsp_collection)
