"""
This file contains the plotting tools for the low level cut study.

Patrick Ecker 2020
"""

import matplotlib.pyplot as plt
from typing import List, Tuple, Dict
import pandas as pd
import copy
import numpy as np
import os

from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors

from rdstar.offline_analysis.selection.selection_info_containers import CutInfo
from rdstar.studies.low_level_cut_study.utilities.low_level_cut_defintions import fsp_cut_dict, D_cut_dict, \
    combined_decay_mode_id_dict, combined_channel_dict

__all__ = [
    "LowLevelCutPlotter"
]


class LowLevelCutPlotter:

    def __init__(self, n_tuple: str, target_path: os.fspath) -> None:
        self.n_tuple = n_tuple
        self.target_path = target_path
        self.df = None
        self.cut_dicts = fsp_cut_dict
        self.cut_dicts.update(D_cut_dict)

    def get_df(self, df: pd.DataFrame) -> None:
        self.df = df

    def apply_cuts(self, pivot: CutInfo, cut_infos: List[CutInfo]) -> pd.DataFrame:
        assert isinstance(self.df, pd.DataFrame), "You need to get a Dataframe first!"
        df_temp = copy.copy(self.df)
        cut_infos = [cut_info.make_root_compatible() for cut_info in cut_infos if cut_info != pivot]
        cut_strings = [cut_info.cut_str for cut_info in cut_infos]
        for cut in cut_strings:
            df_temp.query(cut, inplace=True)

        return df_temp

    def generate_output_dict(self) -> Dict[str, os.fspath]:
        output = {}
        relevant_keys = [key for key in self.cut_dicts if key.startswith(self.n_tuple)]
        for key in relevant_keys:
            for cut_info in self.cut_dicts[key]["cuts"]:
                save_path = os.path.join(self.target_path, f"{key}_{cut_info.cut_name}_evaluation.pdf")
                output.update({f"{key}_{cut_info.cut_name}_evaluation": save_path})

        return output

    def plot_on(self, save: bool = True) -> Dict[str, os.fspath]:
        output = {}
        relevant_keys = [key for key in self.cut_dicts if key.startswith(self.n_tuple)]

        for key in relevant_keys:
            channel = self.cut_dicts[key]["channel"]
            channel_cut = f"extraInfo__bodecayModeID__bc == {combined_decay_mode_id_dict[channel]}" if channel else None

            for cut_info in self.cut_dicts[key]["cuts"]:
                set_matplotlibrc_params()
                temp_cut_infos = copy.copy(self.cut_dicts[key]["cuts"])
                new_cut_info = cut_info.make_root_compatible()
                cut_var = new_cut_info.cut_var

                df = self.apply_cuts(pivot=cut_info, cut_infos=temp_cut_infos)
                if channel_cut:
                    df.query(channel_cut, inplace=True)

                if len(df) == 0:
                    save_path = os.path.join(self.target_path, f"{key}_{new_cut_info.cut_name}_evaluation.pdf")
                    output.update({f"{key}_{new_cut_info.cut_name}_evaluation": save_path})
                    plt.cla()
                    plt.title(cut_info.cut_str, loc="center")
                    plt.title(channel, loc="right")
                    if save:
                        plt.savefig(save_path)
                    continue

                fig, ax = plt.subplots()

                if isinstance(new_cut_info.cut_limits, Tuple):
                    cuts = list(new_cut_info.cut_limits)
                else:
                    cuts = [new_cut_info.cut_limits]
                for cut in cuts:
                    ax.axvline(cut, color=KITColors.kit_red, lw=3.)

                df_bkg = df.query("isSignal == 0")
                bkg_data = df_bkg[cut_var].abs() if new_cut_info.cut_abs else df_bkg[cut_var]
                df_sig = df.query("isSignal == 1")
                sig_data = df_sig[cut_var].abs() if new_cut_info.cut_abs else df_sig[cut_var]

                var_range = (min(bkg_data.min(), sig_data.min()), max(bkg_data.max(), sig_data.max()))

                sig_counts, sig_bins = np.histogram(sig_data, bins=50, range=var_range)
                bkg_counts, bkg_bins = np.histogram(bkg_data, bins=50, range=var_range)
                factor = 0.5 / (sig_counts.max() / ((sig_counts + bkg_counts).max()))
                if factor > 1:
                    factor = round(factor, 0)
                else:
                    factor = 1

                ax.hist(
                    [bkg_data, sig_data],
                    bins=50,
                    range=var_range,
                    histtype="stepfilled",
                    stacked=True,
                    lw=1.,
                    color=[KITColors.kit_cyan, KITColors.kit_orange],
                    label=["Background", "Signal"],
                    edgecolor=KITColors.kit_black
                )
                ax.hist(
                    sig_bins[:-1],
                    sig_bins,
                    weights=sig_counts * factor,
                    range=var_range,
                    histtype="stepfilled",
                    lw=1.,
                    color=KITColors.kit_black,
                    alpha=0.5,
                    edgecolor=KITColors.kit_black,
                    label=r"Signal Shape ($\times$" + f"{factor})" if factor != 1. else "Signal Shape"
                )

                plt.grid(False)
                plt.xlabel(cut_info.cut_label)
                plt.ylabel("Entries")
                plt.title(combined_channel_dict[channel], loc="right")
                plt.legend(loc=1, frameon=True, fancybox=True)

                save_path = os.path.join(self.target_path, f"{key}_{new_cut_info.cut_name}_evaluation.pdf")
                output.update({f"{key}_{new_cut_info.cut_name}_evaluation": save_path})
                if save:
                    plt.savefig(save_path)
                else:
                    plt.show()

        return output


if __name__ == '__main__':
    df = pd.read_hdf("/ceph/pecker/rdstar/low_level_cut_study/CombinedNTuples/dcharged_ntuple_total.h5")
    llcp = LowLevelCutPlotter(n_tuple="dcharged_ntuple", target_path="")
    print(llcp.generate_output_dict())
    llcp.get_df(df=df)
    llcp.plot_on(save=False)
