"""
Useful tools to plot the shapes of the MC templates

Patrick Ecker 2020
"""

from typing import List, Dict, Tuple, Union, Optional
import pandas as pd
import numpy as np
import scipy.stats as stats
import matplotlib.colors as colors
import matplotlib.pyplot as plt

from rdstar.offline_analysis.mc_ids import rdstar_sig_queries, sig_id_to_label
from rdstar.offline_analysis.rdstar_plots import rdstar_dec_modes, rdstar_dec_titles, rdstar_plot_order, \
    DecayModeDictType, DecayTitleDictType, get_rdstar_color_dict
from rdstar.tools.kolmogorov_smirnov import ks2d2s

from templatefitter.plotter.histogram_plots import SimpleHistogramPlot, StackedHistogramPlot
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params, xlabel_pos, ylabel_pos
from templatefitter.plotter.plot_utilities import export

__all__ = [
    "kit_cmp",
    "kit_cmp_inv",
    "ShapePlotter"
]

kit_cmp = colors.LinearSegmentedColormap.from_list(
    "",
    [
        KITColors.kit_red,
        KITColors.kit_orange,
        KITColors.kit_maygreen,
        KITColors.kit_green
    ],
    N=256
)
kit_cmp_inv = colors.LinearSegmentedColormap.from_list(
    "",
    [
        KITColors.kit_green,
        KITColors.kit_maygreen,
        KITColors.kit_orange,
        KITColors.kit_red
    ],
    N=256
)


class ShapePlotter:

    def __init__(
            self,
            df: pd.DataFrame,
            variable: Union[HistVariable, Tuple[HistVariable, HistVariable]],
            plot_order: Optional[List[str]] = None,
            plot_labels: Optional[Dict[int, str]] = None,
            plot_colors: Optional[Dict[int, str]] = None,
            available_modes: Optional[DecayModeDictType] = None,
            mode_labels: Optional[DecayTitleDictType] = None
    ) -> None:
        self._df = df
        self._variable = [*variable] if isinstance(variable, Tuple) else [variable]
        self._plot_order = rdstar_plot_order if plot_order is None else plot_order
        self._plot_labels = sig_id_to_label if plot_labels is None else plot_labels
        self._plot_colors = get_rdstar_color_dict() if plot_colors is None else plot_colors
        self._available_modes = rdstar_dec_modes if available_modes is None else available_modes
        self._mode_labels = rdstar_dec_titles if mode_labels is None else mode_labels

    def get_df_for_modes(self, decay_mode: str) -> pd.DataFrame:
        assert decay_mode in self._available_modes.keys()
        modes = self._available_modes[decay_mode]
        df = self._df.loc[self._df["daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"].isin(modes)]
        return df

    def plot_template_shapes(self, decay_mode: str, export_path: str = None) -> None:
        set_matplotlibrc_params()

        df = self.get_df_for_modes(decay_mode=decay_mode)

        for var in self._variable:
            for comp in self._plot_order:
                sig_id = rdstar_sig_queries[comp]["id"]
                temp_df = df.query(f"SIG_ID == {sig_id}")

                shp = SimpleHistogramPlot(var)

                shp.add_component(
                    label=self._plot_labels[sig_id],
                    data=temp_df,
                    color=self._plot_colors[sig_id]
                )

                fig, ax = plt.subplots()
                shp.plot_on(ax=ax)

                ax.set_title(self._mode_labels[decay_mode], loc="right")

                if export_path:
                    export(fig=fig, filename=f"{decay_mode}_shape_{var.df_label}_{comp}", target_dir=export_path)
                else:
                    plt.show()
                    ax.cla()

    def plot_template_shape_comparison(self, decay_mode: str, export_path: str = None) -> None:
        set_matplotlibrc_params()

        df = self.get_df_for_modes(decay_mode=decay_mode)

        for var in self._variable:
            shp = SimpleHistogramPlot(var)

            for comp in self._plot_order:
                sig_id = rdstar_sig_queries[comp]["id"]
                temp_df = df.query(f"SIG_ID == {sig_id}")

                shp.add_component(
                    label=self._plot_labels[sig_id],
                    data=temp_df,
                    color=self._plot_colors[sig_id],
                    hist_type="step"
                )

            fig, ax = plt.subplots()
            shp.plot_on(ax=ax, normed=True)

            ax.set_title(self._mode_labels[decay_mode], loc="right")

            if export_path:
                export(fig=fig, filename=f"{decay_mode}_shape_comparison_{var.df_label}", target_dir=export_path)
            else:
                plt.show()
                ax.cla()

    def plot_templates_stacked(self, decay_mode: str, export_path: str = None) -> None:
        set_matplotlibrc_params()

        df = self.get_df_for_modes(decay_mode=decay_mode)

        for var in self._variable:
            shp = StackedHistogramPlot(var)

            for comp in self._plot_order:
                sig_id = rdstar_sig_queries[comp]["id"]
                temp_df = df.query(f"SIG_ID == {sig_id}")

                shp.add_component(
                    label=self._plot_labels[sig_id],
                    data=temp_df,
                    color=self._plot_colors[sig_id],
                )

            fig, ax = plt.subplots()
            shp.plot_on(ax=ax)

            ax.set_title(self._mode_labels[decay_mode], loc="right")

            if export_path:
                export(fig=fig, filename=f"{decay_mode}_templates_stacked_{var.df_label}", target_dir=export_path)
            else:
                plt.show()
                ax.cla()

    def compare_shapes(self, decay_mode: str, export_path: str = None, plot_mode: str = "p_value") -> None:
        assert plot_mode in ["p_value", "ks_value"], "plot_mode not supported: has to be p_value or ks_value"
        set_matplotlibrc_params()

        df = self.get_df_for_modes(decay_mode=decay_mode)

        for var in self._variable:
            ks = np.zeros((len(self._plot_order), len(self._plot_order)))
            p_value = np.zeros((len(self._plot_order), len(self._plot_order)))

            for sig_1 in range(0, len(self._plot_order)):
                for sig_2 in range(0, len(self._plot_order)):
                    sig_ids = (rdstar_sig_queries[self._plot_order[sig_1]]["id"],
                               rdstar_sig_queries[self._plot_order[sig_2]]["id"])
                    temp_dfs = (df.query(f"SIG_ID == {sig_ids[0]}"), df.query(f"SIG_ID == {sig_ids[1]}"))

                    shape_1 = np.histogram(
                        temp_dfs[0][var.df_label],
                        bins=var.n_bins,
                        range=var.scope,
                        density=True
                    )[0]
                    shape_2 = np.histogram(
                        temp_dfs[1][var.df_label],
                        bins=var.n_bins,
                        range=var.scope,
                        density=True
                    )[0]

                    ks_value = stats.ks_2samp(shape_1, shape_2)

                    ks[sig_1, sig_2] = ks_value[0]
                    p_value[sig_1, sig_2] = ks_value[1]

            plot_values = p_value
            if plot_mode == "ks_value":
                plot_values = ks

            fig, ax = plt.subplots()

            ax.imshow(plot_values, cmap=kit_cmp_inv if plot_mode == "p_value" else kit_cmp)

            ax.set_xticks(np.arange(len(self._plot_order)))
            ax.set_yticks(np.arange(len(self._plot_order)))
            ax.set_xticklabels(
                [self._plot_labels[rdstar_sig_queries[item]["id"]] for item in self._plot_order],
                rotation=90
            )
            ax.set_yticklabels([self._plot_labels[rdstar_sig_queries[item]["id"]] for item in self._plot_order])

            for i in range(len(self._plot_order)):
                for j in range(len(self._plot_order)):
                    ax.text(j, i, round(plot_values[i, j], 2), ha="center", va="center", color="w")

            ax.set_title(self._mode_labels[decay_mode], loc="right")
            ax.set_title("KS-Value" if plot_mode == "ks_value" else "p-Value", loc="left")

            if export_path:
                export(fig=fig, filename=f"{decay_mode}_kolmogorov_smirnov_{plot_mode}_{var.df_label}",
                       target_dir=export_path)
            else:
                plt.show()
                ax.cla()

    def compare_shapes_2D(self, decay_mode: str, export_path: str = None, plot_mode: str = "p_value") -> None:
        assert plot_mode in ["p_value", "ks_value"], "plot_mode not supported: has to be p_value or ks_value"
        assert len(self._variable) == 2, "There have to be exactly two variables"
        assert self._variable[0].n_bins == self._variable[1].n_bins, "Variables need the same number of bins."
        set_matplotlibrc_params()

        df = self.get_df_for_modes(decay_mode=decay_mode)
        
        ks = np.zeros((len(self._plot_order), len(self._plot_order)))
        p_value = np.zeros((len(self._plot_order), len(self._plot_order)))

        for sig_1 in range(0, len(self._plot_order)):
            for sig_2 in range(0, len(self._plot_order)):
                sig_ids = (rdstar_sig_queries[self._plot_order[sig_1]]["id"],
                           rdstar_sig_queries[self._plot_order[sig_2]]["id"])
                temp_dfs = (df.query(f"SIG_ID == {sig_ids[0]}"), df.query(f"SIG_ID == {sig_ids[1]}"))

                shape_1_var1 = np.histogram(
                    temp_dfs[0][self._variable[0].df_label],
                    bins=self._variable[0].n_bins,
                    range=self._variable[0].scope,
                    density=True
                )[0]
                shape_2_var1 = np.histogram(
                    temp_dfs[1][self._variable[0].df_label],
                    bins=self._variable[0].n_bins,
                    range=self._variable[0].scope,
                    density=True
                )[0]

                shape_1_var2 = np.histogram(
                    temp_dfs[0][self._variable[1].df_label],
                    bins=self._variable[1].n_bins,
                    range=self._variable[1].scope,
                    density=True
                )[0]
                shape_2_var2 = np.histogram(
                    temp_dfs[1][self._variable[1].df_label],
                    bins=self._variable[1].n_bins,
                    range=self._variable[1].scope,
                    density=True
                )[0]

                shapes_1 = np.array([shape_1_var1, shape_1_var2])
                shapes_2 = np.array([shape_2_var1, shape_2_var2])
                ks_value = ks2d2s(shapes_1, shapes_2)

                ks[sig_1, sig_2] = ks_value[0]
                p_value[sig_1, sig_2] = ks_value[1]

        plot_values = p_value
        if plot_mode == "ks_value":
            plot_values = ks

        fig, ax = plt.subplots()

        ax.imshow(plot_values, cmap=kit_cmp_inv if plot_mode == "p_value" else kit_cmp)

        ax.set_xticks(np.arange(len(self._plot_order)))
        ax.set_yticks(np.arange(len(self._plot_order)))
        ax.set_xticklabels(
            [self._plot_labels[rdstar_sig_queries[item]["id"]] for item in self._plot_order],
            rotation=90
        )
        ax.set_yticklabels([self._plot_labels[rdstar_sig_queries[item]["id"]] for item in self._plot_order])

        for i in range(len(self._plot_order)):
            for j in range(len(self._plot_order)):
                ax.text(j, i, round(plot_values[i, j], 2), ha="center", va="center", color="w")

        ax.set_title(self._mode_labels[decay_mode], loc="right")
        ax.set_title("KS-Value" if plot_mode == "ks_value" else "p-Value", loc="left")

        if export_path:
            export(fig=fig, filename=f"{decay_mode}_kolmogorov_smirnov_2D_{plot_mode}", target_dir=export_path)
        else:
            plt.show()
            ax.cla()

    def template_composition(self, decay_mode: str) -> Dict[str, Tuple[int, float]]:
        df = self.get_df_for_modes(decay_mode=decay_mode)

        composition_dict = {}
        total_number_of_events = df.shape[0]
        for comp in self._plot_order:
            sig_id = rdstar_sig_queries[comp]["id"]
            temp_df = df.query(f"SIG_ID == {sig_id}")

            template_events = temp_df.shape[0]

            composition_dict[comp] = (template_events, template_events / total_number_of_events)

        return composition_dict

    def plot_template_composition_pie(self, decay_mode: str, export_path: str = None) -> None:
        composition = self.template_composition(decay_mode=decay_mode)

        fig, ax = plt.subplots()

        ax.pie(
            [composition[comp][1] for comp in self._plot_order],
            colors=[get_rdstar_color_dict()[rdstar_sig_queries[comp]["id"]] for comp in self._plot_order],
            labels=[sig_id_to_label[rdstar_sig_queries[comp]["id"]] for comp in self._plot_order],
            wedgeprops={'edgecolor': 'black', "lw": 0.3}
        )

        ax.set_title(self._mode_labels[decay_mode], loc="center")

        if export_path:
            export(fig=fig, filename=f"{decay_mode}_template_composition_pie", target_dir=export_path)
        else:
            plt.show()
            ax.cla()

    def plot_mva_eff_for_template(self, decay_mode: str, mva_output_col: str, export_path: str = None) -> None:
        df = self.get_df_for_modes(decay_mode=decay_mode)

        fig, ax = plt.subplots()

        for comp in self._plot_order:
            sig_id = rdstar_sig_queries[comp]["id"]
            temp_df = df.query(f"SIG_ID == {sig_id}")

            total_template_events = temp_df.shape[0]
            cut_values = np.linspace(0, 1, 100)
            efficiencies = []

            for value in cut_values:
                eff = temp_df.query(f"{mva_output_col} > {value}").shape[0] / total_template_events
                efficiencies.append(eff)

            ax.plot(
                cut_values,
                efficiencies,
                color=self._plot_colors[sig_id],
                label=self._plot_labels[sig_id]
            )

        ax.set_xlabel("MVA Output Cut", xlabel_pos)
        ax.set_ylabel("Efficiency", ylabel_pos)
        ax.set_title(self._mode_labels[decay_mode], loc="right")

        plt.legend(loc="lower left")

        if export_path:
            export(fig=fig, filename=f"{decay_mode}_mva_cut_efficiency", target_dir=export_path)
        else:
            plt.show()
            ax.cla()
